Package: bsseq
Version: 1.47.1
Encoding: UTF-8
Title: Analyze, manage and store whole-genome methylation data
Description: A collection of tools for analyzing and visualizing whole-genome
    methylation data from sequencing. This includes whole-genome bisulfite
    sequencing and Oxford nanopore data.
Authors@R: c(person(c("Kasper", "Daniel"), "Hansen", role = c("aut", "cre"),
    email = "kasperdanielhansen@gmail.com", comment = c(ORCID = "0000-0003-0086-0687")),
    person("Peter", "Hickey", role = "aut", email = "peter.hickey@gmail.com", comment = c(ORCID = "0000-0002-8153-6258")),
    person("Hervé", "Pagès", role = "ctb"),
    person("Aaron", "Lun", role = "ctb"))
Depends: R (>= 4.0), methods, BiocGenerics, GenomicRanges (>= 1.41.5),
        SummarizedExperiment (>= 1.19.5)
Imports: IRanges (>= 2.23.9), Seqinfo, scales, stats, parallel, tools,
        graphics, Biobase, locfit, gtools, data.table (>= 1.11.8),
        S4Vectors (>= 0.27.12), R.utils (>= 2.0.0), DelayedMatrixStats
        (>= 1.5.2), permute, limma, DelayedArray (>= 0.15.16), Rcpp,
        BiocParallel, BSgenome, Biostrings, utils, HDF5Array (>=
        1.19.11), rhdf5, beachmat (>= 2.23.2)
Suggests: testthat, bsseqData, BiocStyle, rmarkdown, knitr, Matrix,
        doParallel, rtracklayer, BSgenome.Hsapiens.UCSC.hg38,
        batchtools
Collate: utils.R hasGRanges.R BSseq-class.R BSseqTstat_class.R
        BSseq_utils.R combine.R read.bismark.R read.bedMethyl.R
        read.modbam2bed.R read.modkit.R BSmooth.R BSmooth.tstat.R
        dmrFinder.R gof_stats.R plotting.R fisher.R permutations.R
        BSmooth.fstat.R BSseqStat_class.R getStats.R hdf5_utils.R
        DelayedArray_utils.R collapseBSseq.R FWIRanges-class.R
        FWGRanges-class.R findLoci.R Likelihood_functions.R
License: Artistic-2.0
VignetteBuilder: knitr
URL: https://github.com/kasperdanielhansen/bsseq
BugReports: https://github.com/kasperdanielhansen/bsseq/issues
biocViews: DNAMethylation
LinkingTo: Rcpp, beachmat, assorthead (>= 1.1.4)
SystemRequirements: C++17
NeedsCompilation: yes
RoxygenNote: 7.1.0
git_url: https://git.bioconductor.org/packages/bsseq
git_branch: devel
git_last_commit: 46b4295
git_last_commit_date: 2025-10-31
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
Packaged: 2025-11-04 21:34:55 UTC; biocbuild
Author: Kasper Daniel Hansen [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-0086-0687>),
  Peter Hickey [aut] (ORCID: <https://orcid.org/0000-0002-8153-6258>),
  Hervé Pagès [ctb],
  Aaron Lun [ctb]
Maintainer: Kasper Daniel Hansen <kasperdanielhansen@gmail.com>

##
## Importing
##

import(methods)
import(S4Vectors)
import(IRanges)
import(GenomicRanges)
import(SummarizedExperiment)
import(DelayedArray)
import(HDF5Array)
import(BiocParallel)
import(limma)
importFrom(BiocGenerics, "anyDuplicated", "cbind", "colnames",
           "combine", "density", "intersect", "lapply", "ncol",
           "nrow", "order", "paste", "pmax", "pmin", "rbind",
           "Reduce", "rep.int", "rownames", "sapply", "setdiff",
           "strand", "strand<-", "union", "unique", "updateObject", "unstrand")
importFrom(stats, "approxfun", "fisher.test", "ppoints",
           "predict", "preplot", "qchisq",
           "qqplot", "qunif", "cov2cor",
           "setNames")
importFrom(graphics, "abline", "axis", "layout", "legend", "lines",
           "mtext", "par", "plot", "points", "polygon", "rect", "rug", "text", "hist")
import(parallel)
importFrom(locfit, "locfit", "lp")
importFrom(DelayedMatrixStats, "rowSds", "rowVars", "colMeans2", "colSums2",
           "rowSums2", "rowMeans2", "rowAlls")

importFrom(scales, "alpha")
importClassesFrom(Biobase, "AnnotatedDataFrame")
importMethodsFrom(Biobase, "annotatedDataFrameFrom",
                  "pData", "pData<-",
                  "sampleNames", "sampleNames<-")
importFrom(Biobase, "validMsg")
importMethodsFrom(Seqinfo, "seqlengths", "seqlengths<-", "seqinfo",
                  "seqinfo<-", "seqnames", "seqnames<-", "seqlevels",
                  "seqlevels<-", "sortSeqlevels")
importFrom(Seqinfo, "Seqinfo")
importFrom(gtools, "combinations")
importFrom(beachmat, initializeCpp)
importFrom(Rcpp, sourceCpp)

# NOTE: data.table has some NAMESPACE clashes with functions in Bioconductor,
#       e.g., shift(). If new ones are discovered, add them to this list.
import(data.table, except = c(shift, first, second))
importFrom(permute, "shuffleSet", "how")
importFrom(Biostrings, "DNAString", "vmatchPattern", "reverseComplement")
importFrom(utils, "read.delim", "read.table")
importFrom(BSgenome, "vmatchPattern")
importFrom(tools, "file_path_as_absolute")
importFrom(R.utils, "isGzipped", "isBzipped", "gunzip", "bunzip2")

##
## Exporting
##

exportClasses("hasGRanges",
              "BSseq",
              "BSseqTstat",
              "BSseqStat",
              "matrixOrNULL")

exportMethods("[", "show",
              "seqnames", "seqnames<-",
              "seqlevels", "seqlevels<-",
              "seqlengths", "seqlengths<-",
              "start", "start<-",
              "end", "end<-",
              "width", "width<-",
              "strand", "strand<-",
              "granges",
              "dim", "nrow", "ncol",
              "sampleNames", "sampleNames<-",
              "pData", "pData<-",
              "findOverlaps", "overlapsAny", "subsetByOverlaps",
              "combine", "updateObject")

export("BSseq", "getMeth", "getCoverage", "getBSseq", "getStats",
       "collapseBSseq", "orderBSseq", "hasBeenSmoothed", "chrSelectBSseq",
       "BSmooth", "BSmooth.tstat", "dmrFinder", "fisherTests",
       "combineList", "strandCollapse",
       "plotRegion", "plotManyRegions",
       "read.bismark", "read.bedMethyl",
       "read.modbam2bed", "read.modkit",
       "poissonGoodnessOfFit", "binomialGoodnessOfFit",
       "data.frame2GRanges", "BSseqTstat",
       "BSseqStat",
       "findLoci","estimateErrorRate",
       "getCpGs",
       "getMaxLikelihoodMatrix","getCpGMatrix")

S3method("print", "chisqGoodnessOfFit")
S3method("plot", "chisqGoodnessOfFit")
S3method("print", "summary.BSseqTstat")
S3method("summary", "BSseqTstat")
S3method("plot", "BSseqTstat")

## Fixes, to be removed

exportMethods("assays", "assayNames")

# C++ code registration
useDynLib(bsseq, .registration = TRUE, .fixes = "cxx_")

# R package: bsseq

[![Bioc release status](http://www.bioconductor.org/shields/build/release/bioc/bsseq.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/bsseq)
[![Bioc devel status](http://www.bioconductor.org/shields/build/devel/bioc/bsseq.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/bsseq)
[![Bioc downloads rank](https://bioconductor.org/shields/downloads/release/bsseq.svg)](http://bioconductor.org/packages/stats/bioc/bsseq/)
[![Bioc support](https://bioconductor.org/shields/posts/bsseq.svg)](https://support.bioconductor.org/tag/bsseq)
[![Bioc history](https://bioconductor.org/shields/years-in-bioc/bsseq.svg)](https://bioconductor.org/packages/release/bioc/html/bsseq.html#since)
[![Bioc last commit](https://bioconductor.org/shields/lastcommit/devel/bioc/bsseq.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/bsseq/)
[![Bioc dependencies](https://bioconductor.org/shields/dependencies/release/bsseq.svg)](https://bioconductor.org/packages/release/bioc/html/bsseq.html#since)

This is the developer version of Bioconductor package [bsseq](http://bioconductor.org/packages/devel/bioc/html/bsseq.html).  Install in R as:

```r
install.packages("BiocManager")
BiocManager::install("bsseq", version = "devel")
```

getStats <- function(bstat, regions = NULL, ...) {
    stopifnot(is(bstat, "BSseqTstat") || is(bstat, "BSseqStat"))
    if(is(bstat, "BSseqTstat"))
        return(getStats_BSseqTstat(bstat, regions = regions, ...))
    getStats_BSseqStat(bstat, regions = regions, ...)
}

# NOTE: Realises in memory a matrix with nrow = length(hits) and ncol = 1
.getRegionStats <- function(stat, hits, na.rm = FALSE) {
    stat_by_region <- split(as.array(stat[queryHits(hits), ]),
                            subjectHits(hits))
    data.frame(areaStat = vapply(stat_by_region, sum, numeric(1),
                                 na.rm = na.rm, USE.NAMES = FALSE),
               maxStat = vapply(stat_by_region, max, numeric(1),
                                na.rm = na.rm, USE.NAMES = FALSE))
}

getStats_BSseqStat <- function(BSseqStat, regions = NULL, what = NULL) {
    stopifnot(is(BSseqStat, "BSseqStat"))
    if(!is.null(what)) {
        stopifnot(what %in% names(BSseqStat@stats))
        return(BSseqStat@stats[[what]])
    }
    if(is.null(regions))
        return(BSseqStat@stats)
    ## Now we have regions and no what
    if(is(regions, "data.frame")) {
        regions <- data.frame2GRanges(regions)
    }
    hits <- findOverlaps(BSseqStat, regions)
    regionStats <- .getRegionStats(stat = BSseqStat@stats$stat,
                                   hits = hits)
    regionStats
}

# NOTE: Realises in memory a matrix with nrow = length(hits) and ncol = 3
.getRegionStats_ttest <- function(stats, hits, na.rm = FALSE) {
    stat_names <- c("group1.means", "group2.means", "tstat.sd")
    stats <- as.array(stats[queryHits(hits), stat_names])
    stats_by_region <- lapply(seq_along(stat_names), function(j) {
        split(stats[, j], subjectHits(hits))
    })
    names(stats_by_region) <- stat_names
    meanDiff <- mapply(function(g1, g2, na.rm) {
        mean(g1 - g2, na.rm = na.rm)
    }, g1 = stats_by_region[["group1.means"]],
    g2 = stats_by_region[["group2.means"]],
    MoreArgs = list(na.rm = na.rm), SIMPLIFY = TRUE, USE.NAMES = FALSE)
    group1.mean <- vapply(stats_by_region[["group1.means"]], mean, numeric(1),
                          na.rm = na.rm, USE.NAMES = FALSE)
    group2.mean <- vapply(stats_by_region[["group2.means"]], mean, numeric(1),
                          na.rm = na.rm, USE.NAMES = FALSE)
    tstat.sd <- vapply(stats_by_region[["tstat.sd"]], mean, numeric(1),
                       na.rm = na.rm, USE.NAMES = FALSE)
    data.frame(meanDiff = meanDiff,
               group1.mean = group1.mean,
               group2.mean = group2.mean,
               tstat.sd = tstat.sd)
}

getStats_BSseqTstat <- function(BSseqTstat, regions = NULL, stat = "tstat.corrected") {
    stopifnot(is(BSseqTstat, "BSseqTstat"))
    if(is.null(regions))
        return(BSseqTstat@stats)
    if(is(regions, "data.frame")) {
        regions <- data.frame2GRanges(regions)
    }
    stopifnot(stat %in% colnames(BSseqTstat@stats))
    stopifnot(length(stat) == 1)
    stopifnot(is(regions, "GenomicRanges"))
    hits <- findOverlaps(BSseqTstat, regions)
    out <- .getRegionStats(stat = BSseqTstat@stats[, stat, drop = FALSE],
                           hits = hits)
    if(! stat %in% c("tstat.corrected", "tstat"))
        return(out)
    stats_ttest <- .getRegionStats_ttest(stats = BSseqTstat@stats,
                                         hits = hits)
    out <- cbind(out, stats_ttest)
    out
}

BSmooth.tstat <- function(BSseq, group1, group2, estimate.var = c("same", "paired", "group2"),
                          local.correct = TRUE, maxGap = NULL, qSd = 0.75, k = 21, mc.cores = 1, verbose = TRUE){
    smoothSd <- function(Sds, k) {
        k0 <- floor(k/2)
        if(all(is.na(Sds))) return(Sds)
        thresSD <- pmax(Sds, quantile(Sds, qSd, na.rm = TRUE), na.rm = TRUE)
        addSD <- rep(median(Sds, na.rm = TRUE), k0)
        sSds <- as.vector(runmean(Rle(c(addSD, thresSD, addSD)), k = k))
        sSds
    }
    compute.correction <- function(idx, qSd = 0.75) {
        xx <- start(BSseq)[idx]
        yy <- tstat[idx]
        suppressWarnings({
            drange <- diff(range(xx, na.rm = TRUE))
        })
        if(drange <= 25000)
            return(yy)
        tstat.function <- approxfun(xx, yy)
        xx.reg <- seq(from = min(xx), to = max(xx), by = 2000)
        yy.reg <- tstat.function(xx.reg)
        fit <- locfit(yy.reg ~ lp(xx.reg, h = 25000, deg = 2, nn = 0),
                      family = "huber", maxk = 50000)
        correction <- predict(fit, newdata = data.frame(xx.reg = xx))
        yy - correction
    }

    estimate.var <- match.arg(estimate.var)
    stopifnot(is(BSseq, "BSseq"))
    stopifnot(hasBeenSmoothed(BSseq))
    if(is.character(group1)) {
        stopifnot(all(group1 %in% sampleNames(BSseq)))
        group1 <- match(group1, sampleNames(BSseq))
    }
    if(is.numeric(group1)) {
        stopifnot(min(group1) >= 1 & max(group1) <= ncol(BSseq))
    } else stop("problems with argument 'group1'")
    if(is.character(group2)) {
        stopifnot(all(group2 %in% sampleNames(BSseq)))
        group2 <- match(group2, sampleNames(BSseq))
    }
    if(is.numeric(group2)) {
        stopifnot(min(group2) >= 1 & max(group2) <= ncol(BSseq))
    } else stop("problems with argument 'group2'")
    stopifnot(length(intersect(group1, group2)) == 0)
    stopifnot(length(group1) > 0)
    stopifnot(length(group2) > 0)
    stopifnot(length(group1) + length(group2) >= 3)
    if(estimate.var == "paired")
        stopifnot(length(group1) == length(group2))

    if (!.isHDF5ArrayBacked(getCoverage(BSseq))) {
        ## TODO: This check can be super expensive if Cov is a HDF5Matrix, so
        ##       we skip it for the time being. Really noticeable when Cov
        ##       needs row subsetted/reordered (i.e. has Cov@index[[1]] exists)
        ##
        if(any(rowSums2(getCoverage(BSseq)[, c(group1, group2)]) == 0)) {
            warning("Computing t-statistics at locations where there is no data; consider subsetting the 'BSseq' object first")
        }
    }

    if(is.null(maxGap))
        maxGap <- BSseq@parameters$maxGap
    if(is.null(maxGap))
        stop("need to set argument 'maxGap'")

    if(verbose) cat("[BSmooth.tstat] preprocessing ... ")
    ptime1 <- proc.time()
    clusterIdx <- makeClusters(BSseq, maxGap = maxGap)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))

    if(verbose) cat("[BSmooth.tstat] computing stats within groups ... ")
    ptime1 <- proc.time()
    allPs <- getMeth(BSseq, type = "smooth", what = "perBase",
                     confint = FALSE)
    group1.means <- rowMeans2(allPs[, group1, drop = FALSE], na.rm = TRUE)
    group2.means <- rowMeans2(allPs[, group2, drop = FALSE], na.rm = TRUE)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))

    if(verbose) cat("[BSmooth.tstat] computing stats across groups ... ")
    ptime1 <- proc.time()
    switch(estimate.var,
           "group2" = {
               rawSds <- .rowSds(x = allPs, cols = group2, na.rm = TRUE)
               smoothSds <- do.call(c, mclapply(clusterIdx, function(idx) {
                   smoothSd(rawSds[idx], k = k)
               }, mc.cores = mc.cores))
               scale <- sqrt(1/length(group1) + 1/length(group2))
               tstat.sd <- smoothSds * scale
           },
           "same" = {
               rawSds <- sqrt( ((length(group1) - 1) * .rowVars(x = allPs, cols = group1) +
                                    (length(group2) - 1) * .rowVars(x = allPs, cols = group2)) /
                                   (length(group1) + length(group2) - 2))
               smoothSds <- do.call(c, mclapply(clusterIdx, function(idx) {
                   smoothSd(rawSds[idx], k = k)
               }, mc.cores = mc.cores))
               scale <- sqrt(1/length(group1) + 1/length(group2))
               tstat.sd <- smoothSds * scale
           },
           "paired" = {
               rawSds <- .rowSds(x = allPs[, group1, drop = FALSE] - allPs[, group2, drop = FALSE])
               smoothSds <- do.call(c, mclapply(clusterIdx, function(idx) {
                   smoothSd(rawSds[idx], k = k)
               }, mc.cores = mc.cores))
               scale <- sqrt(1/length(group1))
               tstat.sd <- smoothSds * scale
           })
    tstat <- (group1.means - group2.means) / tstat.sd
    is.na(tstat)[tstat.sd == 0] <- TRUE
    if(local.correct) {
        tstat.corrected <- do.call(c, mclapply(clusterIdx,
                                               compute.correction, qSd = qSd,
                                               mc.cores = mc.cores))
    }
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))

    if(local.correct) {
        stats <- cbind(rawSds, tstat.sd, group2.means, group1.means,
                       tstat, tstat.corrected)
        colnames(stats) <- c("rawSds", "tstat.sd",
                             "group2.means", "group1.means", "tstat",
                             "tstat.corrected")

    } else {
        stats <- cbind(rawSds, tstat.sd, group2.means, group1.means,
                       tstat)
        colnames(stats) <- c("rawSds", "tstat.sd",
                             "group2.means", "group1.means", "tstat")
    }

    parameters <- c(BSseq@parameters,
                    list(tstatText = sprintf("BSmooth.tstat (local.correct = %s, maxGap = %d)",
                                             local.correct, maxGap),
                         group1 = group1, group2 = group2, k = k, qSd = qSd,
                         local.correct = local.correct, maxGap = maxGap))
    out <- BSseqTstat(gr = granges(BSseq), stats = stats, parameters = parameters)
    out
}

# Exported functions -----------------------------------------------------------

findLoci <- function(pattern, subject, include = seqlevels(subject),
                     strand = c("*", "+", "-"), fixed = "subject",
                     resize = TRUE) {
    pattern <- DNAString(pattern)
    strand <- match.arg(strand)

    if (is.character(subject) || is(subject, "RTLFile")) {
        if (!requireNamespace("rtracklayer", quietly = TRUE)) {
            stop("rtracklayer package required for importing 'subject'.")
        }
        subject <- try(rtracklayer::import(subject), silent = TRUE)
        if (!is(subject, "DNAStringSet")) {
            stop("Unable to import 'subject' as a DNAStringSet.")
        }
    }

    if (is(subject, "BSgenome")) {
        # NOTE: vmatchPattern,BSgenome-method returns a GRanges instance and
        #       automatically checks both forward and reverse strands.
        gr <- vmatchPattern(
            pattern = pattern,
            subject = subject,
            # NOTE: This must be a regular expression;
            #       see https://github.com/Bioconductor/BSgenome/issues/1.
            exclude = paste0("^", setdiff(seqnames(subject), include), "$"),
            fixed = fixed)
        if (identical(strand, "+")) {
            gr <- gr[strand(gr) == "+"]
        } else if (identical(strand, "-")) {
            gr <- gr[strand(gr) == "-"]
        }
    } else if (is(subject, "DNAStringSet")) {
        subject <- subject[include]
        if (strand %in% c("*", "+")) {
            fwd_gr <- as(
                vmatchPattern(
                    pattern = pattern,
                    subject = subject,
                    fixed = fixed),
                "GRanges")
            strand(fwd_gr) <- "+"
        } else {
            fwd_gr <- GRanges()
        }
        if (strand %in% c("*", "-")) {
            rev_gr <- as(
                vmatchPattern(
                    pattern = reverseComplement(pattern),
                    subject = subject,
                    fixed = fixed),
                "GRanges")
            strand(rev_gr) <- "-"
        } else {
            rev_gr <- GRanges()
        }
        gr <- c(fwd_gr, rev_gr)
    } else {
        stop("Cannot handle 'subject' of class ", class(subject), ".")
    }

    if (resize) {
        gr <- resize(gr, width = 1L, fix = "start")
    }
    gr
}

# TODOs ------------------------------------------------------------------------

# TODO: Default value of seqlevels may not be working.
# TODO: What happens if subject is a non-filepath character (e.g., "CATGCG") or
#       a DNAString?
# TODO: Allow passing of seqinfo and/or autogenerate?

poissonGoodnessOfFit <- function(BSseq, nQuantiles = 10^5) {
    Cov <- getBSseq(BSseq, type = "Cov")
    lambda <- rowMeans2(Cov)
    out <- list(chisq = NULL, df = ncol(Cov) - 1)
    out$chisq <- rowSums2((Cov - lambda)^2 / lambda)
    if(nQuantiles > length(out$chisq))
        nQuantiles <- length(out$chisq)
    probs <- seq(from = 0, to = 1, length.out = nQuantiles)
    out$quantiles <- quantile(out$chisq, prob = probs, na.rm = TRUE)
    class(out) <- "chisqGoodnessOfFit"
    return(out)
}

print.chisqGoodnessOfFit <- function(x, ...) {
    cat("Chisq Goodness of Fit object\n")
    cat(" ", length(x$chisq), "tests\n")
    cat(" ", x$df, "degrees of freedom\n")
}

binomialGoodnessOfFit <- function(BSseq, method = c("MLE"), nQuantiles = 10^5) {
    method <- match.arg(method)
    M <- getCoverage(BSseq, type = "M", what = "perBase")
    Cov <- getCoverage(BSseq, type = "Cov", what = "perBase")
    if(method == "MLE") # This is the MLE under iid
        p <- rowSums2(M) / rowSums2(Cov)
    ## p <- rowMeans(M / Cov, na.rm = TRUE)
    out <- list(chisq = NULL, quantiles = NULL, df = ncol(Cov) - 1)
    ## This gets tricky.  For the binomial model, we need Cov > 0, but
    ## this implies that a given location may have less then nCol(BSseq)
    ## number of observations, and hence the degrees of freedom will
    ## vary.
    out$chisq <- rowSums2((M - Cov*p)^2 / sqrt(Cov * p * (1-p)))
    probs <- seq(from = 0, to = 1, length.out = nQuantiles)
    ## Next line will remove all locations where not all samples have coverage
    out$quantiles <- quantile(out$chisq, prob = probs, na.rm = TRUE)
    class(out) <- "chisqGoodnessOfFit"
    return(out)
}

plot.chisqGoodnessOfFit <- function(x, type = c("chisq", "pvalue"),
                                    plotCol = TRUE, qqline = TRUE, pch = 16, cex = 0.75, ...) {
    type <- match.arg(type)
    switch(type,
           "chisq" = {
               yy <- x$quantiles
               xx <- qchisq(ppoints(yy), df = x$df)
           },
           "pvalue" = {
               yy <- sort(1 - qchisq(x$quantiles, df = x$df))
               xx <- qunif(ppoints(yy), min = 0, max = 1)
           })
    if(plotCol) {
        colors <- rep("black", length(yy))
        colors[floor(length(colors)*.95):length(colors)] <- "red"
        colors[floor(length(colors)*.99):length(colors)] <- "violet"
        colors[floor(length(colors)*.999):length(colors)] <- "orange"
    } else {
        colors <- "black"
    }
    qqplot(xx, yy, xlab = "Theoretical quantiles",
           ylab = "Observed quantiles", col = colors, pch = pch, cex = cex, ...)
    if(qqline) {
        yyqt <- quantile(yy, c(0.25, 0.75))
        xxqt <- quantile(xx, c(0.25, 0.75))
        slope <- diff(yyqt) / diff(xxqt)
        int <- yyqt[1L] - slope * xxqt[1L]
        abline(int, slope, col = "blue")
    }
    abline(0, 1, col = "blue", lty = 2)
}

## The following script downloads and constructs the BS.chr22 dataset,
## included in the bsseq package

library(bsseq)

## First we download.  Each file is slightly less than 200 MB

download.file(
    url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_imr90_r1.tar.gz",
    destfile = "mc_imr90_r1.tar.gz")
untar("mc_imr90_r1.tar.gz",  "mc_imr90_r1/mc_imr90_r1_22", compressed = TRUE)
download.file(
    url = "ftp://ftpuser3:s3qu3nc3@neomorph.salk.edu/mc/mc_imr90_r2.tar.gz",
    destfile = "mc_imr90_r2.tar.gz")
untar("mc_imr90_r2.tar.gz",  "mc_imr90_r2/mc_imr90_r2_22", compressed = TRUE)

## Now the workhorse function

read.lister <- function(file) {
    dat <- read.table(
        file,
        skip = 1,
        row.names = NULL,
        col.names = c("chr", "pos", "strand", "context", "M", "Cov"),
        colClasses = c("character", "integer", "character", "character",
                       "integer", "integer"))
    ## we remove all non-CpG calls.  This includes SNPs
    dat <- dat[dat$context == "CG", ]
    dat$context <- NULL
    dat$chr <- paste("chr", dat$chr, sep = "")
    ## Now we need to handle that the data has separate lines for each strand
    ## We join these
    tmp <- dat[dat$strand == "+",]
    BS.forward <- BSseq(
        pos = tmp$pos,
        chr = tmp$chr,
        M = as.matrix(tmp$M, ncol = 1),
        Cov = as.matrix(tmp$Cov, ncol = 1),
        sampleNames = "forward")
    tmp <- dat[dat$strand == "-",]
    BS.reverse <- BSseq(
        pos = tmp$pos - 1L,
        chr = tmp$chr,
        M = as.matrix(tmp$M, ncol = 1),
        Cov = as.matrix(tmp$Cov, ncol = 1),
        sampleNames = "reverse")
    BS <- combine(BS.forward, BS.reverse)
    BS <- collapseBSseq(BS, group = c("a", "a"), type = "integer")
    BS
}

BS.r1 <- read.lister("mc_imr90_r1/mc_imr90_r1_22")
sampleNames(BS.r1) <- "r1"
BS.r2 <- read.lister("mc_imr90_r2/mc_imr90_r2_22")
sampleNames(BS.r2) <- "r2"

BS.chr22 <- combine(BS.r1, BS.r2)
pData(BS.chr22)$Rep <- c("replicate1", "replicate2")
validObject(BS.chr22)
pData(BS.chr22)

save(BS.chr22, file = "BS.chr22.rda")
tools::resaveRdaFiles("BS.chr22.rda")

fisherTests <- function(BSseq, group1, group2, lookup = NULL,
                        returnLookup = TRUE, mc.cores = 1,
                        verbose = TRUE) {
    stopifnot(is(BSseq, "BSseq"))
    if(is.character(group1)) {
        stopifnot(all(group1 %in% sampleNames(BSseq)))
        group1 <- match(group1, sampleNames(BSseq))
    }
    if(is.numeric(group1)) {
        stopifnot(min(group1) >= 1 & max(group1) <= ncol(BSseq))
    } else stop("problems with argument 'group1'")
    if(is.character(group2)) {
        stopifnot(all(group2 %in% sampleNames(BSseq)))
        group2 <- match(group2, sampleNames(BSseq))
    }
    if(is.numeric(group2)) {
        stopifnot(min(group2) >= 1 & max(group2) <= ncol(BSseq))
    } else stop("problems with argument 'group2'")
    stopifnot(length(intersect(group1, group2)) == 0)
    if(is.null(lookup)) {
        oldkeys <- ""
    } else {
        stopifnot(is.matrix(lookup) && ncol(lookup) == 2 &&
                  setequal(colnames(lookup), c("p.value", "log2OR")))
        oldkeys <- rownames(lookup)
    }
    if(verbose) cat("[fisherTests] setup ... ")
    ptime1 <- proc.time()
    M1 <- rowSums2(getCoverage(BSseq, type = "M")[, group1, drop = FALSE])
    M2 <- rowSums2(getCoverage(BSseq, type = "M")[, group2, drop = FALSE])
    Cov1 <- rowSums2(getCoverage(BSseq, type = "Cov")[, group1, drop = FALSE])
    Cov2 <- rowSums2(getCoverage(BSseq, type = "Cov")[, group2, drop = FALSE])
    keys <- paste(Cov1, Cov2, M1, M2, sep = "_")
    newkeys <- setdiff(keys, oldkeys)
    expand <- matrix(as.integer(do.call(rbind, strsplit(newkeys, "_"))), ncol = 4)
    colnames(expand) <- c("Cov1", "Cov2", "M1", "M2")
    expand2 <- expand
    colnames(expand2) <- c("U1", "U2", "M1", "M2")
    expand2[,"U1"] <- expand[,"Cov1"] - expand[,"M1"]
    expand2[,"U2"] <- expand[,"Cov2"] - expand[,"M2"]
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(round(stime,1), "done in secs\n")
    if(verbose) cat(sprintf("[fisherTests] performing %d tests (using %d cores) ... ",
                            nrow(expand2), mc.cores))
    ptime1 <- proc.time()
    tests <- mclapply(1:nrow(expand2), function(ii) {
        unlist(fisher.test(matrix(expand2[ii,], ncol = 2))[c("p.value", "estimate")])
    }, mc.cores = mc.cores)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(round(stime,1), "done in secs\n")
    if(verbose) cat("[fisherTests] finalizing ... ")
    ptime1 <- proc.time()
    tests <- do.call(rbind, tests)
    tests[,2] <- log2(tests[,2])
    colnames(tests) <- c("p.value", "log2OR")
    rownames(tests) <- newkeys
    lookup <- rbind(tests, lookup)
    results <- lookup[keys,]
    rownames(results) <- NULL
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(round(stime,1), "done in secs\n")
    if(returnLookup)
        return(list(results = results, lookup = lookup))
    else
        return(results)
}

chrSelectBSseq <- function(BSseq, seqnames = NULL, order = FALSE) {
    seqlevels(BSseq, pruning.mode = "coarse") <- seqnames
    if (order) BSseq <- orderBSseq(BSseq, seqOrder = seqnames)
    BSseq
}

orderBSseq <- function(BSseq, seqOrder = NULL) {
    if (!is.null(seqOrder)) {
        seqlevels(BSseq, pruning.mode = "coarse") <- seqOrder
    }
    BSseq[order(granges(BSseq))]
}

# TODO: getMeth() realises the result in memory iff regions is not NULL;
#       discuss with Kasper
# TODO: Whether or not colnames are added to returned value depends on whether
#       regions is non-NULL; discuss with Kasper
# TODO: Add parallel support
# TODO: Document withDimnames
# TODO: Should is be an explicit withDimnames arg or simply passed via `...`?
getMeth <- function(BSseq, regions = NULL, type = c("smooth", "raw"),
                    what = c("perBase", "perRegion"), confint = FALSE,
                    alpha = 0.95, withDimnames = TRUE) {
    p.conf <- function(p, n, alpha) {
        z <- abs(qnorm((1 - alpha)/2, mean = 0, sd = 1))
        upper <- (p + z ^ 2 / (2 * n) +
                      z * sqrt((p * (1 - p) + z ^ 2 / (4 * n)) / n)) /
            (1 + z ^ 2 / n)
        lower <- (p + z ^ 2 / (2 * n) -
                      z * sqrt((p * (1 - p) + z ^ 2 / (4 * n)) / n)) /
            (1 + z ^ 2 / n)
        return(list(meth = p, lower = lower, upper = upper))
    }

    stopifnot(is(BSseq, "BSseq"))
    type <- match.arg(type)
    if (type == "smooth" & !hasBeenSmoothed(BSseq)) {
        stop("'type=smooth' requires the object to have been smoothed.")
    }
    what <- match.arg(what)
    if (what == "perRegion" & is.null(regions)) {
        stop("'what=perRegion' but no 'regions' supplied")
    }
    z <- abs(qnorm((1 - alpha)/2, mean = 0, sd = 1))
    if (is.null(regions) && type == "smooth") {
        coef <- getBSseq(BSseq, "coef", withDimnames)
        meth <- getBSseq(BSseq, "trans", withDimnames)(coef)
        if (confint) {
            upper <- meth + z * getBSseq(BSseq, "se.coef", withDimnames)
            lower <- meth - z * getBSseq(BSseq, "se.coef", withDimnames)
            return(list(meth = meth, lower = lower, upper = upper))
        } else {
            return(meth)
        }
    }
    if (is.null(regions) && type == "raw") {
        meth <- getBSseq(BSseq, "M", withDimnames) /
            getBSseq(BSseq, "Cov", withDimnames)
        if (confint) {
            return(p.conf(meth, getBSseq(BSseq, "Cov", withDimnames), alpha))
        } else {
            return(meth)
        }
    }

    ## At this point, regions have been specified
    if (is(regions, "data.frame")) {
        regions <- data.frame2GRanges(regions)
    }
    stopifnot(is(regions, "GenomicRanges"))
    if (confint) {
        stop("'confint = TRUE' is not supported by 'getMeth' when regions is given")
    }
    grBSseq <- granges(BSseq)
    ov <- findOverlaps(grBSseq, regions)
    # NOTE: This realises a large object in memory (`meth`) - could do it in
    #       chunks if what = perRegion
    if (type == "smooth") {
        meth <- as.matrix(
            getBSseq(BSseq, "trans", withDimnames)(
                getBSseq(BSseq, "coef", withDimnames))[
                    queryHits(ov), , drop = FALSE])
    } else if (type == "raw") {
        meth <- as.matrix(
            (getBSseq(BSseq, "M", withDimnames) /
                 getBSseq(BSseq, "Cov", withDimnames))[
                     queryHits(ov), , drop = FALSE])
    }
    out <- lapply(split(meth, subjectHits(ov)), matrix, ncol = ncol(meth))
    if (what == "perBase") {
        # TODO: Don't really understand the logic of the remaining code; how
        #       could the results end up in the wrong order wrt to regions?
        outList <- vector("list", length(regions))
        outList[as.integer(names(out))] <- out
        return(outList)
    } else if (what == "perRegion") {
        out <- do.call(rbind, lapply(out, colMeans2, na.rm = TRUE))
        # TODO: Don't really understand the logic of the remaining code; how
        #       could the rows end up in the wrong order?
        outMatrix <- matrix(NA, ncol = ncol(BSseq), nrow = length(regions))
        if (withDimnames) colnames(outMatrix) <- sampleNames(BSseq)
        outMatrix[as.integer(rownames(out)), ] <- out
        outMatrix
    }
}

# TODO: getCoverage() realises the result in memory iff regions is not NULL;
#       discuss with Kasper
# TODO: Whether or not colnames are added to returned value depends on whether
#       regions is non-NULL; discuss with Kasper
# TODO: Document withDimnames
# TODO: Should is be an explicit withDimnames arg or simply passed via `...`?
getCoverage <- function(BSseq, regions = NULL, type = c("Cov", "M"),
                        what = c("perBase", "perRegionAverage",
                                 "perRegionTotal"),
                        withDimnames = TRUE) {
    stopifnot(is(BSseq, "BSseq"))
    type <- match.arg(type)
    what <- match.arg(what)
    if (is.null(regions)) {
        if (what == "perBase") {
            return(getBSseq(BSseq, type, withDimnames))
        }
        if (what == "perRegionTotal") {
            return(colSums2(getBSseq(BSseq, type, withDimnames)))
        }
        if (what == "perRegionAverage") {
            return(colMeans2(getBSseq(BSseq, type, withDimnames)))
        }
    }
    if (is(regions,  "data.frame")) {
        regions <- data.frame2GRanges(regions)
    }
    stopifnot(is(regions, "GenomicRanges"))
    grBSseq <- granges(BSseq)
    ov <- findOverlaps(grBSseq, regions)
    coverage <- getBSseq(BSseq, type, withDimnames)[
        queryHits(ov), , drop = FALSE]
    out <- lapply(split(coverage, subjectHits(ov)), matrix,
                  ncol = ncol(coverage))

    if (what == "perBase") {
        # TODO: Don't really understand the logic of the remaining code; how
        #       could the results end up in the wrong order wrt to regions?
        outList <- vector("list", length(regions))
        outList[as.integer(names(out))] <- out
        return(outList)
    } else if (what == "perRegionAverage") {
        out <- do.call(rbind, lapply(out, colMeans2, na.rm = TRUE))
    } else if (what == "perRegionTotal") {
        out <- do.call(rbind, lapply(out, colSums2, na.rm = TRUE))
    }
    # TODO: Don't really understand the logic of the remaining code; how
    #       could the rows end up in the wrong order?
    outMatrix <- matrix(NA, ncol = ncol(BSseq), nrow = length(regions))
    if (withDimnames) colnames(outMatrix) <- sampleNames(BSseq)
    outMatrix[as.integer(rownames(out)), ] <- out
    outMatrix
}

context("Test fixes")

test_that("test_fixes", {
    gr1 <- GRanges(seqnames = "chr1", ranges = IRanges(start = 1:2, width = 1))
    gr2 <- GRanges(seqnames = "chr1", ranges = IRanges(start = 2:3, width = 1))
    ov <- findOverlaps(gr1, gr2)
    expect_equal(queryHits(ov), 2)
    expect_equal(subjectHits(ov), 1)
    expect_equal(subsetByOverlaps(gr1, gr2), gr1[2])
})

test_that("BSseq", {
    M <- matrix(1:9, 3, 3)
    colnames(M) <- c("A1", "A2", "A3")
    BStest <- BSseq(pos = 1:3, chr = rep("chr1", 3), M = M, Cov = M + 2)
    expect_equal(seqnames(BStest), Rle(factor(rep("chr1", 3))))
    expect_equal(start(BStest), c(1, 2, 3))
    expect_equal(end(BStest), c(1, 2, 3))
    expect_equal(width(BStest), rep(1, 3))
    expect_equal(strand(BStest),
                 Rle(factor(rep("*", 3), levels = c("+", "-", "*"))))
    expect_equal(dim(BStest), c(3, 3))
    expect_equal(nrow(BStest), 3)
    expect_equal(ncol(BStest), 3)
    expect_equal(unname(as.array(getCoverage(BStest, type = "M"))), unname(M))
    expect_equal(unname(as.array(getCoverage(BStest, type = "Cov"))),
                 unname(M + 2))
    expect_equal(sampleNames(BStest), colnames(M))

    BStest2 <- BSseq(pos = 3:1, chr = rep("chr1", 3), M = M[3:1, ],
                     Cov = M[3:1, ] + 2)
    # NOTE: BSseq() does not reorder inputs
    expect_true(!isTRUE(all.equal(BStest, BStest2)))
    expect_equal(BStest, sort(BStest2))
    M2 <- rbind(M[3:1, ], c(1, 1, 1))
    Cov2 <- rbind(M[3:1, ] + 2, c(1, 1, 1))
    M2[3, ] <- M2[3, ] - 1
    Cov2[3, ] <- Cov2[3, ] - 1
    BStest3 <- suppressWarnings(
        BSseq(pos = c(3:1,1), chr = rep("chr1", 4), M = M2, Cov = Cov2))
    # NOTE: BSseq() does not reorder inputs
    expect_true(!isTRUE(all.equal(BStest, BStest3)))
    expect_equal(BStest, sort(BStest3))
})

test_that("BSseq with HDF5Backend", {
    M <- matrix(1:9, 3, 3)
    colnames(M) <- c("A1", "A2", "A3")
    Cov <- M + 2L
    BStest <- BSseq(pos = 1:3, chr = rep("chr1", 3), M = M, Cov = Cov)
    hdf5_M <- realize(M, BACKEND = "HDF5Array")
    hdf5_Cov <- realize(Cov, BACKEND = "HDF5Array")
    hdf5_BStest <- BSseq(pos = 1:3, chr = rep("chr1", 3), M = hdf5_M,
                         Cov = hdf5_Cov)
    expect_identical(as.array(getCoverage(BStest)),
                     as.array(getCoverage(hdf5_BStest)))
    expect_identical(as.array(getCoverage(BStest, type = "M")),
                     as.array(getCoverage(hdf5_BStest, type = "M")))
})

test_overlaps <- function() {
    M <- matrix(1:9, 3, 3)
    colnames(M) <- c("A1", "A2", "A3")
    Cov <- M + 2
    gr1 <- GRanges(seqnames = "chr1", ranges = IRanges(start = 1:3, width = 1))
    gr2 <- GRanges(seqnames = "chr1", ranges = IRanges(start = 3:5, width = 1))
    BS1 <- BSseq(gr = gr1, M = M, Cov = Cov)
    BS2 <- BSseq(gr = gr2, M = M, Cov = Cov)
    BSres <- BSseq(gr = gr1[3],
                   M = M[3, ,drop = FALSE],
                   Cov = Cov[3, , drop = FALSE])
    expect_equal(length(BS1), 3L)
    expect_equal(seqlevels(BS1), seqlevels(gr1))
    expect_equal(seqlengths(BS1), seqlengths(gr1))
    expect_equal(seqnames(BS1), seqnames(gr1))
    expect_equal(granges(BS1), gr1)
    expect_equal(start(BS1), start(gr1))
    expect_equal(end(BS1), end(gr1))
    expect_equal(width(BS1), width(gr1))
    expect_equal(strand(BS1), strand(gr1))
    expect_equal(findOverlaps(BS1, BS2), findOverlaps(gr1, gr2))
    expect_equal(findOverlaps(gr1, BS2), findOverlaps(gr1, gr2))
    expect_equal(findOverlaps(BS1, gr2), findOverlaps(gr1, gr2))
    expect_equal(subsetByOverlaps(BS1, gr2), BSres)
    expect_equal(subsetByOverlaps(BS1, BS2), BSres)
    expect_equal(subsetByOverlaps(gr1, BS2), gr1[3])
}

context("combine()")

test_that("combine()", {
    bsseq_fit <- BSmooth(bsseq_test)
    BSSEQ_TEST <- realize(bsseq_test, "HDF5Array")
    BSSEQ_FIT <- realize(bsseq_fit, "HDF5Array")

    ai <- 1:100
    bi <- 51:150
    a <- bsseq_test[ai, 1]
    b <- bsseq_test[bi, 2]
    ab <- combine(a, b)
    A <- BSSEQ_TEST[ai, 1]
    B <- BSSEQ_TEST[bi, 2]
    AB <- combine(A, B)
    expect_equivalent_SE(ab, AB)
    aB <- combine(a, B)
    expect_equivalent_SE(ab, aB)
    Ab <- combine(A, b)
    expect_equivalent_SE(ab, Ab)

    z <- combine(bsseq_fit[, 1], bsseq_fit[, 2])
    expect_equivalent_SE(z, bsseq_fit)
    Z <- combine(BSSEQ_FIT[, 1], BSSEQ_FIT[, 2])
    expect_equivalent_SE(Z, BSSEQ_FIT)
})

test_that("combineList()", {
    bsseq_fit <- BSmooth(bsseq_test)
    BSSEQ_TEST <- realize(x = bsseq_test, "HDF5Array")
    BSSEQ_FIT <- realize(x = bsseq_fit, "HDF5Array")

    ai <- 1:100
    bi <- 51:150
    ci <- 201:300
    a <- bsseq_test[ai, 1]
    b <- bsseq_test[bi, 2]
    c <- bsseq_test[ci, 1]
    colnames(c) <- "r3"
    abc <- combineList(list(a, b, c))
    A <- BSSEQ_TEST[ai, 1]
    B <- BSSEQ_TEST[bi, 2]
    C <- BSSEQ_TEST[ci, 1]
    colnames(C) <- "r3"
    ABC <- combineList(list(A, B, C))
    expect_equivalent_SE(abc, ABC)
    expect_equivalent_SE(combineList(list(a, b, c)),
                         combineList(a, b, c))
    expect_equivalent_SE(combineList(list(a, B, c)),
                         combineList(A, b, C))

    expect_equivalent_SE(combine(a, b),
                         combineList(a, b))
    expect_equivalent_SE(combine(A, B),
                         combineList(A, B))

    z <- combineList(bsseq_fit[, 1], bsseq_fit[, 2])
    expect_equivalent_SE(z, bsseq_fit)
    Z <- combineList(BSSEQ_FIT[, 1], BSSEQ_FIT[, 2])
    expect_equivalent_SE(Z, BSSEQ_FIT)
})

test_that(
    "Test bug fix reported in https://github.com/hansenlab/bsseq/pull/54/", {
        M <- matrix(0:8, 3, 3)
        Cov <- matrix(1:9, 3, 3)

        BS1 <- BSseq(chr = c("chr1", "chr2", "chr1"),
                     pos = c(1, 2, 3),
                     M = M,
                     Cov = Cov,
                     sampleNames = c("A","B", "C"))
        BS2 <- BSseq(chr = c("chr2", "chr1", "chr2"),
                     pos = c(1, 2, 3),
                     M = M,
                     Cov = Cov, sampleNames = c("D", "E", "F"))
        # NOTE: Using combineList()
        combined <- combineList(list(BS1, BS2))
        over <- findOverlaps(BS2, combined, type = "equal")
        methAlone <- getMeth( BS2[queryHits(over), c("D", "E", "F")], type = "raw")
        methCombined <- getMeth(combined[subjectHits(over), c("D", "E", "F")],
                                type = "raw")
        expect_equal(as.matrix(methAlone), as.matrix(methCombined))
        # NOTE: Using combine()
        combined2 <- combine(BS1, BS2)
        over2 <- findOverlaps(BS2, combined2, type = "equal")
        methAlone2 <- getMeth( BS2[queryHits(over2), c("D", "E", "F")],
                               type = "raw")
        methCombined2 <- getMeth(combined2[subjectHits(over2), c("D", "E", "F")],
                                 type = "raw")
        expect_equal(as.matrix(methAlone2), as.matrix(methCombined2))

        # Extended version with 3 BSseq objects
        BS3 <- BSseq(chr = c("chr2", "chr1", "chr2"),
                     pos = c(3, 2, 1),
                     M = M,
                     Cov = Cov,
                     sampleNames = c("G", "H", "I")
        )
        # NOTE: Using combineList()
        combined <- combineList(BS2, BS1, BS3)
        over <- findOverlaps(BS3, combined, type = "equal")
        methAlone <- getMeth(BS3[queryHits(over), c("G", "H", "I")], type = "raw")
        methCombined <- getMeth(combined[subjectHits(over), c("G", "H", "I")],
                                type = "raw")
        expect_equal(as.matrix(methAlone), as.matrix(methCombined))
        # NOTE: Using combine()
        combined2 <- combine(BS2, BS1, BS3)
        over2 <- findOverlaps(BS3, combined2, type = "equal")
        methAlone2 <- getMeth(BS3[queryHits(over2), c("G", "H", "I")], type = "raw")
        methCombined2 <- getMeth(combined2[subjectHits(over2), c("G", "H", "I")],
                                 type = "raw")
        expect_equal(as.matrix(methAlone), as.matrix(methCombined))
    }
)

plotAnnoTrack <- function(gr, annoTrack, cex) {
    ## check may need to be modified
    if(!all(sapply(annoTrack, function(xx) is(xx, "GRanges"))))
        stop("all elements in 'annoTrack' needs to be 'GRanges'")
    plot(start(gr), 1, type = "n", xaxt = "n", yaxt = "n", bty = "n",
         ylim = c(0.5, length(annoTrack) + 0.5), xlim = c(start(gr), end(gr)), xlab = "", ylab = "")
    lapply(seq(along = annoTrack), function(ii) {
        jj <- length(annoTrack) + 1- ii
        ir <- subsetByOverlaps(annoTrack[[ii]], gr)
        if(length(ir) > 0)
            rect(start(ir)-0.5, jj - 0.15, end(ir), jj + 0.15, col = "grey60", border = NA)
        mtext(names(annoTrack)[ii], side = 2, at = jj, las = 1, line = 1,
              cex = cex, adj = 0.3)
        })
}

# Based on plotAnnoTrack() and
# https://jhu-genomics.slack.com/archives/hansen_gtex/p1461760583000142
# TODO: Use standard Bioconductor object for storing gene information rather
#       than this ad hoc data.frame (e.g., based on TxDb.* packages)
plotGeneTrack <- function(gr, geneTrack, cex) {
    geneTrack_gr <- makeGRangesFromDataFrame(geneTrack)
    ol <- findOverlaps(geneTrack_gr, gr)
    genes <- geneTrack[queryHits(ol), ]
    plot(start(gr), 1, type = "n", xaxt = "n", yaxt = "n", bty = "n",
         ylim = c(-1.5, 1.5), xlim = c(start(gr), end(gr)),
         xlab = "", ylab = "", cex.lab = 4, lheight = 2, cex.axis = 1)
    if (nrow(genes) > 0) {
        for (g in 1:nrow(genes)) {
            geneind2 = which(geneTrack$gene_name == genes$gene_name[g])
            geneind2 = geneind2[which(geneTrack$isoforms[geneind2] == 1)]
            direction = unique(geneTrack$strand[geneind2])
            ES = geneTrack$start[geneind2]
            EE = geneTrack$end[geneind2]
            Exons = cbind(ES, EE)
            if (direction == "+") {
                lines(x = c(min(ES), max(EE)),
                      y = c(0.65, 0.65))
                apply(Exons, 1, function(x) {
                    polygon(c(x[1], x[2], x[2], x[1]),
                            c(0.45, 0.45, 0.85, 0.85), col = "darkgrey")
                    })
                text((max(start(gr), min(ES)) +
                          min(end(gr), max(EE))) / 2, 1.2,
                     genes$gene_name[g], cex = cex)
            } else {
                lines(x = c(min(ES), max(EE)),
                      y = c(-0.65, -0.65))
                apply(Exons, 1, function(x)
                    polygon(c(x[1], x[2], x[2], x[1]),
                            c(-0.45, -0.45, -0.85, -0.85),
                            col = "darkgrey"))
                text((max(start(gr), min(ES)
                ) + min(end(gr), max(EE)
                )) / 2, -1.2, genes$gene_name[g], cex = cex)
            }

        }
    }
}

plotManyRegions <- function(BSseq, regions = NULL, extend = 0, main = "", addRegions = NULL,
                            annoTrack = NULL, cex.anno = 1,
                            geneTrack = NULL, cex.gene = 1.5,
                            col = NULL, lty = NULL, lwd = NULL,
                            BSseqStat = NULL, stat = "tstat.corrected", stat.col = "black",
                            stat.lwd = 1, stat.lty = 1, stat.ylim = c(-8,8),
                            mainWithWidth = TRUE, regionCol = alpha("red", 0.1), addTicks = TRUE,
                            addPoints = FALSE, pointsMinCov = 5, highlightMain = FALSE, verbose = TRUE) {
    cat("[plotManyRegions] preprocessing ...")
    if(!is.null(regions)) {
        if(is(regions, "data.frame"))
            gr <- data.frame2GRanges(regions, keepColumns = FALSE)
        else
            gr <- regions
        if(!is(gr, "GRanges"))
            stop("'regions' needs to be either a 'data.frame' (with a single row) or a 'GRanges' (with a single element)")
    } else {
        gr <- granges(BSseq)
    }
    gr <- resize(gr, width = 2*extend + width(gr), fix = "center")
    BSseq <- subsetByOverlaps(BSseq, gr)
    if(!is.null(BSseqStat))
        BSseqStat <- subsetByOverlaps(BSseqStat, gr)

    if(length(start(BSseq)) == 0)
        stop("No overlap between BSseq data and regions")
    if(!is.null(main) && length(main) != length(gr))
        main <- rep(main, length = length(gr))
    cat("done\n")
    for(ii in seq(along = gr)) {
        if(verbose) cat(sprintf("[plotManyRegions]   plotting region %d (out of %d)\n", ii, nrow(regions)))
        plotRegion(BSseq = BSseq, region = regions[ii,], extend = extend,
                   col = col, lty = lty, lwd = lwd, main = main[ii], BSseqStat = BSseqStat,
                   stat = stat, stat.col = stat.col, stat.lwd = stat.lwd,
                   stat.lty = stat.lty, stat.ylim = stat.ylim,
                   addRegions = addRegions, regionCol = regionCol, mainWithWidth = mainWithWidth,
                   annoTrack = annoTrack, cex.anno = cex.anno,
                   geneTrack = geneTrack, cex.gene = cex.gene,
                   addTicks = addTicks, addPoints = addPoints,
                   pointsMinCov = pointsMinCov, highlightMain = highlightMain)
    }
}

.bsPlotLines <- function(x, y, col, lty, lwd, plotRange) {
    if(sum(!is.na(y)) <= 1)
        return(NULL)
    xx <- seq(from = plotRange[1], to = plotRange[2], length.out = 500)
    yy <- approxfun(x, y)(xx)
    lines(xx, yy, col = col, lty = lty, lwd = lwd)
}

.bsPlotPoints <- function(x, y, z, col, pointsMinCov) {
    points(x[z>pointsMinCov], y[z>pointsMinCov], col = col, pch = 16, cex = 0.5)
}

.bsHighlightRegions <- function(regions, gr, ylim, regionCol, highlightMain) {
    if(is.data.frame(regions))
        regions <- data.frame2GRanges(regions)
    if(highlightMain)
        regions <- c(regions, gr)
    if(is.null(regions)) return(NULL)
    ## regions <- pintersect(region, rep(gr, length(regions)))
    ## regions <- regions[width(regions) == 0]
    regions <- subsetByOverlaps(regions, gr)
    regions <- pintersect(regions, rep(gr, length(regions)))
    if(length(regions) == 0)
        return(NULL)
    rect(xleft = start(regions), xright = end(regions), ybottom = ylim[1],
         ytop = ylim[2], col = regionCol, border = NA)
}

.bsGetCol <- function(object, col, lty, lwd) {
    ## Assumes that object has pData and sampleNames methods
    if(is.null(col)) {
        if("col" %in% names(pData(object)))
            col <- pData(object)[["col"]]
        else
            col <- rep("black", nrow(pData(object)))
    }
    if(length(col) != ncol(object))
        col <- rep(col, length.out = ncol(object))
    if(is.null(names(col)))
        names(col) <- sampleNames(object)

    if(is.null(lty)) {
        if("lty" %in% names(pData(object)))
            lty <- pData(object)[["lty"]]
        else
            lty <- rep(1, ncol(object))
    }
    if(length(lty) != ncol(object))
        lty <- rep(lty, length.out = ncol(object))
    if(is.null(names(lty)))
        names(lty) <- sampleNames(object)

    if(is.null(lwd)) {
        if("lwd" %in% names(pData(object)))
            lwd <- pData(object)[["lwd"]]
        else
            lwd <- rep(1, nrow(pData(object)))
    }
    if(length(lwd) != ncol(object))
        lwd <- rep(lwd, length.out = ncol(object))
    if(is.null(names(lwd)))
        names(lwd) <- sampleNames(object)

    return(list(col = col, lty = lty, lwd = lwd))
}

.bsPlotTitle <- function(gr, extend, main, mainWithWidth) {
    if(is.data.frame(gr))
        gr <- data.frame2GRanges(gr)
    if(length(gr) > 1) {
        warning("plotTitle: gr has more than one element")
        gr <- gr[1]
    }
    plotChr <- as.character(seqnames(gr))
    plotRange <- c(start(gr), end(gr))
    regionCoord <- sprintf("%s: %s - %s", plotChr,
                           format(plotRange[1], big.mark = ",", scientific = FALSE),
                           format(plotRange[2], big.mark = ",", scientific = FALSE))
    if(mainWithWidth) {
        regionWidth <- sprintf("width = %s, extended = %s",
                               format(width(gr), big.mark = ",", scientific = FALSE),
                               format(extend, big.mark = ",", scientific = FALSE))
        regionCoord <- sprintf("%s (%s)", regionCoord, regionWidth)
    }
    if(main != "") {
        main <- sprintf("%s\n%s", main, regionCoord)
    } else {
        main <- regionCoord
    }
    main
}

.bsGetGr <- function(object, region, extend) {
    if(is.null(region)) {
        gr <- GRanges(seqnames = seqnames(object)[1],
                      ranges = IRanges(start = min(start(object)),
                      end = max(start(object))))
    } else {
        if(is(region, "data.frame"))
            gr <- data.frame2GRanges(region, keepColumns = FALSE)
        else
            gr <- region
        if(!is(gr, "GRanges") || length(gr) != 1)
            stop("'region' needs to be either a 'data.frame' (with a single row) or a 'GRanges' (with a single element)")
        gr <- resize(gr, width = 2*extend + width(gr), fix = "center")
    }
    gr
}


.plotSmoothData <- function(BSseq, region, extend, addRegions, col, lty, lwd, regionCol,
                            addTicks, addPoints, pointsMinCov, highlightMain) {
    gr <- .bsGetGr(BSseq, region, extend)
    BSseq <- subsetByOverlaps(BSseq, gr)

    ## Extract basic information
    sampleNames <- sampleNames(BSseq)
    names(sampleNames) <- sampleNames
    positions <- start(BSseq)
    smoothPs <- getMeth(BSseq, type = "smooth")
    rawPs <- getMeth(BSseq, type = "raw")
    coverage <- getCoverage(BSseq)

    # Realise in memory data that are to be plotted
    if (addPoints) {
        rawPs <- as.array(rawPs)
        coverage <- as.array(coverage)
    }
    smoothPs <- as.array(smoothPs)

    ## get col, lwd, lty
    colEtc <- .bsGetCol(object = BSseq, col = col, lty = lty, lwd = lwd)

    ## The actual plotting
    plot(positions[1], 0.5, type = "n", xaxt = "n", yaxt = "n",
         ylim = c(0,1), xlim = c(start(gr), end(gr)), xlab = "", ylab = "Methylation")
    axis(side = 2, at = c(0.2, 0.5, 0.8))
    if(addTicks)
        rug(positions)

    .bsHighlightRegions(regions = addRegions, gr = gr, ylim = c(0,1),
                        regionCol = regionCol, highlightMain = highlightMain)

    if(addPoints) {
        sapply(1:ncol(BSseq), function(sampIdx) {
            abline(v = positions[rawPs[, sampIdx] > 0.1], col = "grey80", lty = 1)
        })
    } # This adds vertical grey lines so we can see where points are plotted

    sapply(1:ncol(BSseq), function(sampIdx) {
        .bsPlotLines(positions, smoothPs[, sampIdx], col = colEtc$col[sampIdx],
                     lty = colEtc$lty[sampIdx], lwd = colEtc$lwd[sampIdx],
                     plotRange = c(start(gr), end(gr)))
    })

    if(addPoints) {
        sapply(1:ncol(BSseq), function(sampIdx) {
            .bsPlotPoints(positions, rawPs[, sampIdx], coverage[, sampIdx],
                          col = colEtc$col[sampIdx], pointsMinCov = pointsMinCov)
        })
    }
}


plotRegion <- function(BSseq, region = NULL, extend = 0, main = "", addRegions = NULL,
                       annoTrack = NULL, cex.anno = 1,
                       geneTrack = NULL, cex.gene = 1.5,
                       col = NULL, lty = NULL, lwd = NULL,
                       BSseqStat = NULL, stat = "tstat.corrected", stat.col = "black",
                       stat.lwd = 1, stat.lty = 1, stat.ylim = c(-8,8),
                       mainWithWidth = TRUE,
                       regionCol = alpha("red", 0.1), addTicks = TRUE, addPoints = FALSE,
                       pointsMinCov = 5, highlightMain = FALSE) {

    opar <- par(mar = c(0,4.1,0,0), oma = c(5,0,4,2), mfrow = c(1,1))
    on.exit(par(opar))
    if(is.null(BSseqStat) && is.null(geneTrack)) {
        layout(matrix(1:2, ncol = 1), heights = c(2,1))
    } else if (is.null(geneTrack)) {
        layout(matrix(1:3, ncol = 1), heights = c(2,2,1))
    } else {
        layout(matrix(1:4, ncol = 1), heights = c(2,2,1,0.3))
    }

    .plotSmoothData(BSseq = BSseq, region = region, extend = extend, addRegions = addRegions,
                    col = col, lty = lty, lwd = lwd, regionCol = regionCol,
                    addTicks = addTicks, addPoints = addPoints,
                    pointsMinCov = pointsMinCov, highlightMain = highlightMain)
    gr <- .bsGetGr(BSseq, region, extend)

    if(!is.null(BSseqStat)) {
        BSseqStat <- subsetByOverlaps(BSseqStat, gr)
        if(is(BSseqStat, "BSseqTstat")) {
            stat.values <- as.array(getStats(BSseqStat)[, "tstat.corrected"])
            stat.values <- as.array(stat.values)
            stat.type <- "tstat"
        }
        if(is(BSseqStat, "BSseqStat")) {
            stat.type <- getStats(BSseqStat, what = "stat.type")
            if(stat.type == "tstat") {
                stat.values <- getStats(BSseqStat, what = "stat")
                stat.values <- as.array(stat.values)
            }
            if(stat.type == "fstat") {
                stat.values <- sqrt(getStats(BSseqStat, what = "stat"))
                stat.values <- as.array(stat.values)
            }
        }
        plot(start(gr), 0.5, type = "n", xaxt = "n", yaxt = "n",
             ylim = stat.ylim, xlim = c(start(gr), end(gr)), xlab = "", ylab = stat.type)
        axis(side = 2, at = c(-5,0,5))
        abline(h = 0, col = "grey60")
        .bsPlotLines(start(BSseqStat), stat.values, lty = stat.lty, col = stat.col, lwd = stat.lwd,
                     plotRange = c(start(gr), end(gr)))
    }

    if(!is.null(annoTrack))
        plotAnnoTrack(gr, annoTrack, cex.anno)

    if (!is.null(geneTrack))
        plotGeneTrack(gr, geneTrack, cex.gene)

    if(!is.null(main)) {
        main <- .bsPlotTitle(gr = region, extend = extend, main = main,
                             mainWithWidth = mainWithWidth)
        mtext(side = 3, text = main, outer = TRUE, cex = 1)
    }
    return(invisible(NULL))
}

##     plotP <- function(sample, label = "", col) {
##         plot(positions[1], 0.5, type = "n", xaxt = "n", yaxt = "n",
##              ylim = c(0,1), xlim = plotRange, xlab = "", ylab = "")
##         plotRects(c(0,1))
##         rawp <- getP(BSseq, sample = sample, type = "raw", addPositions = TRUE, addConfint = TRUE)
##         cols <- rep(alpha("black", 0.5), nrow(rawp))
##         segments(x0 = rawp$pos, y0 = rawp$lower, y1 = rawp$upper, col = cols)
##         points(positions, rawp$p, col = cols)
##         if(nrow(BSseq$coef) > 0) {
##             fitp <- getP(BSseq, sample = sample, type = "fit", addPosition = TRUE, addConfint = FALSE)
##             lines(fitp$pos, fitp$p, col = col,, lty = 1)
##             ## lines(fitp$pos, fitp$lower, col = col, lty = 2)
##             ## lines(fitp$pos, fitp$upper, col = col, lty = 2)
##             text(plotRange[1], 0.1, labels = label)
##         }
##     }

##     plotD <- function(sample1, sample2, col) {
##         plot(positions[1], 0.5, type = "n", xaxt = "n", yaxt = "n",
##              ylim = c(-1,1), xlim = plotRange, xlab = "", ylab = "")
##         plotRects(c(-1,1))
##         abline(h = 0, col = alpha("black", 0.6), lty = 2)
##         rawd <- getD(BSseq, sample1 = sample1, sample2 = sample2, type = "raw",
##                      addPositions = TRUE)
##         cols <- rep(alpha("black", 0.5), nrow(rawd))
##         segments(x0 = rawd$pos, rawd$lower, y1 = rawd$upper, col = cols)
##         points(rawd$pos, rawd$d, col = cols)
##         if(nrow(BSseq$coef) > 0) {
##             fitd <- getD(BSseq, sample1 = sample1, sample2 = sample2, type = "fit", addPositions = TRUE)
##             lines(fitd$pos, fitd$d, col = col, lty = 1)
##             ## lines(fitd$pos, fitd$lower, col = "blue", lty = 2)
##             ## lines(fitd$pos, fitd$upper, col = "blue", lty = 2)
##         }
##     }

## ----dependencies, warning=FALSE, message=FALSE-------------------------------
library(bsseq)
library(bsseqData)

## ----showData-----------------------------------------------------------------
data(BS.cancer.ex)
BS.cancer.ex <- updateObject(BS.cancer.ex)
BS.cancer.ex
pData(BS.cancer.ex)

## ----smooth,eval=FALSE--------------------------------------------------------
# BS.cancer.ex.fit <- BSmooth(
#     BSseq = BS.cancer.ex,
#     BPPARAM = MulticoreParam(workers = 1),
#     verbose = TRUE)

## ----showDataFit--------------------------------------------------------------
data(BS.cancer.ex.fit)
BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
BS.cancer.ex.fit

## ----cpgNumbers---------------------------------------------------------------
## The average coverage of CpGs on the two chromosomes
round(colMeans(getCoverage(BS.cancer.ex)), 1)
## Number of CpGs in two chromosomes
length(BS.cancer.ex)
## Number of CpGs which are covered by at least 1 read in all 6 samples
sum(rowSums(getCoverage(BS.cancer.ex) >= 1) == 6)
## Number of CpGs with 0 coverage in all samples
sum(rowSums(getCoverage(BS.cancer.ex)) == 0)

## ----poisson------------------------------------------------------------------
logp <- ppois(0, lambda = 4, lower.tail = FALSE, log.p = TRUE)
round(1 - exp(6 * logp), 3)

## ----smoothSplit,eval=FALSE---------------------------------------------------
# ## Split datag
# BS1 <- BS.cancer.ex[, 1]
# save(BS1, file = "BS1.rda")
# BS2 <- BS.cancer.ex[, 2]
# save(BS1, file = "BS1.rda")
# ## done splitting
# 
# ## Do the following on each node
# 
# ## node 1
# load("BS1.rda")
# BS1.fit <- BSmooth(BS1)
# save(BS1.fit)
# save(BS1.fit, file = "BS1.fit.rda")
# ## done node 1
# 
# ## node 2
# load("BS2.rda")
# BS2.fit <- BSmooth(BS2)
# save(BS2.fit, file = "BS2.fit.rda")
# ## done node 2
# 
# ## join; in a new R session
# load("BS1.fit.rda")
# load("BS2.fit.rda")
# BS.fit <- combine(BS1.fit, BS2.fit)

## ----keepLoci-----------------------------------------------------------------
BS.cov <- getCoverage(BS.cancer.ex.fit)
keepLoci.ex <- which(rowSums(BS.cov[, BS.cancer.ex$Type == "cancer"] >= 2) >= 2 &
                     rowSums(BS.cov[, BS.cancer.ex$Type == "normal"] >= 2) >= 2)
length(keepLoci.ex)
BS.cancer.ex.fit <- BS.cancer.ex.fit[keepLoci.ex,]

## ----BSmooth.tstat------------------------------------------------------------
BS.cancer.ex.tstat <- BSmooth.tstat(BS.cancer.ex.fit, 
                                    group1 = c("C1", "C2", "C3"),
                                    group2 = c("N1", "N2", "N3"), 
                                    estimate.var = "group2",
                                    local.correct = TRUE,
                                    verbose = TRUE)
BS.cancer.ex.tstat

## ----plotTstat,fig.width=4,fig.height=4---------------------------------------
plot(BS.cancer.ex.tstat)

## ----dmrs---------------------------------------------------------------------
dmrs0 <- dmrFinder(BS.cancer.ex.tstat, cutoff = c(-4.6, 4.6))
dmrs <- subset(dmrs0, n >= 3 & abs(meanDiff) >= 0.1)
nrow(dmrs)
head(dmrs, n = 3)

## ----plotSetup----------------------------------------------------------------
pData <- pData(BS.cancer.ex.fit)
pData$col <- rep(c("red", "blue"), each = 3)
pData(BS.cancer.ex.fit) <- pData

## ----plotDmr,fig.width=8,fig.height=4-----------------------------------------
plotRegion(BS.cancer.ex.fit, dmrs[1,], extend = 5000, addRegions = dmrs)

## ----plotManyRegions,eval=FALSE-----------------------------------------------
# pdf(file = "dmrs_top200.pdf", width = 10, height = 5)
# plotManyRegions(BS.cancer.ex.fit, dmrs[1:200,], extend = 5000,
#                 addRegions = dmrs)
# dev.off()

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

# Internal functions -----------------------------------------------------------

# TODO: Test against some of Bismark's other output files. May need a stricter
#       and more accurate heuristic to avoid 'passing' bad files.
# TODO: Test for 'bismark_methylation_extractor' and 'bedGraph' formats
#       (https://github.com/FelixKrueger/Bismark/tree/master/Docs#output-1)
#       and error out (they're not supported by .readBismaskAsDT()).
.guessBismarkFileType <- function(files, n_max = 10L) {
    vapply(files, function(file) {
        x <- read.delim(
            file = file,
            header = FALSE,
            nrows = n_max,
            stringsAsFactors = FALSE)
        n_fields <- ncol(x)
        if (isTRUE(all(n_fields == 6L))) {
            return("cov")
        }
        if (isTRUE(all(n_fields == 7L))) {
            return("cytosineReport")
        }
        # Couldn't guess the file type.
        NA_character_
    }, character(1L))
}


# TODO: Document that the default 'sort = TRUE' applies sort(sortSeqlevels())
#       to the output. This is the default behaviour because it results in
#       the smallest returned object (albeit at the small cost of a sort).
.readBismarkAsFWGRanges <- function(file, rmZeroCov = FALSE,
                                    strandCollapse = FALSE, sort = TRUE,
                                    nThread = 1L, verbose = FALSE) {
    # Argument checks ----------------------------------------------------------

    stopifnot(isTRUEorFALSE(rmZeroCov))
    stopifnot(isTRUEorFALSE(strandCollapse))
    stopifnot(isTRUEorFALSE(sort))

    # Quieten R CMD check about 'no visible binding for global variable'
    M <- U <- NULL

    # Read file to construct data.table of valid loci --------------------------
    if (rmZeroCov) {
        dt <- .readBismarkAsDT(
            file = file,
            col_spec = "BSseq",
            check = TRUE,
            verbose = verbose)
        if (strandCollapse && !is.null(dt[["strand"]]) &&
            !dt[, all(strand == "*")]) {
            # Shift loci on negative strand by 1 to the left and then remove
            # strand since no longer valid.
            dt[strand == "-", start := start - 1L][, strand := NULL]
            # Aggregate counts at loci with the same 'seqnames' and 'start'.
            dt <- dt[,
                     list(M = sum(M), U = sum(U)), by = c("seqnames", "start")]
        }
        # Identify loci with non-zero coverage then drop 'M' and 'U' as no
        # longer required.
        dt <- dt[(M + U) > 0][, c("M", "U") := list(NULL, NULL)]
    } else {
        dt <- .readBismarkAsDT(
            file = file,
            col_spec = "GRanges",
            check = FALSE,
            nThread = nThread,
            verbose = verbose)
        if (strandCollapse && !is.null(dt[["strand"]]) &&
            !dt[, all(strand == "*")]) {
            # Shift loci on negative strand by 1 to the left and then remove
            # strand since no longer valid.
            dt[strand == "-", start := start - 1L][, strand := NULL]
            dt <- data.table:::funique(dt)
        }
    }

    # Construct FWGRanges from 'dt' --------------------------------------------

    # NOTE: Sorting results in a smaller FWGRanges object because the
    #       'seqnames' and 'strand' slots are more compressible in their Rle
    #       representation.
    if (sort) {
        if (is.null(dt[["strand"]])) {
            setkey(dt, seqnames, start)
        } else {
            setkey(dt, seqnames, strand, start)
        }
    }
    seqnames <- Rle(dt[["seqnames"]])
    dt[, seqnames := NULL]
    seqinfo <- Seqinfo(seqnames = levels(seqnames))
    ranges <- .FWIRanges(start = dt[["start"]], width = 1L)
    dt[, start := NULL]
    mcols <- make_zero_col_DFrame(length(ranges))
    if (is.null(dt[["strand"]])) {
        strand <- strand(Rle("*", length(seqnames)))
    } else {
        strand <- Rle(dt[["strand"]])
        dt[, strand := NULL]
    }
    fwgranges <- .FWGRanges(
        seqnames = seqnames,
        ranges = ranges,
        strand = strand,
        seqinfo = seqinfo,
        elementMetadata = mcols)
    # NOTE: Final sort is to re-order with respect to sorted seqlevels.
    if (sort) {
        fwgranges <- sort(sortSeqlevels(fwgranges))
    }
    fwgranges
}

# TODO: Document that this applies sort(sortSeqlevels()) to the output. It is
#       deliberate that there is no option to override this behaviour.
.constructFWGRangesFromBismarkFiles <- function(files,
                                                rmZeroCov,
                                                strandCollapse,
                                                verbose,
                                                nThread,
                                                BPPARAM) {
    subverbose <- max(as.integer(verbose) - 1L, 0L)

    # TODO: Instead of using the 'largest' file, use the largest
    #       'cytosine report' file, which will have all loci in the
    #       reference genome; provided all samples were aligned to the same
    #       reference genome, this means it contains all loci.
    # TODO: Initialise using the 'largest' file (i.e. largest number of lines)?
    #       Would like to do this without reading the data into memory.
    #       Some benchmarks can be found at
    #       https://gist.github.com/peterhurford/0d62f49fd43b6cf078168c043412f70a
    #       My initial tests using /users/phickey/GTExScripts/FlowSortingProject/hdf5/extdata/methylation/nonCG/5248_BA9_neg_CHG_report.txt (32 GB) give:
    #       wc -l:                                       77.000s
    #       R.utils::readLines():                      1165.299s
    #       nrow(fread(..., select = 1, nThread = 1)):  582.721s (359s re-run)
    #       nrow(fread(..., select = 1, nThread = 10)):  82.029s
    #       nrow(fread(..., select = 1, nThread = 40)):  81.408s
    #       file.size():                                  0.000s
    #       Of course, fread() only works directly with non-[b]gzipped files.
    #       And subsequent runs of fread() benefit from some cacheing effect
    #       that I don't fully understand except to know that subsequent runs
    #       are 'artificially' faster.
    #       And using file.size() will be innaccurate if files are a mix of
    #       compressed and uncompressed files.
    # Initalise `loci_dt` using the first file.
    if (verbose) {
        message("[.constructFWGRangesFromBismarkFiles] Extracting loci from ",
                "'", files[1L], "'")
    }
    loci_from_first_file <- .readBismarkAsFWGRanges(
        file = files[[1L]],
        rmZeroCov = rmZeroCov,
        strandCollapse = strandCollapse,
        nThread = nThread,
        verbose = subverbose)
    # Identify loci not found in first file.
    # TODO: Pre-process loci as a GNCList?
    # Set number of tasks to ensure the progress bar gives frequent updates.
    # NOTE: The progress bar increments once per task
    #       (https://github.com/Bioconductor/BiocParallel/issues/54).
    #       Although it is somewhat of a bad idea to overrides a user-specified
    #       bptasks(BPPARAM), the value of bptasks(BPPARAM) doesn't affect
    #       performance in this instance, and so we opt for a useful progress
    #       bar. Only SnowParam (and MulticoreParam by inheritance) have a
    #       bptasks<-() method.
    if (is(BPPARAM, "SnowParam") && bpprogressbar(BPPARAM)) {
        bptasks(BPPARAM) <- length(files) - 1L
    }
    list_of_loci_from_other_files_not_in_first_file <- bplapply(
        files[-1L], function(file, loci_from_first_file) {
            # Read this file.
            loci_from_this_file <- .readBismarkAsFWGRanges(
                file = file,
                rmZeroCov = rmZeroCov,
                strandCollapse = strandCollapse,
                verbose = subverbose)
            subsetByOverlaps(
                x = loci_from_this_file,
                ranges = loci_from_first_file,
                type = "equal",
                invert = TRUE)
        }, loci_from_first_file = loci_from_first_file,
        BPPARAM = BPPARAM)
    # Identify unique FWGRanges.
    loci_non_found_in_first_file <- unique(
        do.call(c, list_of_loci_from_other_files_not_in_first_file))
    loci <- c(loci_from_first_file, loci_non_found_in_first_file)

    # Sort the loci
    sort(sortSeqlevels(loci))
}


# NOTE: In brief benchmarking, readr::read_csv() is ~1.3-1.6x faster than
#       utils::read.delim() when reading a gzipped file, albeit it with ~1.6-2x
#       more total memory allocated. Therefore, there may be times users prefer
#       to trade off faster speed for lower memory usage.
# TODO: (long term) Formalise benchmarks of utils::read.table(),
#       data.table::fread(), and readr::read_tsv() as a document in the bsseq
#       package so that we can readily re-visit these as needed.
.readBismarkAsDT <- function(file,
                             col_spec = c("all", "BSseq", "GRanges"),
                             check = FALSE,
                             showProgress = FALSE,
                             nThread = 1L,
                             verbose = FALSE) {
    # Argument checks ----------------------------------------------------------

    file <- file_path_as_absolute(file)
    file_type <- .guessBismarkFileType(file)
    col_spec <- match.arg(col_spec)
    stopifnot(isTRUEorFALSE(check))
    stopifnot(isTRUEorFALSE(showProgress))
    # NOTE: 'verbose' controls the verbosity of .readBismarkAsDT() and
    #       'fread_verbose' is derived from that to control the verbosity of
    #       data.table::fread().
    fread_verbose <- as.logical(max(verbose - 1L, 0L))

    # Construct the column spec ------------------------------------------------

    if (file_type == "cov") {
        header <- FALSE
        if (col_spec == "BSseq") {
            colClasses <- c("factor", "integer", "NULL", "NULL", "integer",
                            "integer")
            drop <- c(3L, 4L)
            col.names <- c("seqnames", "start", "M", "U")
        } else if (col_spec == "GRanges") {
            colClasses <- c("factor", "integer", "NULL", "NULL", "NULL", "NULL")
            drop <- c(3L, 4L, 5L, 6L)
            col.names <- c("seqnames", "start")
        } else if (col_spec == "all") {
            colClasses <- c("factor", "integer", "integer", "numeric",
                            "integer", "integer")
            drop <- integer(0L)
            col.names <- c("seqnames", "start", "end", "beta", "M", "U")
        }
    } else if (file_type == "cytosineReport") {
        header <- FALSE
        if (col_spec == "BSseq") {
            colClasses <- c("factor", "integer", "factor", "integer",
                            "integer", "NULL", "NULL")
            drop <- c(6L, 7L)
            col.names <- c("seqnames", "start", "strand", "M", "U")
        } else if (col_spec == "GRanges") {
            colClasses <- c("factor", "integer", "factor", "NULL", "NULL",
                            "NULL", "NULL")
            drop <- c(4L, 5L, 6L, 7L)
            col.names <- c("seqnames", "start", "strand")
        } else if (col_spec == "all") {
            colClasses <- c("factor", "integer", "factor", "integer",
                            "integer", "factor", "factor")
            drop <- integer(0L)
            col.names <- c("seqnames", "start", "strand", "M", "U",
                           "dinucleotide_context", "trinucleotide_context")
        }
    }

    # Read the file ------------------------------------------------------------

    if (verbose) {
        message("[.readBismarkAsDT] Parsing '", file, "'")
    }
    ptime1 <- proc.time()
    if (verbose) {
        message("[.readBismarkAsDT] Reading file ...")
    }
    x <- fread(
        # TODO: This should be file = file, but need to wait for data.table
        #       v.1.11.9 for fix.
        input = file,
        sep = "\t",
        header = header,
        verbose = fread_verbose,
        drop = drop,
        colClasses = colClasses,
        col.names = col.names,
        quote = "",
        showProgress = showProgress,
        nThread = nThread)

    # Construct the result -----------------------------------------------------

    if (!is.null(x[["strand"]])) {
        x[, strand := strand(strand)]
    }
    # NOTE: Quieten R CMD check about 'no visible binding for global variable'.
    M <- U <- NULL
    if (check && all(c("M", "U") %in% colnames(x))) {
        if (verbose) {
            message("[.readBismarkAsDT] Checking validity of counts in file.")
        }
        valid <- x[, isTRUE(all(M >= 0L & U >= 0L))]
        if (!valid) {
            stop("[.readBismarkAsDT] Invalid counts detected.\n",
                 "'M' and 'U' columns should be non-negative integers.")
        } else {
            if (verbose) {
                message("[.readBismarkAsDT] All counts in file are valid.")
            }
        }
    }
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if (verbose) {
        message("Done in ", round(stime, 1), " secs")
    }
    x
}

.constructCountsFromSingleFile <- function(b, files, loci, strandCollapse,
                                           grid, M_sink, Cov_sink, sink_lock,
                                           nThread, verbose) {

    # Argument checks ----------------------------------------------------------

    subverbose <- max(as.integer(verbose) - 1L, 0L)
    # NOTE: Quieten R CMD check about 'no visible binding for global variable'.
    M <- U <- NULL

    # Read b-th file to construct data.table of valid loci and their counts ----

    file <- files[b]
    if (verbose) {
        message("[.constructCountsFromBismarkFileAndFWGRanges] Extracting ",
                "counts from ", "'", file, "'")
    }
    dt <- .readBismarkAsDT(
        file = file,
        col_spec = "BSseq",
        check = TRUE,
        nThread = nThread,
        verbose = subverbose)
    # NOTE: Can only collapse by strand if the data are stranded!
    if (strandCollapse && !is.null(dt[["strand"]])) {
        # Shift loci on negative strand by 1 to the left and then remove
        # strand since no longer valid.
        dt[strand == "-", start := start - 1L][, strand := NULL]
        # Aggregate counts at loci with the same 'seqnames' and 'start'.
        dt <- dt[, list(M = sum(M), U = sum(U)), by = c("seqnames", "start")]
    }

    # Construct FWGRanges ------------------------------------------------------

    seqnames <- Rle(dt[["seqnames"]])
    dt[, seqnames := NULL]
    seqinfo <- Seqinfo(seqnames = levels(seqnames))
    ranges <- .FWIRanges(start = dt[["start"]], width = 1L)
    dt[, start := NULL]
    mcols <- make_zero_col_DFrame(length(ranges))
    if (is.null(dt[["strand"]])) {
        strand <- strand(Rle("*", length(seqnames)))
    } else {
        strand <- Rle(dt[["strand"]])
        dt[, strand := NULL]
    }
    loci_from_this_sample <- .FWGRanges(
        seqnames = seqnames,
        ranges = ranges,
        strand = strand,
        seqinfo = seqinfo,
        elementMetadata = mcols)

    # Construct 'M' and 'Cov' matrices -----------------------------------------

    ol <- findOverlaps(loci_from_this_sample, loci, type = "equal")
    M <- matrix(rep(0L, length(loci)), ncol = 1)
    Cov <- matrix(rep(0L, length(loci)), ncol = 1)
    M[subjectHits(ol)] <- dt[queryHits(ol), ][["M"]]
    Cov[subjectHits(ol)] <- dt[queryHits(ol), list(Cov = (M + U))][["Cov"]]

    # Return 'M' and 'Cov' or write them to the RealizationSink objects --------

    if (is.null(M_sink)) {
        return(list(M = M, Cov = Cov))
    }
    # Write to M_sink and Cov_sink while respecting the IPC lock.
    viewport <- grid[[b]]
    ipclock(sink_lock)
    write_block(M_sink, viewport = viewport, block = M)
    write_block(Cov_sink, viewport = viewport, block = Cov)
    ipcunlock(sink_lock)
    NULL
}

.constructCounts <- function(files, loci, strandCollapse, BPPARAM,
                             BACKEND, dir, chunkdim, level, nThread,
                             verbose = FALSE) {

    # Argument checks ----------------------------------------------------------

    subverbose <- max(as.integer(verbose) - 1L, 0L)

    # Construct ArrayGrid ------------------------------------------------------

    # NOTE: Each block contains data for a single sample.
    ans_nrow <- length(loci)
    ans_ncol <- length(files)
    ans_dim <- c(ans_nrow, ans_ncol)
    grid <- RegularArrayGrid(
        refdim = ans_dim,
        spacings = c(ans_nrow, 1L))

    # Construct RealizationSink objects ----------------------------------------

    if (is.null(BACKEND)) {
        M_sink <- NULL
        Cov_sink <- NULL
        sink_lock <- NULL
    } else if (BACKEND == "HDF5Array") {
        h5_path <- file.path(dir, "assays.h5")
        M_sink <- HDF5RealizationSink(
            dim = ans_dim,
            dimnames = NULL,
            type = "integer",
            filepath = h5_path,
            name = "M",
            chunkdim = chunkdim,
            level = level)
        on.exit(close(M_sink), add = TRUE)
        Cov_sink <- HDF5RealizationSink(
            dim = ans_dim,
            dimnames = NULL,
            type = "integer",
            filepath = h5_path,
            name = "Cov",
            chunkdim = chunkdim,
            level = level)
        on.exit(close(Cov_sink), add = TRUE)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)
    } else {
        # TODO: This branch should probably never be entered because we
        #       (implicitly) only support in-memory or HDF5Array backends.
        #       However, we retain it for now (e.g., fstArray backend would use
        #       this until a dedicated branch was implemented).
        M_sink <- DelayedArray::AutoRealizationSink(
            dim = ans_dim,
            type = "integer")
        on.exit(close(M_sink), add = TRUE)
        Cov_sink <- DelayedArray::AutoRealizationSink(
            dim = ans_dim,
            type = "integer")
        on.exit(close(Cov_sink), add = TRUE)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)
    }

    # Apply .constructCountsFromSingleFile() to each file ----------------------

    # Set number of tasks to ensure the progress bar gives frequent updates.
    # NOTE: The progress bar increments once per task
    #       (https://github.com/Bioconductor/BiocParallel/issues/54).
    #       Although it is somewhat of a bad idea to overrides a user-specified
    #       bptasks(BPPARAM), the value of bptasks(BPPARAM) doesn't affect
    #       performance in this instance, and so we opt for a useful progress
    #       bar. Only SnowParam (and MulticoreParam by inheritance) have a
    #       bptasks<-() method.
    if (is(BPPARAM, "SnowParam") && bpprogressbar(BPPARAM)) {
        bptasks(BPPARAM) <- length(grid)
    }
    counts <- bptry(bplapply(
        X = seq_along(grid),
        FUN = .constructCountsFromSingleFile,
        files = files,
        loci = loci,
        strandCollapse = strandCollapse,
        grid = grid,
        M_sink = M_sink,
        Cov_sink = Cov_sink,
        sink_lock = sink_lock,
        verbose = subverbose,
        nThread = nThread,
        BPPARAM = BPPARAM))
    if (!all(bpok(counts))) {
        stop(".constructCounts() encountered errors for these files:\n  ",
             paste(files[!bpok], collapse = "\n  "))
    }

    # Construct 'M' and 'Cov' --------------------------------------------------

    if (is.null(BACKEND)) {
        # Returning matrix objects.
        M <- do.call(c, lapply(counts, "[[", "M"))
        attr(M, "dim") <- ans_dim
        Cov <- do.call(c, lapply(counts, "[[", "Cov"))
        attr(Cov, "dim") <- ans_dim
    } else {
        # Returning DelayedMatrix objects.
        M <- as(M_sink, "DelayedArray")
        Cov <- as(Cov_sink, "DelayedArray")
    }

    return(list(M = M, Cov = Cov))
}

# Exported functions -----------------------------------------------------------

# TODO: If you have N cores available, are you better off using
#       bpworkers() = N in the BPPARAM or nThread = N and use
#       data.table::fread()? Or something in between?
# TODO: Allow user to determine `nThread`?
# TODO: Document that you can't use strandCollapse with files that lack strand
#       (e.g., .cov files).
# TODO: (long term) Report a run-time warning/message if strandCollapse = TRUE
#       is used in conjunction with files without strand information.
# TODO: Try a bpiterate()-based approach to getting the set of loci from files.
read.bismark <- function(files,
                         loci = NULL,
                         colData = NULL,
                         rmZeroCov = FALSE,
                         strandCollapse = TRUE,
                         BPPARAM = bpparam(),
                         BACKEND = NULL,
                         dir = tempfile("BSseq"),
                         replace = FALSE,
                         chunkdim = NULL,
                         level = NULL,
                         nThread = 1L,
                         verbose = getOption("verbose")) {
    # Argument checks ----------------------------------------------------------

    # Check 'files' is valid.
    if (anyDuplicated(files)) {
        stop("'files' cannot have duplicate entries.")
    }
    file_exists <- file.exists(files)
    if (!isTRUE(all(file_exists))) {
        stop("These files cannot be found:\n  ",
             paste(files[!file_exists], collapse = "\n  "))
    }
    guessed_file_types <- .guessBismarkFileType(files)
    if (anyNA(guessed_file_types)) {
        stop("Could not guess Bismark file type for these files:\n  ",
             "  ", paste(files[!is.na(guessed_file_types)], collapse = "\n  "))
    }
    # Check 'loci' is valid.
    if (!is.null(loci)) {
        if (!is(loci, "GenomicRanges")) {
            stop("'loci' must be a GenomicRanges instance if not NULL.")
        }
        if (any(width(loci) != 1L)) {
            stop("All elements of 'loci' must have width equal to 1.")
        }
    }
    # Check 'colData' is valid.
    if (is.null(colData)) {
        colData <- DataFrame(row.names = files)
    }
    if (nrow(colData) != length(files)) {
        stop("Supplied 'colData' must have nrow(colData) == length(files).")
    }
    # Check 'rmZeroCov' and 'strandCollapse' are valid.
    stopifnot(isTRUEorFALSE(rmZeroCov))
    stopifnot(isTRUEorFALSE(strandCollapse))
    # Register 'BACKEND' and return to current value on exit.
    # TODO: Is this strictly necessary?
    current_BACKEND <- getAutoRealizationBackend()
    on.exit(setAutoRealizationBackend(current_BACKEND), add = TRUE)
    setAutoRealizationBackend(BACKEND)
    # Check compatability of 'BPPARAM' with 'BACKEND'.
    if (!.areBackendsInMemory(BACKEND)) {
        if (!.isSingleMachineBackend(BPPARAM)) {
            stop("The parallelisation strategy must use a single machine ",
                 "when using an on-disk realization backend.\n",
                 "See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    } else {
        if (!is.null(BACKEND)) {
            # NOTE: Currently do not support any in-memory realization
            #       backends. If the realization backend is NULL then an
            #       ordinary matrix is returned rather than a matrix-backed
            #       DelayedMatrix.
            stop("The '", BACKEND, "' realization backend is not supported.",
                 "\n  See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    }
    # If using HDF5Array as BACKEND, check remaining options are sensible.
    if (identical(BACKEND, "HDF5Array")) {
        # NOTE: Most of this copied from
        #       HDF5Array::saveHDF5SummarizedExperiment().
        if (!isSingleString(dir)) {
            stop(wmsg("'dir' must be a single string specifying the path to ",
                      "the directory where to save the BSseq object (the ",
                      "directory will be created)."))
        }
        if (!isTRUEorFALSE(replace)) {
            stop("'replace' must be TRUE or FALSE")
        }
        if (!dir.exists(dir)) {
            HDF5Array::create_dir(dir)
        } else {
            HDF5Array::replace_dir(dir, replace)
        }
    }
    # Set verbosity used by internal functions.
    subverbose <- as.logical(max(verbose - 1L, 0L))

    # Construct FWGRanges with all valid loci ----------------------------------

    # NOTE: "Valid" loci are those that remain after collapsing by strand (if
    #       strandCollapse == TRUE) and then removing loci with zero coverage
    #       in all samples (if rmZeroCov == TRUE).
    if (is.null(loci)) {
        ptime1 <- proc.time()
        if (verbose) {
            message(
                "[read.bismark] Parsing files and constructing valid loci ...")
        }
        loci <- .constructFWGRangesFromBismarkFiles(
            files = files,
            rmZeroCov = rmZeroCov,
            strandCollapse = strandCollapse,
            verbose = subverbose,
            nThread = nThread,
            BPPARAM = BPPARAM)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if (verbose) {
            message("Done in ", round(stime, 1), " secs")
        }
    } else {
        if (verbose) {
            message("[read.bismark] Using 'loci' as candidate loci.")
        }
        if (strandCollapse) {
            if (verbose) {
                message("[read.bismark] Collapsing strand of 'loci' ...")
            }
            ptime1 <- proc.time()
            loci <- .strandCollapse(loci)
            ptime2 <- proc.time()
            stime <- (ptime2 - ptime1)[3]
            if (verbose) {
                message("Done in ", round(stime, 1), " secs")
            }
        }
        if (rmZeroCov) {
            if (verbose) {
                message("[read.bismark] Parsing files to identify elements of ",
                        "'loci' with non-zero coverage ...")
            }
            ptime1 <- proc.time()
            # Construct loci with non-zero coverage in files.
            loci_from_files <- .constructFWGRangesFromBismarkFiles(
                files = files,
                rmZeroCov = rmZeroCov,
                strandCollapse = strandCollapse,
                verbose = subverbose,
                nThread = nThread,
                BPPARAM = BPPARAM)
            # Retain those elements of 'loci' with non-zero coverage.
            loci <- subsetByOverlaps(loci, loci_from_files, type = "equal")
            ptime2 <- proc.time()
            stime <- (ptime2 - ptime1)[3]
            if (verbose) {
                message("Done in ", round(stime, 1), " secs")
            }
        }
    }

    # Construct 'M' and 'Cov' matrices -----------------------------------------
    ptime1 <- proc.time()
    if (verbose) {
        message("[read.bismark] Parsing files and constructing 'M' and 'Cov' ",
                "matrices ...")
    }
    counts <- .constructCounts(
        files = files,
        loci = loci,
        strandCollapse = strandCollapse,
        BPPARAM = BPPARAM,
        BACKEND = BACKEND,
        dir = dir,
        chunkdim = chunkdim,
        level = level,
        nThread = nThread,
        verbose = subverbose)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if (verbose) {
        message("Done in ", round(stime, 1), " secs")
    }

    # Construct BSseq object, saving it if it is HDF5-backed -------------------

    if (verbose) {
        message("[read.bismark] Constructing BSseq object ... ")
    }
    se <- SummarizedExperiment(
        assays = counts,
        # NOTE: For now, have to use GRanges instance as rowRanges but
        #       ultimately would like to use a FWGRanges instance.
        rowRanges = as(loci, "GRanges"),
        colData = colData)
    # TODO: Is there a way to use the internal constructor with `check = FALSE`?
    #       Don't need to check M and Cov because this has already happened
    #       when files were parsed.
    # .BSseq(se, trans = function(x) NULL, parameters = list())
    bsseq <- new2("BSseq", se, check = FALSE)
    if (!is.null(BACKEND) && BACKEND == "HDF5Array") {
          # NOTE: Save BSseq object; mimicing
        #       HDF5Array::saveHDF5SummarizedExperiment().
        x <- bsseq
        x@assays <- HDF5Array::shorten_assay2h5_links(x@assays)
        base::saveRDS(x, file = file.path(dir, "se.rds"))
    }
    bsseq
}

# TODOs ------------------------------------------------------------------------

# TODO: Document internal functions for my own sanity. Also, some may be useful
#       to users of bsseq (although I still won't export these for the time
#       being).
# TODO: (long term) Current implementation requires that user can load at least
#       one sample's worth of data into memory per worker. Could instead read
#       chunks of data, write to sink, load next chunk, etc.
# TODO: Document that if 'loci' is NULL and any 'files' (especially the first
#       file) are .cov files, then any loci present in the .cov files will have
#       their strand set to *. If you are mixing and matching .cov and
#       .cytosineReport files and don't want this behaviour (i.e. you want to
#       retain strand) then you'll need to construct your own 'gr' and pass
#       this to the function. Add unit tests for this behaviour.

context("read.bismark")

# TODO: Re-factor read.bismark() and update tests accordingly
test_that("read.bismark() works for 'coverage' file", {
    infile <- system.file("extdata", "test_data.fastq_bismark.bismark.cov.gz",
                          package = "bsseq")
    bsseq <- read.bismark(files = infile,
                          colData = DataFrame(row.names = "test_data"),
                          rmZeroCov = FALSE,
                          strandCollapse = FALSE,
                          verbose = FALSE)
    expect_is(bsseq, "BSseq")
})

test_that("read.bismark() works for 'genome wide cytosine report' file", {
    # Test that read.bismark() works on good input
    infile <- system.file("extdata", "test_data.cytosineReport.gz",
                          package = "bsseq")
    bsseq <- read.bismark(files = infile,
                          rmZeroCov = FALSE,
                          strandCollapse = FALSE,
                          verbose = FALSE)
    expect_is(bsseq, "BSseq")
    # Check that strandCollapse = FALSE works
    expect_equal(nrow(bsseq), 100L)
    expect_true(all(strand(bsseq) %in% c("+", "-")))

    # Check that strandCollapse = TRUE works
    bsseq <- read.bismark(files = infile,
                          rmZeroCov = FALSE,
                          strandCollapse = TRUE,
                          verbose = FALSE)
    expect_equal(nrow(bsseq), 50L)
    expect_true(all(strand(bsseq) == "*"))
})

context("BSmooth")

test_that("Errors on bad input", {
    expect_error(
        object = BSmooth(
            BSseq = as(bsseq_test, "SummarizedExperiment"),
            BPPARAM = SerialParam()),
        regexp = "'BSseq' must be a BSseq object")
    expect_error(
        object =  BSmooth(rev(bsseq_test), BPPARAM = SerialParam()),
        regexp = "'BSseq' must be sorted")
    expect_error(
        object = BSmooth(resize(bsseq_test, 2), BPPARAM = SerialParam()),
        regexp = "All loci in 'BSseq' must have width == 1")
})

test_that("BSmooth properly inherits 'dir'", {
    infile <- system.file(
        "extdata", "test_data.fastq_bismark.bismark.cov.gz",
        package = "bsseq")
    bsseq <- read.bismark(
        files = infile,
        BACKEND = "HDF5Array",
        BPPARAM = SerialParam())
    bsseq <- BSmooth(bsseq, BPPARAM = SerialParam())
    expect_is(bsseq, "BSseq")
    expect_true(hasBeenSmoothed(bsseq))
    expect_error(
        BSmooth(bsseq),
        "already contains a dataset named 'coef' or 'se.coef'")
})

test_that("Checks compatability of 'BPPARAM' with the realization backend", {
        expect_error(
            BSmooth(
                BSseq = realize(bsseq_test, "HDF5Array"),
                BPPARAM = SnowParam(c("node1", "node2"))),
            "The parallelisation strategy must use a single machine when using an on-disk realization backend")
    }
)

test_that("Backends work when using SerialParam parallelisation backend", {
    bsseq_NULL <- BSmooth(
        BSseq = realize(bsseq_test, NULL), BPPARAM = SerialParam())
    bsseq_HDF5Array <- BSmooth(bsseq_test, BPPARAM = SerialParam())
    expect_equivalent_SE(bsseq_NULL, bsseq_HDF5Array)
})

test_that(
    "Expected parallelisation backends work with in-memory realization backend", {
        setAutoRealizationBackend(NULL)

        bsseq_serial_param <- BSmooth(bsseq_test, BPPARAM = SerialParam())
        bsseq_multicore_param <- BSmooth(
            bsseq_test,
            BPPARAM = MulticoreParam(2))
        expect_equivalent_SE(bsseq_serial_param, bsseq_multicore_param)

        bsseq_snow_param_single_machine <- BSmooth(
            bsseq_serial_param,
            BPPARAM = SnowParam(2))
        expect_equivalent_SE(bsseq_test, bsseq_snow_param_single_machine)

        skip_if_not_installed("batchtools")
        bsseq_batch_jobs_param <- BSmooth(
            bsseq_test,
            BPPARAM = BatchtoolsParam(2, progressbar = FALSE))
        expect_equivalent_SE(bsseq_serial_param, bsseq_batch_jobs_param)

        skip_if_not_installed("doParallel")
        doParallel::registerDoParallel(2)
        bsseq_dopar_param <- BSmooth(bsseq_test, BPPARAM = DoparParam())
        expect_equivalent_SE(bsseq_serial_param, bsseq_dopar_param)

        skip("Skipping tests that require SnowParam with multiple machines")
        bsseq_snow_param_multiple_machines <- BSmooth(
            bsseq_test,
            BPPARAM = SnowParam(c("fake1", "fake2")))
        expect_equivalent_SE(
            bsseq_serial_param,
            bsseq_snow_param_multiple_machines)

        skip("Skipping tests that require SnowParam with MPI")
        bsseq_snow_param_mpi <- BSmooth(
            bsseq_test,
            BPPARAM = SnowParam(2, type = "MPI"))
        expect_equivalent_SE(bsseq_serial_param, bsseq_snow_param_mpi)
    }
)

test_that(
    "Expected parallelisation backends work with on-disk realization backend", {
        bsseq_test <- realize(bsseq_test, "HDF5Array")

        bsseq_serial_param <- BSmooth(bsseq_test, BPPARAM = SerialParam())
        bsseq_multicore_param <- BSmooth(
            bsseq_test,
            BPPARAM = MulticoreParam(2))
        expect_equivalent_SE(bsseq_serial_param, bsseq_multicore_param)

        bsseq_snow_param_single_machine <- BSmooth(
            bsseq_test,
            BPPARAM = SnowParam(2))
        expect_equivalent_SE(
            bsseq_serial_param,
            bsseq_snow_param_single_machine)

        # Only single-machine backends support on-disk realization.
        expect_error(
            BSmooth(
                bsseq_serial_param,
                BPPARAM = BatchtoolsParam(2, progressbar = FALSE)))
        expect_error(
            BSmooth(bsseq_test,
                    BPPARAM = SnowParam(workers = c("fake1", "fake2"))))
        expect_error(
            BSmooth(bsseq_test, BPPARAM = SnowParam(2, type = "MPI")))
        skip_if_not_installed("doParallel")
        doParallel::registerDoParallel(2)
        # Only single-machine backends support on-disk realization.
        # TODO: Don't yet have good way to test for single-machine backend when
        #       using DoparaParam; see
        #       https://github.com/Bioconductor/BiocParallel/issues/76
        expect_error(BSmooth(bsseq_test, BPPARAM = DoparParam()))

        skip("Skipping tests that require SnowParam with multiple machines")
        bsseq_snow_param_multiple_machines <- BSmooth(
            bsseq_test,
            BPPARAM = SnowParam(c("fake1", "fake2")))
        expect_equivalent_SE(
            bsseq_serial_param,
            bsseq_snow_param_multiple_machines)

        skip("Skipping tests that require SnowParam with MPI")
        bsseq_snow_param_mpi <- BSmooth(
            bsseq_test,
            BPPARAM = SnowParam(2, type = "MPI"))
        expect_equivalent_SE(bsseq_serial_param, bsseq_snow_param_mpi)
    }
)

test_that("keep.se works", {
    expect_is(
        assay(BSmooth(bsseq_test, keep.se = TRUE, BPPARAM = SerialParam()),
              "se.coef"),
        "matrix")
    expect_error(
        assay(BSmooth(bsseq_test, keep.se = FALSE, BPPARAM = SerialParam()),
              "'se.coef' not in names"))

    expect_is(
        assay(
            BSmooth(
                BSseq = realize(bsseq_test, "HDF5Array"),
                keep.se = TRUE,
                BPPARAM = SerialParam()),
              "se.coef",
            withDimnames = FALSE),
        "HDF5Matrix")
    expect_error(
        assay(BSmooth(bsseq_test, keep.se = FALSE, BPPARAM = SerialParam()),
              "'se.coef' not in names"))
})

setClass("BSseqStat", contains = "hasGRanges",
         representation(stats = "list",
                        parameters = "list")
         )

setValidity("BSseqStat", function(object) {
    msg <- NULL
    if(is.null(names(object@stats)) || any(names(object@stats) == "") ||
       anyDuplicated(names(object@stats)))
        msg <- validMsg(msg, "the 'stats' list needs to be named with unique names.")
    for(name in c("rawSds", "smoothsSds", "stat", "rawTstats")) {
        if(name %in% names(object@stats) && isTRUE(nrow(object@stats[[name]]) != length(object@gr)))
            msg <- validMsg(msg, sprintf("component '%s' of slot 'stats' has to have the same number of rows as slot 'gr' is long", name))
    }
    if(is.null(msg)) TRUE else msg
})

setMethod("show", signature(object = "BSseqStat"),
          function(object) {
              cat("An object of type 'BSseqStat' with\n")
              cat(" ", length(object), "methylation loci\n")
              cat("based on smoothed data:\n")
              cat(" ", object@parameters$smoothText, "\n")
          })

setMethod("[", "BSseqStat", function(x, i, ...) {
    if(missing(i))
        stop("need [i] for subsetting")
    if(missing(i))
        return(x)
    x@gr <- x@gr[i]
    statnames <- names(x@stats)
    names(statnames) <- statnames
    x@stats <- lapply(statnames, function(nam) {
        if(nam %in% c("rawTstats", "modelCoefficients", "rawSds", "smoothSds",
                      "stat")) {
            return(x@stats[[nam]][i,,drop=FALSE])
        }
        x@stats[[nam]]
    })
    x
})

BSseqStat <- function(gr = NULL, stats = NULL, parameters = NULL) {
    out <- new("BSseqStat")
    out@gr <- gr
    out@stats <- stats
    out@parameters <- parameters
    out
}

# TODO: updateObject() to use ordinary matrix instead of DelayedMatrix with
#       in-memory seed.
# setMethod("updateObject", "BSseqStat",
#           function(object, ...) {
#               not_delayed_matrices <- c("cor.coefficients", "stat.type")
#               delayed_matrices <- setdiff(names(stats), not_delayed_matrices)
#               stats <- object@stats
#               stats[delayed_matrices] <- endoapply(stats[delayed_matrices],
#                                                    .DelayedMatrix)
#               object@stats <- stats
#               object
#           }
# )

## summary.BSseqStat <- function(object, ...) {
##     quant <- quantile(getStats(object)[, "tstat.corrected"],
##                       prob = c(0.0001, 0.001, 0.01, 0.5, 0.99, 0.999, 0.9999))
##     quant <- t(t(quant))
##     colnames(quant) <- "quantiles"
##     out <- list(quantiles = quant)
##     class(out) <- "summary.BSseqStat"
##     out
## }

## print.summary.BSseqStat <- function(x, ...) {
##     print(as.matrix(x$quantiles))
## }

## plot.BSseqStat <- function(x, y, ...) {
##     tstat <- getStats(x)[, "tstat"]
##     plot(density(tstat), xlim = c(-10,10), col = "blue", main = "")
##     if("tstat.corrected" %in% colnames(getStats(x))) {
##         tstat.cor <- getStats(x)[, "tstat.corrected"]
##         lines(density(tstat.cor), col = "black")
##         legend("topleft", legend = c("uncorrected", "corrected"), lty = c(1,1),
##                col = c("blue", "black"))
##     } else {
##         legend("topleft", legend = c("uncorrected"), lty = 1,
##                col = c("blue"))
##     }
## }

data.frame2GRanges <- function(df, keepColumns = FALSE, ignoreStrand = FALSE) {
    stopifnot(is(df, "data.frame"))
    stopifnot(all(c("start", "end") %in% names(df)))
    stopifnot(any(c("chr", "seqnames") %in% names(df)))
    if("seqnames" %in% names(df))
        names(df)[names(df) == "seqnames"] <- "chr"
    if(!ignoreStrand && "strand" %in% names(df)) {
        if(is.numeric(df$strand)) {
            strand <- ifelse(df$strand == 1, "+", "*")
            strand[df$strand == -1] <- "-"
            df$strand <- strand
        }
        gr <- GRanges(seqnames = df$chr,
                      ranges = IRanges(start = df$start, end = df$end),
                      strand = df$strand)
    } else {
        gr <- GRanges(seqnames = df$chr,
                      ranges = IRanges(start = df$start, end = df$end))
    }
    if(keepColumns) {
        dt <- as(df[, setdiff(names(df), c("chr", "start", "end", "strand"))],
                 "DataFrame")
        mcols(gr) <- dt
    }
    names(gr) <- rownames(df)
    gr
}

.ON_DISK_SEEDS <- c("HDF5ArraySeed")
.ON_DISK_BACKENDS <- c("HDF5Array")

.areBackendsInMemory <- function(realization_backends) {
    if (is.null(realization_backends)) {
        return(TRUE)
    }
    vapply(realization_backends, function(realization_backend) {
        is.null(realization_backend) ||
            !realization_backend %in% .ON_DISK_BACKENDS
    }, logical(1L))
}

# TODO: The below is a hack, need a more reliable way to do this.
.getBSseqBackends <- function(x) {
    assay_backends <- lapply(assays(x, withDimnames = FALSE), function(assay) {
        if (is.matrix(assay)) return(NULL)
        seed_classes <- .getSeedClasses(assay)
        if (all(vapply(seed_classes, function(x) x == "matrix", logical(1)))) {
            return(NULL)
        }
        backend <- gsub("Seed", "", as.vector(seed_classes))
        if (!identical(backend, "HDF5Array")) {
            stop("Don't know backend of object with seed '", seed_classes, "'.")
        }
        backend
    })
    unique(unlist(assay_backends))
}

# TODO: https://github.com/Bioconductor/BiocParallel/issues/76
.isSingleMachineBackend <- function(BPPARAM) {
    if (is(BPPARAM, "SerialParam") || is(BPPARAM, "MulticoreParam")) {
        return(TRUE)
    } else if (is(BPPARAM, "SnowParam")) {
        if (is.numeric(bpworkers(BPPARAM)) &&
            BPPARAM$.clusterargs$type == "SOCK") {
            return(TRUE)
        } else {
            return(FALSE)
        }
    } else if (is(BPPARAM, "DoparParam")) {
        # TODO: Can't figure this one out, so returning FALSE for now
        return(FALSE)
    } else {
        return(FALSE)
    }
}

# BSseq object used in tests ---------------------------------------------------

data(BS.chr22)
bsseq_test <- BS.chr22[1:1000, ]
seqlevels(bsseq_test) <- "chr21"
bsseq_test <- suppressWarnings(rbind(bsseq_test, BS.chr22[1:1000, ]))

# Helper function used in tests ------------------------------------------------

expect_equivalent_SE <- function(SE1, SE2) {
    stopifnot(require(SummarizedExperiment))
    stopifnot(is(SE1, "SummarizedExperiment"),
              is(SE2, "SummarizedExperiment"))
    assays(SE1) <- endoapply(assays(SE1), as.matrix)
    assays(SE2) <- endoapply(assays(SE2), as.matrix)
    if (isTRUE(all.equal(SE1, SE2))) {
        return(invisible(all.equal(assays(SE1), assays(SE2))))
    }
    invisible(FALSE)
}

# Functions/methods that would be good to have in DelayedArray -----------------

.rowVars <- function(x, rows = NULL, cols = NULL, ...) {
    if (is(x, "DelayedArray")) {
        if (!is.null(rows)) {
            x <- x[rows, ]
        }
        if (!is.null(cols)) {
            x <- x[, cols]
        }
        row_vars <- rowVars(as.array(x), ...)
    } else {
        row_vars <- rowVars(x, rows = rows, cols = cols, ...)
    }
    row_vars
}

.rowSds <- function(x, rows = NULL, cols = NULL, ...) {
    row_vars <- .rowVars(x, rows = rows, cols = cols, ...)
    sqrt(row_vars)
}

.quantile <- function(x, ...) {
    if (is(x, "DelayedArray")) {
        x <- as.array(x)
    }
    quantile(x, ...)
}

.isSimpleDelayedMatrix <- function(x) {
    is(x@seed, "matrix")
}

.zero_type <- function(type) {
    if (identical(type, "integer")) {
        fill <- 0L
    } else if (identical(type, "double")) {
        fill <- 0
    } else {
        stop("'type' = ", type, " is not supported")
    }
}

# A temporary workaround to
# https://github.com/Bioconductor/DelayedArray/issues/41.
.colsum <- function(x, group, reorder = TRUE, na.rm = FALSE, filepath = NULL,
                    name = NULL, chunkdim = NULL, level = NULL,
                    type = c("double", "integer"), BPPARAM = bpparam()) {

    # NOTE: Special case for HDF5Matrix, otherwise defer to rowsum().
    if (is(x, "HDF5Matrix")) {
        # Check arguments ------------------------------------------------------

        type <- match.arg(type)
        if (any(!c(type(x), type) %in% c("integer", "double"))) {
            stop("'type(x)' must be 'integer' or 'double'.")
        }
        if (length(group) != NCOL(x)) {
            stop("incorrect length for 'group'")
        }
        if (anyNA(group)) {
            warning("missing values for 'group'")
        }
        ugroup <- unique(group)
        if (reorder) {
            ugroup <- sort(ugroup, na.last = TRUE, method = "quick")
        }
        # TODO: Default is type = "double" because rowSums2() returns numeric,
        #       but it can be useful to manually override this when you know
        #       the result is integer.

        # Construct RealizationSink --------------------------------------------

        # NOTE: This is ultimately coerced to the output DelayedMatrix
        #       object
        ans_nrow <- nrow(x)
        ans_ncol <- length(ugroup)
        ans_dim <- c(ans_nrow, ans_ncol)
        sink <- HDF5RealizationSink(
            dim = ans_dim,
            dimnames = list(rownames(x), as.character(ugroup)),
            type = type,
            filepath = filepath,
            name = name,
            chunkdim = chunkdim,
            level = level)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)

        # Construct ArrayGrid --------------------------------------------------

        sink_grid <- colAutoGrid(x = sink, ncol = 1L)
        list_of_cols <- split(seq_along(group), group)[ugroup]

        # Compute colsum() -----------------------------------------------------

        bplapply(
            X = seq_along(sink_grid),
            FUN = function(b, x, sink, sink_lock, sink_grid, list_of_cols) {
                cols <- list_of_cols[[b]]
                if (length(cols) == 1L) {
                    ans <- as.matrix(x[, cols, drop = FALSE])
                    if (na.rm) {
                        ans[is.na(ans)] <- 0L
                    }
                } else {
                    ans <- matrix(
                        rowSums2(x, cols = cols, na.rm = na.rm),
                        ncol = 1)
                }
                ipclock(sink_lock)
                write_block(sink, viewport = sink_grid[[b]], block = ans)
                ipcunlock(sink_lock)
                NULL
            },
            x = x,
            sink = sink,
            sink_lock = sink_lock,
            sink_grid = sink_grid,
            list_of_cols = list_of_cols,
            BPPARAM = BPPARAM)
        return(as(sink, "DelayedArray"))
    }

    colsum(x, group, reorder)
}

# A temporary workaround to
# https://github.com/Bioconductor/DelayedArray/issues/41.
.rowsum <- function(x, group, reorder = TRUE, na.rm = FALSE, filepath = NULL,
                    name = NULL, chunkdim = NULL, level = NULL,
                    type = c("double", "integer"), BPPARAM = bpparam()) {

    # NOTE: Special case for HDF5Matrix, otherwise defer to rowsum().
    if (is(x, "HDF5Matrix")) {

        # Check arguments ------------------------------------------------------

        if (any(!c(type(x), type) %in% c("integer", "double"))) {
            stop("'type(x)' must be 'integer' or 'double'.")
        }
        if (length(group) != NROW(x)) {
            stop("incorrect length for 'group'")
        }
        if (anyNA(group)) {
            warning("missing values for 'group'")
        }
        ugroup <- unique(group)
        if (reorder) {
            ugroup <- sort(ugroup, na.last = TRUE, method = "quick")
        }
        # NOTE: Default is type = "double" because colSums2() returns numeric,
        #       but it can be useful to manually override this when you know the
        #       result is integer.
        type <- match.arg(type)

        # Construct RealizationSink --------------------------------------------

        # NOTE: This is ultimately coerced to the output DelayedMatrix
        #       object
        ans_nrow <- length(ugroup)
        ans_ncol <- ncol(x)
        ans_dim <- c(ans_nrow, ans_ncol)
        sink <- HDF5RealizationSink(
            dim = ans_dim,
            dimnames = list(as.character(ugroup), colnames(x)),
            type = type,
            filepath = filepath,
            name = name,
            chunkdim = chunkdim,
            level = level)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)

        # Construct ArrayGrid --------------------------------------------------

        sink_grid <- rowAutoGrid(x = sink, nrow = 1L)
        list_of_rows <- split(seq_along(group), group)[as.character(ugroup)]

        # Compute colsum() -----------------------------------------------------

        bplapply(
            X = seq_along(sink_grid),
            FUN = function(b, x, sink, sink_lock, sink_grid, list_of_rows) {
                rows <- list_of_rows[[b]]
                if (length(rows) == 1L) {
                    ans <- as.matrix(x[rows, , drop = FALSE])
                    if (na.rm) {
                        ans[is.na(ans)] <- 0L
                    }
                } else {
                    ans <- matrix(
                        colSums2(x, rows = rows, na.rm = na.rm),
                        nrow = 1)
                }
                ipclock(sink_lock)
                write_block(sink, viewport = sink_grid[[b]], block = ans)
                ipcunlock(sink_lock)
                NULL
            },
            x = x,
            sink = sink,
            sink_lock = sink_lock,
            sink_grid = sink_grid,
            list_of_rows = list_of_rows,
            BPPARAM = BPPARAM)
        return(as(sink, "DelayedArray"))
    }

    rowsum(x, group, reorder)
}


# Missing methods --------------------------------------------------------------

# NOTE: Copied from minfi
# TODO: Perhaps move this to DelayedMatrixStats?
# TODO: DelayedArray::type() for all RealizationSink subclasses
setMethod("type", "HDF5RealizationSink", function(x) {
    x@type
})
# NOTE: Copied from minfi
# TODO: Perhaps move this to DelayedMatrixStats?
setMethod("type", "arrayRealizationSink", function(x) {
    DelayedArray::type(x@result_envir$result)
})
# NOTE: Copied from minfi
# TODO: Perhaps move this to DelayedMatrixStats?
setMethod("type", "RleRealizationSink", function(x) {
    x@type
})
# NOTE: Copied from minfi
# TODO: Perhaps move this to DelayedMatrixStats?
# TODO: dimnames() for all RealizationSink subclasses
setMethod("dimnames", "arrayRealizationSink", function(x) {
    dimnames(x@result_envir$result)
})

# Advanced block processing routines -------------------------------------------

# NOTE: Copy of minfi:::blockApplyWithRealization()
# TODO: Perhaps move this to DelayedMatrixStats?
# NOTE: DelayedArray::blockApply() with the option to write the blocks to
#       'sink'. Useful, for example, to apply a function across column-blocks
#       of a DelayedMatrix, write these results to disk, and then wrap
#       these in a DelayedMatrix.
# TODO: See https://github.com/Bioconductor/DelayedArray/issues/10
blockApplyWithRealization <- function(x, FUN, ..., sink = NULL, x_grid = NULL,
                                      sink_grid = NULL, BPREDO = list(),
                                      BPPARAM = bpparam()) {
    FUN <- match.fun(FUN)

    # Check conformable dots_grids and sinks_grids
    x_grid <- DelayedArray:::normarg_grid(x_grid, x)
    sink_grid <- DelayedArray:::normarg_grid(sink_grid, sink)
    if (!identical(dim(x_grid), dim(sink_grid))) {
        stop("non-conformable 'x_grid' and 'sink_grid'")
    }

    # Loop over blocks of `x` and write to `sink`
    nblock <- length(x_grid)
    bplapply(seq_len(nblock), function(b) {
        if (DelayedArray:::get_verbose_block_processing()) {
            message("Processing block ", b, "/", nblock, " ... ",
                    appendLF = FALSE)
        }
        x_viewport <- x_grid[[b]]
        sink_viewport <- sink_grid[[b]]
        block <- read_block(x, x_viewport)
        attr(block, "from_grid") <- x_grid
        attr(block, "block_id") <- b
        block_ans <- FUN(block, ...)
        # NOTE: This is the only part different from DelayedArray::blockApply()
        if (!is.null(sink)) {
            write_block(sink, viewport = sink_viewport, block = block_ans)
            block_ans <- NULL
        }
        if (DelayedArray:::get_verbose_block_processing()) {
            message("OK")
        }
    },
    BPREDO = BPREDO,
    BPPARAM = BPPARAM)
}

# TODO: Needed?
.getSEDir <- function(x) {
    paths <- lapply(assays(x, withDimnames = FALSE), function(a) {
        try(path(a), silent = TRUE)
    })
    if (any(vapply(paths, is, logical(1L), "try-error"))) {
        stop("Cannot extract 'dir'.")
    }
    unique_paths <- unique(unlist(paths, use.names = FALSE))
    if (length(unique_paths) > 1) {
        stop("Assay data spread across multiple HDF5 files.")
    }
    dirs <- dirname(unlist(paths, use.names = FALSE))
    unique(dirs)
}

# Should return TRUE for BSseq object created with read.bismark() or saved with
# HDF5Array::saveHDF5SummarizedExperiment().
# TODO: Check dirname(paths[[1L]]) also contains 'se.rds'? It looks like dir
#       can contain other files besides these; check.
.isHDF5BackedBSseqUpdatable <- function(x) {
    stopifnot(is(x, "BSseq"))
    assay_class <- vapply(assays(x, withDimnames = FALSE), class, character(1L))
    if (!all(assay_class == "HDF5Matrix")) {
        return(FALSE)
    }
    paths <- vapply(assays(x, withDimnames = FALSE), path, character(1L))
    if (!all(paths == paths[[1L]]) || !all(basename(paths) == "assays.h5")) {
        return(FALSE)
    }
    TRUE
}

# NOTE: Realises in memory a matrix with nrow = length(bstat) and ncol = 1
dmrFinder <- function(bstat, cutoff = NULL, qcutoff = c(0.025, 0.975),
                      maxGap = 300, stat = "tstat.corrected", verbose = TRUE) {
    if(is(bstat, "BSseqTstat")) {
        if(! stat %in% colnames(bstat@stats))
            stop("'stat' needs to be a column of 'bstat@stats'")
        dmrStat <- as.array(bstat@stats[, stat, drop = FALSE])
    }
    if(is(bstat, "BSseqStat")) {
        dmrStat <- as.array(getStats(bstat, what = "stat"))
    }
    subverbose <- max(as.integer(verbose) - 1L, 0L)
    if(is.null(cutoff))
        cutoff <- .quantile(dmrStat, qcutoff)
    if(length(cutoff) == 1)
        cutoff <- c(-cutoff, cutoff)
    direction <- as.integer(dmrStat >= cutoff[2])
    direction[as.array(dmrStat <= cutoff[1])] <- -1L
    direction[is.na(direction)] <- 0L
    chrs <- as.character(seqnames(bstat))
    positions <- start(bstat)
    regions <- regionFinder3(direction, chr = chrs, positions = positions,
                             maxGap = maxGap, verbose = subverbose)
    if(is.null(regions$down) && is.null(regions$up))
        return(NULL)
    if(verbose) cat("[dmrFinder] creating dmr data.frame\n")
    regions <- do.call(rbind, regions)
    rownames(regions) <- NULL
    regions$width <- regions$end - regions$start + 1
    regions$invdensity <- regions$width / regions$n
    regions$chr <- as.character(regions$chr)
    if(is(bstat, "BSseqTstat")) {
        stats <- getStats(bstat, regions, stat = stat)
        regions <- cbind(regions, stats)
        if(stat %in% c("tstat.corrected", "tstat")) {
            regions$direction <- ifelse(regions$meanDiff > 0, "hyper", "hypo")
        }
        regions <- regions[order(abs(regions$areaStat), decreasing = TRUE),]
    }
    if(is(bstat, "BSseqStat")) {
        stats <- getStats(bstat, regions)
        regions <- cbind(regions, stats)
        regions <- regions[order(abs(regions$areaStat), decreasing = TRUE),]
    }
    regions
}


clusterMaker <- function(chr, pos, order.it=TRUE, maxGap=300){
    nonaIndex <- which(!is.na(chr) & !is.na(pos))
    Indexes <- split(nonaIndex, chr[nonaIndex])
    clusterIDs <- rep(NA, length(chr))
    LAST <- 0
    for(i in seq(along = Indexes)){
        Index <- Indexes[[i]]
        x <- pos[Index]
        if(order.it){
            Index <- Index[order(x)]
            x <- pos[Index]
        }
        y <- as.numeric(diff(x) > maxGap)
        z <- cumsum(c(1, y))
        clusterIDs[Index] <- z + LAST
        LAST <- max(z) + LAST
    }
    clusterIDs
}


regionFinder3 <- function(x, chr, positions, keep, maxGap = 300, verbose = TRUE) {
    getSegments3 <- function(x, cid, verbose = TRUE){
        if(verbose) cat("[regionFinder3] segmenting\n")
        segments <- cumsum(c(1, diff(x) != 0)) +
            cumsum(c(1, diff(cid) != 0))
        names(segments) <- NULL
        if(verbose) cat("[regionFinder3] splitting\n")
        out <- list(up = split(which(x > 0), segments[x > 0]),
                    down = split(which(x < 0), segments[x < 0]),
                    nochange = split(which(x == 0), segments[x == 0]))
        names(out[[1]]) <- NULL
        names(out[[2]]) <- NULL
        names(out[[3]]) <- NULL
        out
    }
    cid0 <- clusterMaker(chr, positions, maxGap = maxGap)
    segments <- getSegments3(x = x, cid = cid0, verbose = verbose)
    out <- vector("list", 2)
    names(out) <- c("up", "down")
    for(ii in 1:2) {
        idx <- segments[[ii]]
        if(length(idx) > 0) {
            out[[ii]] <- data.frame(chr = sapply(idx, function(jj) chr[jj[1]]),
                                    start = sapply(idx, function(jj) min(positions[jj])),
                                    end = sapply(idx, function(jj) max(positions[jj])),
                                    idxStart = sapply(idx, function(jj) min(jj)),
                                    idxEnd = sapply(idx, function(jj) max(jj)),
                                    cluster = sapply(idx, function(jj) cid0[jj[1]]),
                                    n = sapply(idx, length))
        }
    }
    out
}

#Likelihood functions for calling methylation states from bsseq objects generated from read.bedMethyl

#estimateErrorRate
estimateErrorRate <- function(
    BSseq, 
    minCov = 10, 
    maxCov = 100, 
    minRatio = 0.8, 
    plotErrorProfile = FALSE) {
  
  # Input validation of BSseq
  if (!inherits(BSseq, "BSseq")) {
    stop("Error: The input object must be of class 'BSseq'.")
  }
  if (ncol(BSseq) != 1) {
    stop("Error: The input 'BSseq' must have exactly one column.")
  }
  if (!all(c("D", "Cov") %in% names(assays(BSseq)))) {
    stop("Error: The 'BSseq' must contain 'D' and 'Cov' assays.")
  }
  
  # Input validation for thresholds
  if (minCov < 0 || maxCov < 0) {
    stop("Error: Both 'minCov' and 'maxCov' must be non-negative.")
  }
  if (maxCov <= minCov) {
    stop("Error: 'maxCov' must be greater than 'minCov'.")
  }
  if (minRatio < 0 || minRatio > 1) {
    stop("Error: 'minRatio' must be between 0 and 1.")
  }
  
  # Access assay data safely
  cov <- assays(BSseq)[["Cov"]]
  d <- assays(BSseq)[["D"]]
  
  # Calculate CpG-to-non-CpG ratio and filter sites
  ratio <- cov / (cov + d)
  sites <- getCoverage(BSseq) >= minCov & 
    getCoverage(BSseq) <= maxCov & 
    ratio >= minRatio
  
  # Plot the error profile if requested
  if (plotErrorProfile) {
    if (any(sites)) {
      p_all <- hist(ratio, breaks = 100, plot = FALSE)
      p_used <- hist(ratio[sites], breaks = (1 - minRatio) * 100, plot = FALSE)
      plot(p_all, xlim = c(0, 1), xlab = "CpG to non-CpG ratio", main = "Error Profile")
      plot(p_used, col = "black", xlim = c(0, 1), add = TRUE)
    } else {
      warning("No sites satisfy the filtering criteria. No error profile will be plotted.")
    }
  }
  
  # Calculate and return the error rate
  error_rate <- sum(d[sites]) / (sum(d[sites]) + sum(cov[sites]))
  return(error_rate)
}

#genotype_likelihood
#internal function to calculate the likelihood of a genotype given the observed data
.genotype_likelihood <- function(n = 2, g, e, cov, D) {
  if (g == 0) {
    p_mismatch <- n * e
    p_match <- n * (1 - e)
  } else if (g == 1) {
    p_mismatch <- n / 2
    p_match <- n / 2
  } else if (g == 2) {
    p_mismatch <- n * (1 - e)
    p_match <- n * e
  } else {
    stop("Invalid genotype value. Must be 0, 1, or 2.")
  }
  likelihood <- (p_mismatch^D) * (p_match^cov)
  return(likelihood)
}

#.getScaledLikelihoods
#internal function to calculate the scaled likelihoods for each genotype
.getScaledLikelihoods <- function(
    BSseq, 
    e = NULL) {
  # Input validation of BSseq
  if (!inherits(BSseq, "BSseq")) {
    stop("Error: The input object must be of class 'BSseq'.")
  }
  if (ncol(BSseq) != 1) {
    stop("Error: The input 'BSseq' must have exactly one column.")
  }
  if (!all(c("D", "Cov") %in% names(assays(BSseq)))) {
    stop("Error: The 'BSseq' must contain 'D' and 'Cov' assays.")
  }
  # Input validation of e
  if (!is.null(e) && (e < 0 || e > 1)) {
    stop("Error: 'e' must be between 0 and 1.")
  }
  if (is.null(e)) {
    e <- estimateErrorRate(BSseq)
  }
  cov <- assays(BSseq)[["Cov"]]
  d <- assays(BSseq)[["D"]]
  
  l <- matrix(0, nrow = length(cov), ncol = 3)
  l[, 1] <- .genotype_likelihood(g = 0, e = e, cov = cov, D = d)
  l[, 2] <- .genotype_likelihood(g = 1, e = e, cov = cov, D = d)
  l[, 3] <- .genotype_likelihood(g = 2, e = e, cov = cov, D = d)
  
  SL <- l / rowSums(l)
  return(SL)
}

getCpGs <- function(
    BSseq, 
    type = c("homozygous", "heterozygous", "allCpG", "nonCpG"), 
    threshold = 0.99, 
    e = NULL) {
  # Input validation of BSseq
  if (!inherits(BSseq, "BSseq")) {
    stop("Error: The input object must be of class 'BSseq'.")
  }
  if (ncol(BSseq) != 1) {
    stop("Error: The input 'BSseq' must have exactly one column.")
  }
  if (!all(c("D", "Cov") %in% names(assays(BSseq)))) {
    stop("Error: The 'BSseq' must contain 'D' and 'Cov' assays.")
  }
  # Input validation of type
  if (!type %in% c("homozygous", "heterozygous", "allCpG", "nonCpG")) {
    stop("Error: 'type' must be one of 'homozygous', 'heterozygous', 'allCpG', or 'nonCpG'.")
  }
  # Input validation of threshold
  if (threshold < 0 || threshold > 1) {
    stop("Error: 'threshold' must be between 0 and 1.")
  }
  # Input validation of e
  if (!is.null(e) && (e < 0 || e > 1)) {
    stop("Error: 'e' must be between 0 and 1.")
  }
  # Get scaled likelihoods
  SL <- .getScaledLikelihoods(BSseq, e)
  
  # Determine the CpG sites based on the type
  loci <- switch(type,
                   "homozygous" = which(SL[, 1] > threshold),
                   "heterozygous" = which(SL[, 2] > threshold),
                   "allCpG" = which(rowSums(SL[, 1:2]) > threshold),
                   "nonCpG" = which(SL[, 3] > threshold))
  
  return(loci)
}

getCpGMatrix <- function(
    BSseq, 
    e = NULL, 
    allCpG = FALSE) {
  
  # Input validation of BSseq
  if (!inherits(BSseq, "BSseq")) {
    stop("Error: The input object must be of class 'BSseq'.")
  }
  if (!all(c("D", "Cov") %in% names(assays(BSseq)))) {
    stop("Error: The 'BSseq' must contain 'D' and 'Cov' assays.")
  }
  
  # Input validation of e
  if (!is.null(e)) {
    if (!is.numeric(e) || (length(e) != ncol(BSseq))) {
      stop("Error: 'e' must be NULL or a numeric vector with the same length as the number of columns in 'BSseq'.")
    }
    if (any(e < 0 | e > 1)) {
      stop("Error: All elements of 'e' must be between 0 and 1.")
    }
  }
  
  # Initialize CpG matrix
  G <- matrix(nrow = nrow(BSseq), ncol = ncol(BSseq), byrow = FALSE)
  
  for (i in 1:ncol(BSseq)) {
    # Use the provided error rate or estimate it
    e_i <- if (is.null(e)) estimateErrorRate(BSseq[, i]) else e[i]
    
    # Get scaled likelihoods
    SL_i <- .getScaledLikelihoods(BSseq[, i], e_i)
    
    # Calculate CpG matrix
    if (allCpG) {
      G[, i] <- ifelse(SL_i[, 3] >= 1/3, 2, 0)
    } else {
      G[, i] <- max.col(SL_i, ties.method = "last") - 1
    }
  }
  
  return(G)
}

getMaxLikelihoodMatrix <- function(
    BSseq, 
    e = NULL, 
    allCpG = FALSE) {
  
  # Input validation of BSseq
  if (!inherits(BSseq, "BSseq")) {
    stop("Error: The input object must be of class 'BSseq'.")
  }
  if (!all(c("D", "Cov") %in% names(assays(BSseq)))) {
    stop("Error: The 'BSseq' must contain 'D' and 'Cov' assays.")
  }
  
  # Input validation of e
  if (!is.null(e)) {
    if (!is.numeric(e) || (length(e) != ncol(BSseq))) {
      stop("Error: 'e' must be NULL or a numeric vector with the same length as the number of columns in 'BSseq'.")
    }
    if (any(e < 0 | e > 1)) {
      stop("Error: All elements of 'e' must be between 0 and 1.")
    }
  }
  
  # Initialize quality matrix
  Q <- matrix(nrow = nrow(BSseq), ncol = ncol(BSseq), byrow = FALSE)
  
  for (i in 1:ncol(BSseq)) {
    # Use the provided error rate or estimate it
    e_i <- if (is.null(e)) estimateErrorRate(BSseq[, i]) else e[i]
    
    # Get scaled likelihoods
    SL_i <- .getScaledLikelihoods(BSseq[, i], e_i)
    
    # Calculate quality matrix
    if (allCpG) {
      Q[, i] <- ifelse(SL_i[, 3] >= 1/3, SL_i[, 3], (1 - SL_i[, 3]))
    } else {
      Q[, i] <- DelayedMatrixStats::rowMaxs(SL_i)
    }
  }
  
  return(Q)
}

context("read.bedMethyl")

test_that("read.bedMethyl() works for bedMethyl file", {
  infiles <- c(system.file("extdata/HG002_nanopore_test.bedMethyl.gz",
    package = "bsseq"),
  system.file("extdata/HG002_pacbio_test.bedMethyl.gz",
    package = "bsseq"))
  bsseq <- read.bedMethyl(files = infiles,
     colData = DataFrame(row.names = c("test_nanopore","test_pacbio")),
     rmZeroCov = TRUE,
     strandCollapse = TRUE,
     verbose = FALSE)
    expect_is(bsseq, "BSseq")
    expect_equal(nrow(bsseq), 1111L)
    expect_true(all(strand(bsseq) == "*"))
    #test of get CpGs
    bsseq_nano<-bsseq[,1]
    expect_equal(length(getCpGs(bsseq_nano, type="allCpG")), 828L)
    CpGMatrix<-getCpGMatrix(bsseq, allCpG = TRUE)
    MaxLikelihoodMatrix<-getMaxLikelihoodMatrix(bsseq, allCpG = TRUE)
    expect_equal(length(bsseq[which(rowAlls(CpGMatrix == 0) 
                                      & rowMins(MaxLikelihoodMatrix) > 0.99)]), 827L)
})

# Internal functions -----------------------------------------------------------

.collapseColData <- function(x, group, reorder = TRUE) {
    if (length(group) != NROW(x)) {
        stop("incorrect length for 'group'")
    }
    if (anyNA(group)) {
        warning("missing values for 'group'")
    }
    ugroup <- unique(group)
    if (reorder) {
        ugroup <- sort(ugroup, na.last = TRUE, method = "quick")
    }
    idx <- split(seq_along(group), group)[ugroup]
    if (ncol(x)) {
        ans <- lapply(x, function(xx) {
            sapply(X = idx, function(i) unique(xx[i]))
        })
        ## The following only retains columns of colData which
        ## are constant across group
        if(any(!sapply(ans, is.vector)))
            warning("Dropping columns from colData(BSseq) which are not constant within `group`")
        if(length(ans) >= 1)
            ans <- ans[sapply(ans, is.vector)]
        if(length(ans) >= 1)
            ans <- DataFrame(ans, row.names = names(idx))
        return(ans)
    }
    DataFrame(row.names = names(idx))
}

# Exported functions -----------------------------------------------------------

# TODO: Remove `replace`? It'd be a bad idea to use replace = TRUE if `dir` is
#       the same as the current dir for the BSseq object.
# NOTE: This is similar to edgeR::sumTechReps().
# TODO: Make optional the collapsing of colData?
# TODO: Document (and warn) that coef and se.coef aren't collapsed?
collapseBSseq <- function(BSseq, group, BPPARAM = bpparam(),
                          BACKEND = getAutoRealizationBackend(),
                          dir = tempfile("BSseq"), replace = FALSE,
                          chunkdim = NULL, level = NULL,
                          type = c("double", "integer")) {

    # Argument checks ----------------------------------------------------------

    if (!anyDuplicated(group)) {
        return(BSseq)
    }
    if (hasBeenSmoothed(BSseq)) {
        warning("Collapsing a smoothed BSseq object. You will need to ",
                "re-smooth using 'BSmooth()' on the returned object.")
    }
    if (length(group) != NCOL(BSseq)) {
        stop("incorrect length for 'group'")
    }
    if (anyNA(group)) {
        warning("missing values for 'group'")
    }
    # Register 'BACKEND' and return to current value on exit.
    # TODO: Is this strictly necessary?
    # current_BACKEND <- getAutoRealizationBackend()
    # on.exit(setAutoRealizationBackend(current_BACKEND), add = TRUE)
    # setAutoRealizationBackend(BACKEND)
    # Check compatability of 'BPPARAM' with 'BACKEND'.
    if (!.areBackendsInMemory(BACKEND)) {
        if (!.isSingleMachineBackend(BPPARAM)) {
            stop("The parallelisation strategy must use a single machine ",
                 "when using an on-disk realization backend.\n",
                 "See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    } else {
        if (!is.null(BACKEND)) {
            # NOTE: Currently do not support any in-memory realization
            #       backends. If the realization backend is NULL then an
            #       ordinary matrix is returned rather than a matrix-backed
            #       DelayedMatrix.
            stop("The '", BACKEND, "' realization backend is not supported.",
                 "\n  See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    }
    # If using HDF5Array as BACKEND, check remaining options are sensible.
    if (identical(BACKEND, "HDF5Array")) {
        # NOTE: Most of this copied from
        #       HDF5Array::saveHDF5SummarizedExperiment().
        if (!isSingleString(dir)) {
            stop(wmsg("'dir' must be a single string specifying the path to ",
                      "the directory where to save the BSseq object (the ",
                      "directory will be created)."))
        }
        if (!isTRUEorFALSE(replace)) {
            stop("'replace' must be TRUE or FALSE")
        }
        if (!dir.exists(dir)) {
            HDF5Array::create_dir(dir)
        } else {
            HDF5Array::replace_dir(dir, replace)
        }
        h5_path <- file.path(dir, "assays.h5")
    } else if (identical(BACKEND, NULL)) {
        h5_path <- NULL
    }

    # Collapse 'M' and 'Cov' matrices ------------------------------------------

    M <- .colsum(
        x = getCoverage(BSseq, type = "M", withDimnames = FALSE),
        group = group,
        reorder = FALSE,
        BPPARAM = BPPARAM,
        filepath = h5_path,
        name = "M",
        chunkdim = chunkdim,
        level = level,
        type = type)
    Cov <- .colsum(
        x = getCoverage(BSseq, type = "Cov", withDimnames = FALSE),
        group = group,
        reorder = FALSE,
        BPPARAM = BPPARAM,
        filepath = h5_path,
        name = "Cov",
        chunkdim = chunkdim,
        level = level,
        type = type)

    # Collapse 'colData' -------------------------------------------------------

    colData <- .collapseColData(
        x = colData(BSseq),
        group = group,
        reorder = FALSE)

    # Construct BSseq object, saving it if it is HDF5-backed -------------------

    se <- SummarizedExperiment(
        assays = SimpleList(M = unname(M), Cov = unname(Cov)),
        rowRanges = rowRanges(BSseq),
        colData = colData)

    # TODO: Is there a way to use the internal constructor with `check = FALSE`?
    #       Don't need to check M and Cov because this has already happened
    #       when files were parsed.
    bsseq <- new2("BSseq", se, check = FALSE)
    trans <- function() NULL
    environment(trans) <- emptyenv()
    bsseq@trans <- trans
    if (identical(BACKEND, "HDF5Array")) {
        # NOTE: Save BSseq object; mimicing
        #       HDF5Array::saveHDF5SummarizedExperiment().
        x <- bsseq
        x@assays <- HDF5Array::shorten_assay2h5_links(x@assays)
        base::saveRDS(x, file = file.path(dir, "se.rds"))
    }
    bsseq
}

context("GitHub issues")

test_that("Issue 74 is fixed", {

    # Can combine two BSseq objects with same loci but different names.
    # (https://github.com/hansenlab/bsseq/issues/74#issue-341014110)
    nloci <- 68164
    nsamples <- 12
    M <- matrix(rpois(nloci * nsamples, 10), ncol = nsamples)
    Cov <- M + rpois(nloci * nsamples, 2)
    gr <- sort(GRanges(
        seqnames = sample(1:10, nloci, replace = TRUE),
        ranges = IRanges(sample(1:10^7, nloci, replace = FALSE), width = 1)))
    BS <- BSseq(M = M, Cov = Cov, gr = gr, sampleNames = paste0("V", 1:nsamples))
    BS.avg <- collapseBSseq(BS, group = paste0("W", c(rep(1, 5), 2:8)))
    BS.all <- combine(BS, BS.avg)
    colnames(BS.all)

    # As above, but when one of the BSseq objects contains additional colData.
    # https://github.com/hansenlab/bsseq/issues/74#issue-341014110
    BS2 <- BS
    BS2$rep <- paste0("A", c(rep(1, 5), 2:8))
    BS2.avg <- collapseBSseq(BS, group = paste0("W", c(rep(1, 5), 2:8)))
    BS2.all <- combine(BS2,BS2.avg)
    colnames(BS2.all)
    colData(BS2.all)
})

setClass("BSseqTstat", contains = "hasGRanges",
         representation(stats = "matrix",
                        parameters = "list")
         )
setValidity("BSseqTstat", function(object) {
    msg <- NULL
    if(length(object@gr) != nrow(object@stats))
        msg <- c(msg, "length of 'gr' is different from the number of rows of 'stats'")
    if(is.null(msg)) TRUE else msg
})

setMethod("show", signature(object = "BSseqTstat"),
          function(object) {
              cat("An object of type 'BSseqTstat' with\n")
              cat(" ", length(object), "methylation loci\n")
              cat("based on smoothed data:\n")
              cat(" ", object@parameters$smoothText, "\n")
              cat("with parameters\n")
              cat(" ", object@parameters$tstatText, "\n")
          })

setMethod("[", "BSseqTstat", function(x, i, ...) {
    if(missing(i))
        stop("need [i] for subsetting")
    if(missing(i))
        return(x)
    x@gr <- x@gr[i]
    x@stats <- x@stats[i,, drop = FALSE]
    x
})

BSseqTstat <- function(gr = NULL, stats = NULL, parameters = NULL) {
    out <- new("BSseqTstat")
    out@gr <- gr
    out@stats <- stats
    out@parameters <- parameters
    out
}

summary.BSseqTstat <- function(object, ...) {
    quant <- .quantile(getStats(object)[, "tstat.corrected"],
                       prob = c(0.0001, 0.001, 0.01, 0.5, 0.99, 0.999, 0.9999))
    quant <- t(t(quant))
    colnames(quant) <- "quantiles"
    out <- list(quantiles = quant)
    class(out) <- "summary.BSseqTstat"
    out
}

print.summary.BSseqTstat <- function(x, ...) {
    print(as.matrix(x$quantiles))
}

plot.BSseqTstat <- function(x, y, ...) {
    tstat <- getStats(x)[, "tstat"]
    tstat <- as.array(tstat)
    plot(density(tstat), xlim = c(-10,10), col = "blue", main = "")
    if("tstat.corrected" %in% colnames(getStats(x))) {
        tstat.cor <- getStats(x)[, "tstat.corrected"]
        tstat.cor <- as.array(tstat.cor)
        lines(density(tstat.cor), col = "black")
        legend("topleft", legend = c("uncorrected", "corrected"), lty = c(1,1),
               col = c("blue", "black"))
    } else {
        legend("topleft", legend = c("uncorrected"), lty = 1,
               col = c("blue"))
    }
}

setMethod("updateObject", "BSseqTstat",
          function(object, ...) {
              stats <- object@stats
              stats <- stats
              object@stats <- stats
              object
          }
)

# Exported classes -------------------------------------------------------------

# NOTE: This create a 'classGeneratorFunction' (internal constructor), .BSseq()
.BSseq <- setClass(
    "BSseq",
    slots = representation(
        trans = "function",
        parameters = "list"),
    contains = "RangedSummarizedExperiment"
)

# Validity methods -------------------------------------------------------------

.checkAssayNames <- function(object, names) {
    if (all(names %in% assayNames(object))) {
        return(NULL)
    } else {
        paste0(
            "object of class ",
            class(object),
            " needs to have assay slots with names ",
            paste0(names, collapse = ", "))

    }
}

.checkMandCov <- function(M, Cov) {
    msg <- NULL
    validMsg(msg, .Call(cxx_check_M_and_Cov, initializeCpp(M), initializeCpp(Cov), 1))
}

# TODO: Benchmark validity method
setValidity2("BSseq", function(object) {
    msg <- NULL

    if (identical(object, .BSseq())) {
        # No validity checks for object returned by internal constructor
        return(msg)
    }
    msg <- validMsg(msg, .checkAssayNames(object, c("Cov", "M")))

    msg <- validMsg(
        msg = msg,
        result = .checkMandCov(
            M = assay(object, "M", withDimnames = FALSE),
            Cov = assay(object, "Cov", withDimnames = FALSE)))

    if (is.null(msg)) {
        TRUE
    } else {
        msg
    }
})

# Internal functions -----------------------------------------------------------

.oldTrans <- function(x) {
    y <- x
    ix <- which(x < 0)
    ix2 <- which(x > 0)
    y[ix] <- exp(x[ix])/(1 + exp(x[ix]))
    y[ix2] <- 1/(1 + exp(-x[ix2]))
    y
}

# Exported functions -----------------------------------------------------------

# TODO: BSseq() is arguably a bad constructor. It doesn't return a valid BSseq
#       object when called without any arguments. It also does some pretty
#       complicated parsing of the inputs. But we're stuck with it because it's
#       been around for a long time.

# Update the BSseq() function allowing it to store Filtered (ambiguous modification
# status) bases generated by modbam2bed.
BSseq <- function(M = NULL, Cov = NULL, Filtered = NULL, coef = NULL, se.coef = NULL,
                  trans = NULL, parameters = NULL, pData = NULL, gr = NULL,
                  pos = NULL, chr = NULL, sampleNames = NULL,
                  rmZeroCov = FALSE) {

    # Argument checks ----------------------------------------------------------

    # Process assays.
    # NOTE: Nothing to do for 'coef', and 'se.coef'.
    if (is.null(M) || is.null(Cov)) {
        stop("Need 'M' and 'Cov'.")
    }
    # Process 'trans' and 'parameters'.
    if (is.null(trans)) {
        trans <- function() NULL
        environment(trans) <- emptyenv()
    }
    if (is.null(parameters)) {
        parameters <- list()
    }
    # Process 'sampleNames' and 'pData'.
    if (is.null(sampleNames)) {
        if (is.null(pData)) {
            # BSseq object will have no colnames.
            pData <- make_zero_col_DFrame(ncol(M))
        } else {
            # BSseq object will have 'sampleNames' as colnames.
            pData <- DataFrame(row.names = sampleNames)
        }
    } else {
        if (is.null(pData)) {
            # BSseq object will have 'sampleNames' as colnames.
            pData <- DataFrame(row.names = sampleNames)
        } else {
            if (is.null(rownames(pData))) {
                rownames(pData) <- sampleNames
            } else {
                stopifnot(identical(rownames(pData), sampleNames))
            }
        }
    }
    # Process 'gr', 'pos', and 'chr'.
    if (is.null(gr)) {
        if (is.null(pos) || is.null(chr)) {
            stop("Need 'pos' and 'chr' if 'gr' not supplied.")
        }
        gr <- GRanges(seqnames = chr, ranges = IRanges(start = pos, width = 1L))
    }
    if (!is(gr, "GRanges")) {
        stop("'gr' needs to be a GRanges.")
    }
    # Process 'rmZeroCov'.
    stopifnot(isTRUEorFALSE(rmZeroCov))

    # Collapse duplicate loci --------------------------------------------------

    is_duplicated <- duplicated(gr)
    if (any(is_duplicated)) {
        warning("Detected duplicate loci. Collapsing counts in 'M' and 'Cov' ",
                "at these positions.")
        if (!is.null(coef) || !is.null(se.coef)) {
            stop("Cannot collapse when 'coef' or 'se.coef' are non-NULL.")
        }
        loci <- gr[!is_duplicated]
        ol <- findOverlaps(gr, loci, type = "equal")
        M <- rowsum(x = M, group = subjectHits(ol), reorder = FALSE)
        rownames(M) <- NULL
        Cov <- rowsum(x = Cov, group = subjectHits(ol), reorder = FALSE)
        rownames(Cov) <- NULL
        if (!is.null(Filtered)){
          Filtered <- rowsum(x = Filtered, group = subjectHits(ol), reorder = FALSE)
          rownames(Filtered) <- NULL}
    } else {
        loci <- gr
    }

    # Optionally, remove positions with zero coverage --------------------------

    if (rmZeroCov) {
        loci_with_zero_cov <- rowAlls(Cov, value = 0)
        if (any(loci_with_zero_cov)) {
            loci_with_nonzero_cov <- !loci_with_zero_cov
            gr <- gr[loci_with_nonzero_cov]
            M <- M[loci_with_nonzero_cov, , drop = FALSE]
            Cov <- Cov[loci_with_nonzero_cov, , drop = FALSE]
            coef <- coef[loci_with_nonzero_cov, , drop = FALSE]
            se.coef <- se.coef[loci_with_nonzero_cov, , drop = FALSE]
            Filtered <- Filtered[loci_with_nonzero_cov, , drop = FALSE]
        }
    }

    # Construct BSseq object ---------------------------------------------------

    assays <- SimpleList(M = M, Cov = Cov, Filtered = Filtered,
                         coef = coef, se.coef = se.coef)
    assays <- assays[!S4Vectors:::sapply_isNULL(assays)]
    se <- SummarizedExperiment(
        assays = assays,
        rowRanges = loci,
        colData = pData)
    .BSseq(se, trans = trans, parameters = parameters)
}

# Move to BSseq-utils?
hasBeenSmoothed <- function(BSseq) {
    "coef" %in% assayNames(BSseq)
}

# Move to BSseq-utils?
getBSseq <- function(BSseq,
                     type = c("Cov", "M", "gr","coef", "se.coef","trans",
                              "parameters"),
                     withDimnames = TRUE) {
    type <- match.arg(type)
    if (type %in% c("M", "Cov")) {
        return(assay(BSseq, type, withDimnames = withDimnames))
    }
    if (type %in% c("coef", "se.coef")) {
        if (type %in% assayNames(BSseq)) {
            return(assay(BSseq, type, withDimnames = withDimnames))
        } else {
            return(NULL)
        }
    }
    if (type == "trans") {
        return(BSseq@trans)
    }
    if (type == "parameters") {
        return(BSseq@parameters)
    }
    if (type == "gr") {
        return(BSseq@rowRanges)
    }
}

# Move to BSseq-utils?
strandCollapse <- function(BSseq, shift = TRUE, BPPARAM = bpparam(),
                           BACKEND = getAutoRealizationBackend(),
                           dir = tempfile("BSseq"), replace = FALSE,
                           chunkdim = NULL, level = NULL,
                           type = c("double", "integer")) {

    # Argument checks ----------------------------------------------------------

    if (all(runValue(strand(BSseq)) == "*")) {
        warning("All loci are unstranded, nothing to collapse.", call. = FALSE)
        return(BSseq)
    }
    if (!(all(runValue(strand(BSseq)) %in% c("+", "-")))) {
        stop("'BSseq' object has a mix of stranded and unstranded loci.")
    }
    # Register 'BACKEND' and return to current value on exit.
    # TODO: Is this strictly necessary?
    current_BACKEND <- getAutoRealizationBackend()
    on.exit(setAutoRealizationBackend(current_BACKEND), add = TRUE)
    setAutoRealizationBackend(BACKEND)
    # Check compatability of 'BPPARAM' with 'BACKEND'.
    if (!.areBackendsInMemory(BACKEND)) {
        if (!.isSingleMachineBackend(BPPARAM)) {
            stop("The parallelisation strategy must use a single machine ",
                 "when using an on-disk realization backend.\n",
                 "See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    } else {
        if (!is.null(BACKEND)) {
            # NOTE: Currently do not support any in-memory realization
            #       backends. If the realization backend is NULL then an
            #       ordinary matrix is returned rather than a matrix-backed
            #       DelayedMatrix.
            stop("The '", BACKEND, "' realization backend is not supported.",
                 "\n  See help(\"read.bismark\") for details.",
                 call. = FALSE)
        }
    }
    # If using HDF5Array as BACKEND, check remaining options are sensible.
    if (identical(BACKEND, "HDF5Array")) {
        # NOTE: Most of this copied from
        #       HDF5Array::saveHDF5SummarizedExperiment().
        if (!isSingleString(dir)) {
            stop(wmsg("'dir' must be a single string specifying the path to ",
                      "the directory where to save the BSseq object (the ",
                      "directory will be created)."))
        }
        if (!isTRUEorFALSE(replace)) {
            stop("'replace' must be TRUE or FALSE")
        }
        if (!dir.exists(dir)) {
            HDF5Array::create_dir(dir)
        } else {
            HDF5Array::replace_dir(dir, replace)
        }
        h5_path <- file.path(dir, "assays.h5")
    } else if (identical(BACKEND, NULL)) {
        h5_path <- NULL
    }

    # Collapse loci ------------------------------------------------------------

    loci <- rowRanges(BSseq)
    if (shift) {
        loci <- shift(loci, shift = as.integer(-1L * (strand(loci) == "-")))
    }
    collapsed_loci <- reduce(loci, min.gapwidth = 0L, ignore.strand = TRUE)

    # Collapse 'M' and 'Cov' matrices ------------------------------------------

    ol <- findOverlaps(loci, collapsed_loci, type = "equal")
    group <- subjectHits(ol)
    M <- .rowsum(
        x = assay(BSseq, "M", withDimnames = FALSE),
        group = group,
        # NOTE: reorder = TRUE to ensure same row-order as collapsed_loci.
        reorder = TRUE,
        BPPARAM = BPPARAM,
        filepath = h5_path,
        name = "Cov",
        chunkdim = chunkdim,
        level = level,
        type = type)
    Cov <- .rowsum(
        x = assay(BSseq, "Cov", withDimnames = FALSE),
        group = group,
        # NOTE: reorder = TRUE to ensure same row-order as collapsed_loci.
        reorder = TRUE,
        BPPARAM = BPPARAM,
        filepath = h5_path,
        name = "Cov",
        chunkdim = chunkdim,
        level = level,
        type = type)

    # Construct BSseq object, saving it if it is HDF5-backed -------------------

    se <- SummarizedExperiment(
        assays = SimpleList(M = unname(M), Cov = unname(Cov)),
        rowRanges = collapsed_loci,
        colData = colData(BSseq))
    # TODO: Is there a way to use the internal constructor with `check = FALSE`?
    #       Assuming input was valid, the output is valid, too.
    # .BSseq(se, trans = function(x) NULL, parameters = list())
    bsseq <- new2("BSseq", se, check = FALSE)
    if (!is.null(BACKEND) && BACKEND == "HDF5Array") {
        # NOTE: Save BSseq object; mimicing
        #       HDF5Array::saveHDF5SummarizedExperiment().
        x <- bsseq
        x@assays <- HDF5Array::shorten_assay2h5_links(x@assays)
        base::saveRDS(x, file = file.path(dir, "se.rds"))
    }
    bsseq
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "BSseq"), function(object) {
    cat("An object of type 'BSseq' with\n")
    cat(" ", nrow(object), "methylation loci\n")
    cat(" ", ncol(object), "samples\n")
    if (hasBeenSmoothed(object)) {
        cat("has been smoothed with\n")
        cat(" ", object@parameters$smoothText, "\n")
    } else {
        cat("has not been smoothed\n")
    }
    if (.isHDF5ArrayBacked(object)) {
        cat("Some assays are HDF5Array-backed\n")
    } else {
        cat("All assays are in-memory\n")
    }
})

setMethod("pData", "BSseq", function(object) {
    object@colData
})

setReplaceMethod(
    "pData",
    signature = signature(object = "BSseq", value = "data.frame"),
    function(object, value) {
        colData(object) <- as(value, "DataFrame")
        object
    }
)

setReplaceMethod(
    "pData",
    signature = signature(object = "BSseq", value = "DataFrame"),
    function(object, value) {
        colData(object) <- value
        object
    }
)

setMethod("sampleNames", "BSseq", function(object) colnames(object))

setReplaceMethod(
    "sampleNames",
    signature = signature(object = "BSseq", value = "ANY"),
    function(object, value) {
        colnames(object) <- value
        object
    }
)

setMethod("updateObject", "BSseq",
          function(object, ...) {
              # NOTE: identical() is too strong
              if (hasBeenSmoothed(object) &&
                  isTRUE(all.equal(getBSseq(object, "trans"), .oldTrans))) {
                  object@trans <- plogis
              }
              if (is(object, "SummarizedExperiment")) {
                  # NOTE: Call method for SummarizedExperiment objects
                  object <- callNextMethod()
                  return(object)
              } else {
                  BSseq(
                      gr = object@gr,
                      M = object@M,
                      Cov = object@Cov,
                      coef = object@coef,
                      se.coef = object@se.coef,
                      trans = object@trans,
                      parameters = object@parameters,
                      pData = object@phenoData@data)
              }
          }
)

context("BSseq-class")

test_that("BSseq() allows loci of width != 1", {
    bsseq <- BSseq(
        M = matrix(1:10),
        Cov = matrix(1:10),
        gr = GRanges("chr1", IRanges(1:10, width = 10), strand = "*"))
    expect_true(validObject(bsseq))
})

test_that("strandCollapse() does nothing for unstranded data", {
    bsseq <- BSseq(
        M = matrix(1:10),
        Cov = matrix(1:10),
        gr = GRanges("chr1", IRanges(1:10, width = 1), strand = "*"))
    expect_warning(strandCollapse(bsseq))
})

test_that("strandCollapse() works on good input", {
    nrow <- 20
    ncol <- 5
    M <- matrix(sample(0:10, size = nrow * ncol, replace = TRUE), ncol = ncol)
    Cov <- M + sample(0:10, size = nrow * ncol, replace = TRUE)
    gr_pos <- GRanges(
        seqnames = c(rep(1, nrow / 2), rep(2, nrow / 2)),
        ranges = seq(1, 2 * nrow, by = 2),
        strand = "+")
    gr_neg <- invertStrand(shift(gr_pos, 1L))
    bsseq_pos <- BSseq(M = M, Cov = Cov, gr = gr_pos)
    bsseq_neg <- BSseq(M = M, Cov = Cov, gr = gr_neg)
    bsseq <- rbind(bsseq_pos, bsseq_neg)

    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_pos, type = "integer"),
        SE2 = unstrand(bsseq_pos))
    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_neg, shift = FALSE, type = "integer"),
        SE2 = unstrand(bsseq_neg))
    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_neg, type = "integer"),
        SE2 = unstrand(shift(bsseq_neg, -1L)))
    bsseq_strand_collapsed <- strandCollapse(bsseq, type = "integer")
    expect_identical(
        object = rowRanges(bsseq_strand_collapsed),
        expected = unstrand(rowRanges(bsseq_pos)))
    expect_identical(
        object = assay(bsseq_strand_collapsed, "M"),
        expected = assay(bsseq_pos, "M") + assay(bsseq_neg, "M"))
    expect_identical(
        object = assay(bsseq_strand_collapsed, "Cov"),
        expected = assay(bsseq_pos, "Cov") + assay(bsseq_neg, "Cov"))

    bsseq_strand_collapsed2 <- strandCollapse(bsseq[sample(nrow(bsseq))])
    expect_equivalent_SE(bsseq_strand_collapsed, bsseq_strand_collapsed2)

    bsseq_pos <- realize(bsseq_pos, "HDF5Array")
    bsseq_neg <- realize(bsseq_neg, "HDF5Array")
    bsseq <- realize(bsseq, "HDF5Array")
    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_pos, type = "integer"),
        SE2 = unstrand(bsseq_pos))
    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_neg, shift = FALSE),
        SE2 = unstrand(bsseq_neg))
    expect_equivalent_SE(
        SE1 = strandCollapse(bsseq_neg),
        SE2 = unstrand(shift(bsseq_neg, -1L)))
    bsseq_strand_collapsed <- strandCollapse(bsseq, type = "integer")
    expect_identical(
        object = rowRanges(bsseq_strand_collapsed),
        expected = unstrand(rowRanges(bsseq_pos)))
    expect_identical(
        object = as.matrix(assay(bsseq_strand_collapsed, "M")),
        expected = as.matrix(assay(bsseq_pos, "M") + assay(bsseq_neg, "M")))
    expect_identical(
        object = as.matrix(assay(bsseq_strand_collapsed, "Cov")),
        expected = as.matrix(assay(bsseq_pos, "Cov") + assay(bsseq_neg, "Cov")))

    bsseq_strand_collapsed2 <- strandCollapse(
        BSseq = bsseq[sample(nrow(bsseq))],
        type = "integer")
    expect_equivalent_SE(bsseq_strand_collapsed, bsseq_strand_collapsed2)
})

test_that("strandCollapse() errors on bad input", {
    bsseq <- BSseq(
        M = matrix(1:3),
        Cov = matrix(1:3),
        gr = GRanges(
            seqnames = "chr1",
            ranges = IRanges(1:3, width = 1),
            strand = c("+", "-", "*")))
    expect_error(strandCollapse(bsseq))
})

test_that("strandCollapse() will unstrand loci and may re-order them", {
    bsseq <- BSseq(
        M = matrix(1:10),
        Cov = matrix(1:10),
        gr = GRanges("chr1", IRanges(10:1, width = 1), strand = "+"))
    expect_true(all(strand(strandCollapse(bsseq)) == "*"))
    expect_false(expect_equivalent_SE(strandCollapse(bsseq), bsseq))
    expect_equivalent_SE(strandCollapse(bsseq), bsseq[10:1])
})

context("utils")

test_that("plogis()", {
    set.seed(666)
    x <- seq(-100, 100, 0.01)
    x <- x + runif(length(x), -0.001, 0.001)
    expect_equal(plogis(x), bsseq:::.oldTrans(x))
    # NOTE: There was a bug in .oldTrans() when x = 0 that mean it returned 0
    #       instead of 0.5
    expect_equal(bsseq:::.oldTrans(0), 0)
    expect_equal(plogis(0), 0.5)
})

context(".checkMandCov")

test_that(".checkMandCov() works", {
    # Create data
    nrow <- 10L
    ncol <- 2L
    M <- matrix(seq.int(nrow * ncol), nrow = nrow, ncol = ncol)
    Cov <- M + 1L
    M_with_NA_at_start <- M
    M_with_NA_at_start[1L] <- NA_integer_
    Cov_with_NA_at_start <- M
    Cov_with_NA_at_start[1L] <- NA_integer_
    M_with_NA_at_end <- M
    M_with_NA_at_end[nrow * ncol] <- NA_integer_
    Cov_with_NA_at_end <- M
    Cov_with_NA_at_end[nrow * ncol] <- NA_integer_
    M_with_impossible_value_at_start <- M
    M_with_impossible_value_at_start[1L] <- Cov[1L] + 2L
    M_with_impossible_value_at_end <- M
    M_with_impossible_value_at_end[nrow * ncol] <- Cov[nrow * ncol] + 2L
    Cov_with_impossible_value_at_start <- Cov
    Cov_with_impossible_value_at_start[1L] <- M[1L] - 2L
    Cov_with_impossible_value_at_end <- Cov
    Cov_with_impossible_value_at_end[nrow * ncol] <- M[nrow * ncol] - 2L

    # Returns NULL on valid input
    expect_identical(.checkMandCov(M, Cov), NULL)
    expect_identical(.checkMandCov(DelayedArray(M), DelayedArray(Cov)), NULL)
    # NOTE: No check is made that input is strictly integer.
    expect_identical(.checkMandCov(M + 0.1, Cov), NULL)
    expect_identical(.checkMandCov(M, Cov + 0.1), NULL)
    # NOTE: Recall that Matrix::Matrix() stores all data using doubles
    expect_identical(.checkMandCov(Matrix::Matrix(M), Cov), NULL)
    expect_identical(.checkMandCov(M, Matrix::Matrix(Cov)), NULL)

    # Returns msg if anyNA(M) || anyNA(Cov)
    # NOTE: Returns a character vector rather than message() because
    #       .checkMandCov() is used within a validity method.
    expect_identical(.checkMandCov(M_with_NA_at_start, Cov),
                     "'M' must not contain NAs.")
    expect_identical(.checkMandCov(DelayedArray(M_with_NA_at_start), Cov),
                     "'M' must not contain NAs.")
    expect_identical(.checkMandCov(M_with_NA_at_end, Cov),
                     "'M' must not contain NAs.")
    expect_identical(.checkMandCov(DelayedArray(M_with_NA_at_end), Cov),
                     "'M' must not contain NAs.")
    expect_identical(.checkMandCov(M, Cov_with_NA_at_start),
                     "'Cov' must not contain NAs.")
    expect_identical(.checkMandCov(M, DelayedArray(Cov_with_NA_at_start)),
                     "'Cov' must not contain NAs.")
    expect_identical(.checkMandCov(M, Cov_with_NA_at_end),
                     "'Cov' must not contain NAs.")
    expect_identical(.checkMandCov(M, DelayedArray(Cov_with_NA_at_end)),
                     "'Cov' must not contain NAs.")

    # Returns msg if any(M > Cov)
    msg <- "All values of 'M' must be less than or equal to the corresponding value of 'Cov'."
    expect_identical(.checkMandCov(M_with_impossible_value_at_start, Cov), msg)
    expect_identical(
        .checkMandCov(DelayedArray(M_with_impossible_value_at_start), Cov),
        msg)
    expect_identical(.checkMandCov(M_with_impossible_value_at_end, Cov), msg)
    expect_identical(
        .checkMandCov(DelayedArray(M_with_impossible_value_at_end), Cov),
        msg)
    expect_identical(.checkMandCov(M, Cov_with_impossible_value_at_start), msg)
    expect_identical(
        .checkMandCov(M, DelayedArray(Cov_with_impossible_value_at_start)),
        msg)
    expect_identical(.checkMandCov(M, Cov_with_impossible_value_at_end), msg)
    expect_identical(
        .checkMandCov(M, DelayedArray(Cov_with_impossible_value_at_end)),
        msg)
})

# Internal functions -----------------------------------------------------------

.rowTickmarks <- function(hasGRanges, maxGap) {
    gr <- granges(hasGRanges)
    # NOTE: This relies on 'gr' being sorted, so really want to be sure of this!
    stopifnot(!is.unsorted(gr))
    clusters <- reduce(gr, min.gapwidth = maxGap, with.revmap = TRUE)
    cumsum(lengths(mcols(clusters)[["revmap"]]))
}

# TODO: Remove other uses of this function, if possible.
# TODO: Rename to .makeClusters() since it's an internal function
makeClusters <- function(hasGRanges, maxGap = 10^8) {
    chrOrder <- as.character(runValue(seqnames(hasGRanges)))
    if(anyDuplicated(chrOrder))
        stop("argument 'hasGRanges' is not properly order")
    grBase <- granges(hasGRanges)
    clusters <- reduce(resize(grBase, width = 2*maxGap + 1, fix = "center"))
    start(clusters) <- pmax(rep(1, length(clusters)), start(clusters))
    clusters.sp <- split(clusters, seqnames(clusters))
    stopifnot(all(sapply(clusters.sp, function(cluster.gr) {
        if(length(cluster.gr) <= 1) return(TRUE)
        all(start(cluster.gr)[-length(cluster.gr)] < end(cluster.gr)[-1])
    }))) # are the clusters ordered within the chromosome? This is probably guranteed
    clusters <- Reduce(c, clusters.sp[chrOrder])
    stopifnot(all(chrOrder == runValue(seqnames(clusters))))
    ov <- findOverlaps(grBase, clusters)
    clusterIdx <- split(as.matrix(ov)[,1], as.matrix(ov)[,2])
    names(clusterIdx) <- NULL
    clusterIdx
}

.BSmooth <- function(b, M, Cov, pos, coef_sink, se.coef_sink, sink_lock, grid,
                     pos_grid, ns, h, keep.se) {
    # Load packages on worker (required for SnowParam) -------------------------

    suppressPackageStartupMessages({
        requireNamespace("BiocParallel")
    })
    suppressPackageStartupMessages({
        requireNamespace("locfit")
    })
    suppressPackageStartupMessages({
        requireNamespace("DelayedArray")
    })

    # Construct inputs for smoothing -------------------------------------------

    # NOTE: 'bb' indexes over elements of pos_grid by cycling `ncol(M)` times
    #       over 1, ..., nrow(pos_grid).
    bb <- (b - 1L) %% nrow(pos_grid) + 1L
    # NOTE: unname() is necessary because M and Cov may carry colnames
    sdata <- data.frame(
        pos = read_block(pos, pos_grid[[bb]]),
        M = unname(read_block(M, grid[[b]])),
        Cov = unname(read_block(Cov, grid[[b]])))
    # Ensure 0 < M < Cov to avoid boundary issues (only relevant at loci with
    # non-zero coverage, so doesn't matter what M is for loci with zero
    # coverage).
    sdata[["M"]] <- pmin(pmax(sdata[["M"]], 0.01), pmax(sdata[["Cov"]] - 0.01))
    n_loci <- nrow(sdata)
    n_loci_with_non_zero_coverage <- sum(sdata[["Cov"]] > 0)

    # Fit local binomial regression model --------------------------------------

    # NOTE: Require (ns + 1) loci with non-zero coverage to run smooth.
    if (n_loci_with_non_zero_coverage <= ns) {
        coef <- rep(NA_real_, n_loci)
        if (keep.se) {
            se.coef <- rep(NA_real_, n_loci)
        } else {
            se.coef <- NULL
        }
    } else {
        # Set 'nearest neighbour' smoothing parameter.
        nn <- ns / n_loci_with_non_zero_coverage
        # Fit model only using loci with non-zero coverage.
        fit <- locfit(
            M ~ lp(pos, nn = nn, h = h),
            data = sdata,
            weights = Cov,
            family = "binomial",
            subset = Cov > 0,
            maxk = 10000)
        # Evaluate smooth at all loci (regardless of coverage).
        pp <- preplot(
            object = fit,
            newdata = sdata["pos"],
            band = "local")
        coef <- pp$fit
        if (keep.se) {
            se.coef <- pp$se.fit
        } else {
            se.coef <- NULL
        }
    }

    # Return the results of the smooth or write them to the RealizationSink ----

    if (is.null(coef_sink)) {
        return(list(coef = coef, se.coef = se.coef))
    }
    # Write to coef_sink and se.coef_sink while respecting the IPC lock.
    ipclock(sink_lock)
    write_block(coef_sink, viewport = grid[[b]], block = as.matrix(coef))
    if (keep.se) {
        write_block(
            se.coef_sink,
            viewport = grid[[b]],
            block = as.matrix(se.coef))
    }
    ipcunlock(sink_lock)
    NULL
}

# Exported functions -----------------------------------------------------------

# TODO: verbose = TRUE should report timings.
# TODO: BSmooth() should warn if BSseq object contains mix of strands.
# TODO: Consider having BSmooth() create a 'smoothed' assay in addition to or
#       instead of the 'coef' and 'se.coef' assays.
BSmooth <- function(BSseq,
                    ns = 70,
                    h = 1000,
                    maxGap = 10^8,
                    keep.se = FALSE,
                    BPPARAM = bpparam(),
                    chunkdim = NULL,
                    level = NULL,
                    verbose = getOption("verbose")) {

    # Argument checks-----------------------------------------------------------

    # Check validity of 'BSseq' argument
    if (!is(BSseq, "BSseq")) {
        stop("'BSseq' must be a BSseq object.")
    }
    if (!isTRUE(all(width(BSseq) == 1L))) {
        stop("All loci in 'BSseq' must have width == 1.")
    }
    if (is.unsorted(BSseq)) {
        stop("'BSseq' must be sorted before smoothing. Use 'sort(BSseq)'.")
    }
    stopifnot(isTRUEorFALSE(keep.se))
    # Set appropriate BACKEND and check compatability with BPPARAM.
    BACKEND <- .getBSseqBackends(BSseq)
    if (!.areBackendsInMemory(BACKEND)) {
        if (!.isSingleMachineBackend(BPPARAM)) {
            stop("The parallelisation strategy must use a single machine ",
                 "when using an on-disk realization backend.\n",
                 "See help(\"BSmooth\") for details.",
                 call. = FALSE)
        }
    } else {
        if (!is.null(BACKEND)) {
            # NOTE: Currently do not support any in-memory realization
            #       backends. If 'BACKEND' is NULL then an ordinary matrix
            #       is returned rather than a matrix-backed DelayedMatrix.
            stop("The '", BACKEND, "' realization backend is not ",
                 "supported.\n",
                 "See help(\"BSmooth\") for details.",
                 call. = FALSE)
        }
    }
    # If using HDF5Array, check if BSseq object is updateable.
    if (identical(BACKEND, "HDF5Array")) {
        is_BSseq_updateable <- .isHDF5BackedBSseqUpdatable(BSseq)
        if (is_BSseq_updateable) {
            h5_path <- path(assay(BSseq, withDimnames = FALSE))
            if (any(c("coef", "se.coef") %in% rhdf5::h5ls(h5_path)[, "name"])) {
                # TODO: Better error message; what should be done in this case?
                stop("The HDF5 file '", h5_path, "' already contains a ",
                          "dataset named 'coef' or 'se.coef'.")
            }
        } else {
            h5_path <- NULL
            warning(
                wmsg("'BSseq' was not created with either read.bismark() or ",
                     "HDF5Array::saveHDF5SummarizedExperiment(). BSmooth() is ",
                     "using automatically created HDF5 file(s) (see ",
                     "?HDF5Array::setHDF5DumpFile) to store smoothing result. ",
                     "You will need to run ",
                     "HDF5Array::saveHDF5SummarizedExperiment() on the ",
                     "returned object if you wish to save the returned ",
                     "object."))
        }
    }
    stopifnot(isTRUEorFALSE(verbose))

    # Smoothing ----------------------------------------------------------------

    # Extract pieces of BSseq object required for smoothing
    M <- getCoverage(BSseq, type = "M", withDimnames = FALSE)
    Cov <- getCoverage(BSseq, type = "Cov", withDimnames = FALSE)
    pos <- matrix(start(BSseq), ncol = 1)
    # Set up ArrayGrid so that each block contains data for a single sample and
    # single cluster of loci.
    row_tickmarks <- .rowTickmarks(BSseq, maxGap)
    col_tickmarks <- seq_len(ncol(M))
    grid <- ArbitraryArrayGrid(list(row_tickmarks, col_tickmarks))
    # Set up "parallel" ArrayGrid over pos
    pos_grid <- ArbitraryArrayGrid(list(row_tickmarks, 1L))
    # Construct RealizationSink objects (as required)
    if (is.null(BACKEND)) {
        coef_sink <- NULL
        se.coef_sink <- NULL
        sink_lock <- NULL
    } else if (BACKEND == "HDF5Array") {
        coef_sink <- HDF5RealizationSink(
            dim = dim(M),
            dimnames = NULL,
            type = "double",
            filepath = h5_path,
            name = "coef",
            chunkdim = chunkdim,
            level = level)
        on.exit(close(coef_sink), add = TRUE)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)
        if (keep.se) {
            se.coef_sink <- HDF5RealizationSink(
                dim = dim(M),
                dimnames = NULL,
                type = "double",
                filepath = h5_path,
                name = "se.coef",
                chunkdim = chunkdim,
                level = level)
            on.exit(close(se.coef_sink), add = TRUE)
        } else {
            se.coef_sink <- NULL
        }
    } else {
        # TODO: This branch should probably never be entered because we
        #       (implicitly) only support in-memory or HDF5Array backends.
        #       However, we retain it for now (e.g., fstArray backend would
        #       use this until a dedicated branch was implemented).
        coef_sink <- DelayedArray::AutoRealizationSink(dim(M), type = "double")
        on.exit(close(coef_sink), add = TRUE)
        sink_lock <- ipcid()
        on.exit(ipcremove(sink_lock), add = TRUE)
        if (keep.se) {
            se.coef_sink <- DelayedArray::AutoRealizationSink(
                dim(M),
                type = "double")
            on.exit(close(se.coef_sink), add = TRUE)
        } else {
            se.coef_sink <- NULL
        }
    }

    # Set number of tasks to ensure the progress bar gives frequent updates.
    # NOTE: The progress bar increments once per task
    #       (https://github.com/Bioconductor/BiocParallel/issues/54).
    #       Although it is somewhat of a bad idea to overrides a user-specified
    #       bptasks(BPPARAM), the value of bptasks(BPPARAM) doesn't affect
    #       performance in this instance, and so we opt for a useful progress
    #       bar. Only SnowParam (and MulticoreParam by inheritance) have a
    #       bptasks<-() method.
    if (is(BPPARAM, "SnowParam") && bpprogressbar(BPPARAM)) {
        bptasks(BPPARAM) <- length(grid)
    }
    smooth <- bptry(bplapply(
        X = seq_along(grid),
        FUN = .BSmooth,
        M = M,
        Cov = Cov,
        pos = pos,
        coef_sink = coef_sink,
        se.coef_sink = se.coef_sink,
        sink_lock = sink_lock,
        grid = grid,
        pos_grid = pos_grid,
        ns = ns,
        h = h,
        keep.se = keep.se,
        BPPARAM = BPPARAM))
    if (!all(bpok(smooth))) {
        stop("BSmooth() encountered errors: ",
             sum(!bpok(smooth)), " of ", length(smooth),
             " smoothing tasks failed.")
    }
    # Construct coef and se.coef from results of smooth().
    if (is.null(BACKEND)) {
        # Returning matrix objects.
        coef <- do.call(c, lapply(smooth, "[[", "coef"))
        attr(coef, "dim") <- dim(M)
        if (keep.se) {
            se.coef <- do.call(c, lapply(smooth, "[[", "se.coef"))
            attr(se.coef, "dim") <- dim(M)
        } else {
            se.coef <- NULL
        }
    } else {
        # Returning DelayedMatrix objects.
        coef <- as(coef_sink, "DelayedArray")
        if (keep.se) {
            se.coef <- as(se.coef_sink, "DelayedArray")
        } else {
            se.coef <- NULL
        }
    }

    # Construct BSseq object, saving it if it is HDF5-backed -------------------

    # NOTE: Using BiocGenerics:::replaceSlots(..., check = FALSE) to avoid
    #       triggering the validity method.
    assays <- c(assays(BSseq, withDimnames = FALSE)[c("M", "Cov")],
                SimpleList(coef = coef))
    if (keep.se) {
        assays <- c(assays, SimpleList(se.coef = se.coef))
    }
    BSseq <- BiocGenerics:::replaceSlots(
        object = BSseq,
        assays = Assays(assays),
        trans = plogis,
        parameters = list(
            smoothText = sprintf(
                "BSmooth (ns = %d, h = %d, maxGap = %d)", ns, h, maxGap),
            ns = ns,
            h = h,
            maxGap = maxGap),
        check = FALSE)
    if (identical(BACKEND, "HDF5Array") && is_BSseq_updateable) {
        # NOTE: Save BSseq object; mimicing
        #       HDF5Array::saveHDF5SummarizedExperiment().
        dir <- dirname(h5_path)
        x <- BSseq
        x@assays <- HDF5Array::shorten_assay2h5_links(x@assays)
        base::saveRDS(x, file = file.path(dir, "se.rds"))
    }
    BSseq
}

# TODOs ------------------------------------------------------------------------

# TODO: Use the logging facilities of BiocParallel. This is a longterm goal.
#       For example, we could set custom messages within .BSmooth() using the
#       futile.logger syntax; see the BiocParalell vignette 'Errors, Logs and
#       Debugging in BiocParallel'.

#-------------------------------------------------------------------------------
# Is a DelayedMatrix object (or the assays of a SummarizedExperiment object)
# backed by a HDF5 file?
#

.getSeedClasses <- function(seed) {
    if (is(seed, "DelayedOp")) {
        seeds <- try(seed@seeds, silent = TRUE)
        if (is(seeds, "try-error")) {
            seed <- seed@seed
            return(.getSeedClasses(seed))
        }
        return(lapply(seeds, .getSeedClasses))
    } else if (is(seed, "DelayedArray")) {
        # A DelayedArray can have another DelayedArray as a seed
        seed <- seed@seed
        return(.getSeedClasses(seed))
    }
    else {
        # Pick the first element returned by class() (starting with R 4.0,
        # 'class(matrix())' is 'c("matrix", "array")').
        class(seed)[[1L]]
    }
}

# NOTE: Returns TRUE if *any* assay is HDF5Array-backed and FALSE if *all*
#       assays are not HDF5Array-backed
.isHDF5ArrayBacked <- function(object) {
    if (is(object, "SummarizedExperiment")) {
        return(all(vapply(X = assays(object, withDimnames = FALSE),
                          FUN = .isHDF5ArrayBacked,
                          FUN.VALUE = logical(1L))))
    }
    if (is(object, "DelayedArray")) {
        seed <- object@seed
        seed_classes <- .getSeedClasses(seed)
        is_hdf5_backed <- vapply(unlist(seed_classes, use.names = FALSE),
                                 extends, class2 = "HDF5ArraySeed",
                                 logical(1L))
        return(any(is_hdf5_backed))
    } else if (is.matrix(object)) {
        FALSE
    } else if (is.null(object)) {
        FALSE
    } else {
        stop("Don't know how to handle object of class ", class(object))
    }
}

test_that(".isHDF5ArrayBacked()", {
    matrix <- matrix(1:10, ncol = 2)
    da <- DelayedArray(matrix)
    ha <- realize(da, BACKEND = "HDF5Array")
    ha_ha <- rbind(ha, ha)
    da_da <- rbind(da, da)
    ha_ha_ha_ha <- rbind(ha_ha, ha_ha)
    da_da_da_da <- rbind(da_da, da_da)
    da_da_ha_ha <- rbind(da_da, ha_ha)

    expect_true(!bsseq:::.isHDF5ArrayBacked(matrix))
    expect_true(bsseq:::.isHDF5ArrayBacked(ha))
    expect_true(!bsseq:::.isHDF5ArrayBacked(da))
    expect_true(bsseq:::.isHDF5ArrayBacked(ha_ha))
    expect_true(!bsseq:::.isHDF5ArrayBacked(da_da))
    expect_true(bsseq:::.isHDF5ArrayBacked(ha_ha_ha_ha))
    expect_true(!bsseq:::.isHDF5ArrayBacked(da_da_da_da))
    expect_true(bsseq:::.isHDF5ArrayBacked(da_da_ha_ha))

    # A complicated DelayedArray with another DelayedArray as a seed
    skip("TODO: Remove these tests if no longer required")
    da_with_da_seed <- bsseq:::.collapseDelayedMatrix(x = da,
                                                      sp = list(1:2, 3:5),
                                                      MARGIN = 2)
    ha_with_ha_seed <- bsseq:::.collapseDelayedMatrix(
        x = ha,
        sp = list(1:2, 3:5),
        MARGIN = 2,
        BACKEND = "HDF5Array")
    expect_true(!bsseq:::.isHDF5ArrayBacked(da_with_da_seed))
    expect_true(bsseq:::.isHDF5ArrayBacked(ha_with_ha_seed))
})

# Internal classes -------------------------------------------------------------

# The FWGRanges class is a container for the fixed-width genomic locations and
# their associated annotations.
# NOTE: The intention is to make this a fully-fledged GenomicRanges subclass
#       that is part of the GenomicRRanges package. For now, this class is
#       internal to bsseq and only a subset of methods are properly implemented.
.FWGRanges <- setClass(
    "FWGRanges",
    contains = "GenomicRanges",
    representation(
        seqnames = "Rle",
        ranges = "FWIRanges",
        strand = "Rle",
        seqinfo = "Seqinfo"
    ),
    prototype(
        seqnames = Rle(factor()),
        strand = Rle(strand())
    )
)

# Internal methods -------------------------------------------------------------

# NOTE: Combine the new "parallel slots" with those of the parent class. Make
#       sure to put the new parallel slots **first**. See R/Vector-class.R file
#       in the S4Vectors package for what slots should or should not be
#       considered "parallel".
setMethod(
    "parallel_slot_names",
    "FWGRanges",
    function(x) {
        c(GenomicRanges:::extraColumnSlotNames(x), "seqnames", "ranges",
          "strand", callNextMethod())
    }
)

setMethod("seqnames", "FWGRanges", function(x) x@seqnames)

setMethod("strand", "FWGRanges", function(x) x@strand)

setMethod("seqinfo", "FWGRanges", function(x) x@seqinfo)

setMethod(
    "ranges",
    "FWGRanges",
    function(x, use.names = TRUE, use.mcols = FALSE) {
        if (!isTRUEorFALSE(use.names)) stop("'use.names' must be TRUE or FALSE")
        if (!isTRUEorFALSE(use.mcols)) stop("'use.mcols' must be TRUE or FALSE")
        # TODO: This is a work-around the requirement in the GenomicRanges
        #       validity method that ranges(GenomicRanges) returns a IRanges or
        #       IPos instance.
        # ans <- x@ranges
        ans <- as(x@ranges, "IRanges")
        if (!use.names) names(ans) <- NULL
        if (use.mcols) mcols(ans) <- mcols(x)
        ans
    }
)

# TODO: These wouldn't be necessary if ranges(FWGRanges) returned a FWIRanges
#       instead of an IRanges.
setMethod("start", "FWGRanges", function(x) start(x@ranges))
setMethod("width", "FWGRanges", function(x) width(x@ranges))
setMethod("end", "FWGRanges", function(x) {
    if (x@ranges@width == 1L) {
        start(x)
    } else {
        width(x) - 1L + start(x)
    }
})

# TODO: Follow shift,GRanges-method
setMethod("shift", "FWGRanges", function(x, shift = 0L, use.names = TRUE) {
    stopifnot(use.names)
    new_ranges <- shift(x@ranges, shift, use.names)
    x@ranges <- new_ranges
    validObject(x)
    x
})

# NOTE: Needed for seqinfo<-,FWGRanges-method
setMethod("update", "FWGRanges", function(object, ...) {
    BiocGenerics:::replaceSlots(object, ...)
})

# NOTE: This is pieced together in order to get something up and running. It
#       should, however, be possible to this 'properly' via inheritance. Most
#       of this hack is to work around the fact that ranges(FWGRanges)
#       currently has to return a IRanges instance instead of a FWIRanges
#       instance.
.findOverlaps_FWGRanges <- function(query, subject, maxgap = -1L,
                                    minoverlap = 0L,
                                    type = c("any", "start", "end", "within",
                                             "equal"),
                                    select = c("all", "first", "last",
                                               "arbitrary"),
                                    ignore.strand = FALSE) {
    # findOverlaps,GenomicRanges,GenomicRanges-method ----------------------
    type <- match.arg(type)
    select <- match.arg(select)

    # GenomicRanges:::findOverlaps_GNCList() -------------------------------
    if (!(is(query, "FWGRanges") && is(subject, "FWGRanges"))) {
        stop("'query' and 'subject' must be FWGRanges objects")
    }
    if (!isTRUEorFALSE(ignore.strand)) {
        stop("'ignore.strand' must be TRUE or FALSE")
    }

    si <- merge(seqinfo(query), seqinfo(subject))
    q_seqlevels <- seqlevels(query)
    s_seqlevels <- seqlevels(subject)
    common_seqlevels <- intersect(q_seqlevels, s_seqlevels)
    NG <- length(common_seqlevels)
    q_group_idx <- match(common_seqlevels, q_seqlevels)
    s_group_idx <- match(common_seqlevels, s_seqlevels)

    # Extract 'q_groups' and 's_groups' (both of length NG).
    q_groups <- GenomicRanges:::.extract_groups_from_GenomicRanges(
        query)[q_group_idx]
    s_groups <- GenomicRanges:::.extract_groups_from_GenomicRanges(
        subject)[s_group_idx]

    # Extract 'nclists' and 'nclist_is_q' (both of length NG).
    if (is(subject, "GNCList")) {
        nclists <- subject@nclists[s_group_idx]
        nclist_is_q <- rep.int(FALSE, NG)
    } else if (is(query, "GNCList")) {
        nclists <- query@nclists[q_group_idx]
        nclist_is_q <- rep.int(TRUE, NG)
    } else {
        # We'll do "on-the-fly preprocessing".
        nclists <- vector(mode = "list", length = NG)
        nclist_is_q <- rep.int(NA, NG)
    }

    # Extract 'circle_length' (of length NG).
    circle_length <- GenomicRanges:::.get_circle_length(si)[q_group_idx]

    ##Extract 'q_space' and 's_space'.
    if (ignore.strand) {
        q_space <- s_space <- NULL
    } else {
        q_space <- as.integer(strand(query)) - 3L
        s_space <- as.integer(strand(subject)) - 3L
    }

    # IRanges:::find_overlaps_in_groups_NCList() ---------------------------
    q <- query@ranges
    s <- subject@ranges
    circle.length <- circle_length
    if (!(is(q, "IntegerRanges") && is(s, "IntegerRanges"))) {
        stop("'q' and 's' must be IntegerRanges object")
    }
    if (!is(q_groups, "CompressedIntegerList")) {
        stop("'q_groups' must be a CompressedIntegerList object")
    }
    if (!is(s_groups, "CompressedIntegerList")) {
        stop("'s_groups' must be a CompressedIntegerList object")
    }

    if (!isSingleNumber(maxgap)) {
        stop("'maxgap' must be a single integer")
    }
    if (!is.integer(maxgap)) {
        maxgap <- as.integer(maxgap)
    }

    if (!isSingleNumber(minoverlap)) {
        stop("'minoverlap' must be a single integer")
    }
    if (!is.integer(minoverlap)) {
        minoverlap <- as.integer(minoverlap)
    }

    q_circle_len <- circle.length
    q_circle_len[which(nclist_is_q)] <- NA_integer_
    q <- IRanges:::.shift_ranges_in_groups_to_first_circle(
        x = q,
        x_groups = q_groups,
        circle.length = q_circle_len)
    s_circle_len <- circle.length
    s_circle_len[which(!nclist_is_q)] <- NA_integer_
    s <- IRanges:::.shift_ranges_in_groups_to_first_circle(
        x = s,
        x_groups = s_groups,
        circle.length = s_circle_len)
    .Call2("C_find_overlaps_in_groups_NCList",
           start(q), end(q), q_space, q_groups,
           start(s), end(s), s_space, s_groups,
           nclists, nclist_is_q,
           maxgap, minoverlap, type, select, circle.length,
           PACKAGE = "IRanges")
}

setMethod("findOverlaps", c("FWGRanges", "FWGRanges"), .findOverlaps_FWGRanges)

# TODO: A dedicated duplicated method would be faster than
#       duplicated,GenomicRanges, because it could use end(x) instead of
#       width(x) in call to duplicatedIntegerQuads().

# Internal functions -----------------------------------------------------------

# TODO: Warn if x contains mix of + and * or - and * loci?
.strandCollapse <- function(x) {
    stopifnot(is(x, "GenomicRanges"))
    if (all(strand(x) == "*")) return(x)
    # Shift loci on negative strand by 1 to the left
    x[strand(x) == "-"] <- IRanges::shift(x[strand(x) == "-"], -1L)
    # Unstrand all loci
    x <- unstrand(x)
    # Extract unique loci
    unique(x)
}

# TODOs ------------------------------------------------------------------------

# TODO: Document internal classes, methods, and functions for my own sanity.
#       Also, some may be useful to users of bsseq (although I still won't
#       export these for the time being). Ultimately, I'd like to promote
#       FWGRanges and FWIRanges to 'official' GenomicRanges and IntegerRanges
#       subclasses.

setClassUnion("matrixOrNULL", c("matrix", "NULL"))

setClass("hasGRanges",
         representation(gr = "GRanges"))

setMethod("seqnames", signature(x = "hasGRanges"), function(x) {
    seqnames(x@gr)
})
setReplaceMethod("seqnames", "hasGRanges", function(x, value) {
    seqnames(x@gr) <- value
    x
})

setMethod("seqlevels", signature(x = "hasGRanges"), function(x) {
    seqlevels(x@gr)
})
setReplaceMethod("seqlevels", "hasGRanges", function(x, value) {
    seqlevels(x@gr) <- value
    x
})

setMethod("seqlengths", signature(x = "hasGRanges"), function(x) {
    seqlengths(x@gr)
})
setReplaceMethod("seqlengths", "hasGRanges", function(x, value) {
    seqlengths(x@gr) <- value
    x
})

setMethod("granges", signature(x = "hasGRanges"),
    function(x, use.names = TRUE, use.mcols = FALSE, ...)
        granges(x@gr, use.names = use.names, use.mcols = use.mcols, ...)
)

## FIXME: might want a granges replacement function

setMethod("start", "hasGRanges", function(x, ...) {
    start(x@gr, ...)
})
setReplaceMethod("start", "hasGRanges", function(x, check = TRUE, value) {
    start(x@gr, check = check) <- value
    x
})

setMethod("end", "hasGRanges", function(x, ...) {
    end(x@gr, ...)
})
setReplaceMethod("end", "hasGRanges", function(x, check = TRUE, value) {
    end(x@gr, check = check) <- value
    x
})

setMethod("width", "hasGRanges", function(x) {
    width(x@gr)
})
setReplaceMethod("width", "hasGRanges", function(x, check = TRUE, value) {
    width(x@gr, check = check) <- value
    x
})

setMethod("strand", "hasGRanges", function(x) {
    strand(x@gr)
})
setReplaceMethod("strand", "hasGRanges", function(x, value) {
    strand(x@gr) <- value
    x
})

setMethod("length", "hasGRanges", function(x) length(x@gr))

setMethod("[", "hasGRanges", function(x, i, ...) {
    if(missing(i))
        stop("need [i] for subsetting")
    if(missing(i))
        return(x)
    x@gr <- x@gr[i]
    x
})

setMethod("findOverlaps",
          signature(query = "hasGRanges", subject = "GenomicRanges"),
          function (query, subject, maxgap = -1L, minoverlap = 0L,
                    type = c("any", "start", "end", "within", "equal"),
                    select = c("all", "first", "last", "arbitrary"),
                    ignore.strand = FALSE, ...) {
              findOverlaps(query = query@gr, subject = subject,
                           maxgap = maxgap, minoverlap = minoverlap,
                           type = match.arg(type), select = match.arg(select),
                           ignore.strand = ignore.strand, ...)
          })

setMethod("findOverlaps",
          signature(query = "GenomicRanges", subject = "hasGRanges"),
          function (query, subject, maxgap = -1L, minoverlap = 0L,
                    type = c("any", "start", "end", "within", "equal"),
                    select = c("all", "first", "last", "arbitrary"),
                    ignore.strand = FALSE, ...) {
              findOverlaps(query = query, subject = subject@gr,
                           maxgap = maxgap, minoverlap = minoverlap,
                           type = match.arg(type), select = match.arg(select),
                           ignore.strand = ignore.strand, ...)
          })

setMethod("findOverlaps",
          signature(query = "hasGRanges", subject = "hasGRanges"),
          function (query, subject, maxgap = -1L, minoverlap = 0L,
                    type = c("any", "start", "end", "within", "equal"),
                    select = c("all", "first", "last", "arbitrary"),
                    ignore.strand = FALSE, ...) {
              findOverlaps(query = query@gr, subject = subject@gr,
                           maxgap = maxgap, minoverlap = minoverlap,
                           type = match.arg(type), select = match.arg(select),
                           ignore.strand = ignore.strand, ...)
          })

setMethod("overlapsAny", c("hasGRanges", "GenomicRanges"),
    IRanges:::default_overlapsAny
)
setMethod("overlapsAny", c("GenomicRanges", "hasGRanges"),
    IRanges:::default_overlapsAny
)
setMethod("overlapsAny", c("hasGRanges", "hasGRanges"),
    IRanges:::default_overlapsAny
)

setMethod("subsetByOverlaps", c("hasGRanges", "GenomicRanges"),
    IRanges:::default_subsetByOverlaps
)
setMethod("subsetByOverlaps", c("GenomicRanges", "hasGRanges"),
    IRanges:::default_subsetByOverlaps
)
setMethod("subsetByOverlaps", c("hasGRanges", "hasGRanges"),
    IRanges:::default_subsetByOverlaps
)

read.umtab2 <- function(dirs, sampleNames = NULL, rmZeroCov = FALSE,
                        readCycle = FALSE, keepFilt = FALSE,
                        pattern = NULL, keepU, keepM, verbose = TRUE) {
    filesPerDir <- lapply(dirs, function(xx) sort(list.files(xx, pattern = pattern, full.names = TRUE)))
    if(!all(sapply(filesPerDir, function(xx) all(basename(xx) == basename(filesPerDir[[1]])))))
        warning("'dirs' does not contain the exact same file names")
    allChrs <- as.list(seq_along(filesPerDir[[1]]))
    for(ii in seq_along(filesPerDir[[1]])) {
        chrRead <- read.umtab2.chr(sapply(filesPerDir, function(xx) xx[ii]),
                                   keepM = keepM, keepU = keepU, sampleNames = sampleNames,
                                   readCycle = readCycle, keepFilt = keepFilt,
                                   verbose = verbose)
        if(rmZeroCov && length(chrRead) != 0){
            wh <- which(rowSums2(chrRead$U + chrRead$M) > 0)
            chrRead$M <- chrRead$M[wh,]
            chrRead$U <- chrRead$U[wh,]
            if(readCycle) {
                chrRead$Mcy <- chrRead$Mcy[wh,]
                chrRead$Ucy <- chrRead$Ucy[wh,]
            }
            chrRead$chr <- chrRead$chr[wh]
            chrRead$pos <- chrRead$pos[wh]
            if(keepFilt) {
                chrRead$filt_cycle <- chrRead$filt_cycle[wh,]
                chrRead$filt_allele <- chrRead$filt_allele[wh,]
                chrRead$filt_mapq <- chrRead$filt_mapq[wh,]
                chrRead$filt_baseq <- chrRead$filt_baseq[wh,]
            }
        }
        allChrs[[ii]] <- chrRead
        print(gc())
    }
    chr <- do.call(c, lapply(allChrs, function(xx) xx$chr))
    pos <- do.call(c, lapply(allChrs, function(xx) xx$pos))
    M <- do.call(rbind, lapply(allChrs, function(xx) xx$M))
    Cov <- do.call(rbind, lapply(allChrs, function(xx) xx$M)) +
        do.call(rbind, lapply(allChrs, function(xx) xx$U))
    if(readCycle) {
        Mcy <- do.call(rbind, lapply(allChrs, function(xx) xx$Mcy))
        Ucy <- do.call(rbind, lapply(allChrs, function(xx) xx$Ucy))
    } else {
        Mcy <- NULL
        Ucy <- NULL
    }
    if(keepFilt) {
        filt_cycle <- do.call(rbind, lapply(allChrs, function(xx) xx$filt_cycle))
        filt_allele <- do.call(rbind, lapply(allChrs, function(xx) xx$filt_allele))
        filt_mapq <- do.call(rbind, lapply(allChrs, function(xx) xx$filt_mapq))
        filt_baseq <- do.call(rbind, lapply(allChrs, function(xx) xx$filt_baseq))
    } else {
        filt_cycle <- filt_allele <- filt_mapq <- filt_baseq <- NULL
    }
    print(gc())
    csums <- lapply(allChrs, function(xx) xx$csums)
    names(csums) <- basename(filesPerDir[[1]])
    rm(allChrs)
    BSdata <- BSseq(chr = chr, pos = pos, M = M, Cov = Cov, rmZeroCov = rmZeroCov)
    print(gc())
    return(list(BSdata = BSdata, filt_cycle = filt_cycle, filt_allele = filt_allele,
                filt_mapq = filt_mapq, filt_baseq = filt_baseq,
                Mcy = Mcy, Ucy = Ucy, csums = csums))
}


read.umtab2.chr <- function(files, sampleNames = NULL,
                            keepM, keepU, readCycle = FALSE, keepFilt = FALSE,
                            verbose = TRUE) {
    columnHeaders <- strsplit(readLines(files[1], n = 1), "\t")[[1]]
    if("strand" %in% columnHeaders)
        stranded <- TRUE
    what0 <- replicate(length(columnHeaders), integer(0))
    names(what0) <- columnHeaders
    what0[["ref"]] <- character(0)
    if(stranded)
        what0[["strand"]] <- character(0)
    if(!readCycle) {
        what0 <- what0[! what0 %in% c("Mcy", "Ucy")]
    }
    if(!keepFilt) {
        what0 <- what0[!grepl("^filt", what0)]
    }
    if(verbose) cat("[read.umtab2.chr] reading", files[1], "\n")
    scanPars <- list(sep = "", quote = "", quiet = TRUE, skip = 1,
                     what = what0, na.strings = c("NA", "?"))
    intab <- do.call(scan, c(scanPars, file = files[1]))
    if(length(intab[[1]]) == 0)
        return(list())
    pos <- intab[["off"]]
    chr <- intab[["ref"]]
    nPos <- length(pos)
    nSamples <- length(files)
    M <- matrix(0L, nrow = nPos, ncol = nSamples)
    colnames(M)  <- sampleNames
    U <- M
    if(readCycle)
        Mcy <- Ucy <- M
    else
        Mcy <- Ucy <- NULL
    if(keepFilt)
        filt_cycle <- filt_allele <- filt_mapq <- filt_baseq <- M
    else
        filt_cycle <- filt_allele <- filt_mapq <- filt_baseq <- NULL
    allMnames <- grep("^M[[:digit:]][[:digit:]]*", names(intab), value = TRUE)
    allUnames <- grep("^U[[:digit:]][[:digit:]]*", names(intab), value = TRUE)
    csums <- matrix(0, nrow = length(c(allMnames, allUnames)), ncol = length(files))
    colnames(csums) <- sampleNames
    rownames(csums) <- c(allMnames, allUnames)
    if(missing(keepM)) keepM <- allMnames
    if(missing(keepU)) keepU <- allUnames
    stopifnot(all(c(keepM, keepU) %in% c(allMnames, allUnames)))
    M[,1] <- as.integer(Reduce("+", intab[keepM]))
    U[,1] <- as.integer(Reduce("+", intab[keepU]))
    csums[,1] <- as.integer(sapply(intab[c(allMnames, allUnames)], sum))
    if(readCycle) {
        Mcy[,1] <- intab[["Mcy"]]
        Ucy[,1] <- intab[["Ucy"]]
    }
    if(keepFilt) {
        filt_cycle[,1] <- intab[["filt_cycle"]]
        filt_allele[,1] <- intab[["filt_allele"]]
        filt_mapq[,1] <- intab[["filt_mapq"]]
        filt_baseq[,1] <- intab[["filt_baseq"]]
    }
    for(ii in seq_along(files[-1]) + 1) {
        if(verbose) cat("reading", files[ii], "\n")
        intab <- do.call(scan, c(scanPars, file = files[ii]))
        if(length(intab[[1]]) == 0)
            next
        stopifnot(all(pos == intab[["Off"]]) && all(chr == intab[["Chr"]]))
        M[,ii] <- as.integer(Reduce("+", intab[keepM]))
        U[,ii] <- as.integer(Reduce("+", intab[keepU]))
        csums[,ii] <- as.integer(sapply(intab[c(allMnames, allUnames)], sum))
        if(readCycle) {
            Mcy[,ii] <- intab[["Mcy"]]
            Ucy[,ii] <- intab[["Ucy"]]
        }
        if(keepFilt) {
            filt_cycle[,ii] <- intab[["filt_cycle"]]
            filt_allele[,ii] <- intab[["filt_allele"]]
            filt_mapq[,ii] <- intab[["filt_mapq"]]
            filt_baseq[,ii] <- intab[["filt_baseq"]]
        }
    }
    return(list(chr = chr, pos = pos, M = M, U = U,
                Mcy = Mcy, Ucy = Ucy,
                filt_cycle = filt_cycle, filt_allele = filt_allele,
                filt_mapq = filt_mapq, filt_baseq = filt_baseq,
                csums = csums))
}

read.umtab2.chr2 <- function(files, sampleNames = NULL,
                             keepM, keepU, verbose = TRUE) {
    columnHeaders <- strsplit(readLines(files[1], n = 1), "\t")[[1]]
    if("strand" %in% columnHeaders)
        stranded <- TRUE
    what0 <- replicate(length(columnHeaders), integer(0))
    names(what0) <- columnHeaders
    what0[["ref"]] <- character(0)
    if(stranded)
        what0[["strand"]] <- character(0)
    if(verbose) cat("[read.umtab2.chr] parsing all samples first\n")
    scanPars <- list(sep = "", quote = "", quiet = TRUE, skip = 1,
                     what = what0, na.strings = c("NA", "?"))
    ## First we get the locations of everything in all files
    getLocs <- function(file) {
        if(stranded) {
            intab <- scan(file, what = list("ref" = character(0),
                                "off" = integer(0),
                                "strand" = character(0)),
                          flush = TRUE, skip = 1, quote = "", sep = "")
            gr <- GRanges(seqnames = intab[["ref"]],
                          strand = ifelse(intab[["strand"]]== "W", 1L, -1L),
                          ranges = IRanges(start = intab[["off"]], width = 1))
        } else {
            intab <- scan(file, what = list("ref" = character(0),
                                "off" = integer(0)),
                          flush = TRUE, quote = "", sep = "")
            gr <- GRanges(seqnames = intab[["ref"]],
                          ranges = IRanges(start = intab[["pos"]], width = 1))
        }
        gr
    }
    grLoc <- as.list(files)
    grLoc[[1]] <- getLocs(files[1])
    grLoc <- Reduce(function(gr, file) {
        union(gr, getLocs(file))
    }, grLoc)
    nPos <- length(grLoc)
    nSamples <- length(files)

    M <- matrix(0L, nrow = nPos, ncol = nSamples)
    colnames(M)  <- sampleNames
    U <- Mcy <- Ucy <- M
    filt_cycle <- filt_allele <- filt_mapq <- filt_baseq <- M
    allMnames <- grep("^M[[:digit:]][[:digit:]]*", columnHeaders, value = TRUE)
    allUnames <- grep("^U[[:digit:]][[:digit:]]*", columnHeaders, value = TRUE)
    csums <- matrix(0, nrow = length(c(allMnames, allUnames)), ncol = length(files))
    colnames(csums) <- sampleNames
    rownames(csums) <- c(allMnames, allUnames)
    if(missing(keepM)) keepM <- allMnames
    if(missing(keepU)) keepU <- allUnames
    stopifnot(all(c(keepM, keepU) %in% c(allMnames, allUnames)))

    for(ii in seq_along(files)) {
        if(verbose) cat("[read.umtab2.chr] reading", files[ii], "\n")
        intab <- do.call(scan, c(scanPars, file = files[ii]))
        if(length(intab[[1]]) == 0)
            next
        if(stranded) {
            gr <- GRanges(seqnames = intab[["ref"]],
                          strand = ifelse(intab[["strand"]] == "W", 1L, -1L),
                          ranges = IRanges(start = intab[["off"]], width = 1))
        } else {
            gr <- GRanges(seqnames = intab[["ref"]],
                          ranges = IRanges(start = intab[["off"]], width = 1))
        }
        sh <- subjectHits(findOverlaps(gr, grLoc))
        stopifnot(length(sh) == length(intab$ref) & !anyDuplicated(sh))
        M[sh,ii] <- as.integer(Reduce("+", intab[keepM]))
        U[sh,ii] <- as.integer(Reduce("+", intab[keepU]))
        Mcy[sh,ii] <- intab[["Mcy"]]
        Ucy[sh,ii] <- intab[["Ucy"]]
        filt_cycle[sh,ii] <- intab[["filt_cycle"]]
        filt_allele[sh,ii] <- intab[["filt_allele"]]
        filt_mapq[sh,ii] <- intab[["filt_mapq"]]
        filt_baseq[sh,ii] <- intab[["filt_baseq"]]
        csums[,ii] <- as.integer(sapply(intab[c(allMnames, allUnames)], sum))
    }
    return(list(chr = as.character(seqnames(grLoc)), pos = start(grLoc), M = M, U = U,
                Mcy = Mcy, Ucy = Ucy,
                filt_cycle = filt_cycle, filt_allele = filt_allele,
                filt_mapq = filt_mapq, filt_baseq = filt_baseq,
                csums = csums))
}



read.umtab <- function(dirs, sampleNames = NULL, rmZeroCov = FALSE,
                       pattern = NULL,
                       keepU = c("U10", "U20", "U30", "U40"),
                       keepM = c("M10", "M20", "M30", "M40"), verbose = TRUE) {
    filesPerDir <- lapply(dirs, function(xx) sort(list.files(xx, pattern = pattern, full.names = TRUE)))
    if(!all(sapply(filesPerDir, function(xx) all(basename(xx) == basename(filesPerDir[[1]])))))
        warning("'dirs' does not contain the exact same file names")
    allChrs <- as.list(seq_along(filesPerDir[[1]]))
    for(ii in seq_along(filesPerDir[[1]])) {
        chrRead <- read.umtab.chr(sapply(filesPerDir, function(xx) xx[ii]),
                                  keepM = keepM, keepU = keepU, sampleNames = sampleNames,
                                  verbose = verbose)
        if(rmZeroCov && length(chrRead) != 0){
            wh <- which(rowSums2(chrRead$U + chrRead$M) > 0)
            chrRead$M <- chrRead$M[wh,]
            chrRead$U <- chrRead$U[wh,]
            chrRead$Mcy <- chrRead$Mcy[wh,]
            chrRead$Ucy <- chrRead$Ucy[wh,]
            chrRead$chr <- chrRead$chr[wh]
            chrRead$pos <- chrRead$pos[wh]
            chrRead$Map <- chrRead$Map[wh]
            chrRead$GC <- chrRead$GC[wh]
        }
        allChrs[[ii]] <- chrRead
    }
    BSdata <- BSseq(chr = do.call(c, lapply(allChrs, function(xx) xx$chr)),
                    pos = do.call(c, lapply(allChrs, function(xx) xx$pos)),
                    M = do.call(rbind, lapply(allChrs, function(xx) xx$M)),
                    Cov = do.call(rbind, lapply(allChrs, function(xx) xx$M)) +
                    do.call(rbind, lapply(allChrs, function(xx) xx$U)),
                    rmZeroCov = rmZeroCov)
    GC <- do.call(c, lapply(allChrs, function(xx) xx$GC))
    Map <- do.call(c, lapply(allChrs, function(xx) xx$Map))
    Mcy <- do.call(rbind, lapply(allChrs, function(xx) xx$Mcy))
    Ucy <- do.call(rbind, lapply(allChrs, function(xx) xx$Ucy))
    csums <- lapply(allChrs, function(xx) xx$csums)
    names(csums) <- basename(filesPerDir[[1]])
    return(list(BSdata = BSdata, GC = GC, Map = Map,
                Mcy = Mcy, Ucy = Ucy, csums = csums))
}


read.umtab.chr <- function(files, sampleNames = NULL,
                           keepM = c("M10", "M20", "M30", "M40"),
                           keepU = c("U10", "U20", "U30", "U40"), verbose = TRUE) {
    columnHeaders <- c("Chr", "Off", "M0", "M10", "M20", "M30", "M40", "Mqual", "Mcy",
                       "U0", "U10", "U20", "U30", "U40", "Uqual", "Ucy",
                       "MapaFW", "MapaRC", "CGContFW", "CGContRC")
    what0 <- c(list(character(0)), replicate(19, integer(0)))
    names(what0) <- columnHeaders
    line1 <- readLines(files[1], n = 1)
    if(length(line1) == 0)
        return(list())
    line1 <- strsplit(line1, "\t")[[1]]
    stopifnot(all(line1 == columnHeaders))
    if(verbose) cat("[read.umtab.chr] reading", files[1], "\n")
    scanPars <- list(sep = "", quote = "", quiet = TRUE, skip = 1,
                  what = what0, na.strings = c("NA", "?"))
    intab <- do.call(scan, c(scanPars, file = files[1]))
    if(length(intab[[1]]) == 0)
        return(list())
    GC <- as.integer(intab[["CGContFW"]] + intab[["CGContRC"]])
    Map <- as.integer(intab[["MapaFW"]] + intab[["MapaRC"]])
    pos <- intab[["Off"]]
    chr <- intab[["Chr"]]
    nPos <- length(pos)
    nSamples <- length(files)
    M <- matrix(0L, nrow = nPos, ncol = nSamples)
    U <- matrix(0L, nrow = nPos, ncol = nSamples)
    Mcy <- matrix(0L, nrow = nPos, ncol = nSamples)
    Ucy <- matrix(0L, nrow = nPos, ncol = nSamples)
    csums <- matrix(0, nrow = 10, ncol = length(files))
    colnames(M) <- colnames(U) <- colnames(csums) <- sampleNames
    allMnames <- c("M0", "M10", "M20", "M30", "M40")
    allUnames <- c("U0", "U10", "U20", "U30", "U40")
    stopifnot(all(c(keepM, keepU) %in% c(allMnames, allUnames)))
    rownames(csums) <- c(allMnames, allUnames)
    M[,1] <- as.integer(Reduce("+", intab[keepM]))
    U[,1] <- as.integer(Reduce("+", intab[keepU]))
    Mcy[,1] <- intab[["Mcy"]]
    Ucy[,1] <- intab[["Ucy"]]
    csums[,1] <- as.integer(sapply(intab[c(allMnames, allUnames)], sum))
    for(ii in seq_along(files[-1]) + 1) {
        if(verbose) cat("[read.umtab2] reading", files[ii], "\n")
        intab <- do.call(scan, c(scanPars, file = files[ii]))
        if(length(intab[[1]]) == 0)
            next
        stopifnot(all(pos == intab[["Off"]]) && all(chr == intab[["Chr"]]))
        M[,ii] <- as.integer(Reduce("+", intab[keepM]))
        U[,ii] <- as.integer(Reduce("+", intab[keepU]))
        Mcy[,ii] <- intab[["Mcy"]]
        Ucy[,ii] <- intab[["Ucy"]]
        csums[,ii] <- as.integer(sapply(intab[c(allMnames, allUnames)], sum))
    }
    return(list(chr = chr, pos = pos, Map = Map, GC = GC, M = M, U = U,
                Mcy = Mcy, Ucy = Ucy, csums = csums))
}

read.bsmoothDirRaw <- function(dir, seqnames = NULL, keepCycle = FALSE, keepFilt = FALSE,
                               header = TRUE, verbose = TRUE) {
    dir <- normalizePath(dir)
    inpattern <- "\\.cpg\\.tsv(|\\.gz)$"
    if(length(dir) != 1 || !file.info(dir)$isdir)
        stop("argument 'dir' needs to be a single directory")
    allChrFiles <- list.files(dir, pattern = inpattern, full.names = TRUE)
    if(!is.null(seqnames))
        allChrFiles <- allChrFiles[sub(inpattern, "", basename(allChrFiles)) %in% seqnames]
    if(length(allChrFiles) == 0) {
        warning(sprintf("dir '%s' is empty or has no output from 'seqnames'", dir))
        return(NULL)
    }
    if(header) {
        columnHeaders <- sapply(allChrFiles, function(thisfile) {
            if(grepl("\\.gz$", thisfile))
                con <- gzfile(thisfile)
            else
                con <- file(thisfile, open = "r")
            out <- readLines(con, n = 1)
            close(con)
            out
        })
        columnHeaders <- strsplit(columnHeaders, "\t")
        if(!all(sapply(columnHeaders, function(xx) all.equal(columnHeaders[[1]], xx))))
            stop(sprintf("input files in dir '%s' does not have the same headers", dir))
        columnHeaders <- columnHeaders[[1]]
    } else
        columnHeaders <- c("ref", "off", "strand", "Mstr", "Mcy", "Ustr", "Ucy",
                           "filt_cycle", "filt_readlen", "filt_allele", "filt_mapq", "filt_baseq")
    what0 <- replicate(length(columnHeaders), character(0))
    names(what0) <- columnHeaders
    int <- c(which(columnHeaders %in% c("off", "Mcy", "Ucy")), grep("^filt", columnHeaders))
    what0[int] <- replicate(length(int), integer(0))
    if(!keepCycle)
        what0[c("Mcy", "Ucy")] <- replicate(2, NULL)
    if(!keepFilt)
        what0[grep("^filt", names(what0))] <- replicate(length(grep("^filt", names(what0))), NULL)
    outList <- lapply(allChrFiles, function(thisfile) {
        if(verbose)
            cat(sprintf("[read.bsmoothDirRaw] Reading '%s'\n", thisfile))
        if(grepl("\\.gz$", thisfile))
            con <- gzfile(thisfile)
        else
            con <- file(thisfile)
        out <- scan(con, skip = header, what = what0, sep = "\t",
                    quote = "", na.strings = "NA", quiet = TRUE)
        close(con)
        out
    })
    listNames <- names(outList[[1]])
    names(listNames) <- listNames
    out <- lapply(listNames, function(name) {
        do.call(c, lapply(outList, function(xx) xx[[name]]))
    })
    rm(outList)
    gr <- GRanges(seqnames = paste0("chr", out[["ref"]]),
                  ranges = IRanges(start = out[["off"]], width = 1))
    out[["ref"]] <- out[["off"]] <- NULL
    names(out)[names(out) == "strand"] <- "bstrand"
    out <- out[!sapply(out, is.null)]
    df <- DataFrame(out)
    mcols(gr) <- df
    gr
}


sampleRawToBSseq <- function(gr, qualityCutoff = 20, sampleName = NULL, rmZeroCov = FALSE) {
    numberQualsGreaterThan <- function(cvec) {
        onestring <- paste(cvec, collapse = "")
        greater <- (as.integer(charToRaw(onestring)) - 33L >= qualityCutoff)
        out <- tapply(greater, rep(1:length(cvec), times = nchar(cvec)), sum)
        out
    }
    strToCov <- function(vec) {
        Cov <- rep(0, length(vec))
        wh <- which(! vec %in% c("", "0"))
        if(length(wh) > 0)
            Cov[wh] <- numberQualsGreaterThan(vec[wh])
        Cov
    }
    M <- matrix(strToCov(mcols(gr)[, "Mstr"]), ncol = 1)
    Cov <- M + strToCov(mcols(gr)[, "Ustr"])
    mcols(gr) <- NULL
    BSseq(gr = gr, M = M, Cov = Cov, sampleNames = sampleName, rmZeroCov = rmZeroCov)
}

read.bsmooth <- function(dirs, sampleNames = NULL, seqnames = NULL, returnRaw = FALSE,
                         qualityCutoff = 20, rmZeroCov = FALSE, verbose = TRUE) {
    dirs <- normalizePath(dirs, mustWork = TRUE)
    if(!(all(file.info(dirs)$isdir)))
        stop("argument 'dirs' has to be directories")
    if(anyDuplicated(dirs))
        stop("duplicate entries in 'dirs'")
    if(is.null(sampleNames) && !anyDuplicated(basename(dirs)))
        sampleNames <- basename(dirs)
    if(is.null(sampleNames))
        sampleNames <- dirs
    if(!is.null(sampleNames) && (length(sampleNames) != length(dirs) || anyDuplicated(sampleNames)))
        stop("argument 'sampleNames' (if not NULL) has to have the same length as argument 'dirs', without duplicate entries")
    idxes <- seq_along(dirs)
    names(idxes) <- sampleNames
    allOut <- lapply(idxes, function(ii) {
        if(verbose) cat(sprintf("[read.bsmooth] Reading dir '%s' ... ", dirs[ii]))
        ptime1 <- proc.time()
        if(returnRaw) {
            out <- read.bsmoothDirRaw(dir = dirs[ii], seqnames = seqnames, keepCycle = TRUE,
                                      keepFilt = TRUE, header = TRUE, verbose = FALSE)
        } else {
            raw <- read.bsmoothDirRaw(dir = dirs[ii], seqnames = seqnames, keepCycle = FALSE,
                                      keepFilt = FALSE, header = TRUE, verbose = FALSE)
            out <- sampleRawToBSseq(raw, qualityCutoff = qualityCutoff, rmZeroCov, sampleName = sampleNames[ii])
        }
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if(verbose) cat(sprintf("done in %.1f secs\n", stime))
        out
    })
    if(!returnRaw) {
        if(verbose) cat(sprintf("[read.bsmooth] Joining samples ... "))
        ptime1 <- proc.time()
        allOut <- combineList(allOut)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if(verbose) cat(sprintf("done in %.1f secs\n", stime))
    }
    allOut
}

parsingPipeline <- function(dirs, qualityCutoff = 20, outDir, seqnames = NULL,
                            subdir = "ev_bt2_cpg_tab", timing = FALSE) {
    if(!all(file.exists(dirs)))
        stop("not all directories in 'dirs' exists.")
    cat("[parsingPipeline] Parsing all files.\n")
    oneDir <- function(dir) {
        cat("[parsingPipeline]  dir", basename(dir), ": ")
        base <- basename(dir)
        cat("parsing, ")
        ptime1 <- proc.time()
        raw <- read.bsmoothDirRaw(file.path(dir, subdir),
                                  keepCycle = TRUE, keepFilt = TRUE,
                                  verbose = FALSE)
        assign(paste0(base, ".raw"), raw)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if(timing) {
            cat(sprintf("\ndone in %.1f secs\n", stime))
            print(gc())
        }
        cat("saving, ")
        save(list = paste0(base, ".raw"),
             file = file.path(outDir, paste0(base, ".raw.rda")))
        cat("converting, ")
        ptime1 <- proc.time()
        bsseq <- sampleRawToBSseq(raw, qualityCutoff = qualityCutoff,
                                  sampleName = base)
        seqlevels(bsseq)[seqlevels(bsseq) == "chrgi|9626243|ref|NC_001416.1|"] <- "chrLambda"
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if(timing) {
            cat(sprintf("\ndone in %.1f secs\n", stime))
            print(gc())
        }
        cat("ordering, ")
        ptime1 <- proc.time()
        bsseq <- chrSelectBSseq(bsseq, order = TRUE,
                                seqnames = seqnames)
        assign(paste0(base, ".bsseq"), bsseq)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        if(timing) {
            cat(sprintf("\nIn %.1f secs\n", stime))
            print(gc())
        }
        cat("saving, ")
        save(list = paste0(base, ".bsseq"),
             file = file.path(outDir, paste0(base, ".bsseq.rda")))
        cat("done\n")
        NULL
    }
    for(dir in dirs)
        oneDir(dir)
    invisible(dirs)
}

getNullDistribution_BSmooth.tstat <- function(BSseq, idxMatrix1, idxMatrix2,
                                              estimate.var, local.correct,
                                              cutoff, stat, maxGap, mc.cores = 1) {
    stopifnot(nrow(idxMatrix1) == nrow(idxMatrix2))
    message(sprintf("[getNullDistribution_BSmooth.tstat] performing %d permutations\n", nrow(idxMatrix1)))
    nullDist <- mclapply(1:nrow(idxMatrix1), function(ii) {
        ptime1 <- proc.time()
        BS.tstat <- BSmooth.tstat(BSseq, estimate.var = estimate.var,
                                  group1 = idxMatrix1[ii,],
                                  group2 = idxMatrix2[ii,],
                                  local.correct = local.correct, maxGap = 10^8,
                                  verbose = FALSE)
        dmrs0 <- dmrFinder(BS.tstat, stat = stat, cutoff = cutoff, maxGap = maxGap)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        message(sprintf("[getNullDistribution_BSmooth.tstat] completing permutation %d in %.1f sec\n", ii, stime))
        dmrs0
    }, mc.cores = min(nrow(idxMatrix1), mc.cores), mc.preschedule = FALSE)
    nullDist
}

getNullDistribution_BSmooth.fstat <- function(BSseq,
                                              idxMatrix,
                                              design, contrasts,
                                              coef = NULL, cutoff,
                                              maxGap.sd, maxGap.dmr,
                                              mc.cores = 1) {

    message(sprintf("[getNullDistribution_BSmooth.fstat] performing %d permutations\n",
                    nrow(idxMatrix)))
    # NOTE: Using mc.preschedule = TRUE
    # TODO: Need some protection for when a core(s) error, otherwise nullDist
    #       contains a mixture of data frame, NULL, and try-error objects
    #       (which will subesequently break when passed to getFWER.fstat())
    nullDist <- mclapply(seq_len(nrow(idxMatrix)), function(ii) {
        ptime1 <- proc.time()
        # NOTE: More efficient to permute design matrix using idxMatrix[ii, ] than
        #       permute the raw data with BSseq[, idxMatrix[ii, ]]
        bstat <- BSmooth.fstat(BSseq, design = design[idxMatrix[ii, ], ],
                               contrasts = contrasts)
        bstat <- smoothSds(bstat, maxGap = maxGap.sd)
        bstat <- computeStat(bstat, coef = coef)
        dmrs0 <- dmrFinder(bstat, cutoff = cutoff, maxGap = maxGap.dmr)
        ptime2 <- proc.time()
        stime <- (ptime2 - ptime1)[3]
        message(sprintf("[getNullDistribution_BSmooth.fstat] completing permutation %d in %.1f sec\n", ii, stime))
        dmrs0
    }, mc.cores = min(nrow(idxMatrix), mc.cores), mc.preschedule = TRUE)
    nullDist
}

permuteAll <- function(nperm, design) {
    message(sprintf("[permuteAll] performing %d unrestricted permutations of the design matrix\n", nperm))

    CTRL <- how(nperm = nperm)
    # NOTE: shuffleSet() returns a nperm * nrow(design) matrix of permutations
    idxMatrix <- shuffleSet(n = nrow(design), control = CTRL)
}

getNullBlocks_BSmooth.tstat <- function(BSseq, idxMatrix1, idxMatrix2, estimate.var = "same",
                          mc.cores = 1) {
    getNullDistribution_BSmooth.tstat(BSseq = BSseq, idxMatrix1 = idxMatrix1,
                       idxMatrix2 = idxMatrix2, local.correct = FALSE,
                       estimate.var = estimate.var,
                       cutoff = c(-2,2), stat = "tstat", maxGap = 10000,
                       mc.cores = mc.cores)
}

getNullDmrs_BSmooth.tstat <- function(BSseq, idxMatrix1, idxMatrix2, estimate.var = "same",
                        mc.cores = 1) {
    getNullDistribution_BSmooth.tstat(BSseq = BSseq, idxMatrix1 = idxMatrix1,
                       idxMatrix2 = idxMatrix2, local.correct = TRUE,
                       estimate.var = estimate.var,
                       cutoff = c(-4.6,4.6), stat = "tstat.corrected", maxGap = 300,
                       mc.cores = mc.cores)
}




subsetDmrs <- function(xx) {
    if(is.null(xx) || is(xx, "try-error"))
        return(NULL)
    out <- xx[ xx[,"n"] >= 3 & abs(xx[, "meanDiff"]) > 0.1 &
                  xx[, "invdensity"] <= 300, ]
    if(nrow(out) == 0)
        return(NULL)
    out
}

subsetBlocks <- function(xx) {
    if(is.null(xx) || is(xx, "try-error"))
        return(NULL)
    out <- subset(xx, width >= 10000)
    if(nrow(out) == 0)
        return(NULL)
    out
}

getFWER <- function(null, type = "blocks") {
    reference <- null[[1]]
    null <- null[-1]
    null <- null[!sapply(null, is.null)]
    better <- sapply(1:nrow(reference), function(ii) {
        # meanDiff <- abs(reference$meanDiff[ii])
        areaStat <- abs(reference$areaStat[ii])
        width <- reference$width[ii]
        n <- reference$n[ii]
        if (type == "blocks") {
            out <- sapply(null, function(nulldist) {
                # any(abs(nulldist$meanDiff) >= meanDiff &
                        # nulldist$width >= width)
                any(abs(nulldist$areaStat) >= areaStat &
                        nulldist$width >= width)
            })
        }
        if (type == "dmrs") {
            out <- sapply(null, function(nulldist) {
                # any(abs(nulldist$meanDiff) >= meanDiff &
                #     nulldist$n >= n)
                any(abs(nulldist$areaStat) >= areaStat &
                        nulldist$n >= n)
            })
        }
        sum(out)
    })
    better
}

# NOTE: Identical to getFWER() except uses areaStat rather than meanDiff
#       to compare regions.
getFWER.fstat <- function(null, type = "blocks") {
    reference <- null[[1]]
    null <- null[-1]
    null <- null[!sapply(null, is.null)]
    # TODO: Will break if null == list(), which can occur in practice (although
    #       rarely).
    better <- sapply(seq_len(nrow(reference)), function(ii) {
        # meanDiff <- abs(reference$meanDiff[ii])
        areaStat <- abs(reference$areaStat[ii])
        width <- reference$width[ii]
        n <- reference$n[ii]
        if (type == "blocks") {
            out <- vapply(null, function(nulldist) {
                # any(abs(nulldist$meanDiff) >= meanDiff &
                # nulldist$width >= width)
                any(abs(nulldist$areaStat) >= areaStat &
                        nulldist$width >= width)
            }, logical(1L))
        }
        if (type == "dmrs") {
            out <- vapply(null, function(nulldist) {
                # any(abs(nulldist$meanDiff) >= meanDiff &
                #     nulldist$n >= n)
                any(abs(nulldist$areaStat) >= areaStat &
                        nulldist$n >= n)
            }, logical(1L))
        }
        sum(out)
    })
    better
}

# TODO: Simplify makeIdxMatrix() by using permute package
makeIdxMatrix <- function(group1, group2, testIsSymmetric = TRUE, includeUnbalanced = TRUE) {
    groupBoth <- c(group1, group2)
    idxMatrix1 <- NULL
    subsetByMatrix <- function(vec, mat) {
        apply(mat, 2, function(xx) vec[xx])
    }
    combineMat <- function(mat1, mat2) {
        tmp <- lapply(1:nrow(mat1), function(ii) {
            t(apply(mat2, 1, function(xx) { c(mat1[ii,], xx) }))
        })
        do.call(rbind, tmp)
    }
    if(length(group1) == 1 && length(group1) == 1) {
        if(testIsSymmetric)
            idxMatrix1 <- as.matrix(group1)
        else
            idxMatrix1 <- as.matrix(c(group1, group2))
    }
    if(length(group1) == 2 && length(group2) == 2) {
        if(testIsSymmetric) {
            idxMatrix1 <- rbind(group1,
                                combineMat(matrix(group1[1], ncol = 1),
                                           matrix(group2, ncol = 1)))
        } else {
            idxMatrix1 <- rbind(group1, group2,
                                combineMat(matrix(group1, ncol = 1), matrix(group2, ncol = 1)))
        }
    }
    if(length(group1) == 3 && length(group1) == 3) {
        if(testIsSymmetric) {
            idxMatrix1 <- rbind(group1,
                                combineMat(subsetByMatrix(group1, combinations(3,2)),
                                           as.matrix(group2, ncol = 1)))

        } else {
            idxMatrix1 <- rbind(group1, group2,
                                combineMat(subsetByMatrix(group1, combinations(3,2)),
                                           as.matrix(group2, ncol = 1)),
                                combineMat(as.matrix(group1, ncol = 1),
                                           subsetByMatrix(group2, combinations(3,2))))
        }
    }
    if(length(group1) == 4 && length(group1) == 4) {
        if(testIsSymmetric) {
            idxMatrix1 <- rbind(group1,
                                combineMat(subsetByMatrix(group1, combinations(3,2)),
                                           subsetByMatrix(group2, combinations(4,2))))
        } else {
            idxMatrix1 <- rbind(group1, group2,
                                combineMat(subsetByMatrix(group1, combinations(4,2)),
                                           subsetByMatrix(group2, combinations(4,2))))
        }
        if(includeUnbalanced) {
            newMatrix <- combineMat(subsetByMatrix(group1, combinations(4,3)),
                                    as.matrix(group2, ncol = 1))
            idxMatrix1 <- rbind(idxMatrix1, newMatrix)
        }
        if(includeUnbalanced && !testIsSymmetric) {
            newMatrix <- combineMat(as.matrix(group1, ncol = 1),
                                    subsetByMatrix(group2, combinations(4,3)))
            idxMatrix1 <- rbind(idxMatrix1, newMatrix)
        }
    }
    if(length(group1) == 5 && length(group1) == 5) {
        if(testIsSymmetric) {
            idxMatrix1 <- rbind(group1,
                                combineMat(subsetByMatrix(group1, combinations(5, 3)),
                                           subsetByMatrix(group2, combinations(5, 2))))
        } else {
            idxMatrix1 <- rbind(group1, group2,
                                combineMat(subsetByMatrix(group1, combinations(5, 3)),
                                           subsetByMatrix(group2, combinations(5, 2))),
                                combineMat(subsetByMatrix(group1, combinations(5, 2)),
                                           subsetByMatrix(group2, combinations(5, 3))))
        }
        if(includeUnbalanced) {
            idxMatrix1 <- rbind(idxMatrix1,
                                combineMat(subsetByMatrix(group1, combinations(5,4)),
                                           as.matrix(group2, ncol = 1)))
        }
        if(includeUnbalanced && !testIsSymmetric) {
            idxMatrix1 <- rbind(idxMatrix1,
                                combineMat(as.matrix(group1, ncol = 1),
                                           subsetByMatrix(group2, combinations(5,4))))
        }
    }
    if(length(group1) == 6 && length(group1) == 6) {
        if(testIsSymmetric) {
            idxMatrix1 <- rbind(group1,
                                combineMat(subsetByMatrix(group1, combinations(5,3)),
                                           subsetByMatrix(group2, combinations(6,3))))
        } else {
            idxMatrix1 <- rbind(group1, group2,
                                combineMat(subsetByMatrix(group1, combinations(6,3)),
                                           subsetByMatrix(group2, combinations(6,3))))
        }
        if(includeUnbalanced) {
            newMatrix1 <- combineMat(subsetByMatrix(group1, combinations(6,4)),
                                    subsetByMatrix(group2, combinations(6,2)))
            newMatrix2 <- combineMat(subsetByMatrix(group1, combinations(6,5)),
                                    as.matrix(group2, ncol = 1))
            idxMatrix1 <- rbind(idxMatrix1, newMatrix1, newMatrix2)
        }
        if(includeUnbalanced && !testIsSymmetric) {
            newMatrix1 <- combineMat(subsetByMatrix(group1, combinations(6,2)),
                                     subsetByMatrix(group2, combinations(6,4)))
            newMatrix2 <- combineMat(as.matrix(group1, ncol = 1),
                                     subsetByMatrix(group2, combinations(6,5)))
            idxMatrix1 <- rbind(idxMatrix1, newMatrix1, newMatrix2)
        }
    }
    if(is.null(idxMatrix1))
        stop("unable to handle this combination of 'group1', 'group2' and 'testIsSymmetric'")
    rownames(idxMatrix1) <- NULL
    idxMatrix2 <- do.call(rbind, lapply(1:nrow(idxMatrix1), function(ii) {
        setdiff(groupBoth, idxMatrix1[ii,])
    }))
    return(list(idxMatrix1 = idxMatrix1, idxMatrix2 = idxMatrix2))
}

context("Permutations")

test_that("makeIdxMatrix() n = 1", {
    grp1 <- paste0("A", 1)
    grp2 <- paste0("B", 1)

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = FALSE)[[1]]
    expect_equal(anyDuplicated(res), 0L)
    expect_equal(res[1, ], grp1)
    expect_equal(res[2, ], grp2)
    expect_equal(nrow(res), choose(2, 1)) # Fail

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0L)
    expect_equal(res[1, ], grp1)
    expect_equal(nrow(res), choose(2, 1) / 2)
})

test_that("makeIdxMatrix() n = 2", {
    grp1 <- paste0("A", 1:2)
    grp2 <- paste0("B", 1:2)

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = FALSE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(res[2,], grp2)
    expect_equal(nrow(res), choose(4, 2))

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(nrow(res), choose(4, 2) / 2)
})

test_that("makeIdxMatrix() n = 3", {
    grp1 <- paste0("A", 1:3)
    grp2 <- paste0("B", 1:3)

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = FALSE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(res[2,], grp2)
    expect_equal(nrow(res), choose(6, 3))

    res <- bsseq:::makeIdxMatrix(grp1, grp2, testIsSymmetric = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(nrow(res), choose(6, 3) / 2)
})

test_that("makeIdxMatrix() n = 4", {
    grp1 <- paste0("A", 1:4)
    grp2 <- paste0("B", 1:4)

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = FALSE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(res[2,], grp2)
    expect_equal(nrow(res), choose(8,4))

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = TRUE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(nrow(res), choose(8,4) / 2) # Fail
})

test_that("makeIdxMatrix() n = 5", {

    grp1 <- paste0("A", 1:5)
    grp2 <- paste0("B", 1:5)

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = FALSE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(res[2,], grp2)
    expect_equal(nrow(res), choose(10,5))

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = TRUE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(nrow(res), choose(10,5) / 2)
})

test_that("makeIdxMatrix() n = 6", {
    grp1 <- paste0("A", 1:6)
    grp2 <- paste0("B", 1:6)

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = FALSE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(res[2,], grp2)
    expect_equal(nrow(res), choose(12,6))

    res <- bsseq:::makeIdxMatrix(
        grp1,
        grp2,
        testIsSymmetric = TRUE,
        includeUnbalanced = TRUE)[[1]]
    expect_equal(anyDuplicated(res), 0)
    expect_equal(res[1,], grp1)
    expect_equal(nrow(res), choose(12,6) / 2)
})

BSmooth.fstat <- function(BSseq, design, contrasts, verbose = TRUE){
    stopifnot(is(BSseq, "BSseq"))
    stopifnot(hasBeenSmoothed(BSseq))

    ## if(any(rowSums(getCoverage(BSseq)[, unlist(groups)]) == 0))
    ##     warning("Computing t-statistics at locations where there is no data; consider subsetting the 'BSseq' object first")

    if(verbose) cat("[BSmooth.fstat] fitting linear models ... ")
    ptime1 <- proc.time()
    allPs <- getMeth(BSseq, type = "smooth", what = "perBase",
                     confint = FALSE)
    allPs <- as.array(allPs)
    fit <- lmFit(allPs, design)
    fitC <- contrasts.fit(fit, contrasts)
    ## Need
    ##   fitC$coefficients, fitC$stdev.unscaled, fitC$sigma, fitC$cov.coefficients
    ## actuall just need
    ##   tstats <- fitC$coefficients / fitC$stdev.unscaled / fitC$sigma
    ##   rawSds <- fitC$sigma
    ##   cor.coefficients <- cov2cor(fitC$cov.coefficients)
    rawSds <- as.matrix(fitC$sigma)
    cor.coefficients <- cov2cor(fitC$cov.coefficients)
    rawTstats <- fitC$coefficients / fitC$stdev.unscaled / fitC$sigma
    names(dimnames(rawTstats)) <- NULL
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))

    parameters <- c(BSseq@parameters,
                    list(design = design, contrasts = contrasts))
    stats <- list(rawSds = rawSds,
                  cor.coefficients = cor.coefficients,
                  rawTstats = rawTstats)
    out <- BSseqStat(gr = granges(BSseq),
                     stats = stats,
                     parameters = parameters)
    out
}

smoothSds <- function(BSseqStat, k = 101, qSd = 0.75, mc.cores = 1,
                      maxGap = 10^8, verbose = TRUE) {
    smoothSd <- function(Sds, k, qSd) {
        k0 <- floor(k/2)
        if(all(is.na(Sds))) return(Sds)
        thresSD <- pmax(Sds, quantile(Sds, qSd, na.rm = TRUE), na.rm = TRUE)
        addSD <- rep(median(Sds, na.rm = TRUE), k0)
        sSds <- as.vector(runmean(Rle(c(addSD, thresSD, addSD)), k = k))
        sSds
    }
    if(is.null(maxGap))
        maxGap <- BSseqStat@parameters[["maxGap"]]
    if(is.null(maxGap))
        stop("need to set argument 'maxGap'")
    if(verbose) cat("[smoothSds] preprocessing ... ")
    ptime1 <- proc.time()
    clusterIdx <- makeClusters(granges(BSseqStat), maxGap = maxGap)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))
    smoothSds <- do.call("c",
                         mclapply(clusterIdx, function(idx) {
                             rawSds <- getStats(BSseqStat,
                                                what = "rawSds")[idx, ]
                             rawSds <- as.array(rawSds)
                             smoothSd(rawSds, k = k, qSd = qSd)
                         }, mc.cores = mc.cores))
    smoothSds <- as.matrix(smoothSds)
    if("smoothSds" %in% names(getStats(BSseqStat)))
        BSseqStat@stats[["smoothSds"]] <- smoothSds
    else
        BSseqStat@stats <- c(getStats(BSseqStat), list(smoothSds = smoothSds))
    BSseqStat
}

# Quieten R CMD check
globalVariables("tstat")

# NOTE: Realises in memory a matrix with nrow = length(BSseqStat) and
#       ncol = length(coef)
computeStat <- function(BSseqStat, coef = NULL) {
    stopifnot(is(BSseqStat, "BSseqStat"))
    if(is.null(coef)) {
        coef <- 1:ncol(getStats(BSseqStat, what = "rawTstats"))
    }
    raw_tstats <- getStats(BSseqStat, what = "rawTstats")[, coef, drop = FALSE]
    scaled_sds <- getStats(BSseqStat, what = "rawSds") /
        getStats(BSseqStat, what = "smoothSds")
    scaled_sds_matrix <- matrix(
        rep(scaled_sds, ncol(raw_tstats)),
        ncol = ncol(raw_tstats))
    tstats <- raw_tstats * scaled_sds_matrix
    if(length(coef) > 1) {
        cor.coefficients <- getStats(BSseqStat,
                                     what = "cor.coefficients")[coef,coef]
        # NOTE: classifyTestsF() calls as.matrix(tstats) and so realises this
        #       array
        stat <- as.matrix(
            classifyTestsF(tstats, cor.coefficients, fstat.only = TRUE))
        stat.type <- "fstat"
    } else {
        stat <- tstats
        stat.type <- "tstat"
    }
    if("stat" %in% names(getStats(BSseqStat))) {
        BSseqStat@stats[["stat"]] <- stat
        BSseqStat@stats[["stat.type"]] <- stat.type
    } else {
        BSseqStat@stats <- c(getStats(BSseqStat),
                             list(stat = stat, stat.type = stat.type))
    }
    BSseqStat
}

localCorrectStat <- function(BSseqStat, threshold = c(-15,15), mc.cores = 1, verbose = TRUE) {
    compute.correction <- function(idx) {
        xx <- start(BSseqTstat)[idx]
        yy <- tstat[idx] ## FIXME
        if(!is.null(threshold)) {
            stopifnot(is.numeric(threshold) && length(threshold) == 2)
            stopifnot(threshold[1] < 0 && threshold[2] > 0)
            yy[yy < threshold[1]] <- threshold[1]
            yy[yy > threshold[2]] <- threshold[2]
        }
        suppressWarnings({
            drange <- diff(range(xx, na.rm = TRUE))
        })
        if(drange <= 25000)
            return(yy)
        tstat.function <- approxfun(xx, yy)
        xx.reg <- seq(from = min(xx), to = max(xx), by = 2000)
        yy.reg <- tstat.function(xx.reg)
        fit <- locfit(yy.reg ~ lp(xx.reg, h = 25000, deg = 2, nn = 0),
                      family = "huber", maxk = 50000)
        correction <- predict(fit, newdata = data.frame(xx.reg = xx))
        yy - correction
    }
    maxGap <- BSseqStat$parameters$maxGap
    if(verbose) cat("[BSmooth.tstat] preprocessing ... ")
    ptime1 <- proc.time()
    clusterIdx <- makeClusters(BSseqStat$gr, maxGap = maxGap)
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if(verbose) cat(sprintf("done in %.1f sec\n", stime))
    stat <- BSseqStat$stat
    stat.corrected <- do.call(c, mclapply(clusterIdx, compute.correction,
                                          mc.cores = mc.cores))
    BSseqTstat@stats <- cbind(getStats(BSseqTstat),
                              "tstat.corrected" = stat.corrected)
    BSseqTstat@parameters$local.local <- TRUE
    BSseqTstat
}

fstat.pipeline <- function(BSseq, design, contrasts, cutoff, fac, nperm = 1000,
                           coef = NULL, maxGap.sd = 10 ^ 8, maxGap.dmr = 300,
                           type = "dmrs", mc.cores = 1) {
    type <- match.arg(type, c("dmrs", "blocks"))
    bstat <- BSmooth.fstat(BSseq = BSseq, design = design,
                           contrasts = contrasts)
    bstat <- smoothSds(bstat)
    bstat <- computeStat(bstat, coef = coef)
    dmrs <- dmrFinder(bstat, cutoff = cutoff, maxGap = maxGap.dmr)
    if (is.null(dmrs)) {
        stop("No DMRs identified. Consider reducing the 'cutoff' from (",
             paste0(cutoff, collapse = ", "), ")")
    }
    idxMatrix <- permuteAll(nperm, design)
    nullDist <- getNullDistribution_BSmooth.fstat(BSseq = BSseq,
                                                  idxMatrix = idxMatrix,
                                                  design = design,
                                                  contrasts = contrasts,
                                                  coef = coef,
                                                  cutoff = cutoff,
                                                  maxGap.sd = maxGap.sd,
                                                  maxGap.dmr = maxGap.dmr,
                                                  mc.cores = mc.cores)
    fwer <- getFWER.fstat(null = c(list(dmrs), nullDist), type = type)
    dmrs$fwer <- fwer
    meth <- getMeth(BSseq, dmrs, what = "perRegion")
    meth <- t(apply(meth, 1, function(xx) tapply(xx, fac, mean)))
    dmrs <- cbind(dmrs, meth)
    dmrs$maxDiff <- rowMaxs(meth) - rowMins(meth)
    list(bstat = bstat, dmrs = dmrs, idxMatrix = idxMatrix, nullDist = nullDist)
}

fstat.comparisons.pipeline <- function(BSseq, design, contrasts, cutoff, fac,
                                       maxGap.sd = 10 ^ 8, maxGap.dmr = 300,
                                       verbose = TRUE) {
    bstat <- BSmooth.fstat(BSseq = BSseq,
                           design = design,
                           contrasts = contrasts,
                           verbose = verbose)
    bstat <- smoothSds(bstat, maxGap = maxGap.sd, verbose = verbose)
    # NOTE: Want to keep the bstat object corresponding to the original fstat
    #       and not that from the last t-tsts in the following lapply()
    bstat_f <- bstat
    if (verbose) {
        cat(paste0("[fstat.comparisons.pipeline] making ", ncol(contrasts),
                   " comparisons ... \n"))
    }
    l <- lapply(seq_len(ncol(contrasts)), function(coef) {
        if (verbose) {
            cat(paste0("[fstat.comparisons.pipeline] ",
                       colnames(contrasts)[coef], "\n"))
        }
        bstat <- computeStat(bstat, coef = coef)
        dmrs <- dmrFinder(bstat, cutoff = cutoff, maxGap = maxGap.dmr,
                          verbose = verbose)
        if (!is.null(dmrs)) {
            meth <- getMeth(BSseq, dmrs, what = "perRegion")
            meth <- t(apply(meth, 1, function(xx) tapply(xx, fac, mean)))
            dmrs <- cbind(dmrs, meth)
            dmrs$maxDiff <- rowMaxs(meth) - rowMins(meth)
        }
        dmrs
    })
    names(l) <- colnames(contrasts)
    list(bstat = bstat, dmrs = l)
}

read.modkit <- function(files,
                        colData = NULL,
                        rmZeroCov = FALSE,
                        strandCollapse = TRUE){
    .Deprecated("read.bedMethyl")
    gr_list <- list()
    sampleNames <- sub("\\.bed.gz$", "", basename(files))
    if (!is.null(colData)){
        rownames(colData) <- sampleNames
        }

    for (i in seq_along(files)){
        data <- fread(files[i], header = FALSE, sep="\t",
                    stringsAsFactors=FALSE, quote="")
        data$V6[data$V6 == "."] <- "*"

        if (length(unique(data$V4)) == 2){
            gr <- GRanges(seqnames = data[data$V4 == "m", ]$V1,
                    ranges = IRanges(start = data[data$V4 == "m", ]$V2+1,
                                     end = data[data$V4 == "m", ]$V3),
                    strand = data[data$V4 == "m", ]$V6)

            mcols(gr)$m <- data[data$V4 == "m", ]$V12
            mcols(gr)$h <- data[data$V4 != "m", ]$V12
            mcols(gr)$u <- data[data$V4 == "m", ]$V13
            mcols(gr)$filter <- data[data$V4 == "m", ]$V16
        }else{
            gr <- GRanges(seqnames = data$V1,
                          ranges = IRanges(start = data$V2+1, end = data$V3),
                          strand = data$V6)

            mcols(gr)$m <- data$V12
            mcols(gr)$u <- data$V13
            mcols(gr)$filter <- data$V16
        }

        names(gr) <- sampleNames[i]
        gr_list[[sampleNames[i]]] <- gr
        }

    overlap_gr <- Reduce(subsetByOverlaps, gr_list)

    m_u_list <- lapply(gr_list, function(gr){
        overlap_data <- gr[gr %over% overlap_gr]
        if (!is.null(gr$h)) {
            data.frame(m = overlap_data$m, u = overlap_data$u,
                       h = overlap_data$h, filter = overlap_data$filter)
            } else {
                data.frame(m = overlap_data$m,  u = overlap_data$u,
                           filter = overlap_data$filter)
                }
        })

    m <- do.call(cbind, lapply(m_u_list, `[[`, "m"))
    u <- do.call(cbind, lapply(m_u_list, `[[`, "u"))
    h <- do.call(cbind, lapply(m_u_list, `[[`, "h"))
    filter <- do.call(cbind, lapply(m_u_list, `[[`, "filter"))

    if (!is.null(h)){
          bsseq_obj <- BSseq(M = as.matrix(m + h),
                             Cov = as.matrix(m + u + h),
                             Filtered = as.matrix(filter),
                             coef = NULL, se.coef = NULL,
                             pos = start(overlap_gr), trans = NULL,
                             parameters = NULL, pData = colData, gr = overlap_gr,
                             chr = as.vector(seqnames(overlap_gr)),
                             sampleNames = sampleNames, rmZeroCov = rmZeroCov)
          if (strandCollapse) {
            bsseq_obj <- strandCollapse(bsseq_obj)
          }
    }else{
      bsseq_obj <- BSseq(M = as.matrix(m), Cov = as.matrix(u + m),
                         Filtered = as.matrix(filter),
                         coef = NULL, se.coef = NULL,
                         pos = start(overlap_gr), trans = NULL,
                         parameters = NULL, pData = colData, gr = overlap_gr,
                         chr = as.vector(seqnames(overlap_gr)),
                         sampleNames = sampleNames, rmZeroCov = rmZeroCov)
      if (strandCollapse) {
        bsseq_obj <- strandCollapse(bsseq_obj)
      }
    }

    return(bsseq_obj)
}

context("read.modkit")

# TODO: Re-factor read.modkit() and update tests accordingly
test_that("read.modkit() works for BED files without 5hmc", {
    infile <- system.file("extdata", "modkit/chr21.chr22.HG002.top1000.bed.gz",
                          package = "bsseq")
    bsseq <- read.modkit(files = infile,
                         colData = NULL,
                         rmZeroCov = FALSE,
                         strandCollapse = TRUE)

    expect_is(bsseq, "BSseq")
})

test_that("read.modkit() works for BED files with 5hmc", {
    infile <- system.file("extdata", "modkit/Hypo1.first50Bed.txt",
                          package = "bsseq")
    bsseq <- read.modkit(files = infile,
                         colData = NULL,
                         rmZeroCov = FALSE,
                         strandCollapse = TRUE)

    expect_is(bsseq, "BSseq")
})

# Internal classes -------------------------------------------------------------

# The FWIRanges class is a simple container for storing a vector of fixed-width
# integer ranges.
# NOTE: The intention is to make this a fully-fledged IntegerRanges subclass
#       that is part of the IRanges package. For now, this class is internal to
#       bsseq and only a subset of methods are properly implemented.
.FWIRanges <- setClass(
    "FWIRanges",
    contains = "IPosRanges",
    representation(
        start = "integer",
        width = "integer",
        NAMES = "character_OR_NULL"  # R doesn't like @names !!
    )
)

# # Validity methods -------------------------------------------------------------

# TODO: Some of these may be redundant
.valid.FWIRanges <- function(x) {
    msg <- NULL
    if (!is.integer(x@width)) {
        msg <- validMsg(msg, "'width' must be an integer.")
    }
    if (length(x@width) != 1L ) {
        msg <- validMsg(msg, "'width' must have length 1.")
    }
    if (x@width < 0) {
        msg <- validMsg(msg, "'width' must be non-negative.")
    }
}

setValidity2("FWIRanges", .valid.FWIRanges)

# Internal methods -------------------------------------------------------------

setMethod("parallel_slot_names", "FWIRanges",
          function(x) c("start", "NAMES", callNextMethod())
)

setMethod("start", "FWIRanges", function(x, ...) x@start)

setMethod("width", "FWIRanges", function(x) rep.int(x@width, length(x)))

setMethod("end", "FWIRanges", function(x) {
    if (x@width == 1L) {
        x@start
    } else {
        width(x) - 1L + start(x)
    }
})

setMethod("names", "FWIRanges", function(x) x@NAMES)

# TODO: Room for optmisation (e.g., write in C to reduce memory allocations).
.set_FWIRanges_start <- function(x, value, check = TRUE) {
    if (!isTRUEorFALSE(check)) stop("'check' must be TRUE or FALSE")
    old_start <- start(x)
    new_start <- S4Vectors:::numeric2integer(value)
    new_width <- x@width - new_start + old_start
    if (any(new_width != new_width[1L])) {
        x <- as(x, "IRanges")
        start(x) <- value
        return(x)
    }
    x@width <- new_width[1L]
    if (check) validObject(x)
    x
}

setReplaceMethod(
    "start",
    "FWIRanges",
    function(x, ..., value) .set_FWIRanges_start(x, value)
)

.set_FWIRanges_end <- function(x, value, check = TRUE) {
    if (!isTRUEorFALSE(check)) stop("'check' must be TRUE or FALSE")
    new_width <- x@width + S4Vectors:::numeric2integer(value) - end(x)
    if (any(new_width != new_width[1L])) {
        x <- as(x, "IRanges")
        end(x) <- value
        return(x)
    }
    x@width <- new_width[1L]
    if (check) validObject(x)
    x
}

setReplaceMethod(
    "end",
    "FWIRanges",
    function(x, ..., value) .set_FWIRanges_end(x, value)
)

.set_FWIRanges_width <- function(x, value, check = TRUE) {
    if (!isTRUEorFALSE(check)) stop("'check' must be TRUE or FALSE")
    new_width <- S4Vectors:::numeric2integer(value)
    if (any(new_width != new_width[1L])) {
        x <- as(x, "IRanges")
        width(x) <- value
        return(x)
    }
    x@width <- new_width[1L]
    if (check) validObject(x)
    x
}

setReplaceMethod(
    "width",
    "FWIRanges",
    function(x, ..., value) .set_FWIRanges_width(x, value)
)

set_FWIRanges_names <- function(x, value) {
    x@NAMES <- S4Vectors:::normarg_names(value, class(x), length(x))
    # NOTE: No need to validate an FWIRanges object after setting its names so
    #       this should be safe.
    x
}

setReplaceMethod("names", "FWIRanges", set_FWIRanges_names)

setMethod("replaceROWS", "FWIRanges", function(x, i, value) {
    x_width <- x@width
    value_width <- value@width
    if (!identical(x_width, value_width)) {
        x <- as(x, "IRanges")
        value <- as(value, "IRanges")
        return(replaceROWS(x, i, value))
    }
    i <- normalizeSingleBracketSubscript(i, x, as.NSBS = TRUE)
    ans_start <- replaceROWS(start(x), i, start(value))
    ans_width <- value@width
    ans_mcols <- replaceROWS(mcols(x), i, mcols(value))
    BiocGenerics:::replaceSlots(
        x,
        start = ans_start,
        width = ans_width,
        mcols = ans_mcols,
        check = FALSE)
})

# TODO: Follow shift,IRanges-method
setMethod("shift", "FWIRanges", function(x, shift = 0L, use.names = TRUE) {
    stopifnot(use.names)
    shift <- recycleIntegerArg(shift, "shift", length(x))
    new_start <- start(x) + shift
    # TODO: Use BiocGenerics:::replaceSlots()
    x@start <- new_start
    validObject(x)
    x
})

##
.readbedMethylAsFWGRanges <- function(file, rmZeroCov = TRUE,
                                     strandCollapse = TRUE, sort = TRUE, 
                                     nThread = 1L, verbose = FALSE) {
  stopifnot(isTRUEorFALSE(rmZeroCov))
  stopifnot(isTRUEorFALSE(strandCollapse))
  stopifnot(isTRUEorFALSE(sort))
  M <- H <- C <- DE <- DI <- NO <- mod <- NULL
  if (rmZeroCov) {
    dt <- .readbedMethylAsDT(file = file, 
                            col_spec = "BSseq", 
                            check = TRUE,
                            verbose = verbose)
    if (strandCollapse && !is.null(dt[["strand"]]) && 
      !dt[, all(strand == "*")]) {
      dt[strand == "-", `:=`(start, start - 1L)][, `:=`(strand, NULL)]
      dt <- unique(dt[, list(M = sum(M), H = sum(H), C = sum(C), D = sum(DE + DI + NO)), 
                      by = c("seqnames", "start")])
    }
    dt <- dt[(M + H + C) > 0][, `:=`(c("M", "H", "C"), 
                                     list(NULL, NULL, NULL))] 
  } else {
    dt <- .readbedMethylAsDT(file = file, 
                            col_spec = "BSseq", 
                            check = TRUE, 
                            nThread = nThread, 
                            verbose = verbose)
    if (strandCollapse && !is.null(dt[["strand"]]) && !dt[,all(strand == "*")]) 
    {
      dt[strand == "-", `:=`(start, start - 1L)][, `:=`(strand, NULL)]
      dt <- data.table:::funique(dt)
    }
  }
  if (sort) {
    if (is.null(dt[["strand"]])) {
      setkey(dt, seqnames, start)
    } else {
      setkey(dt, seqnames, strand, start)
    }
  }
  seqnames <- Rle(dt[["seqnames"]])
  dt[, `:=`(seqnames, NULL)]
  seqinfo <- Seqinfo(seqnames = levels(seqnames))
  ranges <- .FWIRanges(start = dt[["start"]], width = 1L) #Check how this compares with bed format
  dt[, `:=`(start, NULL)]
  mcols <- make_zero_col_DFrame(length(ranges))
  if (is.null(dt[["strand"]])) {
    strand <- strand(Rle("*", length(seqnames)))
  } else {
    strand <- Rle(dt[["strand"]])
    dt[, `:=`(strand, NULL)]
  }
  fwgranges <- .FWGRanges(
    seqnames = seqnames, 
    ranges = ranges, 
    strand = strand, 
    seqinfo = seqinfo, 
    elementMetadata = mcols)
  if (sort) {
    fwgranges <- sort(sortSeqlevels(fwgranges))
  }
  fwgranges
}

## Right now this function use GRanges instead of FWGRanges
.constructFWGRangesFrombedMethylFiles <- function(files, 
                                                 rmZeroCov, 
                                                 strandCollapse,
                                                 verbose, 
                                                 nThread, 
                                                 BPPARAM) {
  subverbose <- max(as.integer(verbose) - 1L, 0L)
  if (verbose) {
    message("[.constructFWGRangesFrombedMethylFiles] Extracting loci from: ")
    message(writeLines(c(files)), sep = "\n")
  }
  loci_from_first_file <- .readbedMethylAsFWGRanges(
    file = files[[1L]], 
    rmZeroCov = rmZeroCov, 
    strandCollapse = strandCollapse, 
    nThread = nThread, 
    verbose = subverbose)
  if (is(BPPARAM, "SnowParam") && bpprogressbar(BPPARAM)) {
    bptasks(BPPARAM) <- length(files) - 1L
  }
  list_of_loci_from_other_files_not_in_first_file <- bplapply(files[-1L], 
    function(file, loci_from_first_file) {
    loci_from_this_file <- .readbedMethylAsFWGRanges(
      file = file, 
      rmZeroCov = rmZeroCov, 
      strandCollapse = strandCollapse, 
      verbose = subverbose)
    subsetByOverlaps(
      x = loci_from_this_file,
      ranges = loci_from_first_file,
      type = "equal", 
      invert = TRUE)
    }, loci_from_first_file = loci_from_first_file, 
    BPPARAM = BPPARAM)
  loci_non_found_in_first_file <- unique(
    do.call(c, list_of_loci_from_other_files_not_in_first_file))
  loci <- c(loci_from_first_file, loci_non_found_in_first_file)
  sort(sortSeqlevels(loci))
}

##
.readbedMethylAsDT <- function(file,
                              col_spec = c("all", "BSseq", "GRanges"),
                              check = FALSE, 
                              showProgress = FALSE,
                              nThread = 1L,
                              verbose = FALSE) {
  file <- file_path_as_absolute(file)
  col_spec <- match.arg(col_spec)
  stopifnot(isTRUEorFALSE(check))
  stopifnot(isTRUEorFALSE(showProgress))
  fread_verbose <- as.logical(max(verbose - 1L, 0L))
  
  header <- FALSE
  colClasses <- c("factor", "integer", "NULL", "factor", "NULL", 
                  "factor", "NULL", "NULL", "NULL", "NULL",
                  "NULL", "integer", "integer", "integer", "integer",
                  "NULL", "integer", "integer")
  drop <- c(3L, 5L, 7L, 8L, 9L, 10L, 11L, 16L)
  col.names <- c("seqnames", "start", "mod", "strand", "M", "C", "H", "DE", "DI", "NO")
  if (verbose) {
    message("[.readbedMethylAsDT] Parsing '", file, "'")
  }
  ptime1 <- proc.time()
  if (verbose) {
    message("[.readbedMethylAsDT] Reading file ...")
  }
  x <- fread(input = file, 
             sep = "\t", 
             header = header,
             verbose = fread_verbose, 
             drop = drop, 
             colClasses = colClasses, 
             col.names = col.names, 
             quote = "", 
             showProgress = showProgress, 
             nThread = nThread)
  
  if (!is.null(x[["strand"]])) {
    x[, `:=`(strand, strand(strand))]
  }
  setkey(x, mod)
  x<-x[.("m")] # remove all other mods except 5mC (they are still included in the count) 
  if (verbose) {
    message("Reading in 5mC and 5hmC")
  }
  
  M <- H <- C <- DE <- DI <- NO <- mod <- . <- NULL
  if (check && all(c("M", "H", "C") %in% colnames(x))) {
    if (verbose) {
      message("[.readbedMethylAsDT] Checking validity of counts in file.")
    }
    valid <- x[, isTRUE(all(M >= 0L & C >= 0L & H >= 0L & DE >= 0L & DI >= 0L & NO >= 0L))]
    if (!valid) {
      stop("[.readbedMethylAsDT] Invalid counts detected.\n", 
           "'M', 'H' and 'C' columns should be non-negative integers.")
    }
    else {
      if (verbose) {
        message("[.readbedMethylAsDT] All counts in file are valid.")
      }
    }
  }
  ptime2 <- proc.time()
  stime <- (ptime2 - ptime1)[3]
  if (verbose) {
    message("Done in ", round(stime, 1), " secs")
  }
  x
}


##
.constructCountsFromSinglebedMethylFile <- function(b, files, loci, strandCollapse,
                                                   grid, M_sink, H_sink, C_sink, D_sink,
                                                   Cov_sink, sink_lock, 
                                                   nThread, verbose) {
  subverbose <- max(as.integer(verbose) - 1L, 0L)
  M <- H <- C <- DE <- DI <- NO <- mod <- NULL
  file <- files[b]
  if (verbose) {
    message("[.constructCountsFrombedMethylFileAndFWGRanges] Extracting ", 
            "counts from ", "'", file, "'")
  }
  dt <- .readbedMethylAsDT(file = file, col_spec = "BSseq", check = TRUE, 
                          nThread = nThread, verbose = subverbose)
  if (strandCollapse && !is.null(dt[["strand"]]) && 
      !dt[, all(strand == "*")]) {
    dt[strand == "-", `:=`(start, start - 1L)][, `:=`(strand, NULL)]
    dt <- unique(dt[, list(M = sum(M), H = sum(H), C = sum(C), D = sum(DE + DI + NO)), 
                    by = c("seqnames", "start")])
  }
  seqnames <- Rle(dt[["seqnames"]])
  dt[, `:=`(seqnames, NULL)]
  seqinfo <- Seqinfo(seqnames = levels(seqnames))
  ranges <- .FWIRanges(start = dt[["start"]], width = 1L)
  dt[, `:=`(start, NULL)]
  mcols <- make_zero_col_DFrame(length(ranges))
  if (is.null(dt[["strand"]])) {
    strand <- strand(Rle("*", length(seqnames)))
  }
  else {
    strand <- Rle(dt[["strand"]])
    dt[, `:=`(strand, NULL)]
  }
  loci_from_this_sample <- .FWGRanges(
    seqnames = seqnames,
    ranges = ranges, 
    strand = strand, 
    seqinfo = seqinfo, 
    elementMetadata = mcols)
  ol <- findOverlaps(loci_from_this_sample, loci, type = "equal")
  M <- matrix(rep(0L, length(loci)), ncol = 1)
  H <- matrix(rep(0L, length(loci)), ncol = 1)
  C <- matrix(rep(0L, length(loci)), ncol = 1)
  D <- matrix(rep(0L, length(loci)), ncol = 1)
  Cov <- matrix(rep(0L, length(loci)), ncol = 1)
  M[subjectHits(ol)] <- dt[queryHits(ol), ][["M"]]
  H[subjectHits(ol)] <- dt[queryHits(ol), ][["H"]]
  C[subjectHits(ol)] <- dt[queryHits(ol), ][["C"]]
  D[subjectHits(ol)] <- dt[queryHits(ol), ][["D"]]
  Cov[subjectHits(ol)] <- dt[queryHits(ol), list(Cov = (M + H + C))][["Cov"]]
  if (is.null(M_sink)) {
    return(list(M = M, H = H, C = C, D = D, Cov = Cov))
  }
  viewport <- grid[[b]]
  ipclock(sink_lock)
  write_block(M_sink, viewport = viewport, block = M)
  write_block(H_sink, viewport = viewport, block = H)
  write_block(C_sink, viewport = viewport, block = C)
  write_block(D_sink, viewport = viewport, block = D)
  write_block(Cov_sink, viewport = viewport, block = Cov)
  ipcunlock(sink_lock)
  NULL
}

##
.constructbedMethylCounts <- function(files, loci, strandCollapse, BPPARAM,
                            BACKEND, dir, chunkdim, level, nThread,
                            verbose = FALSE) {
  subverbose <- max(as.integer(verbose) - 1L, 0L)
  ans_nrow <- length(loci)
  ans_ncol <- length(files)
  ans_dim <- c(ans_nrow, ans_ncol)
  grid <- RegularArrayGrid(refdim = ans_dim, spacings = c(ans_nrow,1L))
  if (is.null(BACKEND)) {
    M_sink <- NULL
    H_sink <- NULL
    C_sink <- NULL
    D_sink <- NULL
    Cov_sink <- NULL
    sink_lock <- NULL
  }
  else if (BACKEND == "HDF5Array") {
    h5_path <- file.path(dir, "assays.h5")
    M_sink <- HDF5RealizationSink(
      dim = ans_dim, 
      dimnames = NULL, 
      type = "integer", 
      filepath = h5_path, 
      name = "M", 
      chunkdim = chunkdim, 
      level = level)
    on.exit(close(M_sink), add = TRUE)
    H_sink <- HDF5RealizationSink(
      dim = ans_dim, 
      dimnames = NULL, 
      type = "integer", 
      filepath = h5_path, 
      name = "H", 
      chunkdim = chunkdim, 
      level = level)
    on.exit(close(H_sink), add = TRUE)
    C_sink <- HDF5RealizationSink(
      dim = ans_dim, 
      dimnames = NULL, 
      type = "integer", 
      filepath = h5_path, 
      name = "C", 
      chunkdim = chunkdim, 
      level = level)
    on.exit(close(C_sink), add = TRUE)
    D_sink <- HDF5RealizationSink(
      dim = ans_dim, 
      dimnames = NULL, 
      type = "integer", 
      filepath = h5_path, 
      name = "D", 
      chunkdim = chunkdim, 
      level = level)
    on.exit(close(D_sink), add = TRUE)
    Cov_sink <- HDF5RealizationSink(
      dim = ans_dim, 
      dimnames = NULL, 
      type = "integer", 
      filepath = h5_path, 
      name = "Cov", 
      chunkdim = chunkdim, 
      level = level)
    on.exit(close(Cov_sink), add = TRUE)
    sink_lock <- ipcid()
    on.exit(ipcremove(sink_lock), add = TRUE)
  } else {
    M_sink <- DelayedArray::AutoRealizationSink(
      dim = ans_dim, 
      type = "integer")
    on.exit(close(M_sink), add = TRUE)
    H_sink <- DelayedArray::AutoRealizationSink(
      dim = ans_dim, 
      type = "integer")
    on.exit(close(H_sink), add = TRUE)
    C_sink <- DelayedArray::AutoRealizationSink(
      dim = ans_dim, 
      type = "integer")
    on.exit(close(C_sink), add = TRUE)
    D_sink <- DelayedArray::AutoRealizationSink(
      dim = ans_dim, 
      type = "integer")
    on.exit(close(D_sink), add = TRUE)
    Cov_sink <- DelayedArray::AutoRealizationSink(
      dim = ans_dim, 
      type = "integer")
    on.exit(close(Cov_sink), add = TRUE)
    sink_lock <- ipcid()
    on.exit(ipcremove(sink_lock), add = TRUE)
  }
  if (is(BPPARAM, "SnowParam") && bpprogressbar(BPPARAM)) {
    bptasks(BPPARAM) <- length(grid)
  }
  counts <- bptry(bplapply(X = seq_along(grid), 
                           FUN = .constructCountsFromSinglebedMethylFile, 
                           files = files, 
                           loci = loci, 
                           strandCollapse = strandCollapse, 
                           grid = grid, 
                           M_sink = M_sink, 
                           H_sink = H_sink,
                           C_sink = C_sink,
                           D_sink = D_sink,
                           Cov_sink = Cov_sink, 
                           sink_lock = sink_lock, 
                           verbose = subverbose, 
                           nThread = nThread, 
                           BPPARAM = BPPARAM))
  if (!all(bpok(counts))) {
    stop(".constructbedMethylCounts() encountered errors for these files:\n  ",
         paste(files[!bpok], collapse = "\n  "))
  }
  if (is.null(BACKEND)) {
    M <- do.call(c, lapply(counts, "[[", "M"))
    attr(M, "dim") <- ans_dim
    H <- do.call(c, lapply(counts, "[[", "H"))
    attr(H, "dim") <- ans_dim
    C <- do.call(c, lapply(counts, "[[", "C"))
    attr(C, "dim") <- ans_dim
    D <- do.call(c, lapply(counts, "[[", "D"))
    attr(D, "dim") <- ans_dim
    Cov <- do.call(c, lapply(counts, "[[", "Cov"))
    attr(Cov, "dim") <- ans_dim
  } else {
    M <- as(M_sink, "DelayedArray")
    H <- as(H_sink, "DelayedArray")
    C <- as(C_sink, "DelayedArray")
    D <- as(D_sink, "DelayedArray")
    Cov <- as(Cov_sink, "DelayedArray")
  }
  return(list(M = M, H = H, C = C, D = D, Cov = Cov))
}

## 
read.bedMethyl <- function (files,
                            loci = NULL,
                            colData = NULL,
                            rmZeroCov = TRUE, 
                            strandCollapse = TRUE,
                            BPPARAM = bpparam(),
                            BACKEND = NULL, 
                            dir = tempfile("BSseq"),
                            replace = FALSE,
                            chunkdim = NULL, 
                            level = NULL,
                            nThread = 1L,
                            verbose = getOption("verbose")) {
  if (anyDuplicated(files)) {
    stop("'files' cannot have duplicate entries.")
  }
  file_exists <- file.exists(files)
  if (!isTRUE(all(file_exists))) {
    stop("These files cannot be found:\n  ", paste(files[!file_exists], 
                                                   collapse = "\n  "))
  }
  if (!is.null(loci)) {
    if (!is(loci, "GenomicRanges")) {
      stop("'loci' must be a GenomicRanges instance if not NULL.")
    }
    if (any(width(loci) != 1L)) {
      stop("All elements of 'loci' must have width equal to 1.")
    }
  }
  if (is.null(colData)) {
    colData <- DataFrame(row.names = files)
  }
  if (nrow(colData) != length(files)) {
    stop("Supplied 'colData' must have nrow(colData) == length(files).")
  }
  stopifnot(isTRUEorFALSE(rmZeroCov))
  stopifnot(isTRUEorFALSE(strandCollapse))
  current_BACKEND <- getAutoRealizationBackend()
  on.exit(setAutoRealizationBackend(current_BACKEND), add = TRUE)
  setAutoRealizationBackend(BACKEND)
  if (!.areBackendsInMemory(BACKEND)) {
    if (!.isSingleMachineBackend(BPPARAM)) {
      stop("The parallelisation strategy must use a single machine ", 
           "when using an on-disk realization backend.\n", 
           "See help(\"read.bismark\") for details.", call. = FALSE)
    }
  }
  else {
    if (!is.null(BACKEND)) {
      stop("The '", BACKEND, "' realization backend is not supported.", 
           "\n  See help(\"read.bismark\") for details.", 
           call. = FALSE)
    }
  }
  if (identical(BACKEND, "HDF5Array")) {
    if (!isSingleString(dir)) {
      stop(wmsg("'dir' must be a single string specifying the path to ", 
                "the directory where to save the BSseq object (the ", 
                "directory will be created)."))
    }
    if (!isTRUEorFALSE(replace)) {
      stop("'replace' must be TRUE or FALSE")
    }
    if (!dir.exists(dir)) {
      HDF5Array::create_dir(dir)
    }
    else {
      HDF5Array::replace_dir(dir, replace)
    }
  }
  subverbose <- as.logical(max(verbose - 1L, 0L))
  if (is.null(loci)) {
    ptime1 <- proc.time()
    if (verbose) {
      message("[read.bedMethyl] Parsing files and constructing valid loci ...")
    }
    loci <- .constructFWGRangesFrombedMethylFiles(
      files = files, 
      rmZeroCov = rmZeroCov, 
      strandCollapse = strandCollapse, 
      verbose = verbose, 
      nThread = nThread, 
      BPPARAM = BPPARAM) 
    ptime2 <- proc.time()
    stime <- (ptime2 - ptime1)[3]
    if (verbose) {
      message("Done in ", round(stime, 1), " secs")
    }
  }
  else {
    if (verbose) {
      message("[read.bedMethyl] Using 'loci' as candidate loci.")
    }
    if (strandCollapse) {
      if (verbose) {
        message("[read.bedMethyl] Collapsing strand of 'loci' ...")
      }
      ptime1 <- proc.time()
      loci <- .strandCollapse(loci)
      ptime2 <- proc.time()
      stime <- (ptime2 - ptime1)[3]
      if (verbose) {
        message("Done in ", round(stime, 1), " secs")
      }
    }
    if (rmZeroCov) {
      if (verbose) {
        message("[read.bedMethyl] Parsing files to identify elements of ", 
                "'loci' with non-zero coverage ...")
      }
      ptime1 <- proc.time()
      loci_from_files <- .constructFWGRangesFrombedMethylFiles(
        files = files, 
        rmZeroCov = rmZeroCov, 
        strandCollapse = strandCollapse, 
        verbose = subverbose, 
        nThread = nThread, 
        BPPARAM = BPPARAM)
      loci <- subsetByOverlaps(loci, loci_from_files, type = "equal")
      ptime2 <- proc.time()
      stime <- (ptime2 - ptime1)[3]
      if (verbose) {
        message("Done in ", round(stime, 1), " secs")
      }
    }
  }
  ptime1 <- proc.time()
  if (verbose) {
    message("[read.bedMethyl] Parsing files and constructing", 
            "'M', 'H', 'C', 'D' and 'Cov' ", "matrices ...")
  }
  counts <- .constructbedMethylCounts(
    files = files, 
    loci = loci, 
    strandCollapse = strandCollapse, 
    BPPARAM = BPPARAM, 
    BACKEND = BACKEND, 
    dir = dir, 
    chunkdim = chunkdim, 
    level = level, 
    nThread = nThread, 
    verbose = subverbose)
  ptime2 <- proc.time()
  stime <- (ptime2 - ptime1)[3]
  if (verbose) {
    message("Done in ", round(stime, 1), " secs")
  }
  if (verbose) {
    message("[read.bedMethyl] Constructing BSseq object ... ")
  }
  se <- SummarizedExperiment(assays = counts, rowRanges = as(loci, 
                                                             "GRanges"), colData = colData)
  bsseq <- new2("BSseq", se, check = FALSE)
  if (!is.null(BACKEND) && BACKEND == "HDF5Array") {
    x <- bsseq
    x@assays <- HDF5Array::shorten_assay2h5_links(x@assays)
    saveRDS(x, file = file.path(dir, "se.rds"))
  }
  bsseq
}

# Internal functions -----------------------------------------------------------

# TODO: Option to specify that objects should be combined into DelayedArray
#       with a given backend (e.g., even if all inputs are matrix)?
.combineList <- function(x, x_rowRanges, ans_rowRanges) {
    x_is_matrix <- vapply(x, is.matrix, logical(1L))
    if (isTRUE(all(x_is_matrix))) {
        return(.combineList_matrix(x, x_rowRanges, ans_rowRanges))
    }
    .combineList_matrix_like(x, x_rowRanges, ans_rowRanges)
}

.combineList_matrix <- function(x, x_rowRanges, ans_rowRanges) {
    stopifnot(identical(length(x), length(x_rowRanges)))
    # Set up output matrix with appropriate dimension and type
    # TODO: Modify minfi:::.highestType() to accept a list. Can then do
    #       minfi:::.highestType(x)
    x_types <- vapply(x, DelayedArray::type, character(1L))
    ans_type <- typeof(do.call("c", lapply(x_types, vector)))
    ans <- matrix(
        data = .zero_type(ans_type),
        nrow = length(ans_rowRanges),
        ncol = sum(vapply(x, ncol, integer(1L))))

    # Fill output matrix
    col_offset <- 0L
    for (k in seq_along(x_rowRanges)) {
        ol <- findOverlaps(x_rowRanges[[k]], ans_rowRanges, type = "equal")
        # row_idx <- findOverlaps(
        #     query = x_rowRanges[[k]],
        #     subject = ans_rowRanges,
        #     type = "equal",
        #     select = "first")
        col_idx <- col_offset + seq_len(ncol(x[[k]]))
        ans[subjectHits(ol), col_idx] <- x[[k]][queryHits(ol), , drop = FALSE]
        col_offset <- col_offset + ncol(x[[k]])
    }
    # Return output matrix
    ans
}

.combineList_matrix_like <- function(x, x_rowRanges, ans_rowRanges) {
    # Argument checks
    stopifnot(identical(length(x), length(x_rowRanges)))

    # Construct data frame mapping each sample to an element of x and its
    # corresponding column number.
    x_ncol <- vapply(x, ncol, integer(1L))
    x_samples_df <- data.frame(
        sample = seq(1, sum(x_ncol)),
        x_idx = rep(seq_along(x), times = x_ncol),
        column_idx = unlist(lapply(x_ncol, function(end) seq(1, end))))

    # Set up intermediate RealizationSink object of appropriate dimensions
    # and type
    # TODO: Modify minfi:::.highestType() to accept a list. Can then do
    #       minfi:::.highestType(x)
    x_types <- vapply(x, DelayedArray::type, character(1L))
    ans_type <- typeof(do.call("c", lapply(x_types, vector)))
    sink <- DelayedArray::AutoRealizationSink(
        dim = c(length(ans_rowRanges), sum(vapply(x, ncol, integer(1L)))),
        type = ans_type)
    on.exit(close(sink))

    # Set up ArrayGrid instance over columns of `sink`.
    sink_grid <- colAutoGrid(sink)

    # Loop over column grid of 'sink', identify samples required for that
    # block, bring that data into memory, pass down to .combineList_matrix(),
    # and write result to sink.
    for (k in seq_along(sink_grid)) {
        sink_viewport <- sink_grid[[k]]
        block_samples <- seq(start(sink_viewport)[2L], end(sink_viewport)[2L])
        block_samples_df <- x_samples_df[
            x_samples_df[["sample"]] %in% block_samples, ]
        x_to_load <- unique(block_samples_df[["x_idx"]])
        block_x <- lapply(x_to_load, function(idx) {
            columns_to_load <- block_samples_df[
                block_samples_df[["x_idx"]] == idx, "column_idx"]
            as.matrix(x[[idx]][, columns_to_load, drop = FALSE])
        })
        block_rowRanges <- x_rowRanges[x_to_load]
        block_ans <- .combineList_matrix(
            x = block_x,
            x_rowRanges = block_rowRanges,
            ans_rowRanges = ans_rowRanges)
        write_block(sink, viewport = sink_viewport, block = block_ans)
    }

    as(sink, "DelayedArray")
}

# Exported methods -------------------------------------------------------------

setMethod("combine", signature(x = "BSseq", y = "BSseq"), function(x, y, ...) {
    combineList(list(x, y))
})

# Exported functions -----------------------------------------------------------

# TODO: Check sampleNames are disjoint
# TODO: Should this have a BACKEND argument?
combineList <- function(x, ..., BACKEND = NULL) {
    # Argument checks ----------------------------------------------------------

    # Check inputs are BSseq objects
    if (is(x, "BSseq")) x <- list(x, ...)
    x_is_BSseq <- vapply(x, is, logical(1L), "BSseq")
    stopifnot(isTRUE(all(x_is_BSseq)))
    # Check inputs are combinable
    x_trans <- lapply(x, getBSseq, "trans")
    x_has_same_trans <- vapply(
        x_trans,
        function(t) isTRUE(all.equal(t, x_trans[[1L]], check.environment = FALSE)),
        logical(1L))
    stopifnot(isTRUE(all(x_has_same_trans)))
    x_parameters <- lapply(x, getBSseq, "parameters")
    x_has_same_parameters <- vapply(
        x_parameters,
        function(p) isTRUE(all.equal(p, x_parameters[[1L]])),
        logical(1L))
    stopifnot(isTRUE(all(x_has_same_parameters)))
    # Check if all inputs have the same set of loci
    x_rowRanges <- lapply(x, rowRanges)
    # TODO: Check if all loci are identical()/all.equal(). This is must faster
    #       that doing all these reduce()-ing intersect()-ing
    ans_rowRanges <- Reduce(
        f = function(x, y) {
            reduce(c(x, y), drop.empty.ranges = TRUE, min.gapwidth = 0L)
        },
        x = x_rowRanges)
    intersect_rowRanges <- Reduce(intersect, x_rowRanges)
    all_x_have_same_loci <- identical(ans_rowRanges, intersect_rowRanges)
    # Check if safe to combine smoothed representations (coef and se.coef).
    # It's only safe if all objects contain the same set of loci.
    x_has_been_smoothed <- vapply(x, hasBeenSmoothed, logical(1L))
    if (all_x_have_same_loci && all(x_has_been_smoothed)) {
        combine_smooths <- TRUE
        ans_trans <- x_trans[[1L]]
        ans_parameters <- x_parameters[[1L]]
    } else {
        if (any(x_has_been_smoothed)) {
            warning("Combining smoothed and unsmoothed BSseq objects. You ",
                    "will need to re-smooth using 'BSmooth()' on the ",
                    "returned object.")
        }
        if (!all_x_have_same_loci && any(x_has_been_smoothed)) {
            warning("Combining BSseq objects with different loci. You ",
                    "will need to re-smooth using 'BSmooth()' on the ",
                    "returned object.")
        }
        combine_smooths <- FALSE
        ans_coef <- NULL
        ans_se.coef <- NULL
        ans_trans <- function() NULL
        environment(ans_trans) <- emptyenv()
        ans_parameters <- list()
    }

    # Combine elements of BSseq objects ----------------------------------------

    # Combine colData
    ans_colData <- as(
        Reduce(combine, lapply(x, function(xx) as.data.frame(colData(xx)))),
        "DataFrame")
    # Extract assays to be combined
    # NOTE: Use of `assay(..., withDimnames = FALSE)` is deliberate
    x_M <- lapply(x, assay, "M", withDimnames = FALSE)
    x_Cov <- lapply(x, assay, "Cov", withDimnames = FALSE)
    if (combine_smooths) {
        x_coef <- lapply(x, getBSseq, "coef")
        x_se.coef <- lapply(x, getBSseq, "se.coef")
    }
    # Combine assay data
    if (all_x_have_same_loci) {
        x_has_same_rowRanges_as_ans_rowRanges <- vapply(
            x_rowRanges,
            function(r) isTRUE(all.equal(r, ans_rowRanges)),
            logical(1L))
        if (all(x_has_same_rowRanges_as_ans_rowRanges)) {
            # If all rowRanges are identical then can just cbind assays.
            ans_M <- do.call(cbind, x_M)
            ans_Cov <- do.call(cbind, x_Cov)
            if (combine_smooths) {
                ans_coef <- do.call(cbind, x_coef)
                ans_se.coef <- do.call(cbind, x_se.coef)
            }
        } else {
            # If all rowRanges contain same loci but aren't identical then they
            # must differ by how loci are ordered. So order all inputs using a
            # common order and then cbind assays.
            x_order <- lapply(x_rowRanges, function(rr) {
                seqlevels(rr) <- seqlevels(ans_rowRanges)
                order(rr)
            })
            ans_M <- do.call(cbind, Map(extractROWS, x_M, x_order))
            ans_Cov <- do.call(cbind, Map(extractROWS, x_Cov, x_order))
            if (combine_smooths) {
                ans_coef <- do.call(cbind, Map(extractROWS, x_coef, x_order))
                ans_se.coef <- do.call(cbind, Map(extractROWS, x_se.coef, x_order))
            }
        }
    } else {
        ans_M <- .combineList(x_M, x_rowRanges, ans_rowRanges)
        ans_Cov <- .combineList(x_Cov, x_rowRanges, ans_rowRanges)
    }

    # Construct BSseq object ---------------------------------------------------

    ans_assays <- SimpleList(M = ans_M, Cov = ans_Cov)
    if (combine_smooths) {
        ans_assays <- c(ans_assays, SimpleList(coef = ans_coef))
        if (!is.null(ans_se.coef)) {
            ans_assays <- c(ans_assays, SimpleList(se.coef = ans_se.coef))
        }
    }
    se <- SummarizedExperiment(
        assays = ans_assays,
        rowRanges = ans_rowRanges,
        colData = ans_colData)
    # TODO: Avoid validity check.
    .BSseq(se, parameters = ans_parameters, trans = ans_trans)
}

read.modbam2bed <- function(files, colData = NULL, rmZeroCov = FALSE,
                            strandCollapse = TRUE) {
    .Deprecated("read.bedMethyl")
    gr_list <- list()
    sampleNames <- sub("\\.bed$","",basename(files))
    if (!is.null(colData)) {
        rownames(colData) <- sampleNames
        }

    for (i in seq_along(files)) {
        data <- read.table(files[i],header = FALSE, sep="\t",
                    stringsAsFactors=FALSE, quote="")
        gr <- GRanges(seqnames = data$V1,
            ranges = IRanges(start = data$V2+1, end = data$V3))
        mcols(gr)$m <- data$V13
        mcols(gr)$cov <- data$V12 + data$V13
        mcols(gr)$filter <- data$V14
        names(gr) <- sampleNames[i]
        gr_list[[sampleNames[i]]] <- gr
        }

    overlap_gr <- Reduce(subsetByOverlaps, gr_list)

    m_cov_list <- lapply(gr_list, function(gr) {
        overlap_data <- gr[gr %over% overlap_gr]
        data.frame(m = overlap_data$m, cov = overlap_data$cov,
            filter = overlap_data$filter)})

    m <- do.call(cbind,lapply(m_cov_list,`[[`, "m"))
    cov <- do.call(cbind, lapply(m_cov_list, `[[`, "cov"))
    filter <- do.call(cbind, lapply(m_cov_list, `[[`, "filter"))

    bsseq_obj <- BSseq(M = as.matrix(m), Cov = as.matrix(cov),
                     Filtered = as.matrix(filter),
                     coef = NULL,se.coef = NULL,
                     pos = start(overlap_gr),trans = NULL,
                     parameters = NULL, pData = colData, gr = NULL,
                     chr = as.vector(seqnames(overlap_gr)),
                     sampleNames = sampleNames,rmZeroCov = rmZeroCov)

    if (strandCollapse) {
        strandCollapse(bsseq_obj)
    }

    return(bsseq_obj)
}

context("read.modbam2bed")

# TODO: Re-factor read.modbam2bed() and update tests accordingly
test_that("read.modbam2bed() works for BED files", {
  infile <- system.file("extdata", "modbam2bed/ctr1.chr10.chr11.bed.gz",package = "bsseq")
  bsseq <- read.modbam2bed(files = infile,
                        colData = NULL,
                        rmZeroCov = FALSE,
                        strandCollapse = TRUE)
  expect_is(bsseq, "BSseq")
})

---
title: "The bsseq User's Guide"
shorttitle: "bsseq guide"
author: 
  - Kasper Daniel Hansen
package: bsseq
bibliography: bsseq.bib
abstract: >
  A comprehensive guide to using the bsseq package for analyzing whole-genome bisulfite sequencing
  (WGBS) datta.
vignette: >
  %\VignetteIndexEntry{bsseq User's Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output: 
  BiocStyle::html_document:
    toc_float: true
---

# Introduction

This R package is the reference implementation of the BSmooth algorithm for analyzing whole-genome bisulfite sequencing (WGBS) data.  This package does **not** contain alignment software, which is available from [](http://rafalab.jhsph.edu/bsmooth).  This package is not intended for analysis of single-loci bisulfite sequencing (typically Sanger bisulfite sequencing or pyro bisulfite sequencing).

The package has been used to analyze capture bisulfite sequencing data.  For this type of data, the most useful parts of the package is its data-handling functions.  The BSmooth algorithm itself may not be useful for a standard capture bisulfite sequencing experiment, since it relies heavily on smoothing, which again requires that we have measured methylation in bigger regions of the genome.

The BSmooth algorithm is described in detail in [@Hansen:2012].  It was applied to human cancer data in [@Hansen:2011] and we have also used it to analyze data from Rhesus Macaque [@Tung:2012].  Specifically, the algorithm uses smoothing to get reliable semi-local methylation estimates from low-coverage bisulfite sequencing.  After smoothing it uses biological replicates to estimate biological variation and identify differentially methylated regions (DMRs).  The smoothing portion could be used even on a single sample, but we believe that variation between individuals is an important aspect of DNA methylation and needs to be accounted for, see also [@Hansen:2011rna] for a relevant discussion.

The main focus for our development has been the analysis of CpG methylation in humans, and we have successfully used it in other primates [@Tung:2012].  It is highly likely that the approach will work in non-human organisms, but care must be taken: the smoothing parameters (which we have selected based on human data) should be evaluated carefully.  Furthermore, it may not be suitable at all for organisms with vastly different (from humans) methylation structures.

With respect to non-CpG methylation, the situation is mixed.  We have never used the algorithm to analyze non-CpG methylation, but it should be straightforward to do so.  However, the data structures used in the current code might not conveniently scale from the 28.2M CpGs in the human genome to the roughly 2x585M Cs (it may well be possible to do an separate analysis for each chromosome).  This should be less of an issue for organisms with smaller genomes.  We are considering changing these underlying data structures to allow for easier analysis of non-CpG methylation in humans.

## System Requirements

The package requires that all the data is loaded into system memory.  By ``data'' we do not mean the individual reads (which is big for a whole-genome experiment).  Instead, what we need are summarized data given us the number of reads supporting methylation as well as the total number of reads at
each loci.  Focusing on human data, we need to work with objects with a maximum of 28.2 million entries, per sample (since there are roughly 28.2 millions CpGs in the human genome).  This provides us with an upper limit on the data object.

Based on this, the 8 samples from [@Hansen:2011] including the smoothed values, take up roughly 1.2GB of RAM, meaning an analysis can easily be done with 8GB of RAM.  In order to improve speed, the package allows for easy parallel processing of samples/chromosomes.  This will require multiple copies of the data object for each core used, increasing the memory usage substantially to perhaps even 32GB.  This can be avoided by processing the samples sequentially at the cost of speed.

On a 64GB node, the 8 samples from [@Hansen:2011] takes roughly one hour to process in parallel using 8 cores (using much less than 64GB of RAM in total).  This does not including parsing the data from the alignment output files.

## Some terminology

Because not all methylation happens at CpG sites, we have tried to use the term ``methylation loci'' instead of CpG.  We use this term to refer to a single-base methylation site.

Some standard terms from the DNA methylation field: differentially methylated region (DMR), hyper methylation (methylation that is higher in one condition/sample than in another), hypo methylation (methylation that is lower in one condition/sample than in another), and finally differentially methylated position (DMP) referring to a single loci.

## Citation

If you use this package, please cite our BSmooth paper [@Hansen:2012].

## Dependencies 

```{r dependencies, warning=FALSE, message=FALSE}
library(bsseq)
```

# Overview

The package assumes that the following data has been extracted from alignments:

- genomic positions, including chromosome and location, for methylation loci.
- a (matrix) of M (Methylation) values, describing the number of read supporting methylation covering a single loci.  Each row in this matrix is a methylation loci and each column is a sample.
- a (matrix) of Cov (Coverage) values, describing the total number of reads covering a single loci.  Each row in this matrix is a methylation loci and each column is a sample.


We can have a look at some data from [@Lister:2009], specifically chromosome 22 from the IMR90 cell line.

```{r BSchr22}
data(BS.chr22)
BS.chr22
```

The genomic positions are stored as a `GRanges` object `GRanges` are general genomic regions; we represent a single base methylation loci as an interval of width 1 (which may seem a bit strange, but there are good reasons for this).  For example, the first 4 loci in the Lister data are

```{r ov-granges}
head(granges(BS.chr22), n = 4)
```

We also have the M and Cov matrices

```{r ov-getCoverage}
head(getCoverage(BS.chr22, type = "M"), n = 4)
head(getCoverage(BS.chr22), n = 4)
``` 

Since CpG methylation is symmetric on the two strands of a chromosome, we aggregated reads on the forward and reverse strand to form a single value, and we assume the genomic position points to the C of the CpG.  It is not crucial in any way to do this, one may easily analyze each strand separately, but CpG methylation is symmetric and this halves the number of loci.

How to input your methylation data into this data structure (called a `BSseq` object) is described in a section below.  We also have a section on how to operate on these types of objects.

An analysis typically consists of the following steps.

1. Smoothing, using the function `BSmooth()`.
2. Compute t-statistics, using the function `BSmooth.tstat()`.  This converts the `BSseq` object into a `BSseqTstat` object.
3. Threshold these t-statistics to identify DMRs, using the function `dmrFinder()`, returning a simple `data.frame`.

It is usually a good idea to look at the smoothed data either before or after identifying DMRs.  This can be done using the functions `plotRegion()` and `plotManyRegions()`.

We also have functions for assessing goodness of fit for binomial and poison models; this is like to be of marginal interest to most users.  See the man page for `binomialGoodnessOfFit()`.

We also allow for easy computation of Fisher's exact test for each loci, using the function `fisherTests()`.

# Using objects of class BSseq

## Basic operations

Objects of class `BSseq` contains a `GRanges` object with the genomic locations.  This `GRanges` object can be obtained by `granges()`.  A number of standard `GRanges` methods works directly on the `BSseq` object, such as `start()`, `end()`, `seqnames()` (chromosome names) etc.

These objects also contains a `phenoData` object for sample pheno data.  Useful methods are `sampleNames()`, `pData()`.

Finally, we have methods such as `dim()`, `ncol()` (number of columns; number of samples), `nrow()` (number of rows; number of methylation loci).

Objects can be subsetted using two indicies `BSseq[i,j]` with the first index being methylation loci and the second index being samples.  Another very useful way of subsetting the object is by using the method `subsetByOverlaps()`.  This selects all methylation loci inside a set of genomic intervals (there is a difference between first and second argument and either can be
`BSseq` or `GRanges`).

Examples:
```{r using-examples}
head(start(BS.chr22), n = 4)
head(seqnames(BS.chr22), n = 4)
sampleNames(BS.chr22)
pData(BS.chr22)
dim(BS.chr22)
BS.chr22[1:6,1]
subsetByOverlaps(BS.chr22, 
                 GRanges(seqnames = "chr22", 
                         ranges = IRanges(start = 1, end = 2*10^7)))
``` 

## Data handling 

We have a number of functions for manipulating one or more `BSseq` objects.

`BSseq()` instantiates an object of class `BSseq`.  Genomic locations are passed in, either as a `GRanges` object (argument `gr`) or as chromosome and location vectors (arguments `chr` and `pos`).  The arguments `M` and `Cov` accepts matrices, and it is possible to directly give it a `phenoData` object.

```{r data-bsseq}
M <- matrix(0:8, 3, 3)
Cov <- matrix(1:9, 3, 3)
BStmp <- BSseq(chr = c("chr1", "chrX", "chr1"), pos = 1:3, 
               M = M, Cov = Cov, sampleNames = c("A1","A2", "B"))
``` 

A `BSseq` object may be ordered by `orderBSseq()`.  This ensures that data from a single chromosome appears in an ordered, contiguous block.  There is also the possibility for specifying the chromosome order (this is less important).  The smoothing functions assumes that the underlying `BSseq` has been ordered.

```{r data-order}
granges(BStmp)
BStmp <- orderBSseq(BStmp, seqOrder = c("chr1", "chrX"))
granges(BStmp)
``` 

`chrSelectBSseq()` performs the often useful task of selecting one or more chromosomes and can also order the output.  In case `order = TRUE`, the output is ordered according to the order of the `seqnames` argument.
```{r data-chrSelect}
chrSelectBSseq(BStmp, seqnames = "chr1", order = TRUE)
``` 
Of course, in this case the function does little, since `BS.chr22` already only contains data
from chromosome 22.

`combine()` combines two `BSseq` objects in the following way: the samples for the return objects is the union of the samples from the two objects, and the methylation loci are the union of the two methylation loci.  The two objects do not need to have measured the same loci (in the example below, `BStmp` has data on chromosome 1 and X).

```{r data-combine}
BStmp2 <- combine(BStmp, BS.chr22[1:3,])
granges(BStmp2)
getCoverage(BStmp2)
``` 

`collapseBSseq` performs the often useful task of adding several columns of a `BSseq` object.  This is often used in the beginning of an analysis where each column may correspond to a lane and several such columns represents the data for a single biological sample.

```{r data-collapse}
collapseBSseq(BStmp, group = c("A", "A", "B"))
``` 

## Obtaining coverage (methylation)

Coverage, either Cov or M values, are obtained by `getCoverage()`, using the `type` argument:

```{r getCoverage1}
head(getCoverage(BS.chr22, type = "Cov"), n = 4)
head(getCoverage(BS.chr22, type = "M"), n = 4)
``` 

This will return a -- possibly very large -- matrix.  It is also possible to get region based coverage by using the `regions` argument to the function.  This argument is either a `data.frame` (with columns `chr`, `start` and `end`) or a `GRanges` object.  Let us do an example
```{r regions}
regions <- GRanges(seqnames = c("chr22", "chr22"), 
                   ranges = IRanges(start = 1.5 * 10^7 + c(0,200000), 
                                    width = 1000))
getCoverage(BS.chr22, regions = regions, what = "perRegionTotal")
``` 
When `what` is `perRegionTotal` the return value is the total coverage of each region (and note that regions without coverage return `NA`).  Similarly, `perRegionAverage` yields the average coverage in the region.  However, it is often useful to get the actual values, like
```{r getCoverage2}
getCoverage(BS.chr22, regions = regions, what = "perBase")
``` 
This is the default behaviour, and it returns a list where each element corresponds to a region.  Note that regions with no coverage gets a `NULL`.

Methylation estimates can be obtained in the same way, using the function `getMeth()`.  If `type` is set to `raw` the function returns simple single-loci methylation estimates (which are M/Cov).  To get smoothed estimates, the `BSseq` object needs to have been smoothed using Bsmooth, and `type` set to `smooth` (default).  The `getCoverage(0`, `getMeth()` has a `regions` and a `what` argument.  For `getMeth()` the `what` argument can be `perBase` or `perRegion` (the later is really the per-region average methylation).  Additionally, confidence intervals can be computed using a method taking possibly low coverage loci into account as well as loci where the methylation percentage might be close to 0 or 1 [@Agresti:1998].  Currently, confidence intervals cannot be computed for region-level summary estimates.  Examples

```{r getMeth}
getMeth(BS.chr22, regions, type = "raw")
getMeth(BS.chr22, regions[2], type = "raw", what = "perBase")
``` 

# Reading data

## Alignment output from the BSmooth alignment suite

It is straightforward to read output files (summarized evidence) from the BSmooth alignment suite, using `read.bsmooth()`.  This function takes as argument a number of directories, usually corresponding to samples, with the alignment output.  Sometimes, one might want to read only certain chromosomes, which can be accomplished by the `seqnames` argument.  Also, the default values of `qualityCutoff = 20` ensures that we do not use methylation evidence with a base quality strictly lower than 20 (since we may not be confident whether the read really supports methylation or not).  The function `read.bsmooth()` allows for both gzipped and non-gzipped input files. It is faster to read gzipped files, so we recommend gzipping post-alignment.

During the development of BSmooth we experimented with a number of different output formats.  These legacy formats can be read with `read.umtab()` and `read.umtab2()`.

## Alignment output from other aligners

We are interested in adding additional parsers to the package; if your favorite alignment software is not supported, feel free to get in touch.

In general, we need summarized methylation evidence.  In short, for each methylation loci, we need to know how many reads cover the loci and how many of those reads support methylation.

As an example, consider the Lister data.  The files posted online looks like
```
> head mc_imr90_r1_22
assembly        position        strand  class   mc      h
22      14430632        +       CG      9       10
22      14430633        -       CG      8       8
22      14430677        +       CG      1       1
22      14430678        -       CG      3       10
22      14430688        -       CG      6       10
22      14430703        -       CG      2       8
22      14431244        +       CG      5       10
22      14431245        -       CG      5       11
```

For these files, the evidence is split by strand.  It is often easiest to read in one sample at a time, instantiate a `BSseq` object and then use `combine()` and `collapseBSseq()` to combine the samples, since these functions deal objects that have different sets of CpGs.  In the Lister data, note that the position is the position of the "C", so basically, if you want to combine the evidence from both strands, CpGs on the "-" strand needs to have 1 subtracted from their position.  A full script showing how these files are used to create `BS.chr22` can be found in `inst/scripts/get\_BS.chr22.R` in the `r Biocpkg("bsseq")` package.  The path of this file may be found by
```{r locateScript, results="hide"}
system.file("scripts", "get_BS.chr22.R", package = "bsseq")
``` 

# Analysis

Computing smoothed methylation levels is simple.  We subset ``BS.chr22` so we only smooth 1 sample, for run-time purpose

```{r BSmooth}
BS.chr22.1 <- BSmooth(BS.chr22[,"r1"], verbose = TRUE)
BS.chr22.1
``` 

A number of arguments pertains to using multiple cores.  This is useful if you have multiple samples or chromosomes.  `mc.cores` tells `BSmooth()` how many cores to use and `parallelBy` describes whether the parallelization is over samples or chromosomes.  If the parallelization is over samples, each core will compute the smoothing of the entire genome.  And if it is over samples, each cores will smooth all samples for one or more chromosomes.  The appropriate choice depends on the number of samples and cpu cores available.  If you have the exact same number of samples as cores, the fastest is `parallelBy = "sample"`.  If you have less samples than cores, the fastest is `parallelBy = "chromosome"`.  The argument `mc.preshedule` should not need to be changed (unless perhaps if a small value of `maxGap` is uses); see the man page for `parallel::mclapply`.  Note that setting `mc.cores` to a value greater than 1 is not support on MS Windows due to a limitation of the operating system.

For a more detailed discussion of the analysis tools, read the companion vignette ``Analyzing WGBS with the bsseq package'', which also finding DMRs and plotting them.

Fisher's exact tests may be efficiently computed using the function `fisherTests`()`.

Binomial and poisson goodness of fit tests statistics may be computed using `binomialGoodnessOfFit()` and `poissonGoodnessOfFit()`.

# sessionInfo()

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

# References

<!--
Local Variables:
% eval: (add-hook 'LaTeX-mode-hook '(lambda () (if (string= (buffer-name) "bsseq.Rnw") (setq fill-column 100))))
% LocalWords: LocalWords bisulfite methylation methylated CpG CpGs DMR
% LocalWords: DMRs differentially
% End:
-->

---
title: "Analyzing WGBS data with bsseq"
shorttitle: "bsseq analysis"
author: 
  - Kasper Daniel Hansen
package: bsseq
bibliography: bsseq.bib
abstract: >
  An informal guide to analyzing WGBS using bsseq.
  (WGBS) datta.
vignette: >
  %\VignetteIndexEntry{Analyzing WGBS data with bsseq}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output: 
  BiocStyle::html_document:
    toc_float: true
---

# Introduction

This document discusses the ins and outs of an analysis of a whole-genome shotgun bisulfite sequencing (WGBS) dataset, using the BSmooth algorithm, which was first used in [@Hansen:2011] and more formally presented and evaluated in [@Hansen:2012].  The intention with the document is to focus on analysis-related tasks and questions.  Basic usage of the `r Biocpkg("bsseq")` package is covered in "The bsseq user's guide".  It may be useful to consult the user's guide while reading this analysis guide.

In this vignette we analyze chromosome 21 and 22 from [@Hansen:2011].  This is primary data from 3 patients with colon cancer.  For each patient we have normal colon tissue as well as cancer colon.  The samples were run on ABI SOLiD and we generated 50bp single-end reads.  The reads were aligned using the Merman aligner in the BSmooth suite [](http://www.rafalab.org/bsmooth).  See the primary publication for more details [@Hansen:2011].

This data is contained in the `r Biocpkg("bsseqData")`

```{r dependencies, warning=FALSE, message=FALSE}
library(bsseq)
library(bsseqData)
```

The `r Biocpkg("bsseqData")` contains a script, `inst/script/create\_BS.cancer.R`, describing how this data is created from the Merman alignment output (also contained in the package).  Note that the current version of the BSmooth pipeline uses a slightly different alignment output format.

The following object contains the unsmoothed "raw" summarized alignment data.
```{r showData}
data(BS.cancer.ex)
BS.cancer.ex <- updateObject(BS.cancer.ex)
BS.cancer.ex
pData(BS.cancer.ex)
``` 

If you use this package, please cite our BSmooth paper [@Hansen:2012].

# Smoothing

The first step of the analysis is to smooth the data

```{r smooth,eval=FALSE}
BS.cancer.ex.fit <- BSmooth(
    BSseq = BS.cancer.ex, 
    BPPARAM = MulticoreParam(workers = 1), 
    verbose = TRUE)
``` 
This particular piece of code is not being run when the vignette is being created.  It takes roughly 2 minutes per sample.  If you have 6 cores available, use `mc.cores = 6` and the total run time will be roughly 2 minutes.  Note that setting `mc.cores` to a value greater than 1 is not support on MS Windows due to a limitation of the operating system.

For ease of use, the `r Biocpkg("bsseqData")` includes the result of this command:
```{r showDataFit}
data(BS.cancer.ex.fit)
BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
BS.cancer.ex.fit
``` 

This step uses parallelization where each sample is run on a separate core using `mclapply()` from the `r CRANpkg("parallel")` package.  This form of parallelization is built into `r Biocpkg("bsseq")`, and (as written) requires access to a machine with 6 cores and enough RAM.  The smoothing step is being done completely independently on each sample, so if you have a lot of samples (or other circumstances), an alternative is to split the computations manually.  A later subsection shows some example code for doing that.

Let us discuss coverage and representation.  The `BS.cancer.ex` object contains all annotated CpGs on human chromosome 21 and 22, whether or not there is actual data.  Since we have multiple samples, we can roughly divide the genome into 3 categories: CpGs where all samples have data, CpGs where none of the samples have data and CpGs where some, but not all, of the samples have data.  Examining the object at hand, we get

```{r cpgNumbers}
## The average coverage of CpGs on the two chromosomes
round(colMeans(getCoverage(BS.cancer.ex)), 1)
## Number of CpGs in two chromosomes
length(BS.cancer.ex)
## Number of CpGs which are covered by at least 1 read in all 6 samples
sum(rowSums(getCoverage(BS.cancer.ex) >= 1) == 6)
## Number of CpGs with 0 coverage in all samples
sum(rowSums(getCoverage(BS.cancer.ex)) == 0)
``` 

The CpG coverage is roughly 4x, so we would expect many zero coverage CpGs by chance. although that should not necessarily occur in all 6 samples at the same CpG.  If we assume that coverage genome-wide is Poisson distributed with a parameter (lambda) of 4, we would expect
```{r poisson}
logp <- ppois(0, lambda = 4, lower.tail = FALSE, log.p = TRUE)
round(1 - exp(6 * logp), 3)
``` 
of the CpGs to have at least one sample with zero coverage.

There are roughly 130k CpGs with no data at all in any of the 6 samples.  This can happen either because of chance (although that is unlikely) or because the CpG is unmappable.  Since we are dealing with bisulfite converted reads, the unmappable portion of the genome is greater than with normal DNA-sequencing.  For this experiment we only used 50bp single-end reads (in our experience using 100bp paired-end reads greatly increases the mappable percentage of the genome).  These CpGs (with zero coverage in all samples) are in some sense easy to deal with: one should of course be careful drawing conclusions about CpGs with no data.

We have roughly 959 - 573 - 136 = 250$ CpGs where some (but not all) of the samples have zero coverage, and these are in some sense harder to deal with.  Since we have very low coverage to begin with, it may happen just by chance that a single sample have zero coverage, and it may be too restrictive to just exclude these CpGs from an analysis.

Smoothing is done separately for each sample, only using the data where the coverage (for that sample) is non-zero.  This estimates a genome-wide methylation profile, which is then **evaluated** in all CpGs in the `BSseq` object.  As a result, after smoothing, every CpG in the object has an estimated methylation value.  This is very nice for the situation where you want to compare a single CpG across multiple samples, but one or two of the samples have zero coverage by chance.  But note that these smoothed methylation profiles makes less sense in the parts of the genome where there are no covered CpGs nearby.  We fix this by removing these CpGs after smoothing, see below.

Other arguments to the `BSmooth()` function are `mc.cores`, `mc.preschedule`, `parallelBy` which controls the parallelization built into the function as well as `ns`, `h`, `maxGap` which controls the smoothing.  `ns` is the minimum number of CpGs contained in each window, `h` is half the minimum window with (the actual window width is either 2 times `h` or wide enough to contain `ns` covered CpGs, whichever is greater).  Note that the window width is different at each position in the genome and may also be different for different samples at the same position, since it depends on how many nearby CpGs with non-zero coverage.  Per default, a smoothing cluster is a whole chromosome.  By "cluster" we mean a set of CpGs which are processed together.  This means that even if there is a large distance between two CpGs, we borrow strength between them.  By setting `maxGap` this can be prevented since the argument describes the longest distance between two CpGs before a cluster is broken up into two clusters.

## Manually splitting the smoothing computation

An example, only showing sample 1 and 2 for brevity, is (this example is not being run when the vignette is being created):

```{r smoothSplit,eval=FALSE}
## Split datag
BS1 <- BS.cancer.ex[, 1]
save(BS1, file = "BS1.rda")
BS2 <- BS.cancer.ex[, 2]
save(BS1, file = "BS1.rda")
## done splitting

## Do the following on each node

## node 1
load("BS1.rda")
BS1.fit <- BSmooth(BS1)
save(BS1.fit)
save(BS1.fit, file = "BS1.fit.rda")
## done node 1

## node 2
load("BS2.rda")
BS2.fit <- BSmooth(BS2)
save(BS2.fit, file = "BS2.fit.rda")
## done node 2

## join; in a new R session
load("BS1.fit.rda")
load("BS2.fit.rda")
BS.fit <- combine(BS1.fit, BS2.fit)
``` 

This still requires that you have one node with enough RAM to hold all samples in memory.


# Computing t-statistics

Before computing t-statistics, we will remove CpGs with little or no coverage.  If this is not done, you may find many DMRs in areas of the genome with very little coverage, which are most likely false positives.  It is open to personal preferences exactly which CpGs to remove, but for this analysis we will only keep CpGs where at least 2 cancer samples and at least 2 normal samples have at least 2x in coverage.  For readability, we store the coverage in a separate matrix (this is just due to line breaks in the printed output).

```{r keepLoci}
BS.cov <- getCoverage(BS.cancer.ex.fit)
keepLoci.ex <- which(rowSums(BS.cov[, BS.cancer.ex$Type == "cancer"] >= 2) >= 2 &
                     rowSums(BS.cov[, BS.cancer.ex$Type == "normal"] >= 2) >= 2)
length(keepLoci.ex)
BS.cancer.ex.fit <- BS.cancer.ex.fit[keepLoci.ex,]
``` 

(the `keepLoci.ex` is also available for direct inspection in the `r Biocpkg("bsseqData")` package.)


We are now ready to compute t-statistics, by
```{r BSmooth.tstat}
BS.cancer.ex.tstat <- BSmooth.tstat(BS.cancer.ex.fit, 
                                    group1 = c("C1", "C2", "C3"),
                                    group2 = c("N1", "N2", "N3"), 
                                    estimate.var = "group2",
                                    local.correct = TRUE,
                                    verbose = TRUE)
BS.cancer.ex.tstat
``` 

(the `BS.cancer.ex.tstat` is also available for direct inspection in the `r Biocpkg("bsseqData")` package.)



The arguments to `BSmooth.tstat()` are simple.  `group1` and `group2` contain the sample names of the two groups being compared (it is always group1 - group2), and indices may be used instead of sample names.  `estimate.var` describes which samples are being used to estimate the variability.  Because this is a cancer dataset, and cancer have higher variability than normals, we only use the normal samples to estimate the variability.  Other choices of `estimate.var` are `same` (assume same variability in each group) and `paired` (do a paired t-test).  The argument `local.correct` describes whether we should use a large-scale (low-frequency) mean correction.  This is especially important in cancer where we have found many large-scale methylation differences between cancer and normals.

We can look at the marginal distribution of the t-statistic by

```{r plotTstat,fig.width=4,fig.height=4}
plot(BS.cancer.ex.tstat)
``` 

The "blocks" of hypomethylation are clearly visible in the marginal distribution of the uncorrected t-statistics.

Even in comparisons where we do not observe these large-scale methylation differences, it often improves the marginal distribution of the t-statistics to locally correct them ("improves" in the sense of making them more symmetric).

# Finding DMRs

Once t-statistics have been computed, we can compute differentially methylated regions (DMRs) by thresholding the t-statistics.  Here we use a cutoff of $4.6$, which was chosen by looking at the quantiles of the t-statistics (for the entire genome)\footnote{See \url{https://support.bioconductor.org/p/78227/} for further discussion on the choice of cutoff. }.
```{r dmrs}
dmrs0 <- dmrFinder(BS.cancer.ex.tstat, cutoff = c(-4.6, 4.6))
dmrs <- subset(dmrs0, n >= 3 & abs(meanDiff) >= 0.1)
nrow(dmrs)
head(dmrs, n = 3)
``` 

Here, we filter out DMRs that do not have at least 3 CpGs in them and at least a mean difference (across the DMR) in methylation between normal and cancers of at least 0.1.  While the exact values of these two filters can be debated, it is surely a good idea to use something like this.

Other arguments to `dmrFinder()` are `qcutoff` which chooses a quantile-based cutoff (for example `qcutoff = c(0.01, 0.99)`) and `maxGap` which makes sure that a DMR is being split if there are two CpGs with more than `maxGap` between them (default of 300bp).

We rank DMRs by the column `areaStat` which is the sum of the t-statistics in each CpG.  This is kind of the area of the DMR, except that it is weighted by the number of CpGs and not by genomic length.  This is currently the best statistic we know, although it is far from perfect (we would like to do something better).

# Plotting

It is **always** a good idea to look at the DMRs.  One way of encoding standard plotting parameters like `col`, `lty`, and `lwd` is to add columns to the `pData`, like

```{r plotSetup}
pData <- pData(BS.cancer.ex.fit)
pData$col <- rep(c("red", "blue"), each = 3)
pData(BS.cancer.ex.fit) <- pData
``` 

Once this is setup, we can plot a single DMR like

```{r plotDmr,fig.width=8,fig.height=4}
plotRegion(BS.cancer.ex.fit, dmrs[1,], extend = 5000, addRegions = dmrs)
``` 

`extend` tells us how many bp to extend to either side of the plotting region.  `addRegions` is a `data.frame` or `GRanges` listing additional regions that should be highlighted.

Typically, we plot hundreds of DMRs in a single PDF file and use external tools to look at them.  For this purpose, `plotManyRegions()` is very useful since it is much faster than plotting individual DMRs with `plotRegion()`.  An example (not run) is

```{r plotManyRegions,eval=FALSE}
pdf(file = "dmrs_top200.pdf", width = 10, height = 5)
plotManyRegions(BS.cancer.ex.fit, dmrs[1:200,], extend = 5000, 
                addRegions = dmrs)
dev.off()
``` 
which plots the top200.

# Question and answers

> The BSmooth algorithm is supposed to give smooth methylation estimates.  Yet, when I plot the smoothed values, I see jagged lines, which do not look smooth to me.

We estimate a genome-wide methylation profile that is a smooth function of the genomic position.  However, this profile is not stored in the `BSseq` objects.  Instead, we evaluate this smooth profile in the methylation loci in the object.  An example (made-up values) is

```
pos  meth
1     0.1
3     0.1
5     0.1
200   0.6
203   0.6
205   0.6
```

For plotting we do linear interpolation between this points.  The end result is that the methylation profile may appear jagged especially if there is a "big" distance between two CpGs (between pos `5` and `200` above).  If we wanted to plot truly smooth profiles we would have to store the methylation profile evaluated at a regular grid across the genome.  This takes up a lot of space and would add complications to the internal data structures.



# sessionInfo()

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

# References

<!--
% Local Variables:
% LocalWords: LocalWords bisulfite methylation methylated CpG CpGs DMR bsseq bp 
% LocalWords: DMRs differentially ABI SOLiD dataset WGBS BSmooth 
% LocalWords: parallelization Bioconductor hypomethylation genomic pos PDF thresholding
% LocalWords: mappable unmappable indices normals quantile
% End:
-->

\name{BSseq-class}
\Rdversion{1.1}
\docType{class}
\alias{BSseq-class}
\alias{[,BSseq-method}
\alias{combine,BSseq,BSseq-method}
\alias{combineList}
\alias{length,BSseq-method}
\alias{pData,BSseq-method}
\alias{pData<-,BSseq,data.frame-method}
\alias{pData<-,BSseq,DataFrame-method}
\alias{sampleNames,BSseq-method}
\alias{sampleNames<-,BSseq,ANY-method}
\alias{updateObject,BSseq-method}
\alias{assays,BSseq-method}
\alias{assayNames,BSseq-method}
\alias{show,BSseq-method}
\alias{getBSseq}
\alias{collapseBSseq}
\alias{strandCollapse}
\alias{orderBSseq}
\alias{chrSelectBSseq}
\alias{hasBeenSmoothed}

\title{Class BSseq}
\description{
  A class for representing whole-genome or capture bisulfite sequencing
  data.
}
\section{Objects from the Class}{
  An object from the class links together several pieces of information.
  (1) genomic locations stored as a \code{GRanges} object, a location by
  samples matrix of M values, a location by samples matrix of Cov
  (coverage) values, a location by samples matrix of Filtered (ambiguous
  modification status) values, and phenodata information.In addition,
  there are slots for representing smoothed data. This class is an
  extension of \link[SummarizedExperiment]{RangedSummarizedExperiment}
  from the \pkg{SummarizedExperiment} package.
}
\section{Slots}{
  \describe{
    \item{\code{trans}:}{Object of class \code{function}.  This
      function transforms the \code{coef} slot from the scale the
      smoothing was done to the 0-1 methylation scale.}
    \item{\code{parameters}:}{Object of class \code{list}.  A list of
      parameters representing for example how the data was smoothed.}
  }
}
\section{Methods}{
  \describe{
    \item{[}{\code{signature(x = "BSseq")}: Subsetting by location
      (using integer indices) or sample (using integers or sample
      names).}
    \item{length}{Unlike for \code{RangedSummarizedExperiment},
      \code{length()} is the number of methylation loci (equal to
      \code{length(granges(x))}).}
    \item{sampleNames,sampleNames<-}{Sample names and its replacement
      function for the object.  This is an alias for \code{colnames}.}
    \item{pData,pData<-}{Obtain and replace the \code{pData} slot of the
      \code{phenoData} slot. This is an alias for \code{colData}.}
    \item{show}{The show method.}
    \item{combine}{This function combines two \code{BSSeq} objects.
      The genomic locations of the new object is the union of the
      genomic locations of the individual objects.  In addition, the
      methylation data matrices are placed next to each other (as
      appropriate wrt. the new genomic locations) and zeros are entered
      into the matrices as needed.}
}}
\section{Utilities}{
  This class extends
  \link[SummarizedExperiment]{RangedSummarizedExperiment}
  from the \pkg{SummarizedExperiment} package and therefore
  inherits a number of useful \code{GRanges} methods that operate on the
  \code{rowRanges} slot, used for accessing and setting the genomic locations
  and also do \code{subsetByOverlaps}.

  There are a number of almost methods-like functions for operating on
  objects of class \code{BSseq}, including \code{getBSseq},
  \code{collapseBSseq}, and \code{orderBSseq}.  They are detailed below.

  \describe{
    \item{\code{collapseBSseq(BSseq, columns)}}{

      is used to collapse an object of class \code{BSseq}.  By
      collapsing we simply mean that certain columns (samples) are merge
      together by summing up the methylation evidence and coverage.
      This is a useful function if you start by reading in a dataset
      based on say flowcells and you (after QC) want to simply add a
      number of flowcells into one sample.  The argument \code{columns}
      specify which samples are to be merged, in the following way: it
      is a character vector of new sample names, and the names of the
      column vector indicates which samples in the \code{BSseq} object
      are to be collapsed.  If \code{columns} have the same length as
      the number of rows of \code{BSseq} (and has no names) it is
      assumed that the ordering corresponds to the sample ordering in
      \code{BSseq}.}

    \item{\code{orderBSseq(BSseq, seqOrder = NULL)}}{

      simply orders an object of class \code{BSseq} according to
      (increasing) genomic locations.  The \code{seqOrder} vector is a
      character vector of \code{seqnames(BSseq)} describing the order of
      the chromosomes.  This is useful for ordering \code{chr1} before
      \code{chr10}.}

    \item{\code{chrSelectBSseq(BSseq, seqnames = NULL, order = FALSE)}}{

      subsets and optionally reorders an object of class \code{BSseq}.
      The \code{seqnames} vector is a character vector of
      \code{seqnames(BSseq)} describing which chromosomes should be
      retained.  If \code{order} is \code{TRUE}, the chromosomes are
      also re-ordered using \code{orderBSseq}.}

    \item{\code{getBSseq(BSseq, type = c("Cov", "M", "gr", "coef",
	"se.coef", "trans",  "parameters"))}}{

      is a general accessor: is used to obtain a specific slot of an
      object of class \code{BSseq}.  It is primarily intended for
      internal use in the package, for users we recommend \code{granges}
      to get the genomic locations, \code{getCoverage} to get the
      coverage slots and \code{getMeth} to get the smoothed values (if
      they exist).  }

    \item{\code{hasBeenSmoothed(BSseq)}}{

      This function returns a logical depending on whether or not the
      \code{BSseq} object has been smoothed using \code{BSmooth}.  }

    \item{\code{combineList(list, BACKEND = NULL)}}{

      This function function is a faster way of using \code{combine} on
      multiple \linkS4class{BSseq} objects. The input is a list, with each
      component an object of class \linkS4class{BSseq}. The (slower)
      alternative is to use \code{Reduce(combine, list)}.

      The \code{BACKEND} argument determines which backend should be used for
      the 'M' and 'Cov' matrices and, if present, the 'coef' and 'se.coef'
      matrices (the latter two can only be combined if all objects have the
      same rowRanges). The default, \code{BACKEND = NULL}, corresponds to using
      \link[base]{matrix} objects. See
      \code{\link[DelayedArray]{setAutoRealizationBackend}} (in the
      \pkg{DelayedArray} package) for alternative backends.}

    \item{\code{strandCollapse(BSseq, shift = TRUE)}}{

      This function operates on a \code{BSseq} objects which has
      stranded loci (i.e. loci where the strand is one of \sQuote{+} or
      \sQuote{-}). It will collapse the methylation and coverage
      information across the two strands, unstranding the loci in the process
      and potentially re-ordering them.

      The argument \code{shift} indicates whether the positions for the loci
      on the reverse strand should be shifted one (i.e. the positions for
      these loci are the positions of the \sQuote{G} in the
      \sQuote{CpG}; this is the case for Bismark output for example).}
  }
}
\section{Coercion}{
  Package versions 1.5.2 and 1.11.1 introduced a new version of representing
  \sQuote{BSseq} objects.  You can update old serialized (saved)
  objects by invoking \code{x <- updateObject(x)}.
}
\section{Assays}{
  This class overrides the default implementation of \code{assays} to
  make it faster.  Per default, no names are added to the returned data
  matrices.

  Assay names can conveniently be obtained by the function
  \code{assayNames(x)}
}

\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{

  The package vignette.  \code{\link{BSseq}} for the constructor
  function.
  \code{\link[SummarizedExperiment]{RangedSummarizedExperiment}}
  (in the \pkg{SummarizedExperiment} package) for the underlying class.
  \code{\link{getBSseq}}, \code{\link{getCoverage}}, and
  \code{\link{getMeth}} for accessing the data stored in the object and
  finally \code{\link{BSmooth}} for smoothing the bisulfite sequence
  data.

}
\examples{
M <- matrix(1:9, 3,3)
colnames(M) <- c("A1", "A2", "A3")
BStest <- BSseq(pos = 1:3, chr = c("chr1", "chr2", "chr1"), M = M, Cov = M + 2)
chrSelectBSseq(BStest, seqnames = "chr1", order = TRUE)
collapseBSseq(BStest, group = c("A", "A", "B"))

#-------------------------------------------------------------------------------
# An example using a HDF5-backed BSseq object
#
hdf5_BStest <- realize(BStest, "HDF5Array")
chrSelectBSseq(hdf5_BStest, seqnames = "chr1", order = TRUE)
collapseBSseq(
    BSseq = hdf5_BStest,
    group = c("A", "A", "B"),
    BACKEND = "HDF5Array",
    type = "integer")
}
\keyword{classes}

\name{bsseq-deprecated}
\alias{bsseq-deprecated}
\title{Deprecated functions in package \sQuote{bsseq}}

\description{
  These functions are provided for compatibility with older versions
  of \sQuote{bsseq} only, and will be defunct at the next release.
}

\details{

  The following functions are deprecated and will be made defunct; use
  the replacement indicated below:
  \itemize{

    \item{read.modbam2bed: \code{\link{read.bedMethyl}}}
    \item{read.modkit: \code{\link{read.bedMethyl}}}

  }
}

\name{fisherTests}
\alias{fisherTests}
\title{
  Compute Fisher-tests for a BSseq object
}
\description{
  A function to compute Fisher-tests for an object of class \code{BSseq}.
}
\usage{
fisherTests(BSseq, group1, group2, lookup = NULL,
  returnLookup = TRUE, mc.cores = 1, verbose = TRUE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{group1}{A vector of sample names or indexes for the
    \sQuote{treatment} group.}
  \item{group2}{A vector of sample names or indexes for the
    \sQuote{control} group.}
  \item{lookup}{A \sQuote{lookup} object, see details.}
  \item{returnLookup}{Should a \sQuote{lookup} object be returned, see
    details.}
  \item{mc.cores}{The number of cores used.  Note that setting
    \code{mc.cores} to a value greater than 1 is not supported on MS
    Windows, see the help page for \code{mclapply}.}
  \item{verbose}{Should the function be verbose.}
}
\details{
  This function computes row-wise Fisher's exact tests.  It uses an
  internal lookup table so rows which forms equivalent 2x2 tables are
  group together and only a single test is computed.  If
  \code{returnLookup} is \code{TRUE} the return object contains the
  lookup table which may be feed to another call to the function using
  the \code{lookup} argument.

  If \code{group1}, \code{group2} designates more than 1 sample, the
  samples are added together before testing.

  This function can use multiple cores on the same computer.

  This test cannot model biological variability.
}
\value{
  if \code{returnLookup} is \code{TRUE}, a list with components
  \code{results} and \code{lookup}, otherwise just the \code{results}
  component.  The \code{results} (component) is a matrix with the same
  number of rows as the \code{BSseq} argument and 2 columns
  \code{p.value} (the unadjusted p-values) and \code{log2OR} (log2
  transformation of the odds ratio).
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  \link[stats]{fisher.test} for information about Fisher's test.
  \code{\link{mclapply}} for the \code{mc.cores} argument.
}
\examples{
M <- matrix(1:9, 3,3)
colnames(M) <- c("A1", "A2", "A3")
BStest <- BSseq(pos = 1:3, chr = c("chr1", "chr2", "chr1"),
                M = M, Cov = M + 2)
results <- fisherTests(BStest, group1 = "A1", group2 = "A2",
                       returnLookup = TRUE)
results

#-------------------------------------------------------------------------------
# An example using a HDF5Array-backed BSseq object
#
library(HDF5Array)
# See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
hdf5_BStest <- saveHDF5SummarizedExperiment(x = BStest,
                                            dir = tempfile())
results <- fisherTests(hdf5_BStest,
                       group1 = "A1",
                       group2 = "A2",
                       returnLookup = TRUE)
results
}

\name{BSmooth.tstat}
\alias{BSmooth.tstat}
\title{
  Compute t-statistics based on smoothed whole-genome bisulfite
  sequencing data.
}
\description{
  Compute t-statistics based on smoothed whole-genome bisulfite
  sequencing data.
}
\usage{
BSmooth.tstat(BSseq, group1, group2,
  estimate.var = c("same", "paired", "group2"), local.correct = TRUE,
  maxGap = NULL, qSd = 0.75, k = 21, mc.cores = 1, verbose = TRUE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{group1}{A vector of sample names or indexes for the \sQuote{treatment}
    group.}
  \item{group2}{A vector of sample names or indexes for the \sQuote{control}
    group.}
  \item{estimate.var}{How is the variance estimated, see details.}
  \item{local.correct}{A logical; should local correction be used, see
    details.}
  \item{maxGap}{A scalar greater than 0, see details.}
  \item{qSd}{A scalar between 0 and 1, see details.}
  \item{k}{A positive scalar, see details.}
  \item{mc.cores}{The number of cores used.  Note that setting
    \code{mc.cores} to a value greater than 1 is not supported on MS
    Windows, see the help page for \code{mclapply}.}
  \item{verbose}{Should the function be verbose?}
}
\details{
  T-statistics are formed as the difference in means between group 1 and
  group 2 divided by an estimate of the standard deviation, assuming
  that the variance in the two groups are the same (\code{same}), that
  we have paired samples (\code{paired}) or only estimate the variance
  based on group 2 (\code{group2}).  The standard deviation estimates
  are then smoothed (using a running mean with a width of \code{k}) and
  thresholded (using \code{qSd} which sets the minimum standard
  deviation to be the \code{qSd}-quantile).  Optionally, the
  t-statistics are corrected for low-frequency patterns.

  It is sometimes useful to use \code{local.correct} even if no large
  scale changes in methylation have been found; it makes the marginal
  distribution of the t-statistics more symmetric.

  Additional details in the reference.
}
\value{
  An object of class \code{BSseqTstat}.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\references{
 KD Hansen, B Langmead, and RA Irizarry.
  \emph{BSmooth: from whole genome bisulfite sequencing reads to
    differentially methylated regions}.
  Genome Biology (2012) 13:R83.
  doi:\href{http://www.dx.doi.org/10.1186/gb-2012-13-10-r83}{10.1186/gb-2012-13-10-r83}.
}
\seealso{
  \code{\link{BSmooth}} for the input object and
  \code{\linkS4class{BSseq}} for its class.
  \code{\linkS4class{BSseqTstat}} describes the return class.  This
  function is likely to be followed by the use of
  \code{\link{dmrFinder}}.  And finally, see the package vignette(s) for
  more information on how to use it.
}
\examples{
if(require(bsseqData)) {
  data(keepLoci.ex)
  data(BS.cancer.ex.fit)
  BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
  ## Remember to subset the BSseq object, see vignette for explanation
  BS.tstat <- BSmooth.tstat(BS.cancer.ex.fit[keepLoci.ex,],
                            group1 = c("C1", "C2", "C3"),
                            group2 = c("N1", "N2", "N3"),
                            estimate.var = "group2")
  BS.tstat
  ## This object is also stored as BS.cancer.ex.tstat in the
  ## bsseqData package

  #---------------------------------------------------------------------------
  # An example using a HDF5Array-backed BSseq object
  #
  \donttest{
    library(HDF5Array)
    # See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
    hdf5_BS.cancer.ex.fit <- saveHDF5SummarizedExperiment(
        x = BS.cancer.ex.fit[keepLoci.ex, ],
        dir = tempfile())
    hdf5_BS.tstat <- BSmooth.tstat(hdf5_BS.cancer.ex.fit,
                                   group1 = c("C1", "C2", "C3"),
                                   group2 = c("N1", "N2", "N3"),
                                   estimate.var = "group2")
  }
}
}

\name{findLoci}
\alias{findLoci}
\title{
  Find methylation loci in a genome
}
\description{
    This is a convenience function to find methylation loci, such as CpGs, in a reference genome.
    The result is useful as the value of the \code{loci} argument for \code{\link{read.bismark}()}.
}
\usage{
findLoci(pattern,
         subject,
         include = seqlevels(subject),
         strand = c("*", "+", "-"),
         fixed = "subject",
         resize = TRUE)
}
\arguments{
  \item{pattern}{
    A string specifying the pattern to search for, e.g. \code{"CG"}.
    Can contain IUPAC ambiguity codes, e.g., \code{"CH"}.
  }
  \item{subject}{
    A string containing a file path to the genome sequence, in FASTA or 2bit format, to be searched.
    Alternatively, a \code{\link[BSgenome]{BSgenome}} or \code{\link[Biostrings]{DNAStringSet}} object storing the genome sequence to be searched.
  }
  \item{include}{
    A character vector indicating the seqlevels of \code{subject} to be used.
  }
  \item{strand}{
    A character scaler specifying the strand of \code{subject} to be used.
    If \code{strand = "*"}, then both the positive (\code{strand = "+"}) and negative (\code{strand = "-"} strands will be searched.)
    It is assumed that \code{subject} contains the sequence with respect to the \code{+} strand.
  }
  \item{fixed}{
    If \code{"subject"} (the default), IUPAC ambiguity codes in the pattern only are interpreted as wildcards, e.g., a \code{pattern} containing \code{CH} matches a \code{subject} containing \code{CA} but not vice versa.
    See \code{?Biostrings::`\link[Biostrings]{lowlevel-matching}`} for more information
  }
  \item{resize}{
    A logical scalar specifying whether the ranges should be shifted to have width 1 and anchored by the start of the locus, e.g., resize a CpG dinucleotide to give the co-ordinates of the cytosine.
  }
}
\details{
  This function provides a convenience wrapper for finding methylation loci in
  a genome, based on running
  \code{\link[Biostrings:matchPattern]{vmatchPattern}()}.
  Users requiring finer-grained control should directly use the
  \code{\link[Biostrings:matchPattern]{vmatchPattern}()} function and coerce
  the result to a \code{\link[GenomicRanges]{GRanges}} object.
}
\value{
  A \code{\link[GenomicRanges]{GRanges}} object storing the found loci.
}
\author{
  Peter F. Hickey
}
\seealso{
  \itemize{
    \item \code{Biostrings::\link[Biostrings]{vmatchPattern}()}
    \item \code{?BSgenome::`\link[BSgenome]{BSgenome-utils}`}
    }
}
\examples{
  library(Biostrings)
  # Find CpG dinucleotides on the both strands of an artificial sequence
  my_seq <- DNAStringSet("ATCAGTCGC")
  names(my_seq) <- "test"
  findLoci(pattern = "CG", subject = my_seq)
  # Find CHG trinucleotides on the both strands of an artificial sequence
  findLoci(pattern = "CHG", subject = my_seq)

  # Find CpG dinucleotides on the + strand of chr17 from the human genome (hg38)
  if (requireNamespace("BSgenome.Hsapiens.UCSC.hg38")) {
    findLoci(
        pattern = "CG",
        subject = BSgenome.Hsapiens.UCSC.hg38::BSgenome.Hsapiens.UCSC.hg38,
        include = "chr17",
        strand = "+")
  }
}

\name{getCoverage}
\alias{getCoverage}
\title{
  Obtain coverage for BSseq objects.
}
\description{
  Obtain coverage for BSseq objects.
}
\usage{
getCoverage(BSseq, regions = NULL, type = c("Cov", "M"),
  what = c("perBase", "perRegionAverage", "perRegionTotal"),
  withDimnames = TRUE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{regions}{An optional \code{data.frame} or
    \code{GenomicRanges} object specifying a number of genomic regions.}
  \item{type}{This returns either coverage or the total
    evidence for methylation at the loci.}
  \item{what}{The type of return object, see details.}
  \item{withDimnames}{A \code{logical(1)}, indicating whether dimnames should
    be applied to extracted coverage elements. Setting
    \code{withDimnames = FALSE} increases the speed and memory efficiency with
    which coverage is extracted.}
}
\value{
  \strong{NOTE:} The return type of \code{getCoverage} varies depending on its
  arguments.

  If \code{regions} are not specified (\code{regions = NULL}) a
  \code{\link[DelayedArray]{DelayedMatrix}} object (\code{what =
  "perBase"}) is returned.  This will either contain the per-base
  coverage, the average coverage, or the genome total coverage
  (depending on value of \code{what}).

  If \code{what = "perBase"} and \code{regions} are specified, a list is
  returned.  Each element of the list is a
  \code{\link[DelayedArray]{DelayedMatrix}} object corresponding to the
  genomic loci inside the region.  It is conceptually the same as
  splitting the coverage by region.

  If \code{what = "perRegionAverage"} or \code{what = "perRegionTotal"}
  and \code{regions} are specified the return value is a
  \code{\link[DelayedArray]{DelayedMatrix}} object. Each row of the
  \code{\link[DelayedArray]{DelayedMatrix}} corresponds to a region and
  contains either the average coverage or the total coverage in the
  region.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{BSseq}} for the \code{BSseq} class.
}
\examples{
data(BS.chr22)
head(getCoverage(BS.chr22, type = "M"))
reg <- GRanges(seqnames = c("chr22", "chr22"),
  ranges = IRanges(start = c(1, 2*10^7), end = c(2*10^7 +1, 4*10^7)))
getCoverage(BS.chr22, regions = reg, what = "perRegionAverage")
  cList <- getCoverage(BS.chr22, regions = reg)
length(cList)
head(cList[[1]])

#-------------------------------------------------------------------------------
# An example using a HDF5Array-backed BSseq object
#

library(HDF5Array)
# See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
hdf5_BS.chr22 <- saveHDF5SummarizedExperiment(x = BS.chr22,
                                              dir = tempfile())
head(getCoverage(hdf5_BS.chr22, type = "M"))
reg <- GRanges(seqnames = c("chr22", "chr22"),
               ranges = IRanges(start = c(1, 2 * 10 ^ 7),
               end = c(2 * 10 ^ 7 + 1, 4 * 10 ^ 7)))
getCoverage(hdf5_BS.chr22, regions = reg, what = "perRegionAverage")
hdf5_cList <- getCoverage(hdf5_BS.chr22, regions = reg)
length(hdf5_cList)
head(hdf5_cList[[1]])
}

\name{estimateErrorRate}
\alias{estimateErrorRate}
\title{
  Estimate CpG-specific error rate from BSseq object.
}
\description{
  This function estimates the CpG-specific error rate from a single sample BSseq object generated using \code{read.bedMethyl}.
}
\usage{
estimateErrorRate(BSseq, minCov = 10, maxCov = 100, minRatio = 0.8, plotErrorProfile = FALSE)
}
\arguments{
  \item{BSseq}{A single sample object of class \code{BSseq}.}
  \item{minCov}{A non-negative integer specifying the minimum coverage required for CpG loci to be considered.}
  \item{maxCov}{A non-negative integer specifying the maximum coverage allowed for CpG loci to be considered.}
  \item{minRatio}{A numeric value between 0 and 1 specifying the minimum ratio of CpG sites to non-CpG sites for a loci to be considered.}
  \item{plotErrorProfile}{A logical value indicating whether to plot the CpG to non-CpG ratio distribution for the filtered sites.}
}
\value{
  A numeric value representing the estimated CpG-specific error rate the BSseq object.
}
\author{
  Søren Blikdal Hansen (soren.blikdal.hansen@sund.ku.dk)
}
\seealso{
  \code{\linkS4class{BSseq}} for the \code{BSseq} class, 
  \code{\link{read.bedMethyl}} for details on reading data into a \code{BSseq} object.
}
\examples{
# Example input files
infiles <- c(system.file("extdata/HG002_nanopore_test.bedMethyl.gz",
                         package = "bsseq"),
             system.file("extdata/HG002_pacbio_test.bedMethyl.gz",
                         package = "bsseq"))

# Run the function to import data
bsseq <- read.bedMethyl(files = infiles,
                        colData = DataFrame(row.names = c("test_nanopore", 
                                                          "test_pacbio")),
                        strandCollapse = TRUE,
                        verbose = TRUE)

# Estimate error rate
estimateErrorRate(bsseq[, 1], plotErrorProfile = FALSE)
}

\name{dmrFinder}
\alias{dmrFinder}
\title{
  Finds differentially methylated regions for whole genome bisulfite
  sequencing data.
}
\description{
  Finds differentially methylated regions for whole genome bisulfite
  sequencing data.  Essentially identifies regions of the genome where
  all methylation loci have an associated t-statistic that is beyond a
  (low, high) cutoff.
}
\usage{
dmrFinder(bstat, cutoff = NULL, qcutoff = c(0.025, 0.975),
  maxGap=300, stat = "tstat.corrected", verbose = TRUE)
}
\arguments{
  \item{bstat}{An object of class \code{BSseqStat} or \code{BSseqTstat}.}
  \item{cutoff}{The cutoff of the t-statistics.  This should be a vector
    of length two giving the (low, high) cutoff.  If \code{NULL}, see
    \code{qcutoff}.}
  \item{qcutoff}{In case \code{cutoff} is \code{NULL}, compute the
    cutoff using these quantiles of the t-statistic.}
  \item{maxGap}{If two methylation loci are separated by this distance,
    break a possible DMR.  This guarantees that the return DMRs have CpGs
    that are this distance from each other.}
  \item{stat}{Which statistic should be used?}
  \item{verbose}{Should the function be verbose?}
}
\details{
  The workhorse function is \code{BSmooth.tstat} which sets up a
  t-statistic for a comparison between two groups.

  Note that post-processing of these DMRs are likely to be necessary,
  filtering for example for length (or number of loci).
}
\value{
  A \code{data.frame} with columns
    \item{start,end,width,chr}{genomic locations and width.}
    \item{n}{The number of methylation loci.}
    \item{invdensity}{Average length per loci.}
    \item{group1.mean}{The mean methylation level across samples and
      loci in 'group1'.}
    \item{group2.mean}{The mean methylation level across samples and
      loci in 'group2'.}
    \item{meanDiff}{The mean difference in methylation level; the
      difference between \code{group1.mean} and \code{group2.mean}.}
    \item{idxStart, idxEnd, cluster}{Internal use.}
    \item{areaStat}{The 'area' of the t-statistic; equal to the sum of
      the t-statistics for the individual methylation loci.}
    \item{direction}{either \sQuote{hyper} or \sQuote{hypo}.}
    \item{areaStat.corrected}{Only present if \code{column =
	"tstat.corrected"}, contains the area of the corrected t-statistics.}
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\references{
  KD Hansen, B Langmead, and RA Irizarry.
  \emph{BSmooth: from whole genome bisulfite sequencing reads to
    differentially methylated regions}.
  Genome Biology (2012) 13:R83.
  doi:\href{http://www.dx.doi.org/10.1186/gb-2012-13-10-r83}{10.1186/gb-2012-13-10-r83}.
}
\seealso{
  \code{\link{BSmooth.tstat}} for the function constructing the input
    object, and \code{\linkS4class{BSseqTstat}} for its class.  In the
    example below, we use \code{\link[bsseqData]{BS.cancer.ex.tstat}} as
    the actual input object.  Also see the package vignette(s) for a
    detailed example.
}
\examples{
if(require(bsseqData)) {
  dmrs0 <- dmrFinder(BS.cancer.ex.tstat, cutoff = c(-4.6, 4.6), verbose = TRUE)
  dmrs <- subset(dmrs0, abs(meanDiff) > 0.1 & n >= 3)
}
}

\name{plotRegion}
\alias{plotRegion}
\alias{plotManyRegions}
\title{
  Plotting BSmooth methylation estimates
}
\description{
  Functions for plotting BSmooth methylation estimates.  Typically used
  to display differentially methylated regions.
}
\usage{
plotRegion(BSseq, region = NULL, extend = 0, main = "",
  addRegions = NULL, annoTrack = NULL, cex.anno = 1,
  geneTrack = NULL, cex.gene = 1.5, col = NULL, lty = NULL,
  lwd = NULL, BSseqStat = NULL, stat = "tstat.corrected",
  stat.col = "black", stat.lwd = 1, stat.lty = 1, stat.ylim = c(-8, 8),
  mainWithWidth = TRUE, regionCol = alpha("red", 0.1), addTicks = TRUE,
  addPoints = FALSE, pointsMinCov = 5, highlightMain = FALSE)

plotManyRegions(BSseq, regions = NULL, extend = 0, main = "",
  addRegions = NULL, annoTrack = NULL, cex.anno = 1,
  geneTrack = NULL, cex.gene = 1.5, col = NULL, lty = NULL,
  lwd = NULL, BSseqStat = NULL, stat = "tstat.corrected",
  stat.col = "black", stat.lwd = 1, stat.lty = 1, stat.ylim = c(-8, 8),
  mainWithWidth = TRUE, regionCol = alpha("red", 0.1), addTicks = TRUE,
  addPoints = FALSE, pointsMinCov = 5, highlightMain = FALSE,
  verbose = TRUE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{region}{A \code{data.frame} (with start, end and chr columns)
    with 1 row or \code{GRanges} of length 1.  If \code{region} is
    \code{NULL} the entire \code{BSseq} argument is plotted.}
  \item{regions}{A \code{data.frame} (with start, end and chr columns)
    or \code{GRanges}.}
  \item{extend}{Describes how much the plotting region should be
  extended in either direction.  The total width of the plot is equal to
  the width of the region plus twice \code{extend}.}
  \item{main}{The plot title.  The default is to construct a title with
  information about which genomic region is being plotted.}
  \item{addRegions}{A set of additional regions to be highlighted on the
  plots.  As the \code{regions} argument.}
  \item{annoTrack}{A named list of \code{GRanges} objects.  Each
  component is a track and the names of the list are the track names.
  Each track will be plotted as solid bars, and we routinely display
  information such as CpG islands, exons, etc.}
  \item{cex.anno}{\code{cex} argument when plotting \code{annoTrack}.}
  \item{geneTrack}{\strong{EXPERIMENTAL}: A \code{data.frame} with columns:
  \code{chr}, \code{start}, \code{end}, \code{gene_ID}, \code{exon_number},
  \code{strand}, \code{gene_name}, \code{isoforms}. This interface is under
  active development and subject to change.}
  \item{cex.gene}{\code{cex} argument when plotting \code{geneTrack}.}
  \item{col}{The color of the methylation estimates, see details.}
  \item{lty}{The line type of the methylation estimates, see details.}
  \item{lwd}{The line width of the methylation estimates, see details.}
  \item{BSseqStat}{An object of class \code{BSseqStat}.  If present,
    a new panel will be shown with the t-statistics.}
  \item{stat}{Which statistics will be plotted (only used is
    \code{BSseqStat} is not \code{NULL}.)}
  \item{stat.col}{color for the statistics plot.}
  \item{stat.lwd}{line width for the statistics plot.}
  \item{stat.lty}{line type for the statistics plot.}
  \item{stat.ylim}{y-limits for the statistics plot.}
  \item{mainWithWidth}{Should the default title include information
  about width of the plot region.}
  \item{regionCol}{The color used for highlighting the region.}
  \item{addTicks}{Should tick marks showing the location of methylation
  loci, be added?}
  \item{addPoints}{Should the individual unsmoothed methylation
  estimates be plotted.  This usually leads to a very confusing plot,
  but may be useful for diagnostic purposes.}
  \item{pointsMinCov}{The minimum coverage a methylation loci need in
  order for the raw methylation estimates to be plotted.  Useful for
  filtering out low coverage loci.  Only used if \code{addPoints =
  TRUE}.}
  \item{highlightMain}{Should the plot region be highlighted?}
  \item{verbose}{Should the function be verbose?}
}
\details{
  The correct choice of aspect ratio depends on the width of the
  plotting region.  We tend to use \code{width = 10, height = 5}.

  \code{plotManyRegions} is used to plot many regions (hundreds or
  thousands), and is substantially quicker than repeated calls to
  \code{plotRegion}.

  This function has grown to be rather complicated over time.  For
  custom plotting, it is sometimes useful to use the function
  definition as a skeleton and directly modify the code.
}
\value{This function is invoked for its side effect: producing a plot.}
\seealso{
  The package vignette has an extended example.
}
\author{Kasper Daniel Hansen \email{khansen@jhsph.edu}}

\name{BSseqStat-class}
\Rdversion{1.1}
\docType{class}
\alias{BSseqStat}
\alias{BSseqStat-class}
\alias{[,BSseqStat-method}
\alias{[,BSseqStat,ANY,ANY,ANY-method}
\alias{show,BSseqStat-method}
\alias{updateObject,BSseqStat-method}
\title{Class BSseqStat}
\description{
    A class for representing statistics for smoothed whole-genome
    bisulfite sequencing data.
}
\usage{
    BSseqStat(gr = NULL, stats = NULL, parameters = NULL)
}
\arguments{
    \item{gr}{The genomic locations as an object of class \code{GRanges}.}
    \item{stats}{The statistics, as a list of matrix-like objects (see
        'Details' below).}
    \item{parameters}{A list of parameters.}
}
\details{
    The matrix-like elements of the list in the 'stats' slot will be
    coerced to \code{\link[DelayedArray]{DelayedMatrix}} objects; see
    \code{\link[DelayedArray]{DelayedMatrix}} in the \pkg{DelayedArray}
    package for the full list of supported matrix-like objects. We
    recommend using \code{\link[base]{matrix}} objects for in-memory
    storage of data and \code{\link[HDF5Array]{HDF5Matrix}} for on-disk
    storage of data.
}
\section{Objects from the Class}{
    Objects can be created by calls of the form \code{BSseqStat(...)}.
    However, usually objects are returned by \code{BSmooth.fstat(...)} and
    not constructed by the user.
}
\section{Slots}{
  \describe{
    \item{\code{stats}:}{This is a list of   \code{\link[DelayedArray]{DelayedMatrix}}
      objects with list elements representing various statistics for
      methylation loci along the genome.}
    \item{\code{parameters}:}{Object of class \code{list}.  A list of
      parameters representing how the statistics were computed.}
    \item{\code{gr}:}{Object of class \code{GRanges} giving genomic
      locations.}
  }
}
\section{Extends}{
    Class \code{\linkS4class{hasGRanges}}, directly.
}
\section{Methods}{
    \describe{
        \item{[}{The subsetting operator; one may only subset in one
            dimension, corresponding to methylation loci.}
        \item{show}{The show method.}
    }
}
\section{Utilities}{
    This class extends \code{hasGRanges} and therefore inherits a number
    of useful \code{GRanges} methods that operate on the \code{gr} slot,
    used for accessing and setting the genomic locations and also do
    \code{subsetByOverlaps}.
}
\section{Coercion}{
  Package version 1.11.1 introduced a new version of representing
  \sQuote{BSseqStat} objects. You can update old serialized (saved)
  objects by invoking \code{x <- updateObject(x)}.
}
\author{
    Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
    \code{\linkS4class{hasGRanges}} for accessing the genomic locations.
    \code{\link{BSmooth.fstat}} for a function
    that returns objects of class \code{BSseqStat}, and \code{\link{smoothSds}},
    \code{\link{computeStat}} and \code{\link{dmrFinder}}
    for functions that operate based on these statistics.  Also see
    the more specialised \code{\linkS4class{BSseqTstat}}.
}
\keyword{classes}

\name{FWGRanges-class}
\Rdversion{1.1}
\docType{class}
\alias{FWIRanges-class}
\alias{end,FWIRanges-method}
\alias{end<-,FWIRanges-method}
\alias{names,FWIRanges-method}
\alias{names<-,FWIRanges-method}
\alias{start,FWIRanges-method}
\alias{start<-,FWIRanges-method}
\alias{width,FWIRanges-method}
\alias{width<-,FWIRanges-method}
\alias{FWGRanges-class}
\alias{findOverlaps,FWGRanges,FWGRanges-method}
\alias{end,FWGRanges-method}
\alias{start,FWGRanges-method}
\alias{strand,FWGRanges-method}
\alias{seqnames,FWGRanges-method}
\alias{width,FWGRanges-method}
\alias{dimnames,arrayRealizationSink-method}
\title{Classes FWIRanges and FWGRanges}
\description{
  Classes for fixed-width IRanges and GRanges, ie. objects where all
  ranges have the same width. The intention is for these classes to be
  added to GenomicRanges. Documented here temporarily.
}
\details{
  See description. Otherwise works like IRanges and GRanges, except
  there are many unimplemented methods.

  This is really a private class, with private methods, but R's
  NAMESPACE handling means they get unintentionally exported. Hence this
  documentation. 
}
\examples{
showClass("FWIRanges")
}
\keyword{classes}

\name{getMaxLikelihoodMatrix}
\alias{getMaxLikelihoodMatrix}
\title{
  Generate a matrix of the scaled likelihood of most likely CpG status for a multi-sample BSseq object.
}
\description{
  This function generates a matrix of the scaled likelihoods for most likely CpG status for a multi-sample BSseq object. Each element of the matrix represents the scaled likelihood of the most likely CpG status for the locus in the sample. If no data is available for a locus in a sample, the entry in the CpGMatrix is 2 (nonCpG) and the corresponding MaxLikelihood is 1/3.   
}
\usage{
getMaxLikelihoodMatrix(BSseq, e = NULL, allCpG = FALSE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{e}{An optional numeric vector representing error rates for each sample. If \code{NULL}, the error rate for each sample is estimated using \code{\link{estimateErrorRate}}.}
  \item{allCpG}{A logical value indicating whether to classify loci as allCpG and non-CpG loci and sum the scaled likelihood of homozygous CpG and heterozygous CpG. Should be the same for \code{getMaxLikelihoodMatrix} and \code{getCpGMatrix} }
}
\value{
  A numeric matrix where each row represents a locus, each column represents a sample, and the values correspond to the quality scores.
}
\author{
  Søren Blikdal Hansen (soren.blikdal.hansen@sund.ku.dk)
}
\seealso{
  \code{\linkS4class{BSseq}} for the \code{BSseq} class,
  \code{\link{read.bedMethyl}} for details on reading data into a \code{BSseq} object,
  \code{\link{estimateErrorRate}} for estimating the CpG-specific error rate.
  \code{\link{getCpGs}} for filtering a single-sample BSseg object.
  \code{\link{getCpGMatrix}} for generating a matrix with the most likely CpG status matching the getMaxLikelihoodMatrix.
}
\examples{
# Example input files
infiles <- c(system.file("extdata/HG002_nanopore_test.bedMethyl.gz",
                         package = "bsseq"),
             system.file("extdata/HG002_pacbio_test.bedMethyl.gz",
                         package = "bsseq"))

# Run the function to import data
bsseq <- read.bedMethyl(files = infiles,
                        colData = DataFrame(row.names = c("test_nanopore", 
                                                          "test_pacbio")),
                        strandCollapse = TRUE,
                        verbose = TRUE)

# Single samples can be filtered using the getCpGs function
bsseq_nano <- bsseq[, 1]
bsseq_nano_99All_filtered <- bsseq[getCpGs(bsseq_nano, 
                                           type = "allCpG", threshold = 0.99)]

bsseq_pacbio <- bsseq[, 2]
bsseq_pacbio_99All_filtered <- bsseq[getCpGs(bsseq_pacbio, 
                                             type = "allCpG", threshold = 0.99)]

# For filtering multiple samples, we can use a CpGMatrix and a MaxLikelihoodMatrix
# Construct the CpGMatrix and QualityMatrix for the bsseq object
CpGMatrix <- getCpGMatrix(bsseq, allCpG = TRUE)
MaxLikelihoodMatrix <- getMaxLikelihoodMatrix(bsseq, allCpG = TRUE)

# Filter for allCpG loci with a likelihood > 0.99 in both samples
bsseq_combined_99All_filtered <- bsseq[which(rowAlls(CpGMatrix == 0) 
                                          & rowMins(MaxLikelihoodMatrix) > 0.99)]
}

\name{computeStat}
\alias{computeStat}
\title{
    Compute a test statistic based on smoothed whole-genome bisulfite
    sequencing data.
 }
\description{
    Compute a test statistic based on smoothed whole-genome bisulfite
    sequencing data.
 }
\usage{
    computeStat(BSseqStat, coef = NULL)
}

\arguments{
    \item{BSseqStat}{An object of class \code{BSseqStat}, typically an object
        returned by \code{\link{smoothSds}(...)} and not constructed by
        the user.}
    \item{coef}{A vector indicating for which coefficients the statistic is to
        be computed (\code{coef = NULL} corresponds to testing all
        coefficients). If the length of the \code{coef} is 1 then the
        corresponding t-statistic is computed, otherwise the corresponding
        F-statistic is computed.}
}
\details{
    \strong{TODO}
}
\value{
  An object of class \linkS4class{BSseqStat}. More speciically, the input
  \linkS4class{BSseqStat} object with the computed statistics added to the
  \code{stats} slot (accessible with \code{\link{getStats}}).
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}

\seealso{
  \code{\link{smoothSds}} for the function to create the appropriate
  \code{\linkS4class{BSseqStat}} input object.
  \code{\linkS4class{BSseqStat}} also describes the return class.  This
  function is likely to be followed by the use of \code{\link{dmrFinder}}.}
\examples{
  \donttest{
    if(require(bsseqData)) {
        data(keepLoci.ex)
        data(BS.cancer.ex.fit)
        BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
        ## Remember to subset the BSseq object, see vignette for explanation
        ## TODO: Kind of a forced example
        design <- model.matrix(~0 + BS.cancer.ex.fit$Type)
        colnames(design) <- gsub("BS\\\.cancer\\\.ex\\\.fit\\\$Type", "",
                                 colnames(design))
        contrasts <- makeContrasts(
            cancer_vs_normal = cancer - normal,
            levels = design
        )
        BS.stat <- BSmooth.fstat(BS.cancer.ex.fit[keepLoci.ex,],
                                  design,
                                  contrasts)
        BS.stat <- smoothSds(BS.stat)
        BS.stat <- computeStat(BS.stat)
        BS.stat
    }
  }
}
\keyword{internal}

\name{read.bismark}
% TODO: alias required?
\alias{read.bismark}
\title{
  Parsing output from the Bismark alignment suite.
}
\description{
  Parsing output from the Bismark alignment suite.
}
\usage{read.bismark(files,
             loci = NULL,
             colData = NULL,
             rmZeroCov = FALSE,
             strandCollapse = TRUE,
             BPPARAM = bpparam(),
             BACKEND = NULL,
             dir = tempfile("BSseq"),
             replace = FALSE,
             chunkdim = NULL,
             level = NULL,
             nThread = 1L,
             verbose = getOption("verbose"))
}
\arguments{
  \item{files}{The path to the files created by running Bismark's methylation
    extractor, one sample per file.
    Files ending in \code{.gz} or \code{.bz2} will be automatically
    decompressed to \code{\link{tempfile}()}.
    We strongly recommend you use the 'genome wide cytosine report' output files.
    See section 'File formats' for further details.}
  \item{loci}{\code{NULL} (default) or a
    \code{\link[GenomicRanges]{GenomicRanges}} instance
    containing methylation loci (all with width equal to 1).
    If \code{loci = NULL}, then \code{read.bismark()} will perform a first pass over the Bismark file to identify candidate loci.
    If \code{loci} is a \code{\link[GenomicRanges]{GenomicRanges}}
    instance, then these form the candidate loci. 
    In either case, the candidate loci will be filtered if
    \code{rmZeroCov = TRUE} and collapsed if \code{strandCollapse = TRUE} to form the final set of methylation loci that form the \code{\link[SummarizedExperiment:RangedSummarizedExperiment-class]{rowRanges}} of the returned \linkS4class{BSseq} object.
    See section 'Efficient use of \code{read.bismark()}' for further details.}
  \item{colData}{An optional \code{\link[S4Vectors]{DataFrame}}
    describing the samples. 
    Row names, if present, become the column names of the \linkS4class{BSseq}
    object. If \code{NULL}, then a \code{\link[S4Vectors]{DataFrame}}
    will be created with 
    \code{files} used as the row names.}
  \item{rmZeroCov}{A \code{logical(1)} indicating whether methylation
    loci that have zero coverage in all samples be removed.  For several
    reasons, the default \code{rmZeroCov = FALSE} is recommended even in
    cases where you ultimately want to remove such loci.  See section
    'Efficient use of \code{read.bismark()}' for further details.}
  \item{strandCollapse}{A \code{logical(1)} indicating whether
    strand-symmetric methylation loci (i.e. CpGs) should be collapsed
    across strands.  This is only applicable for stranded methylation
    loci, e.g., loci extracted from 'genome wide cytosine reports' (see
    section 'File formats' for further details).}
  \item{BPPARAM}{An optional
    \code{\link[BiocParallel]{BiocParallelParam}} instance determining
    the parallel back-end to be used during evaluation.  Currently
    supported are \code{\link[BiocParallel]{SerialParam}} (Unix, Mac,
    Windows), \code{\link[BiocParallel]{MulticoreParam}} (Unix and Mac),
    \code{\link[BiocParallel]{SnowParam}} (Unix, Mac, and Windows,
    limited to single-machine clusters), and
    \code{\link[BiocParallel]{BatchtoolsParam}} (Unix, Mac, Windows,
    only with the in-memory realization backend).  See sections
    'Parallelization and progress monitoring' and 'Realization backends'
    for further details.}
  \item{BACKEND}{\code{NULL} or a single string specifying the name of
    the realization backend.  When the backend is set to \code{NULL},
    the \code{M} and \code{Cov} assays are realized in memory as
    ordinary matrices, otherwise these are realized with the given
    \code{BACKEND}.  See section 'Realization backends' for further
    details.}
  \item{dir}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    The path (as a single string) to the directory where to save the HDF5-based
    \linkS4class{BSseq} object. The directory will be created so should not
    already exist, unless \code{replace} is set to \code{TRUE}.}
  \item{replace}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    If directory \code{dir} already exists, should it be replaced
    with a new one? The content of the existing directory will be lost!}
  \item{chunkdim}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    The dimensions of the chunks to use for writing the data to
    disk. By default,
    \code{\link[HDF5Array]{getHDF5DumpChunkDim}()} using the dimensions of the returned
    \linkS4class{BSseq} object will be used. See
    \code{?\link[HDF5Array]{getHDF5DumpChunkDim}} for more information.}
  \item{level}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    The compression level to use for writing the data to disk. By
    default,
    \code{\link[HDF5Array]{getHDF5DumpCompressionLevel}()} will be used. See
    \code{?\link[HDF5Array]{getHDF5DumpCompressionLevel}} for more information.}
  \item{nThread}{The number of threads used by \code{\link[data.table]{fread}}
    when reading the \code{files}. Be careful when combining a parallel backend
    specified with \code{BPPARAM} with \code{nThread} > 1 because each worker
    will use \code{nThread}.}
  \item{verbose}{A \code{logical(1)} indicating whether progress messages
    should be printed (default \code{TRUE}).}
}

\section{File formats}{
  The format of each file is automatically detected using the internal function \code{bsseq:::.guessBismarkFileType()}.
  Files ending in \code{.gz}, \code{.bz2}, \code{.xz}, or \code{.zip} will be automatically decompressed to \code{\link{tempdir}()}.
  \subsection{Supported file formats}{
    Bismark's 'genome wide cytosine report' (\url{https://github.com/FelixKrueger/Bismark/tree/master/Docs#the-genome-wide-cytosine-report-optional-is-tab-delimited-in-the-following-format-1-based-coords}) and 'coverage' (\url{https://github.com/FelixKrueger/Bismark/tree/master/Docs#the-coverage-output-looks-like-this-tab-delimited-1-based-genomic-coords}) formats are both supported.
    If setting \code{loci = NULL}, then we strongly recommend using the 'genome wide cytosine report' output format because this includes strand information for each locus.
     The 'coverage' output does not contain strand information and so the \code{\link[BiocGenerics]{strand}} of the returned \linkS4class{BSseq} object will be set to \code{*} unless stranded \code{loci} are supplied.
  }

  \subsection{Unsupported file formats}{
     Neither the 'bedGraph' output format (\url{https://github.com/FelixKrueger/Bismark/tree/master/Docs#the-bedgraph-output-optional-looks-like-this-tab-delimited-0-based-start-1-based-end-coords}) nor the 'bismark_methylation_extractor' output format (\url{https://github.com/FelixKrueger/Bismark/tree/master/Docs#the-bismark_methylation_extractor-output-is-in-the-form-tab-delimited-1-based-coords}) are supported.
     The former does not include the required counts of methylated and unmethylated reads hile the is an intermediate file containing read-level, rather than locus-level, data on methylation.
  }
  \subsection{One-based vs. zero-based genomic co-ordinates}{
    The genomic co-ordinates of the Bismark output files may be zero-based or one-based depending on whether the \code{--zero_based} argument was used when running Bismark's methylation extractor.
    Furthermore, the default co-ordinate counting system varies by version of Bismark.
    \pkg{bsseq} makes no assumptions about the basis of the genomic co-ordinates and it is left to the user to ensure that the appropriate basis is used in the analysis of their data.

    Since Bioconductor packages typically use one-based co-ordinates, we strongly recommend that your Bismark files are also one-based.
  }
}

\section{Efficient use of \code{read.bismark()}}{
  We recommend the following to achieve fast and efficient importing of Bismark files:

  \itemize{
    \item Specify the set of methylation loci via the \code{loci} argument.
    \item Use Bismark files in the 'coverage' output format.
    \item Leave \code{rmZeroCov = FALSE}.
    \item Use a \code{BPPARAM} with a moderate number of workers (cores).
    \item Use \code{BACKEND = "HDF5Array"}.
    \item Use multiple threads per worker (i.e. \code{nThread} > 1).
  }

  Each point is discussed below.
  \subsection{Specifying \code{loci}}{
    Specifying the set of methylation loci via the \code{loci} argument means that \code{read.bismark()} does not need to first parse all \code{files} to identify the set of candidate loci.
    Provided that \code{rmZeroCov = FALSE}, this means that each file is only read once.
    This may be a considerable saving when there are a large number of \code{files}.

    \strong{If you are unsure whether the below-described shortcuts apply to your data, leave \code{loci = NULL} and let \code{read.bismark()} identify the set of candidate loci from \code{files}.}

    You may wish to use the \code{\link{findLoci}()} function to find all methylation loci of interest in your reference genome (e.g., all CpGs) and then pass the result via the \code{loci} argument.

    Alternatively, if all \code{files} are 'genome wide cytosine reports' for samples aligned to the same reference genome, then all \code{files} contain the exact same set of methylation loci.
    In this case, you may wish to first construct \code{loci} using the internal function \code{bsseq:::.readBismarkAsFWGRanges()} applied to a single file, e.g., \code{loci = bsseq:::.readBismarkAsFWGRanges(files[1], rmZeroCov, strandCollapse)}.
  }

  \subsection{Using the 'coverage' Bismark files}{
    It will generally be faster to parse Bismark files in the 'coverage' output format than those in the 'genome wide cytosine report' format This is because the former only includes loci with non-zero coverage and so the file size is often considerably smaller, particularly for shallowly sequenced samples (e.g., those from single-cell bisulfite sequencing).
  }

  \subsection{Leaving \code{rmZeroCov = FALSE}}{
    If you set \code{rmZeroCov = TRUE}, then \code{read.bismark()} must first parse all the \code{files} to identify which loci have zero coverage in all samples and then filter these out from the set of candidate loci.
    \strong{This will happen even if you supply \code{loci} with a
    \code{\link[GenomicRanges]{GenomicRanges}} of candidate loci.}

    Furthermore, any coverage-based filtering of methylation loci is best left until you have constructed your final \linkS4class{BSseq} object.
    In our experience, the final \linkS4class{BSseq} object is often the product of combining multiple \linkS4class{BSseq} objects, each constructed with a separate call to \code{read.bismark()}.
    In such cases, it is premature to use \code{rmZeroCov = TRUE} when running each \code{read.bismark()}; regretably, combining these objects will often then lead to an inefficiently stored \linkS4class{BSseq} object.
  }

  \subsection{Using a \code{BPPARAM} with a moderate number of workers (cores)}{

    Each file can be processed on its own, so you can process in
    parallel as many files as you have workers. However, if using the
    HDF5Array backend, then writing to the HDF5 file cannot be performed
    in parallel and so becomes the bottleneck. Nonetheless, by using a
    moderate number of workers (2 - 10), we can ensure there is
    processed data available to write to disk as soon as the current
    write is completed.
  }

  \subsection{Using \code{BACKEND = "HDF5Array"}}{
    By using the \link[HDF5Array]{HDF5Array} realization backend from
    \pkg{HDF5Array}, we reduce the amount of data that is kept in-memory
    at any one time.  Once each file is parsed, the data are written to
    the HDF5 file and are no longer needed in-memory.  When combined
    with multiple workers (cores), this means that each file will only
    need to read and retain in-memory 1 sample's worth of data at a
    time.

    Conversely, if you opt for all data to be in-memory (via
    \code{BACKEND = NULL}), then each worker will pass each file's data
    back to the main process and memory usage will steadily accumulate
    to often unreasonable levels.  }

  \subsection{Using \code{nThread} > 1}{
    \code{read.bismark} uses \code{data.table::\link[data.table]{fread}} to
    read each file, which supports threaded-parallisation. Depending on the
    system, increasing \code{nThread} can achieve near-linear speed-ups in the
    number of threads for reading each file. Care needs to be taken when
    \code{nThread} > 1 is used in conjunction with a parallel backend via
    \code{BPPARAM} to ensure the system isn't overloaded. For example, using
    \code{BPPARAM = MulticoreParam(workers = 10)} with \code{nThread = 4} may
    use up to 40 workers simultaneously.
  }
}

\section{Realization backends}{

  The \code{read.bismark()} function creates a \linkS4class{BSseq}
  object with two assays, \code{M} and \code{Cov}.  The choice of
  \emph{realization backend} controls whether these assays are stored
  in-memory as an ordinary \link[base]{matrix} or on-disk as a
  \code{\link[HDF5Array]{HDF5Array}}, for example.  The choice of
  realization backend is controlled by the \code{BACKEND} argument,
  which defaults to the current value of
  \code{DelayedArray::\link[DelayedArray]{getAutoRealizationBackend}()}.

  \code{read.bismark()} supports the following realization backends:

  \itemize{
    \item \code{NULL} (in-memory): This stores each new assay in-memory using an ordinary \link[base]{matrix}.
    \item \code{HDF5Array} (on-disk): This stores each new assay on-disk in a HDF5 file using an \code{\link[HDF5Array]{HDF5Matrix}} from \pkg{HDF5Array}.
  }

  Please note that certain combinations of realization backend and
  parallelization backend are currently not supported.  For example, the
  \link[HDF5Array]{HDF5Array} realization backend is currently only
  compatible when used with a single-machine parallelization backend
  (i.e. it is not compatible with a \code{\link[BiocParallel]{SnowParam}} that
  specifies an \emph{ad hoc} cluster of \strong{multiple} machines).
  % TODO: Test this
  \code{BSmooth()} will issue an error when given such
  incompatible realization and parallelization backends.

  Additional arguments related to the realization backend can be passed
  via the \code{...} argument.  These arguments must be named and are
  passed to the relevant \code{\link[DelayedArray]{RealizationSink}} constructor.  For
  example, the \code{...} argument can be used to specify the path to
  the HDF5 file to be used by \code{BSmooth()}.  Please see the examples
  at the bottom of the page.
}
\section{Parallelization, progress monitoring, and logging}{
  \code{read.bismark()} now uses the \pkg{BiocParallel} package to implement
  parallelization. This brings some notable improvements:

  \itemize{
    \item Imported files can now be written directly to an on-disk
      realization backend by the worker. This dramatically reduces memory
      usage compared to previous versions of \pkg{bsseq} that required all
      results be retained in-memory.
    \item Parallelization is now supported on Windows through the use of
      a \code{\link[BiocParallel]{SnowParam}} object as the value of
	\code{BPPARAM}.
    \item Detailed and extensive job logging facilities.
  }

  All parallelization options are controlled via the \code{BPPARAM}
  argument.  In general, we recommend that users combine multicore
  (single-machine) parallelization with an on-disk realization backend
  (see section, 'Realization backend'). For Unix and Mac users, this
  means using a \code{\link[BiocParallel]{MulticoreParam}}. For Windows
  users, this means using a single-machine
  \code{\link[BiocParallel]{SnowParam}}. Please consult the
  \pkg{BiocParallel} documentation to take full advantage of the more
  advanced features.

  A useful feature of \pkg{BiocParallel} are progress bars to monitor the
  status of long-running jobs, such as \code{BSmooth()}. Progress bars are
  controlled via the \code{progressbar} argument in the
  \code{\link[BiocParallel]{BiocParallelParam}} constructor.

  \pkg{BiocParallel} also supports extensive and detailed logging facilities.
  Please consult the \pkg{BiocParallel} documentation to take full advantage
  these advanced features.
}

\value{
  A \linkS4class{BSseq} object.
}

\seealso{
  \itemize{
    \item \code{\link{collapseBSseq}()} for collapsing (aggregating) data from sample's with multiple Bismark methylation extractor files (e.g., technical replicates).
  }
}

\examples{
  # Run read.bismark() on a single sample to construct a matrix-backed BSseq
  # object.
  infile <- system.file("extdata/test_data.fastq_bismark.bismark.cov.gz",
                        package = "bsseq")
  bsseq <- read.bismark(files = infile,
                        colData = DataFrame(row.names = "test_data"),
                        rmZeroCov = FALSE,
                        strandCollapse = FALSE,
                        verbose = TRUE)
  # This is a matrix-backed BSseq object.
  sapply(assays(bsseq, withDimnames = FALSE), class)
  bsseq

  \dontrun{
  # Run read.bismark() on a single sample to construct a HDF5Array-backed BSseq
  # object (with data written to 'test_dir')
  test_dir <- tempfile("BSseq")
  bsseq <- read.bismark(files = infile,
                        colData = DataFrame(row.names = "test_data"),
                        rmZeroCov = FALSE,
                        strandCollapse = FALSE,
                        BACKEND = "HDF5Array",
                        dir = test_dir,
                        verbose = TRUE)
  # This is a HDF5Array-backed BSseq object.
  sapply(assays(bsseq, withDimnames = FALSE), class)
  # The 'M' and 'Cov' assays are in the HDF5 file 'assays.h5' (in 'test_dir').
  sapply(assays(bsseq, withDimnames = FALSE), path)
  }
}

\author{
  Peter Hickey \email{peter.hickey@gmail.com}
}

% TODO: Mention and link to findLoci().

\name{Internals}
\alias{matrixOrNULL-class}
\title{Internals}
\description{
   The \code{matrixOrNULL} class is a class union, representing a slot
   which is either a matrix or \code{NULL} (representing missing
   information). 
}
\value{
  Undocumented
}
\keyword{internal}

\name{hasGRanges-class}
\Rdversion{1.1}
\docType{class}
\alias{class:hasGRanges}
\alias{hasGRanges-class}
\alias{hasGRanges}
\alias{seqnames,hasGRanges-method}
\alias{seqnames<-,hasGRanges-method}
\alias{seqlevels,hasGRanges-method}
\alias{seqlevels<-,hasGRanges-method}
\alias{seqlengths,hasGRanges-method}
\alias{seqlengths<-,hasGRanges-method}
\alias{granges,hasGRanges-method}
\alias{start,hasGRanges-method}
\alias{start<-,hasGRanges-method}
\alias{end,hasGRanges-method}
\alias{end<-,hasGRanges-method}
\alias{width,hasGRanges-method}
\alias{width<-,hasGRanges-method}
\alias{strand,hasGRanges-method}
\alias{strand<-,hasGRanges-method}
\alias{strand<-,hasGRanges,ANY-method}
\alias{length,hasGRanges-method}
\alias{[,hasGRanges-method}
\alias{[,hasGRanges,ANY,ANY,ANY-method}
\alias{findOverlaps,hasGRanges,GenomicRanges-method}
\alias{findOverlaps,GenomicRanges,hasGRanges-method}
\alias{findOverlaps,hasGRanges,hasGRanges-method}
\alias{overlapsAny,hasGRanges,GenomicRanges-method}
\alias{overlapsAny,GenomicRanges,hasGRanges-method}
\alias{overlapsAny,hasGRanges,hasGRanges-method}
\alias{subsetByOverlaps,hasGRanges,GenomicRanges-method}
\alias{subsetByOverlaps,GenomicRanges,hasGRanges-method}
\alias{subsetByOverlaps,hasGRanges,hasGRanges-method}

\title{Class hasGRanges}
\description{
  A class with a GRanges slot, used as a building block for other
  classes.  Provides basic accessor functions etc.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{new("hasGRanges", ...)}.
}
\section{Slots}{
  \describe{
    \item{\code{gr}:}{Object of class \code{GRanges}.}
  }
}
\section{Methods}{
  \describe{
    \item{"["}{Subsets a single dimension.}
    \item{granges}{Get the \code{GRanges} object representing 
      genomic locations.}
    \item{start,start<-,end,end<-,width,width<-}{Start, end and width
      for the genomic locations of the object, also replacement
      functions.  This accessor functions operate directly on the
      \code{gr} slot.} 
    \item{strand,strand<-}{Getting and setting the
      \code{strand} of the genomic locations (the \code{gr} slot).} 
    \item{seqlengths,seqlengths<-}{Getting and setting the
      \code{seqlengths} of the genomic locations (the \code{gr} slot).} 
    \item{seqlevels,seqlevels<-}{Getting and setting the
      \code{seqlevels} of the genomic locations (the \code{gr} slot).} 
    \item{seqnames,seqnames<-}{Getting and setting the
      \code{seqnames} of the genomic locations (the \code{gr} slot).} 
    \item{show}{The show method.}
    \item{findOverlaps}{\code{(query = "hasGRanges", subject =
	"hasGRanges")}: finds overlaps between the \code{granges()} of
      the two objects.}
    \item{findOverlaps}{\code{(query = "GenomicRanges", subject =
	"hasGRanges")}: finds overlaps between query and the \code{granges()} of
      the subject.}
    \item{findOverlaps}{\code{(query = "hasGRanges", subject =
	"GenomicRanges")}: finds overlaps between the \code{granges()}
      of the query and the subject.}
    \item{subsetByOverlaps}{\code{(query = "hasGRanges", subject =
	"hasGRanges")}: Subset the query, keeping the genomic locations that 
      overlaps the subject.}
    \item{subsetByOverlaps}{\code{(query = "hasGRanges", subject =
	"GenomicRanges")}: Subset the query, keeping the genomic
      locations that overlaps the subject.}
    \item{subsetByOverlaps}{\code{(query = "GenomicRanges",
	subject = "hasGRanges")}: Subset the query, keeping the genomic
      locations that overlaps the subject.}
  }
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\note{
  If you extend the \code{hasGRanges} class, you should consider writing
  a subset method (\code{[}), and a \code{show} method.  If the new
  class supports single index subsetting, the \code{subsetByOverlaps}
  methods will automatically extend.
}
\examples{
showClass("hasGRanges")
}
\keyword{classes}

\name{getMeth}
\alias{getMeth}
\title{
  Obtain methylation estimates for BSseq objects.
}
\description{
  Obtain methylation estimates for BSseq objects, both smoothed and raw.
}
\usage{
getMeth(BSseq, regions = NULL, type = c("smooth", "raw"),
  what = c("perBase", "perRegion"), confint = FALSE, alpha = 0.95,
  withDimnames = TRUE)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{regions}{An optional \code{data.frame} or
    \code{GenomicRanges} object specifying a number of genomic regions.}
  \item{type}{This returns either smoothed or raw estimates of the methylation level.}
  \item{what}{The type of return object, see details.}
  \item{confint}{Should a confidence interval be return for the
    methylation estimates (see below).  This is only supported if
    \code{what} is equal to \code{perBase}.}
  \item{alpha}{alpha value for the confidence interval.}
  \item{withDimnames}{A \code{logical(1)}, indicating whether dimnames should
    be applied to extracted coverage elements. Setting
    \code{withDimnames = FALSE} increases the speed and memory efficiency with
    which coverage is extracted.}
}
\note{
  A \code{BSseq} object needs to be smoothed by the function
  \code{BSmooth} in order to support \code{type = "smooth"}.
}
\value{
  \strong{NOTE:} The return type of \code{getMeth} varies depending on its
  arguments.

  If \code{region = NULL} the \code{what} argument is ignored.  This is
  also the only situation in which \code{confint = TRUE} is supported.
  The return value is either a \code{\link[DelayedArray]{DelayedMatrix}}
  (\code{confint = FALSE} or a list with three
  \code{\link[DelayedArray]{DelayedMatrix}} components \code{confint =
  TRUE} (\code{meth}, \code{upper} and \code{lower}), giving the
  methylation estimates and (optionally) confidence intervals.

  Confidence intervals for \code{type = "smooth"} is based on standard
  errors from the smoothing algorithm (if present).  Otherwise it is
  based on pointwise confidence intervals for binomial distributions
  described in Agresti (see below), specifically the score confidence
  interval.

  If \code{regions} are specified, \code{what = "perBase"} will make the
  function return a list, each element of the list being a
  \code{\link[DelayedArray]{DelayedMatrix}} corresponding to a genomic
  region (and each row of the \code{\link[DelayedArray]{DelayedMatrix}}
  being a loci inside the region). If \code{what = "perRegion"} the
  function returns a \code{\link[DelayedArray]{DelayedMatrix}}, with
  each row corresponding to a region and containing the average
  methylation level in that region.
}
\references{
  A Agresti and B Coull.
  \emph{Approximate Is Better than "Exact" for Interval Estimation of
    Binomial Proportions}.
  The American Statistician (1998) 52:119-126.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}.
}
\seealso{
  \code{\linkS4class{BSseq}} for the \code{BSseq} class and
  \code{\link{BSmooth}} for smoothing such an object.
}
\examples{
data(BS.chr22)
head(getMeth(BS.chr22, type = "raw"))
reg <- GRanges(seqnames = c("chr22", "chr22"),
  ranges = IRanges(start = c(1, 2*10^7), end = c(2*10^7 +1, 4*10^7)))
head(getMeth(BS.chr22, regions = reg, type = "raw", what = "perBase"))

#-------------------------------------------------------------------------------
# An example using a HDF5Array-backed BSseq object
#

library(HDF5Array)
# See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
hdf5_BS.chr22 <- saveHDF5SummarizedExperiment(x = BS.chr22,
                                              dir = tempfile())
head(getMeth(hdf5_BS.chr22, type = "raw"))
head(getMeth(hdf5_BS.chr22, regions = reg, type = "raw", what = "perBase"))
}

\name{BSseq}
\alias{BSseq}
\title{
  The constructor function for BSseq objects.
}
\description{
  The constructor function for BSseq objects.
}
\usage{
BSseq(M = NULL, Cov = NULL, Filtered = NULL, coef = NULL, se.coef = NULL,
  trans = NULL, parameters = NULL, pData = NULL, gr = NULL,
  pos = NULL, chr = NULL, sampleNames = NULL, rmZeroCov = FALSE)
}
\arguments{
  \item{M}{A matrix-like object of methylation evidence (see 'Details' below).}
  \item{Cov}{A matrix-like object of coverage (see 'Details' below)).}
  \item{Filtered}{A matrix-like object of ambiguous modification bases obtained from modbam2bed.}
  \item{coef}{A matrix-like object of smoothing estimates (see 'Details'
    below).}
  \item{se.coef}{A matrix-like object of smoothing standard errors (see
    'Details' below).}
  \item{trans}{A smoothing transformation.}
  \item{parameters}{A list of smoothing parameters.}
  \item{pData}{An \code{data.frame} or \code{\link[S4Vectors]{DataFrame}}.}
  \item{sampleNames}{A vector of sample names.}
  \item{gr}{An object of type \code{\link[GenomicRanges]{GRanges}}.}
  \item{pos}{A vector of locations.}
  \item{chr}{A vector of chromosomes.}
  \item{rmZeroCov}{Should genomic locations with zero coverage in all
    samples be removed.}
}
\details{
  The 'M', 'Cov', 'coef', and 'se.coef' matrix-like objects will be
  coerced to \code{\link[DelayedArray]{DelayedMatrix}} objects; see
  \code{\link[DelayedArray]{DelayedMatrix}} in the \pkg{DelayedArray}
  package for the full list of supported matrix-like objects. We
  recommend using \link[base]{matrix} objects for in-memory storage of
  data and \code{\link[HDF5Array]{HDF5Matrix}} for on-disk storage of
  data.

  Genomic locations are specified either through \code{gr} or through
  \code{chr} and \code{pos} but not both.  There should be the same
  number of genomic locations as there are rows in the \code{M} and
  \code{Cov} matrix.

  The argument \code{rmZeroCov} may be useful in order to reduce the
  size of the return object be removing methylation loci with zero
  coverage.

  In case one or more methylation loci appears multiple times, the
  \code{M} and \code{Cov} matrices are summed over rows linked to the
  same methylation loci. See the example below.

  Users should never have to specify \code{coef}, \code{se.coef},
  \code{trans}, and \code{parameters}, this is for internal use (they
  are added by \code{BSmooth}).

  \code{phenoData} is a way to specify pheno data (as known from the
  \code{ExpressionSet} and \code{eSet} classes), at a minimum
  \code{sampleNames} should be given (if they are not present, the
  function uses \code{col.names(M)}).
}
\value{
  An object of class \code{BSseq}.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  \code{\linkS4class{BSseq}}
}
\examples{
  M <- matrix(0:8, 3, 3)
  Cov <- matrix(1:9, 3, 3)
  BS1 <- BSseq(chr = c("chr1", "chr2", "chr1"), pos = c(1,2,3),
               M = M, Cov = Cov, sampleNames = c("A","B", "C"))
  BS1
  BS2 <- BSseq(chr = c("chr1", "chr1", "chr1"), pos = c(1,1,1),
               M = M, Cov = Cov, sampleNames = c("A","B", "C"))
  BS2

  #-------------------------------------------------------------------------------
  # An example using a HDF5Array-backed BSseq object
  #

  library(HDF5Array)
  hdf5_M <- realize(M, "HDF5Array")
  hdf5_Cov <- realize(Cov, "HDF5Array")
  hdf5_BS1 <- BSseq(chr = c("chr1", "chr2", "chr1"),
                    pos = c(1, 2, 3),
                    M = hdf5_M,
                    Cov = hdf5_Cov,
                    sampleNames = c("A", "B", "C"))
  hdf5_BS1
  hdf5_BS2 <- BSseq(chr = c("chr1", "chr1", "chr1"),
                    pos = c(1, 1, 1),
                    M = hdf5_M,
                    Cov = hdf5_Cov,
                    sampleNames = c("A", "B", "C"))
  hdf5_BS2
}

\name{GoodnessOfFit}
\alias{poissonGoodnessOfFit}
\alias{poissonGoodnessOfFit}
\alias{chisqGoodnessOfFit}
\alias{binomialGoodnessOfFit}
\alias{print.chisqGoodnessOfFit}
\alias{plot.chisqGoodnessOfFit}
\title{
  Binomial and poisson goodness of fit statistics for BSSeq objects
}
\description{
  Binomial and poisson goodness of fit statistics for BSSeq objects,
  including plotting capability.
}
\usage{
poissonGoodnessOfFit(BSseq, nQuantiles = 10^5)
binomialGoodnessOfFit(BSseq, method = c("MLE"), nQuantiles = 10^5)
\S3method{print}{chisqGoodnessOfFit}(x, \ldots)
\S3method{plot}{chisqGoodnessOfFit}(x, type = c("chisq", "pvalue"), plotCol = TRUE, qqline = TRUE,
  pch = 16, cex = 0.75, \ldots)
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{x}{A chisqGoodnessOfFit object (as produced by
    \code{poissonGoodnessOfFit} or \code{binomialGoodnessOfFit}).}
  \item{nQuantiles}{The number of (evenly-spaced) quantiles stored in
    the return object.}
  \item{method}{How is the parameter estimated.}
  \item{type}{Are the chisq or the p-values being plotted.}
  \item{plotCol}{Should the extreme quantiles be colored.}
  \item{qqline}{Add a \code{qqline}.}
  \item{pch, cex}{Plotting symbols and size.}
  \item{...}{Additional arguments being passed to \code{qqplot} (for
    \code{plot}) or ignored (for \code{print}).}
}
\details{

  These functions compute and plot goodness of fit statistics for
  \code{BSseq} objects.  For each methylation loci, the Poisson
  goodness of fit statistic tests whether the coverage (at that loci) is
  independent and identically Poisson distributed across the samples.
  In a similar fashion, the Binomial goodness of fit statistic tests
  whether the number of reads supporting methylation are independent and
  identically binomial distributed across samples (with different size
  parameters given by the coverage vector).

  These functions do not handle \code{NA} values.
}
\value{
  The plotting method is invoked for its side effect.  Both
  \code{poissonGoodnessOfFit} and \code{binomialGoodnessOfFit} returns
  an object of class \code{chisqGoodnessOfFit} which is a list with components
 \item{chisq}{a vector of Chisq values.}
 \item{quantiles}{a vector of quantiles (of the chisq values).}
 \item{df}{degress of freedom}
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  For the plotting method, see \code{qqplot}.
}
\examples{
if(require(bsseqData)) {
  data(BS.cancer.ex)
  BS.cancer.ex <- updateObject(BS.cancer.ex)
  gof <- poissonGoodnessOfFit(BS.cancer.ex)
  plot(gof)

#-------------------------------------------------------------------------------
# An example using a HDF5Array-backed BSseq object
#
\donttest{
    library(HDF5Array)

    # See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
    hdf5_BS.cancer.ex <- saveHDF5SummarizedExperiment(x = BS.cancer.ex,
                                                      dir = tempfile())
    hdf5_gof <- poissonGoodnessOfFit(hdf5_BS.cancer.ex)
    plot(hdf5_gof)
}
}
}

\name{BS.chr22}
\alias{BS.chr22}
\docType{data}
\title{
  Whole-genome bisulfite sequencing for chromosome 22 from Lister et al.
}
\description{
  This dataset represents chromosome 22 from the IMR90 cell line
  sequenced in Lister et al.  Only CpG methylation are included (there
  were very few non-CpG loci).  The two samples are two different
  extractions from the same cell line (ie. technical replicates), and
  are pooled in the analysis in the original paper.
}
\usage{data(BS.chr22)}
\format{
  An object of class \code{BSseq}.
}
\details{
  All coordinates are in hg18.
}
\source{
  Obtained from \url{http://neomorph.salk.edu/human_methylome/data.html}
  specifically the files \url{mc_imr90_r1.tar.gz} and
  \url{mc_imr90_r2.tar.gz}.  A script which downloads these files and
  constructs the \code{BS.chr22} object may be found in
  \file{inst/scripts/get_BS.chr22.R}, see the example.
}
\references{
  R Lister et al.
  \emph{Human DNA methylomes at base resolution show widespread
    epigenomic differences}.
  Nature (2009) 462, 315-322.
}
\examples{
data(BS.chr22)
BS.chr22

script <- system.file("scripts", "get_BS.chr22.R", package = "bsseq")
script
readLines(script)
}
\keyword{datasets}

\name{BSseqTstat-class}
\Rdversion{1.1}
\docType{class}
\alias{BSseqTstat}
\alias{BSseqTstat-class}
\alias{[,BSseqTstat-method}
\alias{[,BSseqTstat,ANY,ANY,ANY-method}
\alias{show,BSseqTstat-method}
\alias{updateObject,BSseqTstat-method}
\title{Class BSseqTstat}
\description{
  A class for representing t-statistics for smoothed whole-genome
  bisulfite sequencing data.
}
\usage{
  BSseqTstat(gr = NULL, stats = NULL, parameters = NULL)
}
\arguments{
  \item{gr}{The genomic locations as an object of class \code{GRanges}.}
  \item{stats}{The statistics, as a matrix-like object (see 'Details' below).}
  \item{parameters}{A list of parameters.}
}

\details{
    The 'stats' matrix-like object will be coerced to a
    \code{\link[DelayedArray]{DelayedMatrix}} objects; see
    \code{\link[DelayedArray]{DelayedMatrix}} in the \pkg{DelayedArray}
    package for the full list of supported matrix-like objects. We
    recommend using \code{\link[base]{matrix}} objects for in-memory
    storage of data and \code{\link[HDF5Array]{HDF5Matrix}} for on-disk
    storage of data.
}
\section{Objects from the Class}{
Objects can be created by calls of the form \code{BSseqTstat(...)}.
However, usually objects are returned by \code{BSmooth.tstat(...)} and
not constructed by the user..
}
\section{Slots}{
  \describe{
    \item{\code{stats}:}{This is a
      \code{\link[DelayedArray]{DelayedMatrix}} object with columns
      representing  various statistics for methylation loci along the
      genome.}
    \item{\code{parameters}:}{Object of class \code{list}.  A list of
      parameters representing how the t-statistics were computed.}
    \item{\code{gr}:}{Object of class \code{GRanges} giving genomic
      locations.}
  }
}
\section{Extends}{
Class \code{\linkS4class{hasGRanges}}, directly.
}
\section{Methods}{
  \describe{
    \item{[}{The subsetting operator; one may only subset in one
      dimension, corresponding to methylation loci.}
    \item{show}{The show method.}
  }
}
\section{Utilities}{
  This class extends \code{hasGRanges} and therefore inherits a number
  of useful \code{GRanges} methods that operate on the \code{gr} slot,
  used for accessing and setting the genomic locations and also do
  \code{subsetByOverlaps}.
  }
\section{Coercion}{
  Package version 1.11.1 introduced a new version of representing
  \sQuote{BSseqTstat} objects. You can update old serialized (saved)
  objects by invoking \code{x <- updateObject(x)}.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  The package vignette(s).  \code{\linkS4class{hasGRanges}} for accessing
  the genomic locations.  \code{\link{BSmooth.tstat}} for a function
  that returns objects of class \code{BSseqTstat}, and \code{\link{dmrFinder}}
  for a function that computes DMRs based on the t-statistics.  Also see
  \code{\link[bsseqData]{BS.cancer.ex.tstat}} for an example of the
  class in the \pkg{bsseqData} package.
}
\examples{
if(require(bsseqData)) {
    data(BS.cancer.ex.tstat)
    dmrFinder(BS.cancer.ex.tstat)
}
}
\keyword{classes}

\name{BSmooth.fstat}
\alias{BSmooth.fstat}
\title{
    Compute F-statistics based on smoothed whole-genome bisulfite
    sequencing data.
}
\description{
    Compute F-statistics based on smoothed whole-genome bisulfite
    sequencing data.
}
\usage{
BSmooth.fstat(BSseq, design, contrasts, verbose = TRUE)
}
\arguments{
    \item{BSseq}{An object of class \code{BSseq}.}
    \item{design}{The design matrix of the bisulfite-sequencing experiment,
        with rows corresponding to arrays and columns to coefficients to be
        estimated.}
    \item{contrasts}{Numeric matrix with rows corresponding to columns in
        \code{design} and columns containing contrasts. May be a vector if
        there is only one contrast.}
    \item{verbose}{Should the function be verbose?}
}
\details{
    \strong{TODO}
}
\value{
    An object of class \linkS4class{BSseqStat}.
}
\author{
    Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
    \code{\link{BSmooth}} for the input object and
    \linkS4class{BSseq} for its class.
    \linkS4class{BSseqStat} describes the return class. This
    function is likely to be followed by the use of
    \code{\link{smoothSds}}, \code{\link{computeStat}}, and
    \code{\link{dmrFinder}}.
}
\examples{
\donttest{
    if(require(bsseqData)) {
    # limma required for makeContrasts()
    library(limma)
    data(keepLoci.ex)
    data(BS.cancer.ex.fit)
    BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
    ## Remember to subset the BSseq object, see vignette for explanation
    ## TODO: Kind of a forced example
    design <- model.matrix(~0 + BS.cancer.ex.fit$Type)
    colnames(design) <- gsub("BS\\\.cancer\\\.ex\\\.fit\\\$Type", "",
                             colnames(design))
    contrasts <- makeContrasts(
        cancer_vs_normal = cancer - normal,
        levels = design
    )
    BS.stat <- BSmooth.fstat(BS.cancer.ex.fit[keepLoci.ex,],
                             design,
                             contrasts)
    BS.stat

    #---------------------------------------------------------------------------
    # An example using a HDF5Array-backed BSseq object
    #
    library(HDF5Array)
    # See ?SummarizedExperiment::saveHDF5SummarizedExperiment for details
    hdf5_BS.cancer.ex.fit <- saveHDF5SummarizedExperiment(
        x = BS.cancer.ex.fit[keepLoci.ex, ],
        dir = tempfile())
    hdf5_BS.stat <- BSmooth.fstat(hdf5_BS.cancer.ex.fit,
                                  design,
                                  contrasts)
    hdf5_BS.stat
    }
}
}
\keyword{internal}

\name{smoothSds}
\alias{smoothSds}
\title{
    Smooth the standard deviations using a thresholded running mean based on
    smoothed whole-genome bisulfite sequencing data.
}
\description{
    Smooth the standard deviations using a thresholded running mean based on
    smoothed whole-genome bisulfite sequencing data.
}
\usage{
    smoothSds(BSseqStat, k = 101, qSd = 0.75, mc.cores = 1, maxGap = 10^8,
              verbose = TRUE)
}

\arguments{
    \item{BSseqStat}{An object of class \code{BSseqStat}, typically an object
        returned by \code{\link{BSmooth.fstat}(...)} and not constructed by
        the user.}
    \item{k}{A positive scalar, see details.}
    \item{qSd}{A scalar between 0 and 1, see details.}
    \item{mc.cores}{The number of cores used.  Note that setting
        \code{mc.cores} to a value greater than 1 is not supported on MS
        Windows, see the help page for \code{mclapply}.}
    \item{maxGap}{A scalar greater than 0, see details.}
    \item{verbose}{Should the function be verbose?}
}
\details{
    The standard deviation estimates are smoothed using a running mean with a
    width of \code{k} and thresholded using \code{qSd} which sets the minimum
    standard deviation to be the \code{qSd}-quantile.
}
\value{
    An object of class \linkS4class{BSseqStat}. More speciically, the input
    \linkS4class{BSseqStat} object with the computed statistics added to the
    \code{stats} slot (accessible with \code{\link{getStats}}).
}
\author{
    Kasper Daniel Hansen \email{khansen@jhsph.edu}
}

\seealso{
    \code{\link{BSmooth.fstat}} for the function to create the appropriate
    \code{\linkS4class{BSseqStat}} input object.
    \code{\linkS4class{BSseqStat}} also describes the return class.  This
    function is likely to be followed by the use of \code{\link{computeStat}}.}
\examples{
  \donttest{
    if(require(bsseqData)) {
        # library(limma) required for makeContrasts()
        library(limma)
        data(keepLoci.ex)
        data(BS.cancer.ex.fit)
        BS.cancer.ex.fit <- updateObject(BS.cancer.ex.fit)
        ## Remember to subset the BSseq object, see vignette for explanation
        ## TODO: Kind of a forced example
        design <- model.matrix(~0 + BS.cancer.ex.fit$Type)
        colnames(design) <- gsub("BS\\\.cancer\\\.ex\\\.fit\\\$Type", "",
                                 colnames(design))
        contrasts <- makeContrasts(
            cancer_vs_normal = cancer - normal,
            levels = design
        )
        BS.stat <- BSmooth.fstat(BS.cancer.ex.fit[keepLoci.ex,],
                                 design,
                                 contrasts)
        BS.stat <- smoothSds(BS.stat)
        ## Comparing the raw standard deviations to the smoothed standard
        ## deviations
        summary(getStats(BS.stat, what = "rawSds"))
        summary(getStats(BS.stat, what = "smoothSds"))
    }
  }
}
\keyword{internal}

\name{read.umtab}
\alias{read.umtab}
\alias{read.umtab2}
\title{
  Parsing UM tab files (legacy output) containing output from the
  BSmooth aligner. 
}
\description{
  Parsing UM tab files containing output from the bisulfite aligner
  \code{Merman}.  This is two different legacy formats, which we keep
  around.  These functions are likely to be deprecated in the future.
}
\usage{
read.umtab(dirs, sampleNames = NULL, rmZeroCov = FALSE,
  pattern = NULL, keepU = c("U10", "U20", "U30", "U40"),
  keepM = c("M10", "M20", "M30", "M40"), verbose = TRUE)

read.umtab2(dirs, sampleNames = NULL, rmZeroCov = FALSE,
  readCycle = FALSE, keepFilt = FALSE,
  pattern = NULL, keepU, keepM, verbose = TRUE)
}
\arguments{
  \item{dirs}{Input directories.  Usually each sample is in a different
    directory.} 
  \item{pattern}{An optional pattern, see \code{list.files} in the base
    package.} 
  \item{sampleNames}{sample names, based on the order of \code{dirs}.}
  \item{rmZeroCov}{Should methylation loci that have zero coverage in
    all samples be removed.  This will result in a much smaller object
    if the data originates from (targeted) capture bisulfite
    sequencing.} 
  \item{keepU}{A vector of U columns which are kept.}
  \item{keepM}{A vector of M columns which are kept.}
  \item{readCycle}{Should the cycle columns be returned?}
  \item{keepFilt}{Should the filter columns be returned?}
  \item{verbose}{Make the function verbose.}
}
\value{
  Both functions returns lists, the components are
  \item{BSdata}{An object of class \code{BSseq} containing the
    methylation evidence.}
  \item{GC}{A vector of local GC content values.}
  \item{Map}{A vector of local mapability values.}
  \item{Mcy}{A matrix of the number of unique M cycles.}
  \item{Ucy}{A matrix of the number of unique U cycles.}
  \item{chr}{A vector of chromosome names.}
  \item{pos}{A vector of genomic positions.}
  \item{M}{A matrix representing methylation evidence.}
  \item{U}{A matrix representing un-methylation evidence.}
  \item{csums}{Description of 'comp2'}
}
\details{
  \code{read.umtab2} is newer than \code{read.umtab} and both process
  output from older versions of the BSmooth alignment suite (versions
  older than 0.6.1).  These functions are likely to be deprecated in the
  future.  Newer output from the BSmooth alignment suite can be parsed
  using \code{read.bsmooth}.

  A script using this function can be found in the \code{bsseqData}
  package, in the file \file{scripts/create_BS.cancer.R}.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  \code{\link{read.bsmooth}}.
}
\examples{
\dontrun{
require(bsseqData)
umDir <- system.file("umtab", package = "bsseqData")
sampleNames <- list.files(umDir)
dirs <- file.path(umDir, sampleNames, "umtab")
umData <- read.umtab(dirs, sampleNames)
}
}

\name{data.frame2GRanges}
\alias{data.frame2GRanges}
\title{
  Converts a data frame to a GRanges.
}
\description{
  Converting a data.frame to a GRanges object.  The data.frame needs
  columns like chr, start and end (strand is optional).  Additional
  columns may be kept in the GRanges object.
}
\usage{
data.frame2GRanges(df, keepColumns = FALSE, ignoreStrand = FALSE)
}
\arguments{
  \item{df}{A \code{data.frame} with columns \code{chr} or
    \code{seqnames}, \code{start}, \code{end} and optionally a
    \code{strand column}.}    
  \item{keepColumns}{In case \code{df} has additional columns, should
    these columns be stored as metadata columns on the return GRanges
    or should they be discarded.}
  \item{ignoreStrand}{In case \code{df} has a \code{strand} column,
    should this column be ignored.}
}
\note{
  In case \code{df} has \code{rownames}, they will be used as
  \code{names} for the return object.
}
\value{
  An object of class \code{GRanges}
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\examples{
df <- data.frame(chr = "chr1", start = 1:3, end = 2:4,
                 strand = c("+","-","+"))
data.frame2GRanges(df, ignoreStrand = TRUE)
}

\name{getStats}
\alias{getStats}
\title{
  Obtain statistics from a BSseqTstat object
}
\description{
  Essentially an accessor function for the statistics of a \code{BSseqTstat}
  object.
}
\usage{
getStats(bstat, regions = NULL, ...)
}
\arguments{
  \item{bstat}{An object of class \code{BSseqStat} or \code{BSseqTstat}.}
  \item{regions}{An optional \code{data.frame} or
    \code{GenomicRanges} object specifying a number of genomic
    regions.}
  \item{...}{Additional arguments passed to the different backends based
    on the class of \code{bstat}; see Details.}
}
\value{
  An object of class \code{data.frame} possible restricted to the
  regions specified.
}
\details{
  Additional argument when the \code{bstat} object is of class \code{BSseqTstat}:
  \describe{
    \item{stat}{Which statistics column should be obtained.}
  }
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}
\seealso{
  \code{\linkS4class{BSseqTstat}} for the \code{BSseqTstat} class, and
  \code{\link{getCoverage}} and \code{\link{getMeth}} for similar
  functions, operating on objects of class \code{BSseq}.
}
\examples{
if(require(bsseqData)) {
  data(BS.cancer.ex.tstat)
  head(getStats(BS.cancer.ex.tstat))
  reg <- GRanges(seqnames = c("chr22", "chr22"),
     ranges = IRanges(start = c(1, 2*10^7), end = c(2*10^7 +1, 4*10^7)))
  head(getStats(BS.cancer.ex.tstat, regions = reg))
}
}

\name{read.bedMethyl}
\alias{read.bedMethyl}
\title{
  Parsing bedMethyl output from modkit pileup.
}
\description{
  Parsing bedMethyl output from modkit pileup.
}
\usage{
read.bedMethyl(files,
               loci = NULL,
               colData = NULL,
               rmZeroCov = TRUE,
               strandCollapse = TRUE,
               BPPARAM = bpparam(),
               BACKEND = NULL,
               dir = tempfile("BSseq"),
               replace = FALSE,
               chunkdim = NULL,
               level = NULL,
               nThread = 1L,
               verbose = getOption("verbose"))
}
\arguments{
  \item{files}{The path to the files created by running modkit pileup, one sample per file. See the methods section of [link to preprint] for validated output.}
  \item{loci}{\code{NULL} (default) or a \code{\link[GenomicRanges]{GenomicRanges}} instance containing methylation loci (all with width equal to 1). If \code{loci = NULL}, then \code{read.bedMethyl()} will perform a first pass over the bedMethyl files to identify candidate loci. If \code{loci} is a \code{\link[GenomicRanges]{GenomicRanges}} instance, then these form the candidate loci. The candidate loci will be collapsed if \code{strandCollapse = TRUE}.}
  \item{colData}{An optional \code{\link[S4Vectors]{DataFrame}} describing the samples. Row names, if present, become the column names of the \linkS4class{BSseq} object. If \code{NULL}, then a \code{\link[S4Vectors]{DataFrame}} will be created with \code{files} used as the row names.}
  \item{rmZeroCov}{A \code{logical(1)} indicating whether methylation loci that have zero coverage in all samples should be removed. Default setting is \code{rmZeroCov = TRUE}}
  \item{strandCollapse}{A \code{logical(1)} indicating whether strand-symmetric methylation loci (i.e. CpGs) should be collapsed across strands.}
  \item{BPPARAM}{An optional \code{\link[BiocParallel]{BiocParallelParam}} instance determining the parallel back-end to be used during evaluation.}
  \item{BACKEND}{\code{NULL} or a single string specifying the name of the realization backend. Currently, the backend is not supported for downstream applications.}
  \item{dir}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.} The path (as a single string) to the directory where to save the HDF5-based \linkS4class{BSseq} object.}
  \item{replace}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.} If the directory \code{dir} already exists, should it be replaced with a new one?}
  \item{chunkdim}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.} The dimensions of the chunks to use for writing the data to disk.}
  \item{level}{The compression level to use for writing the data to disk.}
  \item{nThread}{The number of threads used by \code{\link[data.table]{fread}} when reading the \code{files}.}
  \item{verbose}{A \code{logical(1)} indicating whether progress messages should be printed (default \code{TRUE}).}
}

\section{File formats}{
  The format of each file should be similar to the examples in [link to preprint]. Files ending in \code{.gz}, \code{.bz2}, \code{.xz}, or \code{.zip} will be automatically decompressed to \code{\link{tempdir}()}.
  \subsection{Supported file formats}{
    Modkit bedMethyl files from modkit pileup. For downstream likelihood functions we recommend running modkit pileup on output from bam files modification/basecalled using a CG context model and not using a reference genome for pileup.     
  }
  \subsection{Unsupported file formats}{
    Other types of output.
  }
  \subsection{One-based vs. zero-based genomic co-ordinates}{
    The genomic co-ordinates of bedMethyl files are zero-based. Since Bioconductor packages typically use one-based co-ordinates, the co-ordinates from the bedMethyl files are converted to one-based in the BSseq object.
  }
}

\examples{
# Example: Reading bedMethyl files included in the bsseq package
# Paths to example bedMethyl files in the package's extdata directory
infiles <- c(system.file("extdata/HG002_nanopore_test.bedMethyl.gz",
                         package = "bsseq"),
             system.file("extdata/HG002_pacbio_test.bedMethyl.gz",
                         package = "bsseq"))

# Run the function to import data
bsseq <- read.bedMethyl(files = infiles,
                        colData = DataFrame(row.names = c("test_nanopore", 
                                                          "test_pacbio")),
                        rmZeroCov = FALSE,
                        strandCollapse = TRUE,
                        verbose = TRUE)

# View the resulting BSseq object
bsseq
}

\author{
  Søren Blikdal Hansen (soren.blikdal.hansen@sund.ku.dk) 
}

\name{read.bsmooth}
\alias{read.bsmooth}
\title{
  Parsing output from the BSmooth alignment suite
}
\description{
  Parsing output from the BSmooth alignment suite.
}
\usage{
read.bsmooth(dirs, sampleNames = NULL, seqnames = NULL,
  returnRaw = FALSE, qualityCutoff = 20, rmZeroCov = FALSE,
  verbose = TRUE)
}
\arguments{
  \item{dirs}{Input directories.  Usually each sample is in a different
    directory, or perhaps each (sample, lane) is a different directory.} 
  \item{sampleNames}{sample names, based on the order of \code{dirs}.
    If \code{NULL} either set to \code{basename(dirs)} (if unique) or
    \code{dirs}.}
  \item{seqnames}{The default is to read all BSmooth output files in
    \code{dirs}.  Using this argument, it is possible to restrict this
    to only files with names in \code{seqnames} (apart from
    \code{.cpg.tsv} and optionally \code{.gz}).}
  \item{returnRaw}{Should the function return the complete information
    in the output files?}
  \item{qualityCutoff}{Only use evidence (methylated and unmethylated
    evidence) for a given methylation loci, if the base in the read has
    a quality greater than this cutoff.}
  \item{rmZeroCov}{Should methylation loci that have zero coverage in
    all samples be removed.  This will result in a much smaller object
    if the data originates from (targeted) capture bisulfite
    sequencing.} 
  \item{verbose}{Make the function verbose.}
}
\note{
  Input files can either be gzipped or not.  Gzipping the input files
  results in much greater speed of reading (and saves space), so it is
  recommended.
  
  We are working on making this function faster and less memory hungry.
}
\value{
  Either an object of class \code{BSseq} (if \code{returnRaw = FALSE})
  or a list of \code{GRanges} which each component coming from a
  directory. 
}
\seealso{
  \code{\link{read.umtab}} for parsing legacy (old) formats from the
  BSmooth alignment suite.  \code{\link{collapseBSseq}} for collapse
  (merging or summing) the data in two different directories.
}
\author{
  Kasper Daniel Hansen \email{khansen@jhsph.edu}
}

\name{read.modbam2bed}
\alias{read.modbam2bed}
\title{Construct BSseq objects from nanopore BED files}
\description{
Construct BSseq objects from nanopore BED files
}
\usage{
read.modbam2bed(
    files,
    colData = NULL,
    rmZeroCov = FALSE,
    strandCollapse = TRUE
)
}
\arguments{
    \item{files}{vector, BED files}
    \item{colData}{data frame, phenotypic data with samples as rows and variables as columns}
    \item{rmZeroCov}{A logical (1) indicating whether methylation loci that have zero coverage in all samples be removed}
    \item{strandCollapse}{A logical (1) indicating whether stand-symmetric methylation loci (i.e. CpGs) should be collapsed across strands}
}

\value{
BSseq object
}
\details{
This function is deprecated in favor of `read.bedMethyl()`.

This function reads in nanopore sequencing modified BED files
to Bsseq objects. Nanopore sequencing data (i.e. aggregated modified base
counts) is stored in modified-base BAM files. These modified-base BAM files
are converted to bedMethyl (BED) files using \href{https://github.com/epi2me-labs/modbam2bed}{modbam2bed}.

\subsection{Details for using modbam2bed}{
    After installing modbam2bed, a conda environment is activated. Index files
for BAM files are created using \code{samtools index}. The code requires aligned reads with the Mm and Ml tags (MM and ML also supported), and the reference sequence used for alignment (<reference.fasta>).

\itemize{
    \item \code{-e, -- extended} to output canonical, modified, and filtered bases;
    \item \code{-m, -- mod_base=BASE} to output modified base of interest, one of: 5mC, 5hmC, 5fC, 5caC, 5hmU, 5fU, 5caU, 6mA, 5oxoG, Xao. (Or modA, modC, modG, modT, modU, modN for generic modified base);
    \item \code{-r, --region=chr:start-end} to output chromosome or genomic region of interest;
    \item \code{-f, --threshold=THRESHOLD} to output filtered bases for
probability lower than threshold (default = 0.66)
}
}

  \subsection{modbam2bed to Bsseq object}{
    After creating BED files using modbam2bed, the BED files are read in and the Bsseq object is constructed via \code{read.modbam2bed()} function. The function reads in BED files, extract genomic regions, methylation, coverage, ambiguous modification status data and sample information and then construct Bsseq object using \code{BSseq} function within the package.}
}
\examples{
files <- c(system.file("extdata/modbam2bed/ctr1.chr10.chr11.bed.gz", package = "bsseq"),
    system.file("extdata/modbam2bed/ctr2.chr10.chr11.bed.gz", package = "bsseq"),
    system.file("extdata/modbam2bed/ctr3.chr10.chr11.bed.gz", package = "bsseq"),
    system.file("extdata/modbam2bed/tret1.chr10.chr11.bed.gz", package = "bsseq"),
    system.file("extdata/modbam2bed/tret2.chr10.chr11.bed.gz", package = "bsseq"),
    system.file("extdata/modbam2bed/tret3.chr10.chr11.bed.gz", package = "bsseq"))
pd <- data.frame(condition = rep(c("control", "treatment"), each = 3),
    replicate = rep(c("rep1", "rep2", "rep3"), times = 2))
## bsseq_nano <- bsseq::read.modbam2bed(files,colData=pd,rmZeroCov = FALSE,
##     strandCollapse=TRUE)
}

\name{getCpGs}
\alias{getCpGs}
\title{
  Get CpG (and non-CpG) loci from a single sample BSseq object.
}
\description{
  This function identifies CpG (and non-CpG) loci from a single sample BSseq object using scaled likelihoods computed from the count of CpG sites and Non-CpG sites mapped to the loci, based on the specified type, and minimum scaled likelihood (threshold). The function only work for BSseq objects generated using read.bedMethyl. 
}
\usage{
getCpGs(BSseq, type = c("homozygous", "heterozygous", "allCpG", "nonCpG"), threshold = 0.99, e = NULL)
}
\arguments{
  \item{BSseq}{An single sample object of class \code{BSseq}.}
  \item{type}{A character string specifying the type of loci to extract. Must be one of \code{"homozygous"}, \code{"heterozygous"}, \code{"allCpG"}, or \code{"nonCpG"}.}
  \item{threshold}{A numeric value between 0 and 1 specifying the minimum likelihood threshold required for loci to be included.}
  \item{e}{An optional numeric value representing the error rate. If \code{NULL}, the error rate is estimated using \code{\link{estimateErrorRate}}.}
}
\value{
  An integer vector of indices representing the loci that match the criteria.
}
\author{
  Søren Blikdal Hansen (soren.blikdal.hansen@sund.ku.dk)
}
\seealso{
  \code{\linkS4class{BSseq}} for the \code{BSseq} class, 
  \code{\link{read.bedMethyl}} for details on reading data into a \code{BSseq} object,
  \code{\link{estimateErrorRate}} for estimating the CpG-specific error rate.
}
\examples{
# Example input files
infiles <- c(system.file("extdata/HG002_nanopore_test.bedMethyl.gz",
                         package = "bsseq"),
             system.file("extdata/HG002_pacbio_test.bedMethyl.gz",
                         package = "bsseq"))

# Run the function to import data
bsseq <- read.bedMethyl(files = infiles,
                        colData = DataFrame(row.names = c("test_nanopore", 
                                                          "test_pacbio")),
                        strandCollapse = TRUE,
                        verbose = TRUE)

# Filter CpG sites for the Nanopore dataset
bsseq_nano <- bsseq[, 1]
bsseq_nano_99All_filtered <- bsseq[getCpGs(bsseq_nano, 
                                           type = "allCpG", threshold = 0.99)]

# Filter CpG sites for the PacBio dataset
bsseq_pacbio <- bsseq[, 2]
bsseq_pacbio_99All_filtered <- bsseq[getCpGs(bsseq_pacbio, 
                                             type = "allCpG", threshold = 0.99)]
}

\name{BSmooth}
\alias{BSmooth}
\title{
  BSmooth, smoothing bisulfite sequence data
}
\description{
  This implements the BSmooth algorithm for estimating methylation levels from bisulfite sequencing data.}
\usage{
BSmooth(BSseq,
        ns = 70,
        h = 1000,
        maxGap = 10^8,
        keep.se = FALSE,
        BPPARAM = bpparam(),
        chunkdim = NULL,
        level = NULL,
        verbose = getOption("verbose"))
}
\arguments{
  \item{BSseq}{An object of class \code{BSseq}.}
  \item{ns}{The minimum number of methylation loci in a smoothing window.}
  \item{h}{The minimum smoothing window, in bases.}
  \item{maxGap}{The maximum gap between two methylation loci, before the smoothing is broken across the gap.
  The default smoothes each chromosome separately.}
  \item{keep.se}{Should the estimated standard errors from the smoothing
    algorithm be kept. This will make the return object roughly 30
    percent bigger and is currently not be used for anything in \pkg{bsseq}.}
  \item{BPPARAM}{An optional
    \code{\link[BiocParallel]{BiocParallelParam}} instance determining
    the parallel back-end to be used during evaluation. Currently
    supported are \code{\link[BiocParallel]{SerialParam}} (Unix, Mac,
    Windows), \code{\link[BiocParallel]{MulticoreParam}} (Unix and Mac),
    \code{\link[BiocParallel]{SnowParam}} (Unix, Mac, and Windows,
    limited to single-machine clusters), and
    \code{\link[BiocParallel]{BatchtoolsParam}} (Unix, Mac, Windows,
    only with the in-memory realization backend). See sections
    'Parallelization and progress monitoring' and 'Realization backends'
    for further details.}
  \item{chunkdim}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    The dimensions of the chunks to use for writing the data to
    disk. By default, \code{\link[HDF5Array]{getHDF5DumpChunkDim}()}
    using the dimensions of the returned \linkS4class{BSseq} object will
    be used. See \code{?{getHDF5DumpChunkDim}} for more information.}
  \item{level}{\strong{Only applicable if \code{BACKEND == "HDF5Array"}.}
    The compression level to use for writing the data to disk. By
    default, \code{\link[HDF5Array]{getHDF5DumpCompressionLevel}()} will be used. See
    \code{?\link[HDF5Array]{getHDF5DumpCompressionLevel}} for more information.}
  \item{verbose}{A \code{logical(1)} indicating whether progress messages
    should be printed (default \code{TRUE}).}
}
\details{
  \code{ns} and \code{h} are passed to the \code{locfit} function. The
  bandwidth used is the maximum (in genomic distance) of the \code{h}
  and a width big enough to contain \code{ns} number of methylation
  loci.
}

\section{Realization backends}{
  The \code{BSmooth()} function creates a new assay to store the coefficients
  used to construct the smoothed methylation estimates ((\code{coef}). An
  additional assay is also created if \code{keep.se == TRUE} (\code{se.coef}).

  The choice of \emph{realization backend} controls whether these assay(s) are
  stored in-memory as an ordinary \link[base]{matrix} or on-disk as a
  \code{\link[HDF5Array]{HDF5Array}}, for example.

  The choice of realization backend is controlled by the \code{BACKEND}
  argument, which defaults to the current value of \code{DelayedArray::\link[DelayedArray]{getAutoRealizationBackend}()}.

  \code{BSmooth} supports the following realization backends:

  \itemize{
    \item \code{NULL} (in-memory): This stores each new assay in-memory using
      an ordinary \link[base]{matrix}.
    \item \code{HDF5Array} (on-disk): This stores each new assay on-disk in a
    HDF5 file using an \code{\link[HDF5Array]{HDF5Matrix}} from \pkg{HDF5Array}.
  }

  Please note that certain combinations of realization backend and
  parallelization backend are currently not supported. For example, the
  \code{\link[HDF5Array]{HDF5Array}} realization backend is currently only compatible when
  used with a single-machine parallelization backend (i.e. it is not compatible
  with a \code{\link[BiocParallel]{SnowParam}} that specifies an \emph{ad hoc} cluster of
  \strong{multiple} machines). \code{BSmooth()} will issue an error when given
  such incompatible realization and parallelization backends. Furthermore, to
  avoid memory usage blow-ups, \code{BSmooth()} will issue an error if an
  in-memory realization backend is used when smoothing a disk-backed
  \linkS4class{BSseq} object.

  Additional arguments related to the realization backend can be passed via the
  \code{...} argument. These arguments must be named and are passed to the
  relevant \code{\link[DelayedArray]{RealizationSink}} constructor. For example, the
  \code{...} argument can be used to specify the path to the HDF5 file to be
  used by \code{BSmooth()}. Please see the examples at the bottom of the page.
}

\section{Parallelization and progress monitoring}{
  \code{BSmooth()} now uses the \pkg{BiocParallel} package to implement
  parallelization. This brings some notable improvements:

  \itemize{
    \item Smoothed results can now be written directly to an on-disk
      realization backend by the worker. This dramatically reduces memory
      usage compared to previous versions of \pkg{bsseq} that required all
      results be retained in-memory.
    \item Parallelization is now supported on Windows through the use of a
    \code{\link[BiocParallel]{SnowParam}} object as the value of \code{BPPARAM}.
    \item Detailed and extensive job logging facilities.
  }

  All parallelization options are controlled via the \code{BPPARAM} argument.
  In general, we recommend that users combine multicore (single-machine)
  parallelization with an on-disk realization backend (see section,
  'Realization backend'). For Unix and Mac users, this means using
  a \code{\link[BiocParallel]{MulticoreParam}}. For Windows users, this means using a
  single-machine \code{\link[BiocParallel]{SnowParam}}. Please consult the \pkg{BiocParallel}
  documentation to take full advantage of the more advanced features.

  \subsection{Deprecated arguments}{
    \code{parallelBy}, \code{mc.cores}, and \code{mc.preschedule} are
    deprecated and will be removed in subsequent releases of \pkg{bsseq}. These
    arguments were necessary when \code{BSmooth()} used the \pkg{parallel}
    package to implement parallelization, but this functionality is superseded
    by the aforementioned use of \pkg{BiocParallel}. We recommend that users
    who previously relied on these arguments switch to
    \code{BPPARAM = MulticoreParam(workers = mc.cores, progressbar = TRUE)}.
  }

  \subsection{Progress monitoring}{
    A useful feature of \pkg{BiocParallel} are progress bars to monitor the
    status of long-running jobs, such as \code{BSmooth()}. Progress bars are
    controlled via the \code{progressbar} argument in the
    \code{\link[BiocParallel]{BiocParallelParam}} constructor. Progress bars replace the
    use of the deprecated \code{verbose} argument to print out information on
    the status of \code{BSmooth()}.

    \pkg{BiocParallel} also supports extensive and detailed logging facilities.
    Please consult the \pkg{BiocParallel} documentation to take full advantage
    these advanced features.
  }
}

\value{
  An object of class \code{BSseq}, containing coefficients used to fit smoothed
  methylation values and optionally standard errors for these.
}
\author{
  Method and original implementation by Kasper Daniel Hansen
  \email{khansen@jhsph.edu}. Updated implementation to support disk-backed
  \linkS4class{BSseq} objects and more general parallelization by Peter Francis
  Hickey.
}
\references{
  KD Hansen, B Langmead, and RA Irizarry.
  \emph{BSmooth: from whole genome bisulfite sequencing reads to
    differentially methylated regions}.
  Genome Biology (2012) 13:R83.
  doi:\href{http://www.dx.doi.org/10.1186/gb-2012-13-10-r83}{10.1186/gb-2012-13-10-r83}.
}
\seealso{
  \code{\link[locfit]{locfit}} in the locfit package, as well as
  \code{\linkS4class{BSseq}}.
}
\examples{
\dontrun{
  # Run BSmooth() on a matrix-backed BSseq object using an in-memory realization
  # backend with serial evaluation.
  data(BS.chr22)
  # This is a matrix-backed BSseq object.
  sapply(assays(BS.chr22, withDimnames = FALSE), class)
  BS.fit <- BSmooth(BS.chr22, BPPARAM = SerialParam(progressbar = TRUE))
  # The new 'coef' assay is an ordinary matrix.
  sapply(assays(BS.fit, withDimnames = FALSE), class)
  BS.fit

  # Run BSmooth() on a disk-backed BSseq object using the HDF5Array realization
  # backend (with data written to the file 'BSmooth_example.h5') with
  # multi-core parallel evaluation.
  BS.chr22 <- realize(BS.chr22, "HDF5Array")
  # This is a disk-backed BSseq object.
  sapply(assays(BS.chr22, withDimnames = FALSE), class)
  BS.fit <- BSmooth(BS.chr22,
              BPPARAM = MulticoreParam(workers = 2, progressbar = TRUE),
              BACKEND = "HDF5Array",
              filepath = "BSmooth_example.h5")
  # The new 'coef' assay is an HDF5Matrix.
  sapply(assays(BS.fit, withDimnames = FALSE), class)
  BS.fit
  # The new 'coef' assay is in the HDF5 file 'BSmooth_example.h5' (in the
  # current working directory).
  sapply(assays(BS.fit, withDimnames = FALSE), path)
}
}

\name{read.modkit}
\alias{read.modkit}
\title{Construct BSseq objects from nanopore BED files}
\description{
Construct BSseq objects from nanopore BED files
}
\usage{
read.modkit(
    files,
    colData = NULL,
    rmZeroCov = FALSE,
    strandCollapse = TRUE
)
}
\arguments{
    \item{files}{vector, BED files}
    \item{colData}{data frame, phenotypic data with samples as rows and variables as columns}
    \item{rmZeroCov}{A logical (1) indicating whether methylation loci that have zero coverage in all samples be removed}
    \item{strandCollapse}{A logical (1) indicating whether stand-symmetric methylation loci (i.e. CpGs) should be collapsed across strands}
}

\value{
BSseq objects
}
\details{
This function is deprecated in favor of `read.bedMethyl()`.
  
This function reads in nanopore sequencing modified BED files
to Bsseq objects. Nanopore sequencing data (i.e. aggregated modified base
counts) is stored in modified-base BAM files. These modified-base BAM files
are converted to bedMethyl (BED) files using \href{https://github.com/nanoporetech/modkit}{modkit}.

\subsection{Details for modkit}{
    Modkit outputs modified reads, unmodified reads, ambiguous modification reads (reads where the probability was below the threshold and usually failing the lowest 10th percentile), and other modified reads.}

\subsection{modkit to Bsseq object}{
    After creating BED files using modkit, the BED files are read in and the Bsseq object is constructed via \code{read.modkit()} function. The function reads in BED files, extract genomic regions, methylation, coverage, ambiguous modification status data and sample information and then construct Bsseq object using \code{BSseq} function within the package. Other modification bases such as hydroxymethylation are extracted and added to the methylation matrix when present.}
}

\examples{
# No other modification present
files <- c(system.file("extdata/modkit/chr21.chr22.HG002.top1000.bed.gz", package = "bsseq"))
## bsseq_nano <- read.modkit(files, rmZeroCov = FALSE, strandCollapse=FALSE)

# Other modification present
files <- c(system.file("extdata/modkit/Hypo1.first50Bed.txt",package = "bsseq"))
## bsseq_nano <- read.modkit(files, rmZeroCov = FALSE, strandCollapse=FALSE)
}
